(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory AbstractRewriting_Exercises imports AbstractRewriting begin

section {* Exercises Concerning Abstract Rewriting *}

text {* Aart's Lecture Notes (ALN), Exercise~1.6.a. *}

lemma NF_trancl_iff_NF:"NF(r^+) = NF r" by (auto simp:Image_mono NF_no_trancl_step)

text {* ALN, Exercise~1.6.b. *}

lemma NF_subset: "r \<subseteq> s \<Longrightarrow> NF s \<subseteq> NF r" by auto

text {* ALN, Exercise~1.6.c. *}

lemma joinability_refl':"(x,x) \<in> r^\<down>" unfolding joinability_def by auto
lemma joinability_refl:"refl(r^\<down>)" unfolding refl_on_def using joinability_refl' by auto
lemma refl_imp_NF_empty: "refl(r) \<Longrightarrow> NF(r) = {}" unfolding refl_on_def by auto
lemma NF_joinability:"NF(r^\<down>) = {}" by (rule refl_imp_NF_empty[OF joinability_refl])

subsubsection {* Exercises *}
text {* ALN, Exercise~1.9.a *}
lemma NF_imp_complete: 
 assumes "x \<in> NF r" shows "complete_elt r x"
proof - {
  from assms have "\<forall>y. (x,y) \<notin> r" by (rule NF_no_step)
  hence "\<not>(\<exists>S. S 0 = x \<and> (\<forall>i. (S i,S(Suc i)) \<in> r))" by auto
 }
 hence "SN_elt r x" by auto {
  fix y z assume "(x,y) \<in> r^* \<and> (x,z) \<in> r^*"
  with `x \<in> NF r` have "x = y" and "x = z" by (auto simp:NF_not_suc)
  hence "(y,z) \<in> r^\<down>" by auto
 }
 hence "CR_elt r x" by auto
 with `SN_elt r x` show ?thesis by auto
qed

text {* ALN, Exercise~1.9.b *}
lemma semi_complete_imp_UNF_elt_exists:
 assumes "semi_complete r"
 shows "\<exists>y. (x,y) \<in> r^! \<and> UNF_elt r x"
proof -
  from assms have "WN_elt r x" and "CR_elt r x" by auto
  {
    fix y z assume "(x,y) \<in> r^!" and "(x,z) \<in> r^!"
    hence "(x,y) \<in> r^*" and "(x,z) \<in> r^*" and "y \<in> NF r" and "z \<in> NF r" by auto
    with `CR_elt r x` have "(y,z) \<in> r^\<down>" by auto
    with `y \<in> NF r` and `z \<in> NF r` have "y = z" by (auto intro: joinability_NF_imp_eq)
  } hence "UNF_elt r x" by auto
  from `WN_elt r x` have "\<exists>y. (x,y) \<in> r^!" by auto
  with `UNF_elt r x` show ?thesis by simp
qed

lemma UNF_elt_exists_imp_conversion_iff_joinability_to_NF:
 assumes "\<forall>x. \<exists>y. (x,y) \<in> r^! \<and> UNF_elt r x"
 shows "r^\<leftrightarrow>* = (r^! O (r^!)\<inverse>)"
proof
 show "r^\<leftrightarrow>* \<subseteq> r^! O (r^!)\<inverse>"
 proof (rule subsetI2)
  fix x y assume "(x,y) \<in> r^\<leftrightarrow>*"
  hence "(x,y) \<in> (r^\<leftrightarrow>)^*" by auto
  then obtain n where "(x,y) \<in> (r^\<leftrightarrow>)^^n" using rtrancl_imp_UN_rel_pow by best
  thus "(x,y) \<in> r^! O (r^!)^-1"
  proof (induct n arbitrary: x y)
   case 0 hence "x = y" by simp
   from assms obtain x' where "(x,x') \<in> r^!" and "(x',x) \<in> (r^!)\<inverse>" by best
   thus ?case unfolding `x = y` by auto
  next
   case (Suc n)
   from `(x,y) \<in> (r^\<leftrightarrow>)^^Suc n`
    obtain x' where "(x,x') \<in> r^\<leftrightarrow>" and "(x',y) \<in> (r^\<leftrightarrow>)^^n" using rel_pow_Suc_D2 by best
   with Suc have "(x',y) \<in> r^! O (r^!)\<inverse>" by simp
   then obtain u where "(x',u) \<in> r^!" and "(y,u) \<in> r^!" by auto
   from assms have "UNF_elt r x" and "UNF_elt r x'" by auto
   from assms obtain v where "(x,v) \<in> r^!" by best
   from `(x,x') \<in> r^\<leftrightarrow>` show ?case
   proof
    assume "(x,x') \<in> r"
    with `(x',u) \<in> r^!` have "(x,u) \<in> r^!" by auto
    with `UNF_elt r x` and `(x,v) \<in> r^!` have "v = u" by auto
    from `(x,v) \<in> r^!` and `(y,u) \<in> r^!` show ?thesis unfolding `v = u` by auto
   next
    assume "(x,x') \<in> r^-1"
    with `(x,v) \<in> r^!` have "(x',v) \<in> r^!" by auto
    with `UNF_elt r x'` and `(x',u) \<in> r^!` have "u = v" by auto
    from `(x,v) \<in> r^!` and `(y,u) \<in> r^!` show ?thesis unfolding `u = v` by auto
   qed
  qed
 qed
next
 show "r^! O (r^!)\<inverse> \<subseteq> r^\<leftrightarrow>*"
 proof (rule subsetI2)
  fix x y assume "(x,y) \<in> r^! O (r^!)\<inverse>"
  then obtain x' where "(x,x') \<in> r^!" and "(y,x') \<in> r^!" by auto
  hence "(x,x') \<in> r^*" and "(y,x') \<in> r^*" by auto
  hence "(x,x') \<in> r^\<leftrightarrow>*" and "(y,x') \<in> r^\<leftrightarrow>*" by (auto simp: conversion_I')
  hence "(x',y) \<in> r^\<leftrightarrow>*" using conversion_sym unfolding sym_def by fast
  with `(x,x') \<in> r^\<leftrightarrow>*` show "(x,y) \<in> r^\<leftrightarrow>*" using conversion_trans unfolding trans_def by fast
 qed
qed

lemma normalizability_trancl_absorb: "(r^!)^+ = r^!"
proof (rule equalityI)
  show "(r^!)^+ \<subseteq> r^!"
  proof (rule subsetI2)
    fix x y assume "(x,y) \<in> (r^!)^+" thus "(x,y) \<in> r^!"
      by (induct) (auto simp: normalizability_def NF_def)
  qed
next
  show "r^! \<subseteq> (r^!)^+"
  proof (rule subsetI2)
    fix x y assume "(x,y) \<in> r^!" thus "(x,y) \<in> (r^!)^+" by auto
  qed
qed

lemma conversion_iff_joinability_to_NF_imp_semi_complete:
 assumes "r^\<leftrightarrow>* = (r^! O (r^!)\<inverse>)"
 shows "semi_complete r"
proof - {
 fix x 
 {
  have "(x,x) \<in> r^\<leftrightarrow>*" by auto
  hence "\<exists>y. (x,y) \<in> r^!" unfolding assms by auto
 }
 hence "WN_elt r x" by auto
 {
  fix y z assume "(x,y) \<in> r^*" and "(x,z) \<in> r^*"
  hence "(x,y) \<in> r^\<leftrightarrow>*" and "(x,z) \<in> r^\<leftrightarrow>*" by (auto simp: conversion_I')
  from `(x,y) \<in> r^\<leftrightarrow>*` have "(y,x) \<in> r^\<leftrightarrow>*" using conversion_sym unfolding sym_def by fast
  with `(x,z) \<in> r^\<leftrightarrow>*` have "(y,z) \<in> r^\<leftrightarrow>*" using conversion_trans unfolding trans_def by fast
  then obtain y' where "(y,y') \<in> r^!" and "(z,y') \<in> r^!" unfolding assms by best
  hence "(y,y') \<in> r^*" and "(z,y') \<in> r^*" by auto
  hence "(y,z) \<in> r^\<down>" by auto
 }
 hence "CR_elt r x" by auto
 with `WN_elt r x` have "semi_complete_elt r x" by auto
} thus ?thesis by auto
qed

text {* ALN, Exercise~1.9.c *}
lemma WN_iff_NF_of_normalizability_empty: "WN r = (NF(r^!) = {})"
proof
 assume "WN r"
 {
  fix x assume "x \<in> NF(r^!)"
  hence "\<forall>y. (x,y) \<notin> r^!" by auto
  from `WN r` have "\<exists>y. (x,y) \<in> r^!" by best
  with `\<forall>y. (x,y) \<notin> r^!` have "x \<in> {}" by simp
 }
 hence "NF(r^!) \<subseteq> {}" by auto
 thus "NF(r^!) = {}" unfolding subset_empty by simp
next
 assume "NF(r^!) = {}"
 {
  fix x have "\<exists>y. (x,y) \<in> r^!"
  proof (rule ccontr)
   assume "\<not>(\<exists>y. (x,y) \<in> r^!)"
   hence "x \<in> NF(r^!)" by auto
   with `NF(r^!) = {}` show False by simp
  qed
  hence "WN_elt r x" by auto
 }
 thus "(WN::'a brel \<Rightarrow> bool) r" by auto
qed

text {* ALN, Exercise~1.10.b *}
lemma CR_elt_imp_CR_elt_reduct:
 assumes "CR_elt r x" and "(x,y) \<in> r^*"
 shows "CR_elt r y"
proof - {
 fix y' z' assume "(y,y') \<in> r^*" and "(y,z') \<in> r^*"
 with `(x,y) \<in> r^*` have "(x,y') \<in> r^*" and "(x,z') \<in> r^*" by auto
 with `CR_elt r x` have "(y',z') \<in> r^\<down>" by auto
} thus ?thesis by auto
qed

lemma UNF_elt_imp_UNF_elt_reduct:
 assumes "UNF_elt r x" and "(x,y) \<in> r^*"
 shows "UNF_elt r y"
proof - {
 fix y' z' assume "(y,y') \<in> r^!" and "(y,z') \<in> r^!"
 with `(x,y) \<in> r^*` have "(x,y') \<in> r^!" and "(x,z') \<in> r^!" by auto
 with `UNF_elt r x` have "y' = z'" by auto
} thus ?thesis by auto
qed

lemma complete_elt_imp_complete_elt_reduct:
 assumes "complete_elt r x" and "(x,y) \<in> r^*"
 shows "complete_elt r y"
proof - {
 from `complete_elt r x` have "SN_elt r x" and "CR_elt r x" by auto
 have "SN_elt r y" and "CR_elt r y"
  using SN_imp_SN_reduct[OF `SN_elt r x` `(x,y) \<in> r^*`]
    CR_elt_imp_CR_elt_reduct[OF `CR_elt r x` `(x,y) \<in> r^*`] by auto
} thus ?thesis by auto
qed

text {* ALN, Exercise~1.10.c *}
lemma WN_elt_imp_WN_elt_anc:
 assumes "WN_elt r x" and "(y,x) \<in> r^*"
 shows "WN_elt r y"
proof - 
 from `WN_elt r x` obtain z where"(x,z) \<in> r^!"  by auto
 with `(y,x) \<in> r^*`
 have "(y,z) \<in> r^!" ..
 thus ?thesis ..
qed

text {* ALN, Exercise~1.13 *}
definition NFP :: "'a brel \<Rightarrow> bool" where
 "NFP r \<equiv> (\<forall>(x::'a). WN_elt r x \<longrightarrow> CR_elt r x)"

lemma NFP_I[intro]: "(\<And>(x::'a). WN_elt r x \<Longrightarrow> CR_elt r x) \<Longrightarrow> (NFP::'a brel \<Rightarrow> bool) r" 
unfolding NFP_def by simp

lemma NFP_E[elim]: "NFP r \<Longrightarrow> ((\<And>(x::'a). WN_elt r x \<Longrightarrow> CR_elt r x) \<Longrightarrow> P) \<Longrightarrow> P"
unfolding NFP_def by simp

text {* ALN, Exercise~1.13 a) *}
lemma CR_imp_NFP:
 assumes "CR (r::'a brel)"
 shows "NFP r"
proof
 fix x :: 'a
 assume "WN_elt r x"
 show "CR_elt r x :: bool" using assms unfolding CR_def by auto
qed

text {* ALN, Exercise~1.13 c) *}
text {* If A has the NFP then \leftarrow\cdot \to^! \subseteq \to^!*}
lemma one_to_two:
 assumes "NFP (r::'a brel)"
 shows "(r\<inverse> O r^!) \<subseteq> r^!"
proof (rule subsetI2)
 fix y z assume "(y, z) \<in> (r\<inverse> O r^!)"
 then obtain x where "(x,y) \<in> r" and "(x,z) \<in> r^!" by best
 with assms have "CR_elt r x" and "WN_elt r x" unfolding NFP_def by auto
 from `(x,y) \<in> r` and `(x,z) \<in> r^!` have " (x,y) \<in> r^*" and "(x,z) \<in> r^*" by auto
 with `CR_elt r x` obtain u where "(y,u) \<in> r^*" and "(z,u) \<in> r^*" by best
 from `(x,z) \<in> r^!` have "z \<in> NF r" by auto
 with `(z,u) \<in> r^*` have "z=u" by (rule NF_not_suc)
 with `z \<in> NF r` and `(y,u) \<in> r^*` show  "(y,z) \<in> r^!" by auto
qed

text {* If \leftarrow\cdot \to^! \subseteq \to^! then any element
convertible to a NF rewrites to that NF *}
lemma two_to_three:
 assumes "r\<inverse> O r^! \<subseteq> r^!"
 shows "\<forall>x y. ((x,y) \<in> r^\<leftrightarrow>* \<and> y \<in> NF r \<longrightarrow> (x,y) \<in> r^* )"
proof (intro allI)
 fix x y 
 show "(x, y) \<in> r^\<leftrightarrow>* \<and> y \<in> NF r \<longrightarrow> (x, y) \<in> r^*"
 proof
  assume "(x, y) \<in> r^\<leftrightarrow>* \<and> y \<in> NF r"
  hence "(x,y) \<in> r^\<leftrightarrow>*" and "y  \<in> NF r" by auto
  then obtain n where "(x,y) \<in> (r^\<leftrightarrow>)^^n" unfolding conversion_def rtrancl_is_UN_rel_pow by auto
  thus "(x, y) \<in> r^*"
  proof (induct n arbitrary: x)
   case 0
   assume "(x,y) \<in> r^\<leftrightarrow> ^^ 0" hence "x = y" by simp show ?case unfolding `x = y` by simp
  next
   case (Suc n)
   from `(x,y) \<in> r^\<leftrightarrow> ^^ Suc n` obtain  z where "(x,z) \<in> r^\<leftrightarrow>" and "(z,y) \<in> r^\<leftrightarrow> ^^ n" using rel_pow_Suc_D2 by best
   with Suc have "(z,y) \<in> r^*" by simp
   with `y \<in> NF r` have  "(z,y) \<in> r^!" by auto
   from `(x,z) \<in> r^\<leftrightarrow>` show ?case
   proof
    assume "(x,z) \<in> r" with `(z,y) \<in> r^!` show ?thesis by auto
   next
    assume "(x,z) \<in> r^-1"
    with `(z,y) \<in> r^!` have "(x,y) \<in> (r\<inverse> O r^!)" by best
    with assms show ?thesis by auto
   qed
  qed
 qed
qed

text {* ALN, Exercise~1.13 d) *}

text {* NFP implies UNC and thus also UNF. The reverse implications do not hold. *}
lemma NFP_imp_UNC:
 assumes "NFP r"
 shows "UNC r"
proof
 fix x y
 assume "x \<in> NF r" and "y \<in> NF r" and "(x, y) \<in> r^\<leftrightarrow>*"
 show "x=y"
 proof -
  from `NFP r` have  "(r\<inverse> O r^!) \<subseteq> r^!" by (rule one_to_two)
  with `(x, y) \<in> r^\<leftrightarrow>*` and `y \<in> NF r` have "(x,y) \<in> r^*" using two_to_three by auto
  with this and `x \<in> NF r` show "x=y" using NF_not_suc by auto
 qed
qed

lemma NFP_imp_UN:
 assumes "NFP r"
 shows "UNF r"
proof -
 from assms have "UNC r" by (rule NFP_imp_UNC)
 thus ?thesis by (rule UNC_imp_UNF)
qed

lemma trancl_imp_ne_list_of_steps:
  assumes "(s,t) \<in> R^+"
  shows "\<exists>ts. length ts > 1 \<and> ts!0 = s \<and> ts!(length ts - 1) = t \<and>
         (\<forall>i<length ts - 1. (ts!i,ts!(Suc i)) \<in> R)" (is "\<exists>ts. _ \<and> _ \<and> _ \<and> ?P ts")
using assms proof (induct rule: trancl.induct)
  case (r_into_trancl x y)
  let ?ts = "[x,y]"
  have "length ?ts > 1" by simp
  moreover have "?ts!0 = x" by simp
  moreover have "?ts!(length ?ts - 1) = y" by simp
  moreover from r_into_trancl have "?P ?ts" by auto
  ultimately show ?case by fast
next
  case (trancl_into_trancl x y z)
  then obtain ts where "length ts > 1" and "ts!0 = x"
    and last1: "ts!(length ts - 1) = y" and "?P ts" by auto
  let ?ts = "ts@[z]"
  have len: "length ?ts = length ts + 1" by simp
  from `length ts > 1` have "length ?ts > 1" by auto
  moreover with `ts!0 = x` have "?ts!0 = x" by (induct ts) auto
  moreover have last2: "?ts!(length ?ts - 1) = z" by simp
  moreover have "?P ?ts"
  proof (intro allI impI)
    fix i assume A: "i < length ?ts - 1"
    show "(?ts!i,?ts!(Suc i)) \<in> R"
    proof (cases "i < length ts - 1")
      case True with `?P ts` A show ?thesis unfolding len unfolding nth_append by auto
    next
      case False with A have "i = length ts - 1" by simp
      with last1 `length ts > 1` have "?ts!i = y" unfolding nth_append by auto
      have "Suc i = length ?ts - 1" using `i = length ts - 1` using `length ts > 1` by auto
      with last2 have "?ts!(Suc i) = z" by auto
      from `(y,z) \<in> R` show ?thesis unfolding `?ts!(Suc i) = z` `?ts!i = y` .
    qed
  qed
  ultimately show ?case by fast
qed

end
