(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory RedPairImpl imports PolyCarrier RPO SCTProcessors
begin

types 'f lpo_repr = "(('f \<times> nat) \<times> nat)list"

datatype 'f redpair = Nat_carrier "('f,nat)lpoly_interL" 
  | Int_carrier "('f,int)lpoly_interL" 
  | Arctic_carrier "('f,arctic)lpoly_interL" 
  | Int_mat_carrier nat nat "('f,int mat)lpoly_interL"
  | Arc_mat_carrier nat "('f,arctic mat)lpoly_interL"
  | LPO "'f lpo_repr" "'f afs_list"
  | Rat_carrier rat "('f,rat)lpoly_interL"
  | Rat_mat_carrier rat nat nat "('f,rat mat)lpoly_interL"



definition get_max_arity :: "('f,'v)trsL \<Rightarrow> nat"
where "get_max_arity R = max_list (map snd (concat (map (\<lambda> (l,r). funas_term_list l @ funas_term_list r) R)))"

fun lpo_repr_to_pr :: "'f lpo_repr \<Rightarrow> ('f filtered \<Rightarrow> nat \<Rightarrow> nat)"
where "lpo_repr_to_pr prs (FPair f a) ignored = (case lookup (f,a) prs of None \<Rightarrow> 0 | Some x \<Rightarrow> x)"

fun shows_afs :: "('f::shows)afs_list \<Rightarrow> shows"
where "shows_afs af = shows_map (\<lambda>((f,n),e). ''pi('' +#+ shows f +@+ ''/'' +#+ shows n +@+ '') = '' +#+ (case e of
  Collapse i \<Rightarrow> shows(Suc i)
| AFList ids \<Rightarrow> shows(map Suc ids)) +@+ shows_nl) af"
(*
fun afs_to_string :: "('f:: sshow) afs_list \<Rightarrow> string list"
where "afs_to_string af = map (\<lambda> ((f,n), e). ''pi( '' +++ sshow f +++ ''/'' +++ sshow n +++ '' ) = '' +++ (case e of Collapse i \<Rightarrow> sshow (Suc i) | AFList ids \<Rightarrow> show_list_gen ''[]'' ''['' '','' '']'' (map (\<lambda> i. sshow (Suc i)) ids))) af"
*)

fun shows_lpo_repr :: "('f::shows)lpo_repr \<Rightarrow> shows"
where "shows_lpo_repr prs = (''LPO with the following precedence'' +#+ shows_nl +@+ shows_map (\<lambda>((f,n),pr).
  ''precedence('' +#+ shows f +@+ ''['' +#+ shows n +@+ '']) = '' +#+ shows pr +@+ shows_nl
) prs +@+ shows_nl
+@+ ''precedence(_) = 0'' +#+ shows_nl
)"

(*
fun lpo_repr_to_string :: "('f :: sshow) lpo_repr \<Rightarrow> string list"
where "lpo_repr_to_string prs = ''LPO with the following precedence'' # map (\<lambda> ((f,n),pr). ''precedence('' +++ sshow f +++ ''['' +++ sshow n +++ '']) = '' +++ sshow pr) prs @ [''precedence(_) = 0'']"
*)

definition shows_lpo_af :: "('f::shows)lpo_repr \<Rightarrow> 'f afs_list \<Rightarrow> shows"
where "shows_lpo_af repr af \<equiv>
  shows_lpo_repr repr +@+ shows_nl +@+ ''in combination with the following argument filter'' +#+ shows_nl +@+
  shows_afs af
"

(*
definition lpo_af_to_string :: "('f :: sshow) lpo_repr \<Rightarrow> 'f afs_list \<Rightarrow> string list"
where "lpo_af_to_string repr af \<equiv> lpo_repr_to_string repr @ ([] # ''in combination with the following argument filter'' #  afs_to_string af)"
*)

definition lpo_repr where "lpo_repr \<equiv> shows_string ''LPO+AF''"
definition lpoS where "lpoS \<equiv> shows_string ''>_LPO+AF''"
definition lpoNS where "lpoNS \<equiv> shows_string ''>=_LPO+AF''"


context LPO_afs_impl
begin
(* I do not understand why these lemmas are not available directly *)
lemmas mono_redpair_ur_proc = check_mono_redpair_ur_proc_sound
lemmas mono_redpair_proc = check_mono_redpair_proc_sound
lemmas rule_removal = rule_removal_sound
lemmas redpair_ur_af_proc =  check_redpair_ur_af_proc_sound
lemmas redpair_proc = check_redpair_proc_sound
lemmas redpair_ur_proc = check_redpair_ur_proc_sound
lemmas sct_proc =  check_sct_ur_af_proc_sound
end

definition create_rule_removal_proc_comp :: "('a::shows)poly_components \<Rightarrow> ('a \<Rightarrow> bool) \<Rightarrow> ('f::shows,'a)lpoly_interL \<Rightarrow> ('f,'v::shows)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> 
        shows monad"
where "create_rule_removal_proc_comp c cm I R rR  = (
  let pi = to_lpoly_inter (get_comp_1 c) I in (
  check_polo_coeffs_comp c I \<guillemotright>
  checkRuleRemovalProc (check_poly_monoI (get_comp_0 c) (get_comp_ge c) cm I)
           (check_polo_s_comp c pi)
           (check_polo_ns_comp c pi)
           (''polynomial order ('' +#+ get_comp_description c +#+ shows_string '')'')
           (create_lpoly_repr c I)
           R rR))"

fun create_rule_removal_proc :: "('f::shows)redpair \<Rightarrow> ('f,'v::shows)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow>
        shows monad"
where "create_rule_removal_proc (Int_carrier I) = create_rule_removal_proc_comp int_components int_mono I"
  |   "create_rule_removal_proc (Nat_carrier I) = create_rule_removal_proc_comp nat_components nat_mono I"
  |   "create_rule_removal_proc (Rat_carrier \<delta> I) = create_rule_removal_proc_comp (rat_components \<delta>) rat_mono I"
  |   "create_rule_removal_proc (Arctic_carrier I) = (\<lambda> _ _. error(''arctic not monotonic for TRSs'' +#+ shows_nl))"
  |   "create_rule_removal_proc (Arc_mat_carrier n I) = (\<lambda> _ _. error(''arctic not monotonic for TRSs'' +#+ shows_nl))"
  |   "create_rule_removal_proc (Int_mat_carrier n sd I) = create_rule_removal_proc_comp (mat_components n sd int_components) (mat_mono sd int_mono) I"
  |   "create_rule_removal_proc (Rat_mat_carrier \<delta> n sd I) = create_rule_removal_proc_comp (mat_components n sd (rat_components \<delta>)) (mat_mono sd rat_mono) I"
  |   "create_rule_removal_proc (LPO pri pi) = (\<lambda> R rR. let m = get_max_arity R; pr = lpo_repr_to_pr pri; \<pi> = afs_list_to_afs pi in check_wf_afs pi \<guillemotright> checkRuleRemovalProc (check_mono_afs pi) (afs_check lpoS \<pi> (lpo_strict pr m)) (afs_check lpoNS \<pi> (lpo_nstrict pr m)) lpo_repr (shows_lpo_af pri pi) R rR)"

context mono_poly_order_carrier
begin
lemma rule_removal_tt_ci :
  "ci_tt_impl (\<lambda> R rR. isOK(create_rule_removal_proc_comp (carrier_components d) check_mono I R rR))"
unfolding ci_tt_impl_def
proof (intro allI, intro impI)
  fix R rR
  let ?c = "carrier_components d"
  let ?pi = "pi_of I"
  assume check: "isOK(create_rule_removal_proc_comp ?c check_mono I R rR)"
  from check have check1: "isOK(check_polo_coeffs_comp ?c I)" by (auto simp: Let_def create_rule_removal_proc_comp_def)
  from check have accept: "isOK(checkRuleRemovalProc (check_poly_monoI 0 geq check_mono I) (check_polo_s_comp ?c ?pi) (check_polo_ns_comp ?c ?pi) id id R rR)" (is "isOK(?proc)") 
    by (auto simp: Let_def create_rule_removal_proc_comp_def)
  have "ci_tt_impl (\<lambda> R rR. isOK(checkRuleRemovalProc (check_poly_mono I) (check_polo_s_comp ?c ?pi) (check_polo_ns_comp ?c ?pi) id id R rR))"
    by (rule mono_redpair_impl.rule_removal_sound, rule conditional_inter_mono_redpair_impl, rule check1)
  with accept
  show "set rR \<subseteq>ci set R \<and> (\<forall> t. ideriv (set R) t \<longrightarrow>  (\<exists> i. ideriv (set rR) (shift t i)))"
    unfolding ci_tt_impl_def by best
qed
end

lemma create_rule_removal_proc: 
  shows "ci_tt_impl (\<lambda> R rR. isOK(create_rule_removal_proc I R rR))"
proof (cases I)
  case (Nat_carrier nI)
  hence id: "(\<lambda> R rR. isOK(create_rule_removal_proc I R rR)) = 
    (\<lambda> R rR. isOK(create_rule_removal_proc_comp nat_components nat_mono nI R rR))" by auto
  show ?thesis by (simp only: id nat_components_def, rule nat_matrix_carrier.rule_removal_tt_ci)
next
  case (Int_carrier nI)
  hence id: "(\<lambda> R rR. isOK(create_rule_removal_proc I R rR)) = 
    (\<lambda> R rR. isOK(create_rule_removal_proc_comp int_components int_mono nI R rR))" by auto
  show ?thesis by (simp only: id int_components_def, rule int_matrix_carrier.rule_removal_tt_ci)
next
  case (Rat_carrier d nI)
  hence id: "(\<lambda> R rR. isOK(create_rule_removal_proc I R rR)) = 
    (\<lambda> R rR. isOK(create_rule_removal_proc_comp (rat_components d) rat_mono nI R rR))" by auto
  show ?thesis by (simp only: id rat_components_def, rule rat_matrix_carrier.rule_removal_tt_ci[of d])
next
  case (Arctic_carrier nI) thus ?thesis unfolding ci_tt_impl_def by auto
next
  case (Rat_mat_carrier d n sd nI)
  hence id: "(\<lambda> R rR. isOK(create_rule_removal_proc I R rR)) = 
    (\<lambda> R rR. isOK(create_rule_removal_proc_comp (mat_components n sd (rat_components d)) (mat_mono sd rat_mono) nI R rR))" by auto
  show ?thesis by (simp only: id mat_components_def Let_def, simp add: rat_mono_poly_order_carrier.rule_removal_tt_ci[simplified mat_components_def Let_def, simplified])
next
  case (Int_mat_carrier n sd nI)
  hence id: "(\<lambda> R rR. isOK(create_rule_removal_proc I R rR)) = 
    (\<lambda> R rR. isOK(create_rule_removal_proc_comp (mat_components n sd (int_components)) (mat_mono sd int_mono) nI R rR))" by auto
  show ?thesis by (simp only: id mat_components_def Let_def, simp add: int_mono_poly_order_carrier.rule_removal_tt_ci[simplified mat_components_def Let_def, simplified])
next
  case (Arc_mat_carrier n nI) thus ?thesis unfolding ci_tt_impl_def by auto
next
  case (LPO pri pi)
  let ?pr = "lpo_repr_to_pr pri"
  let ?pres = "shows_lpo_af pri pi"
  let ?m = "\<lambda> R. get_max_arity R"
  let ?pi = "afs_list_to_afs pi"
  let ?chS = "afs_check lpoS ?pi"
  let ?ch = "afs_check lpoNS ?pi"
  from LPO have id: "(\<lambda> R rR. isOK(create_rule_removal_proc I R rR)) = 
    (\<lambda> R rR. isOK(check_wf_afs pi \<guillemotright> checkRuleRemovalProc (OK \<guillemotright> check_mono_afs pi) (?chS (lpo_strict ?pr (?m R))) (?ch (lpo_nstrict ?pr (?m R))) lpo_repr ?pres R rR))" by (simp add: Let_def)
  show ?thesis 
  proof (simp only: id, unfold ci_tt_impl_def, intro allI impI)
    fix R rR
    let ?check1 = "check_wf_afs pi"
    let ?check2 = "checkRuleRemovalProc (OK \<guillemotright> check_mono_afs pi) (?chS (lpo_strict ?pr (?m R))) (?ch (lpo_nstrict ?pr (?m R))) lpo_repr ?pres R rR"
    assume check: "isOK(?check1 \<guillemotright> ?check2)"
    from check have c1: "isOK(?check1)" and c2: "isOK(?check2)" by auto
    from c2 LPO_afs_impl.rule_removal[unfolded LPO_afs_impl_def, of pi, OF c1]
    show "set rR \<subseteq>ci set R \<and> (\<forall> t. ideriv (set R) t \<longrightarrow> (\<exists> i. ideriv (set rR) (shift t i)))"
      unfolding ci_tt_impl_def by blast
  qed
qed


definition create_redpair_proc_comp :: "('a::shows)poly_components \<Rightarrow> ('f::shows,'a)lpoly_interL \<Rightarrow> ('f,'v::shows)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> 
        shows monad"
where "create_redpair_proc_comp c I P rP R = (
  let pi = to_lpoly_inter (get_comp_1 c) I in (
  check_polo_coeffs_comp c I \<guillemotright>
  checkRedPairProc 
           (check_polo_s_comp c pi)
           (check_polo_ns_comp c pi)
           (''polynomial order ('' +#+ get_comp_description c +#+ shows_string '')'') 
           (create_lpoly_repr c I)
           P rP R))"

context poly_order_carrier
begin
lemma redpair_proc_comp_ci : 
  "chain_identifying_onlyDP_proc_impl (\<lambda> P R rm rrm rP. isOK(create_redpair_proc_comp (carrier_components d) I P rP R))"
unfolding "chain_identifying_onlyDP_proc_impl_def"
proof (intro allI, intro impI)
  fix P R rP
  let ?c = "carrier_components d"
  let ?pi = "to_lpoly_inter one I"
  assume check: "isOK(create_redpair_proc_comp ?c I P rP R)"
  hence check1: "isOK(check_polo_coeffs_comp ?c I)" by (auto simp: Let_def create_redpair_proc_comp_def)
  from check have accept: "isOK(checkRedPairProc (check_polo_s_comp ?c ?pi) (check_polo_ns_comp ?c ?pi) id id P rP R)" (is "isOK(?proc)") 
    by (auto simp: Let_def create_redpair_proc_comp_def) 
  have "chain_identifying_onlyDP_proc_impl (\<lambda> P R rm rrm rP. isOK(checkRedPairProc (check_polo_s_comp ?c ?pi) (check_polo_ns_comp ?c ?pi) id id P rP R))"
    by (rule redpair_impl.check_redpair_proc_sound, rule conditional_inter_redpair_impl, rule check1)
  thus "\<forall> s t \<sigma>. min_ichain (set P, set R) s t \<sigma> \<longrightarrow> (\<exists> i. ichain (set rP, set R) (shift s i) (shift t i) (shift \<sigma> i))"
    unfolding chain_identifying_onlyDP_proc_impl_def
    using accept by best
qed
end

fun create_redpair_proc :: "('f::shows)redpair \<Rightarrow> ('f,'v::shows)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> 
        shows monad"
where "create_redpair_proc (Int_carrier I) = create_redpair_proc_comp int_components I"
  |   "create_redpair_proc (Nat_carrier I) = create_redpair_proc_comp nat_components I"
  |   "create_redpair_proc (Rat_carrier \<delta> I) = create_redpair_proc_comp (rat_components \<delta>) I"
  |   "create_redpair_proc (Arctic_carrier I) = create_redpair_proc_comp arctic_components I" 
  |   "create_redpair_proc (Int_mat_carrier n sd I) = create_redpair_proc_comp (mat_components n sd int_components) I"
  |   "create_redpair_proc (Rat_mat_carrier \<delta> n sd I) = create_redpair_proc_comp (mat_components n sd (rat_components \<delta>)) I"
  |   "create_redpair_proc (Arc_mat_carrier n I) = create_redpair_proc_comp (mat_arctic_components n) I" 
  |   "create_redpair_proc (LPO pri pi) = (\<lambda> P rP R. let m = max (get_max_arity R) (get_max_arity P); pr = lpo_repr_to_pr pri; \<pi> = afs_list_to_afs pi in check_wf_afs pi \<guillemotright> checkRedPairProc (afs_check lpoS \<pi> (lpo_strict pr m)) (afs_check lpoNS \<pi> (lpo_nstrict pr m)) lpo_repr (shows_lpo_af pri pi) P rP R)"


lemma create_redpair_proc: 
  fixes I :: "('f::shows)redpair"
  shows "chain_identifying_onlyDP_proc_impl (\<lambda> P R rm rrm rP. isOK(create_redpair_proc I P rP R))"
proof (cases I)
  case (Nat_carrier nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_proc I P rP R)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_proc_comp nat_components nI P rP R))" by auto
  show ?thesis by (simp only: id nat_components_def, rule nat_matrix_carrier.redpair_proc_comp_ci)
next
  case (Int_carrier nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_proc I P rP R)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_proc_comp int_components nI P rP R))" by auto
  show ?thesis by (simp only: id int_components_def,  rule int_matrix_carrier.redpair_proc_comp_ci)
next
  case (Rat_carrier d nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_proc I P rP R)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_proc_comp (rat_components d) nI P rP R))" by auto
  show ?thesis by (simp only: id rat_components_def,  rule rat_matrix_carrier.redpair_proc_comp_ci)
next
  case (Arctic_carrier nI) 
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_proc I P rP R)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_proc_comp arctic_components nI P rP R))" by auto
  show ?thesis by (simp only: id arctic_components_def, rule arctic_matrix_carrier.redpair_proc_comp_ci) 
next
  case (Int_mat_carrier n sd nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_proc I P rP R)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_proc_comp (mat_components n sd int_components) nI P rP R))" by auto
  show ?thesis by (simp only: id Let_def mat_components_def, simp add: int_mono_poly_order_carrier.redpair_proc_comp_ci[simplified mat_components_def Let_def, simplified])
next
  case (Rat_mat_carrier d n sd nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_proc I P rP R)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_proc_comp (mat_components n sd (rat_components d)) nI P rP R))" by auto
  show ?thesis
    by (simp only: id Let_def mat_components_def, simp add: rat_mono_poly_order_carrier.redpair_proc_comp_ci[simplified mat_components_def Let_def, simplified])
next
  case (Arc_mat_carrier n nI) 
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_proc I P rP R)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_proc_comp (mat_arctic_components n) nI P rP R))" by auto
  show ?thesis by (simp only: id mat_arctic_components_def Let_def, simp add: mat_arctic_poly_order_carrier.redpair_proc_comp_ci[simplified mat_arctic_components_def Let_def, simplified])
next
  case (LPO pri pi)
  let ?pr = "lpo_repr_to_pr pri"
  let ?pres = "shows_lpo_af pri pi"
  let ?m = "\<lambda> P R. max (get_max_arity R) (get_max_arity P)"
  let ?pi = "afs_list_to_afs pi"
  let ?chS = "afs_check lpoS ?pi"
  let ?ch = "afs_check lpoNS ?pi"
  from LPO have id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_proc I P rP R)) = 
    (\<lambda> P R rm rrm rP. isOK(check_wf_afs pi \<guillemotright> checkRedPairProc (?chS (lpo_strict ?pr (?m P R))) (?ch (lpo_nstrict ?pr (?m P R))) lpo_repr ?pres P rP R))" by (simp add: Let_def)
  show ?thesis 
  proof (simp only: id, unfold chain_identifying_onlyDP_proc_impl_def, intro allI impI)
    fix P R rP :: "('f,'v::shows)trsL" and s t \<sigma>
    let ?check1 = "check_wf_afs pi"
    let ?check2 = "checkRedPairProc (?chS (lpo_strict ?pr (?m P R))) (?ch (lpo_nstrict ?pr (?m P R))) lpo_repr ?pres P rP R"
    assume check: "isOK(?check1 \<guillemotright> ?check2)"
      and chain: "min_ichain (set P, set R) s t \<sigma>"
    from check have c1: "isOK(?check1)" and c2: "isOK(?check2)" by auto
    from chain c2 LPO_afs_impl.redpair_proc[unfolded LPO_afs_impl_def, of pi, OF c1]
    show "\<exists> i. ichain (set rP, set R) (shift s i) (shift t i) (shift \<sigma> i)" unfolding chain_identifying_onlyDP_proc_impl_def by blast
  qed
qed


definition create_mono_redpair_proc_comp :: "('a::shows)poly_components \<Rightarrow> ('a \<Rightarrow> bool) \<Rightarrow> ('f::shows,'a)lpoly_interL \<Rightarrow> ('f,'v::shows)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> 
        shows monad"
  where "create_mono_redpair_proc_comp c cm I P rP R rR = (
  let pi = to_lpoly_inter (get_comp_1 c) I in (
  check_polo_coeffs_comp c I \<guillemotright>
  checkMonoRedPairProc (check_poly_monoI (get_comp_0 c) (get_comp_ge c) cm I)
           (check_polo_s_comp c pi)
           (check_polo_ns_comp c pi)
           (''polynomial order ('' +#+ get_comp_description c +#+ shows_string '')'') 
           (create_lpoly_repr c I)
           P rP R rR))"

context mono_poly_order_carrier
begin



lemma mono_redpair_proc_comp_ci : 
  "chain_identifying_proc_impl (\<lambda> P R rm rP rR. isOK(create_mono_redpair_proc_comp (carrier_components d) check_mono I P rP R rR))"
unfolding "chain_identifying_proc_impl_def"
proof (intro allI, intro impI)
  fix P R rP rR
  let ?c = "carrier_components d"
  let ?pi = "pi_of I"
  assume check: "isOK(create_mono_redpair_proc_comp ?c check_mono I P rP R rR)"
  hence check1: "isOK(check_polo_coeffs_comp ?c I)" by (auto simp: Let_def create_mono_redpair_proc_comp_def)
  from check have accept: "isOK(checkMonoRedPairProc (check_poly_monoI 0 geq check_mono I) (check_polo_s_comp ?c ?pi) (check_polo_ns_comp ?c ?pi) id id P rP R rR)" (is "isOK(?proc)") 
    by (auto simp: Let_def create_mono_redpair_proc_comp_def) 
  have "chain_identifying_proc_impl (\<lambda> P R rm rP rR. isOK(checkMonoRedPairProc (check_poly_mono I) (check_polo_s_comp ?c ?pi) (check_polo_ns_comp ?c ?pi) id id P rP R rR))"
    by (rule mono_redpair_impl.check_mono_redpair_proc_sound, rule conditional_inter_mono_redpair_impl, rule check1)
  thus "set rR \<subseteq>ci set R \<and> (\<forall> s t \<sigma>. min_ichain (set P, set R) s t \<sigma> \<longrightarrow> (\<exists> i. ichain (set rP, set rR) (shift s i) (shift t i) (shift \<sigma> i)))"
    unfolding chain_identifying_proc_impl_def
    using accept by best
qed
end

fun create_mono_redpair_proc :: "('f::shows)redpair \<Rightarrow> ('f,'v::shows)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> 
        shows monad"
where "create_mono_redpair_proc (Int_carrier I) = create_mono_redpair_proc_comp int_components int_mono I"
  |   "create_mono_redpair_proc (Nat_carrier I) = create_mono_redpair_proc_comp nat_components nat_mono I"
  |   "create_mono_redpair_proc (Rat_carrier \<delta> I) = create_mono_redpair_proc_comp (rat_components \<delta>) rat_mono I"
  |   "create_mono_redpair_proc (Arctic_carrier I) = (\<lambda> _ _ _ _. error(''arctic not monotonic for TRSs'' +#+ shows_nl))"
  |   "create_mono_redpair_proc (Arc_mat_carrier n I) = (\<lambda> _ _ _ _. error(''arctic not monotonic for TRSs'' +#+ shows_nl))"
  |   "create_mono_redpair_proc (Int_mat_carrier n sd I) = create_mono_redpair_proc_comp (mat_components n sd int_components) (mat_mono sd int_mono) I"
  |   "create_mono_redpair_proc (Rat_mat_carrier d n sd I) = create_mono_redpair_proc_comp (mat_components n sd (rat_components d)) (mat_mono sd rat_mono) I"
  |   "create_mono_redpair_proc (LPO pri pi) = (\<lambda> P rP R rR. let m = max (get_max_arity R) (get_max_arity P); pr = lpo_repr_to_pr pri; \<pi> = afs_list_to_afs pi in check_wf_afs pi \<guillemotright> checkMonoRedPairProc (check_mono_afs pi) (afs_check lpoS \<pi> (lpo_strict pr m)) (afs_check lpoNS \<pi> (lpo_nstrict pr m)) lpo_repr (shows_lpo_af pri pi) P rP R rR)"


lemma create_mono_redpair_proc: "chain_identifying_proc_impl (\<lambda> P R rm rP rR. isOK(create_mono_redpair_proc I P rP R rR))"
proof (cases I)
  case (Nat_carrier nI)
  hence id: "(\<lambda> P R rm rP rR. isOK(create_mono_redpair_proc I P rP R rR)) = 
    (\<lambda> P R rm rP rR. isOK(create_mono_redpair_proc_comp nat_components nat_mono nI P rP R rR))" by auto
  show ?thesis by (simp only: id nat_components_def, rule nat_matrix_carrier.mono_redpair_proc_comp_ci)
next
  case (Rat_carrier d nI)
  hence id: "(\<lambda> P R rm rP rR. isOK(create_mono_redpair_proc I P rP R rR)) = 
    (\<lambda> P R rm rP rR. isOK(create_mono_redpair_proc_comp (rat_components d) rat_mono nI P rP R rR))" by auto
  show ?thesis by (simp only: id rat_components_def, rule rat_matrix_carrier.mono_redpair_proc_comp_ci)
next
  case (Int_carrier nI)
  hence id: "(\<lambda> P R rm rP rR. isOK(create_mono_redpair_proc I P rP R rR)) = 
    (\<lambda> P R rm rP rR. isOK(create_mono_redpair_proc_comp int_components int_mono nI P rP R rR))" by auto
  show ?thesis by (simp only: id int_components_def, rule int_matrix_carrier.mono_redpair_proc_comp_ci)
next
  case (Arctic_carrier nI) thus ?thesis unfolding chain_identifying_proc_impl_def by auto
next
  case (Rat_mat_carrier d n sd nI) 
  hence id: "(\<lambda> P R rm rP rR. isOK(create_mono_redpair_proc I P rP R rR)) = 
    (\<lambda> P R rm rP rR. isOK(create_mono_redpair_proc_comp (mat_components n sd (rat_components d)) (mat_mono sd rat_mono) nI P rP R rR))" by auto
  show ?thesis by (simp only: id mat_components_def Let_def, simp add: rat_mono_poly_order_carrier.mono_redpair_proc_comp_ci[simplified mat_components_def Let_def, simplified])
next
  case (Int_mat_carrier n sd nI) 
  hence id: "(\<lambda> P R rm rP rR. isOK(create_mono_redpair_proc I P rP R rR)) = 
    (\<lambda> P R rm rP rR. isOK(create_mono_redpair_proc_comp (mat_components n sd int_components) (mat_mono sd int_mono) nI P rP R rR))" by auto
  show ?thesis by (simp only: id mat_components_def Let_def, simp add: int_mono_poly_order_carrier.mono_redpair_proc_comp_ci[simplified mat_components_def Let_def, simplified])
next
  case (Arc_mat_carrier n nI) thus ?thesis unfolding chain_identifying_proc_impl_def by auto
next
  case (LPO pri pi)
  let ?pr = "lpo_repr_to_pr pri"
  let ?pres = "shows_lpo_af pri pi"
  let ?m = "\<lambda> P R. max (get_max_arity R) (get_max_arity P)"
  let ?pi = "afs_list_to_afs pi"
  let ?chS = "afs_check lpoS ?pi"
  let ?ch = "afs_check lpoNS ?pi"
  from LPO have id: "(\<lambda> P R rm rP rR. isOK(create_mono_redpair_proc I P rP R rR)) = 
    (\<lambda> P R rm rP rR. isOK(check_wf_afs pi \<guillemotright> checkMonoRedPairProc (OK \<guillemotright> check_mono_afs pi) (?chS (lpo_strict ?pr (?m P R))) (?ch (lpo_nstrict ?pr (?m P R))) lpo_repr ?pres P rP R rR))" by (simp add: Let_def)
  show ?thesis 
  proof (simp only: id, unfold chain_identifying_proc_impl_def, intro allI impI)
    fix P R rP rR 
    let ?check1 = "check_wf_afs pi"
    let ?check2 = "checkMonoRedPairProc (OK \<guillemotright> check_mono_afs pi) (?chS (lpo_strict ?pr (?m P R))) (?ch (lpo_nstrict ?pr (?m P R))) lpo_repr ?pres P rP R rR"
    assume check: "isOK(?check1 \<guillemotright> ?check2)"
    from check have c1: "isOK(?check1)" and c2: "isOK(?check2)" by auto
    from c2 LPO_afs_impl.mono_redpair_proc[unfolded LPO_afs_impl_def, of pi, OF c1]
    show "set rR \<subseteq>ci set R \<and> (\<forall> s t \<sigma>. min_ichain (set P, set R) s t \<sigma> \<longrightarrow> (\<exists> i. ichain (set rP, set rR) (shift s i) (shift t i) (shift \<sigma> i)))"
      unfolding chain_identifying_proc_impl_def by blast
  qed
qed


definition create_redpair_ur_proc_comp :: "('a::shows)poly_components \<Rightarrow> ('f::shows,'a)lpoly_interL \<Rightarrow> ('f,'v::shows)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> 
  ('f \<Rightarrow> ('f,'v)trsL)option \<Rightarrow> ('f,'v)trsL \<Rightarrow> 
        shows monad"
where "create_redpair_ur_proc_comp c I P rP R rm ur = (
  let pi = to_lpoly_inter (get_comp_1 c) I in (
  check_polo_coeffs_comp c I \<guillemotright>
  checkRedPairUrProc 
           (check_polo_s_comp c pi)
           (check_polo_ns_comp c pi)
           (''polynomial order ('' +#+ get_comp_description c +#+ shows_string '')'') 
           (create_lpoly_repr c I)
           P rP R rm ur))"

context poly_order_carrier
begin

lemma redpair_ur_proc_comp_ci : 
  "chain_identifying_onlyDP_proc_impl (\<lambda> P R rm rrm rP. isOK(create_redpair_ur_proc_comp (carrier_components d) I P rP R rm ur))"
unfolding "chain_identifying_onlyDP_proc_impl_def"
proof (intro allI, intro impI)
  fix P R rP
  let ?c = "carrier_components d"
  let ?pi = "to_lpoly_inter one I"
  assume check: "isOK(create_redpair_ur_proc_comp ?c I P rP R (computeRuleMap R) ur)"
  hence check1: "isOK(check_polo_coeffs_comp ?c I)" by (auto simp: Let_def create_redpair_ur_proc_comp_def)
  from check have accept: "isOK(checkRedPairUrProc (check_polo_s_comp ?c ?pi) (check_polo_ns_comp ?c ?pi) id id P rP R (computeRuleMap R) ur)" (is "isOK(?proc)") 
    by (auto simp: Let_def create_redpair_ur_proc_comp_def) 
  have "chain_identifying_onlyDP_proc_impl (\<lambda> P R rm rrm rP. isOK(checkRedPairUrProc (check_polo_s_comp ?c ?pi) (check_polo_ns_comp ?c ?pi) id id P rP R rm ur))"
    by (rule ce_redpair_impl.check_redpair_ur_proc_sound, rule conditional_inter_ce_redpair_impl, rule check1)
  thus "\<forall> s t \<sigma>. min_ichain (set P, set R) s t \<sigma> \<longrightarrow> (\<exists> i. ichain (set rP, set R) (shift s i) (shift t i) (shift \<sigma> i))"
    unfolding chain_identifying_onlyDP_proc_impl_def
    using accept by best
qed
end

fun create_redpair_ur_proc :: "('f::shows)redpair \<Rightarrow> ('f,'v::shows)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> 
  ('f \<Rightarrow> ('f,'v)trsL)option \<Rightarrow> ('f,'v)trsL \<Rightarrow> 
        shows monad"
where "create_redpair_ur_proc (Int_carrier I) = create_redpair_ur_proc_comp int_components I"
  |   "create_redpair_ur_proc (Nat_carrier I) = create_redpair_ur_proc_comp nat_components I"
  |   "create_redpair_ur_proc (Rat_carrier \<delta> I) = create_redpair_ur_proc_comp (rat_components \<delta>) I"
  |   "create_redpair_ur_proc (Arctic_carrier I) = create_redpair_ur_proc_comp arctic_components I" 
  |   "create_redpair_ur_proc (Int_mat_carrier n sd I) = create_redpair_ur_proc_comp (mat_components n sd int_components) I"
  |   "create_redpair_ur_proc (Rat_mat_carrier \<delta> n sd I) = create_redpair_ur_proc_comp (mat_components n sd (rat_components \<delta>)) I"
  |   "create_redpair_ur_proc (Arc_mat_carrier n I) = create_redpair_ur_proc_comp (mat_arctic_components n) I" 
  |   "create_redpair_ur_proc (LPO pri pi) = (\<lambda> P rP R rm ur. let m = max (get_max_arity R) (get_max_arity P); pr = lpo_repr_to_pr pri; \<pi> = afs_list_to_afs pi in 
                check_wf_afs pi \<guillemotright> checkRedPairUrProc (afs_check lpoS \<pi> (lpo_strict pr m)) (afs_check lpoNS \<pi> (lpo_nstrict pr m)) lpo_repr (shows_lpo_af pri pi) P rP R rm ur)"


lemma create_redpair_ur_proc: 
  fixes I :: "('f::shows)redpair" and ur :: "('f,'v::shows)trsL"
  shows "chain_identifying_onlyDP_proc_impl (\<lambda> P R rm rrm rP. isOK(create_redpair_ur_proc I P rP R rm ur))"
proof (cases I)
  case (Nat_carrier nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_ur_proc I P rP R rm ur)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_ur_proc_comp nat_components nI P rP R rm ur))" by auto
  show ?thesis by (simp only: id nat_components_def,  rule nat_matrix_carrier.redpair_ur_proc_comp_ci)
next
  case (Int_carrier nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_ur_proc I P rP R rm ur)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_ur_proc_comp int_components nI P rP R rm ur))" by auto
  show ?thesis by (simp only: id int_components_def,  rule int_matrix_carrier.redpair_ur_proc_comp_ci)
next
  case (Rat_carrier d nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_ur_proc I P rP R rm ur)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_ur_proc_comp (rat_components d) nI P rP R rm ur))" by auto
  show ?thesis by (simp only: id rat_components_def,  rule rat_matrix_carrier.redpair_ur_proc_comp_ci)
next
  case (Arctic_carrier nI) 
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_ur_proc I P rP R rm ur)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_ur_proc_comp arctic_components nI P rP R rm ur))" by auto
  show ?thesis by (simp only: id arctic_components_def,  rule arctic_matrix_carrier.redpair_ur_proc_comp_ci) 
next
  case (Int_mat_carrier n sd nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_ur_proc I P rP R rm ur)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_ur_proc_comp (mat_components n sd int_components) nI P rP R rm ur))" by auto
  show ?thesis by (simp only: id mat_components_def Let_def, simp add: int_mono_poly_order_carrier.redpair_ur_proc_comp_ci[simplified mat_components_def Let_def, simplified])
next
  case (Rat_mat_carrier d n sd nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_ur_proc I P rP R rm ur)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_ur_proc_comp (mat_components n sd (rat_components d)) nI P rP R rm ur))" by auto
  show ?thesis by (simp only: id mat_components_def Let_def, simp add: rat_mono_poly_order_carrier.redpair_ur_proc_comp_ci[simplified mat_components_def Let_def, simplified])
next
  case (Arc_mat_carrier n nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_ur_proc I P rP R rm ur)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_ur_proc_comp (mat_arctic_components n) nI P rP R rm ur))" by auto
  show ?thesis 
    by (simp only: id mat_arctic_components_def Let_def, simp add: mat_arctic_poly_order_carrier.redpair_ur_proc_comp_ci[simplified mat_arctic_components_def Let_def, simplified])
next
  case (LPO pri pi)
  let ?pr = "lpo_repr_to_pr pri"
  let ?pres = "shows_lpo_af pri pi"
  let ?m = "\<lambda> P R. max (get_max_arity R) (get_max_arity P)"
  let ?pi = "afs_list_to_afs pi"
  let ?chS = "afs_check lpoS ?pi"
  let ?ch = "afs_check lpoNS ?pi"
  from LPO have id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_ur_proc I P rP R rm ur)) = 
    (\<lambda> P R rm rrm rP. isOK(check_wf_afs pi \<guillemotright> checkRedPairUrProc (?chS (lpo_strict ?pr (?m P R))) (?ch (lpo_nstrict ?pr (?m P R))) lpo_repr ?pres P rP R rm ur))" by (simp add: Let_def)
  show ?thesis 
  proof (simp only: id, unfold chain_identifying_onlyDP_proc_impl_def, intro allI impI)
    fix P R rP :: "('f::shows,'v)trsL" and s t \<sigma>
    let ?c1 = "check_wf_afs pi"
    let ?c2 = "checkRedPairUrProc (?chS (lpo_strict ?pr (?m P R))) (?ch (lpo_nstrict ?pr (?m P R))) lpo_repr ?pres P rP R (computeRuleMap R) ur"
    assume check: "isOK(?c1 \<guillemotright> ?c2)"
      and min: "min_ichain (set P, set R) s t \<sigma>"
    from check have c1: "isOK(?c1)" and c2: "isOK(?c2)" by auto
    from min c2 LPO_afs_impl.redpair_ur_proc[unfolded LPO_afs_impl_def, OF c1]
    show "\<exists> i. ichain (set rP, set R) (shift s i) (shift t i) (shift \<sigma> i)" unfolding chain_identifying_onlyDP_proc_impl_def by blast
  qed
qed


definition create_mono_redpair_ur_proc_comp :: "('a::shows)poly_components \<Rightarrow> ('a \<Rightarrow> bool) \<Rightarrow> ('f::shows,'a)lpoly_interL \<Rightarrow> ('f,'v::shows)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> 
  ('f \<Rightarrow> ('f,'v)trsL)option \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> 
        shows monad"
where "create_mono_redpair_ur_proc_comp c cm I P rP R rm rR ur = (
  let pi = to_lpoly_inter (get_comp_1 c) I in (
  check_polo_coeffs_comp c I \<guillemotright>
  checkMonoRedPairUrProc (check_poly_monoI (get_comp_0 c) (get_comp_ge c) cm I)
           (check_polo_s_comp c pi)
           (check_polo_ns_comp c pi)
           (''polynomial order ('' +#+ get_comp_description c +#+ shows_string '')'') 
           (create_lpoly_repr c I)
           P rP R rm rR ur))"


context mono_poly_order_carrier
begin

lemma mono_redpair_ur_proc_comp_ci : 
  "chain_identifying_proc_impl (\<lambda> P R rm rP rR. isOK(create_mono_redpair_ur_proc_comp (carrier_components d) check_mono I P rP R rm rR ur))"
unfolding "chain_identifying_proc_impl_def"
proof (intro allI, intro impI)
  fix P R rP rR
  let ?c = "carrier_components d"
  let ?pi = "pi_of I"
  assume check: "isOK(create_mono_redpair_ur_proc_comp ?c check_mono I P rP R (computeRuleMap R) rR ur)"
  hence check1: "isOK(check_polo_coeffs_comp ?c I)" by (auto simp: Let_def create_mono_redpair_ur_proc_comp_def)
  from check have check2: "isOK(check_poly_mono I)" by (auto simp: Let_def create_mono_redpair_ur_proc_comp_def)
  from check have accept: "isOK(checkMonoRedPairUrProc (check_poly_monoI 0 geq check_mono I) (check_polo_s_comp ?c ?pi) (check_polo_ns_comp ?c ?pi) id id P rP R (computeRuleMap R) rR ur)" (is "isOK(?proc)") 
    by (auto simp: Let_def create_mono_redpair_ur_proc_comp_def) 
  have "chain_identifying_proc_impl (\<lambda> P R rm rP rR. isOK(checkMonoRedPairUrProc (check_poly_mono I) (check_polo_s_comp ?c ?pi) (check_polo_ns_comp ?c ?pi) id id P rP R rm rR ur))"
    by (rule ce_mono_redpair_impl.check_mono_redpair_ur_proc_sound,
      rule conditional_inter_ce_mono_redpair_impl, rule check1, rule check2)
  thus "set rR \<subseteq>ci set R \<and> (\<forall> s t \<sigma>. min_ichain (set P, set R) s t \<sigma> \<longrightarrow> (\<exists> i. ichain (set rP, set rR) (shift s i) (shift t i) (shift \<sigma> i)))"
    unfolding chain_identifying_proc_impl_def
    using accept by best
qed
end

fun create_mono_redpair_ur_proc :: "('f::shows)redpair \<Rightarrow> ('f,'v::shows)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> 
  ('f \<Rightarrow> ('f,'v)trsL)option \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> 
        shows monad"
where "create_mono_redpair_ur_proc (Int_carrier I) = create_mono_redpair_ur_proc_comp int_components int_mono I"
  |   "create_mono_redpair_ur_proc (Nat_carrier I) = create_mono_redpair_ur_proc_comp nat_components nat_mono I"
  |   "create_mono_redpair_ur_proc (Rat_carrier \<delta> I) = create_mono_redpair_ur_proc_comp (rat_components \<delta>) rat_mono I"
  |   "create_mono_redpair_ur_proc (Arctic_carrier I) = (\<lambda> _ _ _ _ _ _. error(''arctic not monotonic for TRSs'' +#+ shows_nl))"
  |   "create_mono_redpair_ur_proc (Arc_mat_carrier n I) = (\<lambda> _ _ _ _ _ _. error(''arctic not monotonic for TRSs'' +#+ shows_nl))"
  |   "create_mono_redpair_ur_proc (Int_mat_carrier n sd I) = create_mono_redpair_ur_proc_comp (mat_components n sd int_components) (mat_mono sd int_mono) I"
  |   "create_mono_redpair_ur_proc (Rat_mat_carrier d n sd I) = create_mono_redpair_ur_proc_comp (mat_components n sd (rat_components d)) (mat_mono sd rat_mono) I"
  |   "create_mono_redpair_ur_proc (LPO pri pi) = (\<lambda> P rP R rR rm ur. let m = max (get_max_arity R) (get_max_arity P); pr = lpo_repr_to_pr pri; \<pi> = afs_list_to_afs pi in 
                check_wf_afs pi \<guillemotright> checkMonoRedPairUrProc (check_mono_afs pi) (afs_check lpoS \<pi> (lpo_strict pr m)) (afs_check lpoNS \<pi> (lpo_nstrict pr m)) lpo_repr (shows_lpo_af pri pi) P rP R rR rm ur)" 


lemma create_mono_redpair_ur_proc: "chain_identifying_proc_impl (\<lambda> P R rm rP rR. isOK(create_mono_redpair_ur_proc I P rP R rm rR ur))"
proof (cases I)
  case (Nat_carrier nI)
  hence id: "(\<lambda> P R rm rP rR. isOK(create_mono_redpair_ur_proc I P rP R rm rR ur)) = 
    (\<lambda> P R rm rP rR. isOK(create_mono_redpair_ur_proc_comp nat_components nat_mono nI P rP R rm rR ur))" by auto
  show ?thesis by (simp only: id nat_components_def, rule nat_matrix_carrier.mono_redpair_ur_proc_comp_ci)
next
  case (Int_carrier nI)
  hence id: "(\<lambda> P R rm rP rR. isOK(create_mono_redpair_ur_proc I P rP R rm rR ur)) = 
    (\<lambda> P R rm rP rR. isOK(create_mono_redpair_ur_proc_comp int_components int_mono nI P rP R rm rR ur))" by auto
  show ?thesis by (simp only: id int_components_def, rule int_matrix_carrier.mono_redpair_ur_proc_comp_ci)
next
  case (Int_mat_carrier n sd nI)
  hence id: "(\<lambda> P R rm rP rR. isOK(create_mono_redpair_ur_proc I P rP R rm rR ur)) = 
    (\<lambda> P R rm rP rR. isOK(create_mono_redpair_ur_proc_comp (mat_components n sd int_components) (mat_mono sd int_mono) nI P rP R rm rR ur))" by auto
  show ?thesis by (simp only: id mat_components_def Let_def, simp add: int_mono_poly_order_carrier.mono_redpair_ur_proc_comp_ci[simplified mat_components_def Let_def, simplified])
next
  case (Rat_mat_carrier d n sd nI)
  hence id: "(\<lambda> P R rm rP rR. isOK(create_mono_redpair_ur_proc I P rP R rm rR ur)) = 
    (\<lambda> P R rm rP rR. isOK(create_mono_redpair_ur_proc_comp (mat_components n sd (rat_components d)) (mat_mono sd rat_mono) nI P rP R rm rR ur))" by auto
  show ?thesis by (simp only: id mat_components_def Let_def, simp add: rat_mono_poly_order_carrier.mono_redpair_ur_proc_comp_ci[simplified mat_components_def Let_def, simplified])
next
  case (Rat_carrier d nI)
  hence id: "(\<lambda> P R rm rP rR. isOK(create_mono_redpair_ur_proc I P rP R rm rR ur)) = 
    (\<lambda> P R rm rP rR. isOK(create_mono_redpair_ur_proc_comp (rat_components d) rat_mono nI P rP R rm rR ur))" by auto
  show ?thesis by (simp only: id rat_components_def, rule rat_matrix_carrier.mono_redpair_ur_proc_comp_ci)
next
  case (Arctic_carrier nI) thus ?thesis unfolding chain_identifying_proc_impl_def by auto
next
  case (Arc_mat_carrier n nI) thus ?thesis unfolding chain_identifying_proc_impl_def by auto
next
  case (LPO pri pi)
  let ?pr = "lpo_repr_to_pr pri"
  let ?pres = "shows_lpo_af pri pi"
  let ?m = "\<lambda> P R. max (get_max_arity R) (get_max_arity P)"
  let ?pi = "afs_list_to_afs pi"
  let ?chS = "afs_check lpoS ?pi"
  let ?ch = "afs_check lpoNS ?pi"
  from LPO have id: "(\<lambda> P R rm rP rR. isOK(create_mono_redpair_ur_proc I P rP R rm rR ur)) = 
    (\<lambda> P R rm rP rR. isOK(check_wf_afs pi \<guillemotright> checkMonoRedPairUrProc (OK \<guillemotright> check_mono_afs pi) (?chS (lpo_strict ?pr (?m P R))) (?ch (lpo_nstrict ?pr (?m P R))) lpo_repr ?pres P rP R rm rR ur))" by (simp add: Let_def)
  show ?thesis 
  proof (simp only: id, unfold chain_identifying_proc_impl_def, intro allI impI)
    fix P R rP rR 
    let ?check1 = "check_wf_afs pi"
    let ?check2 = "checkMonoRedPairUrProc (OK \<guillemotright> check_mono_afs pi) (?chS (lpo_strict ?pr (?m P R))) (?ch (lpo_nstrict ?pr (?m P R))) lpo_repr ?pres P rP R (computeRuleMap R) rR ur"
    assume check: "isOK(?check1 \<guillemotright> ?check2)"
    from check have c1: "isOK(?check1)" and c2: "isOK(?check2)" by auto
    from c2 LPO_afs_impl.mono_redpair_ur_proc[unfolded LPO_afs_impl_def, of pi, OF c1]
    show "set rR \<subseteq>ci set R \<and> (\<forall> s t \<sigma>. min_ichain (set P, set R) s t \<sigma> \<longrightarrow> (\<exists> i. ichain (set rP, set rR) (shift s i) (shift t i) (shift \<sigma> i)))"
      unfolding chain_identifying_proc_impl_def by blast
  qed
qed

definition create_redpair_ur_af_proc_comp :: "('a::shows)poly_components \<Rightarrow> ('f::shows,'a)lpoly_interL \<Rightarrow> ('f,'v::shows)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> 
  ('f \<Rightarrow> ('f,'v)trsL)option \<Rightarrow> ('f,'v)trsL \<Rightarrow> 
        shows monad"
where "create_redpair_ur_af_proc_comp c I P rP R rm ur = (
  let pi = to_lpoly_inter (get_comp_1 c) I in (
  check_polo_coeffs_comp c I \<guillemotright>
  checkRedPairUrAfProc 
           (check_polo_s_comp c pi)
           (check_polo_ns_comp c pi)
           (''polynomial order ('' +#+ get_comp_description c +#+ shows_string '')'') 
           (create_lpoly_repr c I)
           P rP R rm ur
           (create_afI (get_comp_0 c) I)))"


context poly_order_carrier
begin

lemma redpair_ur_af_proc_comp_ci : 
  "chain_identifying_onlyDP_proc_impl (\<lambda> P R rm rrm rP. isOK(create_redpair_ur_af_proc_comp (carrier_components d) I P rP R rm ur))"
unfolding "chain_identifying_onlyDP_proc_impl_def"
proof (intro allI, intro impI)
  fix P R rP
  let ?c = "carrier_components d"
  let ?pi = "to_lpoly_inter one I"
  assume check: "isOK(create_redpair_ur_af_proc_comp ?c I P rP R (computeRuleMap R) ur)"
  hence check1: "isOK(check_polo_coeffs_comp ?c I)" by (auto simp: Let_def create_redpair_ur_af_proc_comp_def)
  let ?af = "create_af I"
  from check have accept: "isOK(checkRedPairUrAfProc (check_polo_s_comp ?c ?pi) (check_polo_ns_comp ?c ?pi) id id P rP R (computeRuleMap R) ur ?af)" (is "isOK(?proc)") 
    by (auto simp: Let_def create_redpair_ur_af_proc_comp_def) 
  have "chain_identifying_onlyDP_proc_impl (\<lambda> P R rm rrm rP. isOK(checkRedPairUrAfProc (check_polo_s_comp ?c ?pi) (check_polo_ns_comp ?c ?pi) id id P rP R rm ur ?af))"
    by (rule ce_af_redpair_impl.check_redpair_ur_af_proc_sound, rule conditional_inter_ce_af_redpair_impl, rule check1)
  thus "\<forall> s t \<sigma>. min_ichain (set P, set R) s t \<sigma> \<longrightarrow> (\<exists> i. ichain (set rP, set R) (shift s i) (shift t i) (shift \<sigma> i))"
    unfolding chain_identifying_onlyDP_proc_impl_def
    using accept by best
qed
end


fun create_redpair_ur_af_proc :: "('f::shows)redpair \<Rightarrow> ('f,'v::shows)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> 
  ('f \<Rightarrow> ('f,'v)trsL)option \<Rightarrow> ('f,'v)trsL \<Rightarrow> 
        shows monad"
where "create_redpair_ur_af_proc (Int_carrier I) = create_redpair_ur_af_proc_comp int_components I"
  |   "create_redpair_ur_af_proc (Nat_carrier I) = create_redpair_ur_af_proc_comp nat_components I"
  |   "create_redpair_ur_af_proc (Rat_carrier \<delta> I) = create_redpair_ur_af_proc_comp (rat_components \<delta>) I"
  |   "create_redpair_ur_af_proc (Arctic_carrier I) = create_redpair_ur_af_proc_comp arctic_components I" 
  |   "create_redpair_ur_af_proc (Int_mat_carrier n sd I) = create_redpair_ur_af_proc_comp (mat_components n sd int_components) I"
  |   "create_redpair_ur_af_proc (Rat_mat_carrier \<delta> n sd I) = create_redpair_ur_af_proc_comp (mat_components n sd (rat_components \<delta>)) I"
  |   "create_redpair_ur_af_proc (Arc_mat_carrier n I) = create_redpair_ur_af_proc_comp (mat_arctic_components n) I"
  |   "create_redpair_ur_af_proc (LPO pri pi) = (\<lambda> P rP R rm ur. let m = max (get_max_arity R) (get_max_arity P); pr = lpo_repr_to_pr pri; \<pi> = afs_list_to_afs pi in 
                check_wf_afs pi \<guillemotright> checkRedPairUrAfProc (afs_check lpoS \<pi> (lpo_strict pr m)) (afs_check lpoNS \<pi> (lpo_nstrict pr m)) lpo_repr (shows_lpo_af pri pi) P rP R rm ur (afs_list_to_af pi))"

lemma create_redpair_ur_af_proc: fixes ur :: "('f::shows,'v::shows)trsL"
  shows "chain_identifying_onlyDP_proc_impl (\<lambda> P R rm rrm rP. isOK(create_redpair_ur_af_proc I P rP R rm ur))"
proof (cases I)
  case (Nat_carrier nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_ur_af_proc I P rP R rm ur)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_ur_af_proc_comp nat_components nI P rP R rm ur))" by auto
  show ?thesis by (simp only: id nat_components_def,  rule nat_matrix_carrier.redpair_ur_af_proc_comp_ci)
next 
  case (Int_carrier nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_ur_af_proc I P rP R rm ur)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_ur_af_proc_comp int_components nI P rP R rm ur))" by auto
  show ?thesis by (simp only: id int_components_def,  rule int_matrix_carrier.redpair_ur_af_proc_comp_ci)
next
  case (Rat_carrier d nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_ur_af_proc I P rP R rm ur)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_ur_af_proc_comp (rat_components d) nI P rP R rm ur))" by auto
  show ?thesis by (simp only: id rat_components_def,  rule rat_matrix_carrier.redpair_ur_af_proc_comp_ci)
next
  case (Arctic_carrier nI) 
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_ur_af_proc I P rP R rm ur)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_ur_af_proc_comp arctic_components nI P rP R rm ur))" by auto
  show ?thesis by (simp only: id arctic_components_def,  rule arctic_matrix_carrier.redpair_ur_af_proc_comp_ci)
next
  case (Int_mat_carrier n sd nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_ur_af_proc I P rP R rm ur)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_ur_af_proc_comp (mat_components n sd int_components) nI P rP R rm ur))" by auto
  show ?thesis by (simp only: id mat_components_def Let_def, simp add: int_mono_poly_order_carrier.redpair_ur_af_proc_comp_ci[simplified mat_components_def Let_def, simplified])
next
  case (Rat_mat_carrier d n sd nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_ur_af_proc I P rP R rm ur)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_ur_af_proc_comp (mat_components n sd (rat_components d)) nI P rP R rm ur))" by auto
  show ?thesis by (simp only: id mat_components_def Let_def, simp add: rat_mono_poly_order_carrier.redpair_ur_af_proc_comp_ci[simplified mat_components_def Let_def, simplified])
next
  case (Arc_mat_carrier n nI) 
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_ur_af_proc I P rP R rm ur)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_ur_af_proc_comp (mat_arctic_components n) nI P rP R rm ur))" by auto
  show ?thesis by (simp only: id mat_arctic_components_def Let_def, simp add: mat_arctic_poly_order_carrier.redpair_ur_af_proc_comp_ci[simplified mat_arctic_components_def Let_def, simplified])
next
  case (LPO pri pi)
  let ?pr = "lpo_repr_to_pr pri"
  let ?pres = "shows_lpo_af pri pi"
  let ?m = "\<lambda> P R. max (get_max_arity R) (get_max_arity P)"
  let ?pi = "afs_list_to_afs pi"
  let ?af = "afs_list_to_af pi"
  let ?chS = "afs_check lpoS ?pi"
  let ?ch = "afs_check lpoNS ?pi"
  from LPO have id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_ur_af_proc I P rP R rm ur)) = 
    (\<lambda> P R rm rrm rP. isOK(check_wf_afs pi \<guillemotright> checkRedPairUrAfProc (?chS (lpo_strict ?pr (?m P R))) (?ch (lpo_nstrict ?pr (?m P R))) lpo_repr ?pres P rP R rm ur ?af))" by (simp add: Let_def)
  show ?thesis 
  proof (simp only: id, unfold chain_identifying_onlyDP_proc_impl_def, intro allI impI)
    fix P R rP :: "('f::shows,'v)trsL" and s t \<sigma>
    let ?c1 = "check_wf_afs pi"
    let ?c2 = "checkRedPairUrAfProc (?chS (lpo_strict ?pr (?m P R))) (?ch (lpo_nstrict ?pr (?m P R))) lpo_repr ?pres P rP R (computeRuleMap R) ur ?af"
    assume check: "isOK(?c1 \<guillemotright> ?c2)"
      and min: "min_ichain (set P, set R) s t \<sigma>"
    from check have c1: "isOK(?c1)" and c2: "isOK(?c2)" by auto
    from min c2 LPO_afs_impl.redpair_ur_af_proc[unfolded LPO_afs_impl_def, OF c1]
    show "\<exists> i. ichain (set rP, set R) (shift s i) (shift t i) (shift \<sigma> i)" unfolding chain_identifying_onlyDP_proc_impl_def by blast
  qed
qed


definition create_sct_ur_af_proc_comp :: "('a::shows)poly_components \<Rightarrow> ('f::shows,'a)lpoly_interL \<Rightarrow> ('f,'v::shows)trsL \<Rightarrow> 
  ('f \<Rightarrow> ('f,'v)trsL)option \<Rightarrow> ('f \<Rightarrow> ('f,'v)trsL)option \<Rightarrow> ('f,'v)trsL \<Rightarrow> (('f,'v)rule \<times> ((nat \<times> nat)list \<times> (nat \<times> nat)list))list \<Rightarrow>
        shows monad"
where "create_sct_ur_af_proc_comp c I P rm rrm ur graphs = (
  let pi = to_lpoly_inter (get_comp_1 c) I in (
  check_polo_coeffs_comp c I \<guillemotright>
  check_sct_ur_af_proc 
           (check_polo_s_comp c pi)
           (check_polo_ns_comp c pi)
           (''polynomial order ('' +#+ get_comp_description c +#+ '')'' +#+ shows_nl +@+
            create_lpoly_repr c I +@+ shows_nl)
           P graphs rm rrm ur
           (create_afI (get_comp_0 c) I)
))"

context poly_order_carrier
begin
lemma sct_ur_af_proc_comp_ci : 
  "chain_identifying_onlyDP_proc_impl (\<lambda> P R rm rrm rP. isOK(create_sct_ur_af_proc_comp (carrier_components d) I P rm rrm ur graphs))"
unfolding "chain_identifying_onlyDP_proc_impl_def"
proof (intro allI, intro impI)
  fix P R rP
  let ?c = "carrier_components d"
  let ?pi = "to_lpoly_inter one I"
  assume check: "isOK(create_sct_ur_af_proc_comp ?c I P (computeRuleMap R) (computeRuleMap (reverse_trs R)) ur graphs)"
  hence check1: "isOK(check_polo_coeffs_comp ?c I)" by (auto simp: Let_def create_sct_ur_af_proc_comp_def)
  let ?af = "create_af I"
  from check have accept: "isOK(check_sct_ur_af_proc (check_polo_s_comp ?c ?pi) (check_polo_ns_comp ?c ?pi) id P graphs (computeRuleMap R) (computeRuleMap (reverse_trs R)) ur ?af)" (is "isOK(?proc)")
    by (cases "computeRuleMap R", auto simp: Let_def create_sct_ur_af_proc_comp_def check_sct_ur_af_proc_def)
  have "chain_identifying_onlyDP_proc_impl (\<lambda> P R rm rrm rP. isOK(check_sct_ur_af_proc (check_polo_s_comp ?c ?pi) (check_polo_ns_comp ?c ?pi) id P graphs rm rrm ur ?af))"
    by (rule ce_af_redpair_impl.check_sct_ur_af_proc_sound, rule conditional_inter_ce_af_redpair_impl, rule check1)
  thus "\<forall> s t \<sigma>. min_ichain (set P, set R) s t \<sigma> \<longrightarrow> (\<exists> i. ichain (set rP, set R) (shift s i) (shift t i) (shift \<sigma> i))"
    unfolding chain_identifying_onlyDP_proc_impl_def
    using accept by best
qed
end

fun create_sct_ur_af_proc :: "('f::shows)redpair \<Rightarrow> ('f,'v::shows)trsL \<Rightarrow> 
  ('f \<Rightarrow> ('f,'v)trsL)option \<Rightarrow> ('f \<Rightarrow> ('f,'v)trsL)option \<Rightarrow> ('f,'v)trsL \<Rightarrow> (('f,'v)rule \<times> ((nat \<times> nat)list \<times> (nat \<times> nat)list))list \<Rightarrow>
        shows monad" 
where "create_sct_ur_af_proc (Int_carrier I) = create_sct_ur_af_proc_comp int_components I"
  |   "create_sct_ur_af_proc (Nat_carrier I) = create_sct_ur_af_proc_comp nat_components I"
  |   "create_sct_ur_af_proc (Rat_carrier \<delta> I) = create_sct_ur_af_proc_comp (rat_components \<delta>) I"
  |   "create_sct_ur_af_proc (Arctic_carrier I) = create_sct_ur_af_proc_comp arctic_components I" 
  |   "create_sct_ur_af_proc (Int_mat_carrier n sd I) = create_sct_ur_af_proc_comp (mat_components n sd int_components) I"
  |   "create_sct_ur_af_proc (Rat_mat_carrier \<delta> n sd I) = create_sct_ur_af_proc_comp (mat_components n sd (rat_components \<delta>)) I"
  |   "create_sct_ur_af_proc (Arc_mat_carrier n I) = create_sct_ur_af_proc_comp (mat_arctic_components n) I" 
  |   "create_sct_ur_af_proc (LPO pri pi) = (\<lambda> P rm rrm ur graphs. let m = max (get_max_arity ur) (get_max_arity P); pr = lpo_repr_to_pr pri; \<pi> = afs_list_to_afs pi in 
                check_wf_afs pi \<guillemotright> check_sct_ur_af_proc (afs_check lpoS \<pi> (lpo_strict pr m)) (afs_check lpoNS \<pi> (lpo_nstrict pr m)) (shows_lpo_af pri pi) P graphs rm rrm ur (afs_list_to_af pi))"

lemma create_sct_ur_af_proc: fixes ur :: "('f::shows,'v::shows)trsL"
  shows "chain_identifying_onlyDP_proc_impl (\<lambda> P R rm rrm rP. isOK(create_sct_ur_af_proc I P rm rrm ur graphs))"
proof (cases I)
  case (Nat_carrier nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_sct_ur_af_proc I P rm rrm ur graphs)) = 
    (\<lambda> P R rm rrm rP. isOK(create_sct_ur_af_proc_comp nat_components nI P rm rrm ur graphs))" by auto
  show ?thesis by (simp only: id nat_components_def,  rule nat_matrix_carrier.sct_ur_af_proc_comp_ci)
next 
  case (Int_carrier nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_sct_ur_af_proc I P rm rrm ur graphs)) = 
    (\<lambda> P R rm rrm rP. isOK(create_sct_ur_af_proc_comp int_components nI P rm rrm ur graphs))" by auto
  show ?thesis by (simp only: id int_components_def,  rule int_matrix_carrier.sct_ur_af_proc_comp_ci)
next 
  case (Rat_carrier d nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_sct_ur_af_proc I P rm rrm ur graphs)) = 
    (\<lambda> P R rm rrm rP. isOK(create_sct_ur_af_proc_comp (rat_components d) nI P rm rrm ur graphs))" by auto
  show ?thesis by (simp only: id rat_components_def,  rule rat_matrix_carrier.sct_ur_af_proc_comp_ci)
next 
  case (Arctic_carrier nI) 
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_sct_ur_af_proc I P rm rrm ur graphs)) = 
    (\<lambda> P R rm rrm rP. isOK(create_sct_ur_af_proc_comp arctic_components nI P rm rrm ur graphs))" by auto
  show ?thesis by (simp only: id arctic_components_def, rule arctic_matrix_carrier.sct_ur_af_proc_comp_ci)
next
  case (Int_mat_carrier n sd nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_sct_ur_af_proc I P rm rrm ur graphs)) = 
    (\<lambda> P R rm rrm rP. isOK(create_sct_ur_af_proc_comp (mat_components n sd int_components) nI P rm rrm ur graphs))" by auto
  show ?thesis by (simp only: id mat_components_def Let_def, simp add: int_mono_poly_order_carrier.sct_ur_af_proc_comp_ci[simplified mat_components_def Let_def, simplified])
next
  case (Rat_mat_carrier d n sd nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_sct_ur_af_proc I P rm rrm ur graphs)) = 
    (\<lambda> P R rm rrm rP. isOK(create_sct_ur_af_proc_comp (mat_components n sd (rat_components d)) nI P rm rrm ur graphs))" by auto
  show ?thesis by (simp only: id mat_components_def Let_def, simp add: rat_mono_poly_order_carrier.sct_ur_af_proc_comp_ci[simplified mat_components_def Let_def, simplified])
next
  case (Arc_mat_carrier n nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_sct_ur_af_proc I P rm rrm ur graphs)) = 
    (\<lambda> P R rm rrm rP. isOK(create_sct_ur_af_proc_comp (mat_arctic_components n) nI P rm rrm ur graphs))" by auto
  show ?thesis by (simp only: id mat_arctic_components_def Let_def, simp add: mat_arctic_poly_order_carrier.sct_ur_af_proc_comp_ci[simplified mat_arctic_components_def Let_def, simplified])
next
  case (LPO pri pi)
  let ?pr = "lpo_repr_to_pr pri"
  let ?pres = "shows_lpo_af pri pi"
  let ?m = "\<lambda> P R. max (get_max_arity R) (get_max_arity P)"
  let ?pi = "afs_list_to_afs pi"
  let ?af = "afs_list_to_af pi"
  let ?chS = "afs_check lpoS ?pi"
  let ?ch = "afs_check lpoNS ?pi"
  from LPO have id: "(\<lambda> P R rm rrm rP. isOK(create_sct_ur_af_proc I P rm rrm ur graphs)) = 
    (\<lambda> P R rm rrm rP. isOK(check_wf_afs pi \<guillemotright> check_sct_ur_af_proc (?chS (lpo_strict ?pr (?m P ur))) (?ch (lpo_nstrict ?pr (?m P ur))) ?pres P graphs rm rrm ur ?af))" by (simp add: Let_def)
  show ?thesis 
  proof (simp only: id, unfold chain_identifying_onlyDP_proc_impl_def, intro allI impI)
    fix P R rP s t \<sigma>
    let ?c1 = "check_wf_afs pi"
    let ?c2 = "check_sct_ur_af_proc (?chS (lpo_strict ?pr (?m P ur))) (?ch (lpo_nstrict ?pr (?m P ur))) ?pres P graphs (computeRuleMap R) (computeRuleMap (reverse_trs R)) ur ?af"
    assume check: "isOK(?c1 \<guillemotright> ?c2)"
      and min: "min_ichain (set P, set R) s t \<sigma>"
    hence c1: "isOK(?c1)" and c2: "isOK(?c2)" by auto
    from min c2 LPO_afs_impl.sct_proc[unfolded LPO_afs_impl_def, OF c1]
    show "\<exists> i. ichain (set rP, set R) (shift s i) (shift t i) (shift \<sigma> i)" unfolding chain_identifying_onlyDP_proc_impl_def by blast
  qed
qed

end