(*
Copyright 2009 Alexander Krauss, Christian Sternagel, René Thieman

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory SCTProcessors
imports SCT Edg SubtermCriterionImpl UsableRules 
begin


(* get_arg delivers the whole term (for 0), or otherwise the i-th argument, where 
   arguments are here indexed starting by 1 *)
fun get_arg :: "('f,'v)term \<Rightarrow> nat \<Rightarrow> ('f,'v)term"
where "get_arg t 0 = t"
    | "get_arg t (Suc n) = args t ! n"

context redpair
begin

lemma sct_semantics: "sct_semantics S NS"
  by (rule sct_semantics.intro, rule SN, rule compat)

lemma sct_with_redpair:
   fixes P :: "('f,'v)trs" and R :: "('f,'v)trsL" and Gs :: "(('f,'v)term \<times> ('f,'v)term \<times> ('f,'v)hterm \<times> ('f,'v)hterm, nat) scg list"
   assumes chain: "ichain (P,set R) s t \<sigma>"
   and no_left_vars: "\<forall> lr \<in> set R. \<forall> x. fst lr \<noteq> Var x"
   and tuple: "\<forall> (s,t) \<in> P. \<not> is_var s \<and> \<not> is_var t \<and> the (root t) \<notin> defs (set R)"
   and rules: "set R \<subseteq> NS"
   and graphs: "\<forall> (s,t) \<in> P. \<exists> stri non_stri. (
         Scg (s,t,capRM (computeRuleMap (reverse_trs R)) s,capRM (computeRuleMap R) t) (s,t,capRM (computeRuleMap (reverse_trs R)) s,capRM (computeRuleMap R) t) stri non_stri \<in> set Gs \<and>
         (\<forall> (i,j) \<in> set stri. i \<le> length (args s) \<and> j \<le> length (args t) \<and> (get_arg s i, get_arg t j) \<in> S) \<and> 
         (\<forall> (i,j) \<in> set non_stri. i \<le> length (args s) \<and> j \<le> length (args t) \<and> (get_arg s i, get_arg t j) \<in> NS))"
   and check: "check_SCT (\<lambda> (s,t,cs,ct) (u,v,cu,cv). gcMatch ct u \<and> gcMatch cu t) Gs" (is "check_SCT ?conn Gs")
   shows False
proof -
  let ?cap = "\<lambda> t. capRM (computeRuleMap R) t"
  let ?capRM = "\<lambda> s. capRM (computeRuleMap (reverse_trs R)) s"
  let ?entry = "\<lambda> k. (s k,t k,?capRM (s k),?cap (t k))"
  let ?pairs = "\<lambda> k. (?entry k, (\<lambda> j. get_arg (s k) j \<cdot> \<sigma> k))"
  let ?Rs = "(rstep (set R))^*"
  let ?R = "{(?pairs k, ?pairs (Suc k)) | k. True}"
  let ?S = "(NS \<union> S)\<^sup>* O S O (NS \<union> S)\<^sup>*"
  let ?NS = "(NS \<union> S)\<^sup>*"
  {
    fix k
    let ?sk = "\<lambda> i. get_arg (s k) i \<cdot> \<sigma> k"
    let ?tk = "\<lambda> i. get_arg (t k) i \<cdot> \<sigma> k"
    let ?sk1 = "\<lambda> i. get_arg (s (Suc k)) i \<cdot> \<sigma> (Suc k)"
    from chain have inP: "(s k, t k) \<in> P" by auto
    with tuple have validDP: "\<not> is_var (t k) \<and> the (root (t k)) \<notin> defs (set R)" by auto
    from chain inP tuple have validDP1: "\<not> is_var (s (Suc k))" by auto
    from chain have steps: "(t k \<cdot> \<sigma> k, s (Suc k) \<cdot> \<sigma> (Suc k)) \<in> ?Rs" by auto
    have argSteps: "\<And> j. j \<le> length (args (t k)) \<longrightarrow> (?tk j, ?sk1 j) \<in> ?Rs"
    proof
      fix j1      
      assume j1: "j1 \<le> length (args (t k))"
      show "(?tk j1, ?sk1 j1) \<in> ?Rs"
      proof (cases j1)
        case 0
        thus ?thesis using steps by auto
      next
        case (Suc j)
        let ?tj = "args (t k) ! j"
        let ?sj = "args (s (Suc k)) ! j"
        from validDP obtain f ts where tk: "t k = Fun f ts \<and> f \<notin> defs (set R)" by (cases "t k", auto)
        hence f: "f \<notin> defs (set R)" ..
        from validDP1 obtain g ss where sk1: "s (Suc k) = Fun g ss" by (cases "s (Suc k)", auto) 
        let ?tss = "(map (\<lambda> t. t \<cdot> \<sigma> k) ts)"
        let ?sss = "(map (\<lambda> s. s \<cdot> \<sigma> (Suc k)) ss)"
        from steps have "(Fun f ts \<cdot> \<sigma> k, Fun g ss \<cdot> \<sigma> (Suc k)) \<in> ?Rs" using tk sk1 by auto
        hence steps2: "(Fun f ?tss, Fun g ?sss) \<in> ?Rs" by auto
        from this no_left_vars f have "\<exists> us. length us = length ?tss \<and> Fun g ?sss = Fun f us \<and> (\<forall> i < length ?tss. (?tss ! i, us ! i) \<in> ?Rs)"
          by (rule nondef_root_imp_arg_steps)
        from this obtain us where nearly: "length us = length ?tss \<and> Fun g ?sss = Fun f us \<and> (\<forall> i < length ?tss. (?tss ! i, us ! i) \<in> ?Rs)" ..
        from nearly j1 Suc tk sk1 show ?thesis by auto
      qed
    qed
    from subNS wmNS rules trs_inclusion have "rstep (set R) \<subseteq> NS \<union> S" by blast
    hence stepsNS: "(rstep (set R))^* \<subseteq> ?NS" by (rule rtrancl_mono)
    from steps have conn: "?conn (?entry k) (?entry (Suc k))"
      using gcMatchCapRMSound gcMatchCapRRMSound by auto
    from bspec[OF graphs inP] obtain stri non_stri where graph: "Scg (?entry k) (?entry k) stri non_stri \<in> set Gs \<and> 
         (\<forall> (i,j) \<in> set stri. i \<le> length (args (s k)) \<and> j \<le> length (args (t k)) \<and> (get_arg (s k) i, get_arg (t k) j) \<in> S) \<and> 
         (\<forall> (i,j) \<in> set non_stri. i \<le> length (args (s k)) \<and> j \<le> length (args (t k)) \<and> (get_arg (s k) i, get_arg (t k) j) \<in> NS)" by simp blast
    let ?g = "Scg (?entry k) (?entry k) stri non_stri"
    from graph have gGs: "?g \<in> set Gs" ..
    let ?check = "\<lambda> g. ((?entry k, ?sk), (?entry (Suc k), ?sk1))
           \<in> sct_semantics.steps S NS ?conn g"
    have "?check ?g" 
    proof (simp add: sct_semantics.steps.simps sct_semantics conn[simplified], rule conjI)
      {
        fix i j
        assume "(i,j) \<in> set non_stri"
        with graph have i: "i \<le> length (args (s k))" and j: "j \<le> length (args (t k))" and NS: "(get_arg (s k) i, get_arg (t k) j) \<in> NS" by auto
        from j argSteps have "(?tk j, ?sk1 j) \<in> ?Rs" by auto
        with stepsNS have astepsNS: "(?tk j, ?sk1 j) \<in> ?NS" by auto
        from NS subNS have non_strict: "(?sk i, ?tk j) \<in> (NS \<union> S)" unfolding sub_closed_def using csI by blast
        have "(?sk i, ?sk1 j) \<in> ?NS" by (rule rtrancl_trans, rule r_into_rtrancl, rule non_strict, rule astepsNS)
      }
      thus "\<forall>ij\<in>set non_stri.(\<lambda>(i, j). (?sk i, ?sk1 j) \<in> ?NS) ij" by auto
    next
      {
        fix i j
        assume "(i,j) \<in> set stri"
        with graph have i: "i \<le> length (args (s k))" and j: "j \<le> length (args (t k))" and S: "(get_arg (s k) i, get_arg (t k) j) \<in> S" by auto
        from j argSteps have "(?tk j, ?sk1 j) \<in> ?Rs" by auto
        with stepsNS have astepsNS: "(?tk j, ?sk1 j) \<in> ?NS" by auto
        from S subS have strict: "(?sk i, ?tk j) \<in> S" unfolding sub_closed_def using csI by blast
        have "(?sk i, ?sk1 j) \<in> ?S" using strict astepsNS by auto
      }
      thus "\<forall>ij\<in>set stri.(\<lambda>(i, j). (?sk i, ?sk1 j) \<in> ?S) ij" by auto
    qed 
    with gGs have "\<exists> g \<in> set Gs. ?check g" by auto
  }
  hence main: "?R \<subseteq> (\<Union>G \<in> set Gs. sct_semantics.steps S NS ?conn G)" by auto
  have "SN ?R"
    by (rule sct_semantics.SCT_correctness2[where S = S and NS = NS], rule sct_semantics, rule main, rule check)
  obtain f where "f = ?pairs" by auto
  hence id: "?R = {(f k, f (Suc k)) | k. True}" by auto
  have "\<not> SN ?R" unfolding SN_defs by (simp only: id, blast) 
  with `SN ?R` show False by auto
qed
end

context ce_af_redpair
begin

lemma sct_ur_af_redpair:
   fixes P :: "('a,'b)trs" and R :: "('a,'b)trsL" and Gs :: "(('a,'b)term \<times> ('a,'b)term \<times> ('a,'b)hterm \<times> ('a,'b)hterm, nat) scg list"
   assumes chain: "min_ichain (P,set R) s t \<sigma>"
   and tuple: "\<forall> (s,t) \<in> P. \<not> is_var s \<and> \<not> is_var t \<and> the (root t) \<notin> defs (set R)"
   and rules: "ur \<subseteq> NS"
   and ur_closed: "ur_closed_af ur us R \<pi>"
   and graphs: "\<forall> (s,t) \<in> P. \<exists> stri non_stri. (
         Scg (s,t,capRM (computeRuleMap (reverse_trs R)) s,capRM (computeRuleMap R) t) (s,t,capRM (computeRuleMap (reverse_trs R)) s,capRM (computeRuleMap R) t) stri non_stri \<in> set Gs \<and>
         (\<forall> (i,j) \<in> set stri. i \<le> length (args s) \<and> j \<le> length (args t) \<and> ur_closed_term_af (get_arg t j) R ur us \<pi> \<and> (get_arg s i, get_arg t j) \<in> S) \<and> 
         (\<forall> (i,j) \<in> set non_stri. i \<le> length (args s) \<and> j \<le> length (args t) \<and> ur_closed_term_af (get_arg t j) R ur us \<pi> \<and> (get_arg s i, get_arg t j) \<in> NS))"
   and check: "check_SCT (\<lambda> (s,t,cs,ct) (u,v,cu,cv). gcMatch ct u \<and> gcMatch cu t) Gs" (is "check_SCT ?conn Gs")
   shows False
proof -
  let ?cap = "\<lambda> t. capRM (computeRuleMap R) t"
  let ?capRM = "\<lambda> s. capRM (computeRuleMap (reverse_trs R)) s"
  let ?entry = "\<lambda> k. (s k,t k,?capRM (s k),?cap (t k))"
  let ?pairs = "\<lambda> k. (?entry k, (\<lambda> j. get_arg (s k) j \<cdot> \<sigma> k))"
  let ?I = "i_trans (R, ur, us, (c,d,n))"
  let ?Is = "\<lambda> i. (i_trans_sub (R,ur,us,(c,d,n)) (\<sigma> i))"
  let ?pairs = "\<lambda> k. (?entry k, (\<lambda> j. get_arg (s k) j \<cdot> ?Is k))"
  let ?Rs = "(rstep (set R))^*"
  let ?R = "{(?pairs k, ?pairs (Suc k)) | k. True}"
  let ?S = "(NS \<union> S)\<^sup>* O S O (NS \<union> S)\<^sup>*"
  let ?NS = "(NS \<union> S)\<^sup>*"
  have no_left_vars: "\<forall> lr \<in> set R. \<forall> x. fst lr \<noteq> Var x"
  proof (rule ccontr)
    assume "\<not> ?thesis"
    then obtain l r x where "(l,r) \<in> set R \<and> (fst (l,r)) = Var x" by force
    hence "\<not> SN_elt (rstep (set R)) (t 0 \<cdot> \<sigma> 0)" using left_var_imp_not_SN by force
    with chain show False by (simp add: minimal_cond_def)
  qed
  {
    fix k
    let ?sk = "\<lambda> i. get_arg (s k) i \<cdot> \<sigma> k"
    let ?tk = "\<lambda> i. get_arg (t k) i \<cdot> \<sigma> k"
    let ?sk1 = "\<lambda> i. args (s (Suc k)) ! i \<cdot> \<sigma> (Suc k)"
    let ?Isk = "\<lambda> i. get_arg (s k) i \<cdot> ?Is k"
    let ?Itk = "\<lambda> i. get_arg (t k) i \<cdot> ?Is k"
    let ?Isk1 = "\<lambda> i. get_arg (s (Suc k)) i \<cdot> ?Is (Suc k)"
    from chain have inP: "(s k, t k) \<in> P" by auto
    with tuple have validDP: "\<not> is_var (t k) \<and> the (root (t k)) \<notin> defs (set R)" by auto
    from chain inP tuple have validDP1: "\<not> is_var (s (Suc k))" by auto
    from chain have steps: "(t k \<cdot> \<sigma> k, s (Suc k) \<cdot> \<sigma> (Suc k)) \<in> ?Rs" by auto
    from chain have tSN: "SN_elt (rstep (set R)) (t k \<cdot> \<sigma> k)" by (simp add: minimal_cond_def)
    have argSteps: "\<And> j. j \<le> length (args (t k)) \<and> ur_closed_term_af (get_arg (t k) j) R ur us \<pi> \<longrightarrow> (?Itk j, ?Isk1 j) \<in> ?NS"
    proof
      fix j1
      assume ass: "j1 \<le> length (args (t k)) \<and> ur_closed_term_af (get_arg (t k) j1) R ur us \<pi>" 
      show "(?Itk j1, ?Isk1 j1) \<in> ?NS"
      proof (cases j1)
        case 0
        with ass have urc: "ur_closed_term_af (t k) R ur us \<pi>" by auto
        from steps have "(t k \<cdot> \<sigma> k, s (Suc k) \<cdot> \<sigma> (Suc k)) \<in> ?Rs" (is ?part1) by auto
        from i_transVI[OF tSN `?part1` ur_closed rules] have II: "(?I (t k \<cdot> \<sigma> k), ?I (s (Suc k) \<cdot> \<sigma> (Suc k))) \<in> NS^*" ..
        from i_transI[OF urc tSN] have tI: "(t k \<cdot> ?Is k, ?I (t k \<cdot> \<sigma> k)) \<in> NS^*" . 
        from i_transII[OF SN_imp_SN_reduct[OF tSN `?part1`]] ce_orient have Is: "(?I (s (Suc k) \<cdot> \<sigma> (Suc k)), s (Suc k) \<cdot> ?Is (Suc k)) \<in> NS^*" by blast
        have "(t k \<cdot> ?Is k, s (Suc k) \<cdot> ?Is (Suc k)) \<in> NS^*" by (rule rtrancl_trans, rule tI, rule rtrancl_trans, rule II, rule Is)
        hence "(t k \<cdot> ?Is k, s (Suc k) \<cdot> ?Is (Suc k)) \<in> ?NS" using rtrancl_mono by blast
        thus ?thesis using 0 by simp
      next
        case (Suc j)
        with ass have "j < length (args (t k)) \<and> ur_closed_term_af (args (t k) ! j) R ur us \<pi>" (is "?jlen \<and> ?urc") by auto
        hence ?jlen and ?urc by auto
        let ?tj = "args (t k) ! j"
        let ?sj = "args (s (Suc k)) ! j"
        from validDP obtain f ts where tk: "t k = Fun f ts \<and> f \<notin> defs (set R)" by (cases "t k", auto)
        hence f: "f \<notin> defs (set R)" ..
        from validDP1 obtain g ss where sk1: "s (Suc k) = Fun g ss" by (cases "s (Suc k)", auto) 
        let ?tss = "(map (\<lambda> t. t \<cdot> \<sigma> k) ts)"
        let ?sss = "(map (\<lambda> s. s \<cdot> \<sigma> (Suc k)) ss)"
        from tSN have SN: "SN_elt (rstep (set R)) (Fun f ?tss)" using tk by auto
        from tk `?jlen` have mem: "?tj \<cdot> \<sigma> k \<in> set ?tss" by auto
        from steps have "(Fun f ts \<cdot> \<sigma> k, Fun g ss \<cdot> \<sigma> (Suc k)) \<in> ?Rs" using tk sk1 by auto
        hence steps2: "(Fun f ?tss, Fun g ?sss) \<in> ?Rs" by auto
        from this no_left_vars f have "\<exists> us. length us = length ?tss \<and> Fun g ?sss = Fun f us \<and> (\<forall> i < length ?tss. (?tss ! i, us ! i) \<in> ?Rs)"
          by (rule nondef_root_imp_arg_steps)
        from this obtain us where nearly: "length us = length ?tss \<and> Fun g ?sss = Fun f us \<and> (\<forall> i < length ?tss. (?tss ! i, us ! i) \<in> ?Rs)" ..
        from nearly `?jlen` tk sk1 have "(?tj \<cdot> \<sigma> k, ?sj \<cdot> \<sigma> (Suc k)) \<in> ?Rs" (is ?part1) by auto
        have SN2: "SN_elt (rstep (set R)) (?tj \<cdot> \<sigma> k)"
          by  (rule SN_imp_SN_arg[where f = f and ts = ?tss], rule SN, rule mem)
        from i_transVI[OF SN2 `?part1` ur_closed rules] have II: "(?I (?tj \<cdot> \<sigma> k), ?I (?sk1 j)) \<in> NS^*" ..
        from i_transI[OF `?urc` SN2] have tI: "(?tj \<cdot> ?Is k, ?I (?tj \<cdot> \<sigma> k)) \<in> NS^*" . 
        from i_transII[OF SN_imp_SN_reduct[OF SN2 `?part1`]] ce_orient have Is: "(?I (?sk1 j), ?sj \<cdot> ?Is (Suc k)) \<in> NS^*" by blast
        have "(?tj \<cdot> ?Is k, ?sj \<cdot> ?Is (Suc k)) \<in> NS^*" by (rule rtrancl_trans, rule tI, rule rtrancl_trans, rule II, rule Is)
        hence "(?tj \<cdot> ?Is k, ?sj \<cdot> ?Is (Suc k)) \<in> ?NS" using rtrancl_mono by blast
        thus ?thesis using Suc by simp
      qed
    qed
    from subNS wmNS rules trs_inclusion have urNS: "rstep ur \<subseteq> NS" by blast
    from steps have conn: "?conn (?entry k) (?entry (Suc k))" using gcMatchCapRMSound gcMatchCapRRMSound by auto
    from inP obtain stri non_stri where graph: "Scg (?entry k) (?entry k) stri non_stri \<in> set Gs \<and> 
         (\<forall> (i,j) \<in> set stri. i \<le> length (args (s k)) \<and> j \<le> length (args (t k)) \<and> ur_closed_term_af (get_arg (t k) j) R ur us \<pi> \<and> (get_arg (s k) i, get_arg (t k) j) \<in> S) \<and> 
         (\<forall> (i,j) \<in> set non_stri. i \<le> length (args (s k)) \<and> j \<le> length (args (t k)) \<and> ur_closed_term_af (get_arg (t k) j) R ur us \<pi> \<and> (get_arg (s k) i, get_arg (t k) j) \<in> NS)" using graphs by force
    let ?g = "Scg (?entry k) (?entry k) stri non_stri"
    from graph have gGs: "?g \<in> set Gs" ..
    let ?check = "\<lambda> g. ((?entry k, ?Isk), (?entry (Suc k)), ?Isk1)
           \<in> sct_semantics.steps S NS ?conn g"
    have "?check ?g"
    proof (simp add: sct_semantics.steps.simps sct_semantics conn[simplified], rule conjI)
      {
        fix i j
        assume "(i,j) \<in> set non_stri"
        with graph have i: "i \<le> length (args (s k))" and j: "j \<le> length (args (t k))"
          and usable: "ur_closed_term_af (get_arg (t k) j) R ur us \<pi>" and NS: "(get_arg (s k) i, get_arg (t k) j) \<in> NS" by auto
        from j usable argSteps have astepsNS: "(?Itk j, ?Isk1 j) \<in> ?NS" by auto
        from NS subNS have non_strict: "(?Isk i, ?Itk j) \<in> (NS \<union> S)" unfolding sub_closed_def using csI by blast
        have "(?Isk i, ?Isk1 j) \<in> ?NS" by (rule rtrancl_trans, rule r_into_rtrancl, rule non_strict, rule astepsNS) 
      }
      thus "\<forall>ij\<in>set non_stri.(\<lambda>(i, j). (?Isk i, ?Isk1 j) \<in> ?NS) ij" by auto
    next
      {
        fix i j
        assume "(i,j) \<in> set stri"
        with graph have i: "i \<le> length (args (s k))" and j: "j \<le> length (args (t k))"
          and usable: "ur_closed_term_af (get_arg (t k) j) R ur us \<pi>" and S: "(get_arg (s k) i, get_arg (t k) j) \<in> S" by auto
        from j usable argSteps have astepsNS: "(?Itk j, ?Isk1 j) \<in> ?NS" by auto
        from S subS have strict: "(?Isk i, ?Itk j) \<in> S" unfolding sub_closed_def using csI by blast
        have "(?Isk i, ?Isk1 j) \<in> ?S" using strict astepsNS by auto
      }
      thus "\<forall>ij\<in>set stri.(\<lambda>(i, j). (?Isk i, ?Isk1 j) \<in> ?S) ij" by auto
    qed
    with gGs have "\<exists> g \<in> set Gs. ?check g" by auto
  }
  hence main: "?R \<subseteq> (\<Union>G \<in> set Gs. sct_semantics.steps S NS ?conn G)" by auto
  have "SN ?R"
    by (rule sct_semantics.SCT_correctness2[where S = S and NS = NS], rule sct_semantics, rule main, rule check)
  obtain f where "f = ?pairs" by auto
  hence id: "?R = {(f k, f (Suc k)) | k. True}" by auto
  have "\<not> SN ?R" unfolding SN_defs by (simp only: id, blast) 
  with `SN ?R` show False by auto
qed
end

lemma sct_with_subterm:
   fixes P::"('f,'v)trs" and R::"('f,'v)trsL" and Gs::"(('f,'v)term \<times> ('f,'v)term \<times> ('f,'v)hterm \<times> ('f,'v)hterm , nat) scg list"
   assumes chain: "min_ichain (P,set R) s t \<sigma>"
   and tuple: "\<forall>(s,t)\<in>P. \<not> is_var s \<and> \<not> is_var t \<and> the (root t) \<notin> defs (set R)"
   and graphs: "\<forall>(s,t)\<in>P.\<exists>stri nstri. (
         Scg (s,t,capRM (computeRuleMap (reverse_trs R)) s,capRM (computeRuleMap R) t) (s,t,capRM (computeRuleMap (reverse_trs R)) s,capRM (computeRuleMap R) t) stri nstri \<in> set Gs \<and>
         (\<forall>(i,j)\<in>set stri. i \<le> length(args s) \<and> j \<le> length(args t) \<and> (get_arg s i, get_arg t j) \<in> supt) \<and> 
         (\<forall>(i,j)\<in>set nstri. i \<le> length(args s) \<and> j \<le> length(args t) \<and> (get_arg s i, get_arg t j) \<in> supteq))"
   and check: "check_SCT (\<lambda> (s,t,cs,ct) (u,v,cu,cv). gcMatch ct u \<and> gcMatch cu t) Gs" (is "check_SCT ?conn Gs")
   shows False
proof -
  let ?cap = "\<lambda> t. capRM (computeRuleMap R) t"
  let ?capRM = "\<lambda> s. capRM (computeRuleMap (reverse_trs R)) s"
  let ?entry = "\<lambda> k. (s (Suc k),t (Suc k),?capRM (s (Suc k)),?cap (t (Suc k)))"
  let ?s = "\<lambda>i. s(Suc i)"
  let ?t = "\<lambda>i. t(Suc i)"
  let ?\<sigma> = "\<lambda>i. \<sigma>(Suc i)"
  let ?rstep = "rstep(set R)"
  let ?supt = "SN_rel ((supt \<union> ?rstep)^+) ?rstep"
  let ?supteq = "SN_rel (?rstep^*) ?rstep"
  let ?pairs = "\<lambda>k. (?entry k,(\<lambda> j. get_arg(?s k) j \<cdot> ?\<sigma> k))"
  let ?Rs = "?rstep^*"
  let ?R = "{(?pairs k,?pairs(Suc k)) | k. True}"
  let ?S = "(?supteq \<union> ?supt)\<^sup>* O ?supt O (?supteq \<union> ?supt)\<^sup>*"
  let ?NS = "(?supteq \<union> ?supt)\<^sup>*"
  have no_left_vars: "\<forall> lr \<in> set R. \<forall> x. fst lr \<noteq> Var x"
  proof (rule ccontr)
    assume "\<not> ?thesis"
    then obtain l r x where "(l,r) \<in> set R \<and> (fst (l,r)) = Var x" by force
    hence "\<not> SN_elt (rstep (set R)) (t 0 \<cdot> \<sigma> 0)" using left_var_imp_not_SN by force
    with chain show False by (simp add: minimal_cond_def)
  qed
  have sct_semantics: "sct_semantics ?supt ?supteq"
  proof
    show "SN ?supt"
    proof -
      have "\<forall>(t::('f,'v)term). SN_elt ?supt t"
      proof (rule ccontr)
        assume "\<not>(\<forall>(t::('f,'v)term). SN_elt ?supt t)"
        then obtain t::"('f,'v)term" where "\<not> SN_elt ?supt t" by auto
        then obtain S where "S 0 = t" and "\<forall>i. (S i,S(Suc i)) \<in> ?supt"
          unfolding SN_defs by auto
        hence "\<forall>i. SN_elt ?rstep (S i)" unfolding SN_rel_def by simp
        hence "SN_elt ?rstep (S 0)" by simp
        hence "SN_elt (supt \<union> ?rstep) t" unfolding `S 0 = t`
          by (rule SN_rstep_seq_imp_SN_supt_union_rstep_seq)
        hence "SN_elt ((supt \<union> ?rstep)^+) t" by (rule SN_elt_imp_SN_elt_trancl)
        moreover have "?supt \<subseteq> (supt \<union> ?rstep)^+" by (rule SN_rel_subset)
        ultimately have "SN_elt ?supt t" by (rule SN_imp_SN_subset)
        with `\<not> SN_elt ?supt t` show False by simp
      qed
      thus ?thesis unfolding SN_defs by simp
    qed
  next
    show "?supteq O ?supt \<subseteq> ?supt"
    proof (rule subsetI2)
      fix s t assume "(s,t) \<in> ?supteq O ?supt"
      then obtain u where fst: "(s,u) \<in> ?supteq" and snd: "(u,t) \<in> ?supt" by auto
      have "(s,u) \<in> ?Rs" and "SN_elt ?rstep s" using fst unfolding SN_rel_def by auto
      moreover have "(u,t) \<in> (supt \<union> ?rstep)^+" and "SN_elt ?rstep u" using snd
        unfolding SN_rel_def by auto
      ultimately show "(s,t) \<in> ?supt"
      proof (induct)
        case base thus ?case unfolding SN_rel_def by simp
      next
        case (step v w)
        from `(v,w) \<in> ?rstep` have "(v,w) \<in> ?rstep^+" by simp
        with trancl_union_right[where r="?rstep" and s="supt"]
          have "(v,w) \<in> (supt \<union> ?rstep)^+" by blast
        with `(w,t) \<in> (supt \<union> ?rstep)^+` have "(v,t) \<in> (supt \<union> ?rstep)^+" by simp
        moreover have "SN_elt ?rstep v" by (rule rtrancl_step_preserves_SN[OF `(s,v) \<in> ?Rs` `SN_elt ?rstep s`])
        ultimately show ?case using step by simp
      qed
    qed
  qed
  from chain have min_t: "\<forall>i. SN_elt ?rstep (t i \<cdot> \<sigma> i)" by (auto simp: minimal_cond_def)
  from chain have inR: "\<forall>k. (t k \<cdot> \<sigma> k,s(Suc k) \<cdot> \<sigma>(Suc k)) \<in> ?rstep^*" by auto
  have "\<forall>i>(0::nat). SN_elt ?rstep (s i \<cdot> \<sigma> i)"
  proof (intro impI allI)
    fix i::nat assume "i > 0"
    then obtain n where "i = Suc n" by (cases i) auto
    from inR have step: "(t(i - 1) \<cdot> \<sigma>(i - 1),s i \<cdot> \<sigma> i) \<in> ?rstep^*" unfolding `i = Suc n` by auto
    from min_t have "SN_elt ?rstep (t(i - 1) \<cdot> \<sigma>(i - 1))" unfolding `i = Suc n` by auto
    from rtrancl_step_preserves_SN[OF step this] show "SN_elt ?rstep (s i \<cdot> \<sigma> i)" .
  qed
  hence min_s: "\<forall>i. SN_elt ?rstep (?s i \<cdot> ?\<sigma> i)" by auto
  {
    fix k
    let ?sk = "\<lambda>i. get_arg(?s k) i \<cdot> ?\<sigma> k"
    let ?tk = "\<lambda>i. get_arg(?t k) i \<cdot> ?\<sigma> k"
    let ?sk1 = "\<lambda>i. get_arg(?s(Suc k)) i \<cdot> ?\<sigma>(Suc k)"
    from chain have inP: "(?s k,?t k) \<in> P" by auto
    from chain have inR: "(?t k \<cdot> ?\<sigma> k,?s(Suc k) \<cdot> ?\<sigma>(Suc k)) \<in> ?rstep^*" by auto
    have tk_supt: "\<forall>i\<le>length(args(?t k)). i > 0 \<longrightarrow> (?t k \<cdot> ?\<sigma> k,?tk i) \<in> supt"
    proof (intro impI allI)
      fix i assume i: "i \<le> length(args(?t k))" and i2: "i > 0" thus "(?t k \<cdot> ?\<sigma> k,?tk i) \<in> supt"
      proof (cases "?t k")
        case (Var x) with tuple and inP show ?thesis by auto
      next
        case (Fun f ss)
        from i2 obtain j where j2: "i = Suc j" by (cases i, auto)
        with i have j: "j < length (args(?t k))" by auto 
        from j have "(?t k,args(?t k)!j) \<in> supt" unfolding Fun
          using supt.arg[of "args(Fun f ss)!j" "ss" f] by auto
        thus ?thesis using j2  suptp_stable[to_set] by auto
      qed
    qed
    have sk_supt: "\<forall>i\<le>length(args(?s k)). i > 0 \<longrightarrow> (?s k \<cdot> ?\<sigma> k,?sk i) \<in> supt"
    proof (intro impI allI)
      fix i assume i: "i \<le> length(args(?s k))" and i2: "i > 0" thus "(?s k \<cdot> ?\<sigma> k,?sk i) \<in> supt"
      proof (cases "?s k")
        case (Var x) with tuple and inP show ?thesis by auto
      next
        case (Fun f ss)
        from i2 obtain j where j2: "i = Suc j" by (cases i, auto)
        with i have j: "j < length (args(?s k))" by auto 
        from j have "(?s k,args(?s k)!j) \<in> supt" unfolding Fun
          using supt.arg[of "args(Fun f ss)!j" "ss" f] by auto
        thus ?thesis using j2 suptp_stable[to_set] by auto
      qed
    qed
    from chain have sntk: "SN_elt ?rstep (?t k \<cdot> ?\<sigma> k)" by (simp add: minimal_cond_def)
    have SN_tki: "\<forall>i \<le> length(args(?t k)). SN_elt ?rstep (?tk i)"
    proof
      fix i
      show "i \<le> length(args(?t k)) \<longrightarrow> SN_elt ?rstep (?tk i)"
        using subterm_preserves_SN[OF sntk] tk_supt sntk by (cases i, auto)
    qed
    have SN_sk: "SN_elt ?rstep (?s k \<cdot> ?\<sigma> k)" using min_s by simp
    have SN_ski: "\<forall>i\<le>length(args(?s k)). SN_elt ?rstep (?sk i)"
    proof
      fix i
      show "i \<le> length(args(?s k)) \<longrightarrow> SN_elt ?rstep (?sk i)"
        using subterm_preserves_SN[OF SN_sk] sk_supt SN_sk by (cases i, auto)
    qed
    from inP tuple have validDP: "\<not> is_var(?t k) \<and> the(root(?t k)) \<notin> defs (set R)" by auto
    from chain inP tuple have validDP1: "\<not> is_var(?s(Suc k))" by auto
    from chain have steps: "(?t k \<cdot> ?\<sigma> k,?s(Suc k) \<cdot> ?\<sigma>(Suc k)) \<in> ?Rs" by auto
    have stepsNS: "\<forall>j\<le>length(args(?t k)). (?tk j,?sk1 j) \<in> ?NS"
    proof (intro allI impI)
      fix j1
      assume j1: "j1 \<le> length(args(?t k))"
      show "(?tk j1, ?sk1 j1) \<in> ?NS"
      proof (cases j1)
        case 0
        with steps show ?thesis unfolding SN_rel_def using sntk by auto        
      next
        case (Suc j)
        let ?tj = "args(?t k)!j"
        let ?sj = "args(?s(Suc k))!j"
        from j1 Suc have j: "j < length(args(?t k))" by simp
        from validDP obtain f ts where tk: "?t k = Fun f ts \<and> f \<notin> defs (set R)" by (cases "?t k", auto)
        hence f: "f \<notin> defs(set R)" ..
        from validDP1 obtain g ss where sk1: "?s(Suc k) = Fun g ss" by (cases "?s(Suc k)") auto
        let ?tss = "map (\<lambda>t. t \<cdot> ?\<sigma> k) ts"
        let ?sss = "map (\<lambda>s. s \<cdot> ?\<sigma>(Suc k)) ss"
        from steps have "(Fun f ts \<cdot> ?\<sigma> k, Fun g ss \<cdot> ?\<sigma>(Suc k)) \<in> ?Rs" using tk sk1 by auto
        hence steps2: "(Fun f ?tss, Fun g ?sss) \<in> ?Rs" by auto
        from this no_left_vars f have "\<exists> us. length us = length ?tss \<and> Fun g ?sss = Fun f us \<and> (\<forall> i < length ?tss. (?tss ! i, us ! i) \<in> ?Rs)"
          by (rule nondef_root_imp_arg_steps)
        from this obtain us where nearly: "length us = length ?tss \<and> Fun g ?sss = Fun f us \<and> (\<forall> i < length ?tss. (?tss ! i, us ! i) \<in> ?Rs)" ..
        from SN_tki have SN_tkj: "SN_elt ?rstep (?tj \<cdot> ?\<sigma> k)" using j1 Suc by auto
        from nearly j tk sk1 have "(?tj \<cdot> ?\<sigma> k,?sj \<cdot> ?\<sigma>(Suc k)) \<in> ?Rs" by auto
        with SN_tkj have "(?tj \<cdot> ?\<sigma> k,?sj \<cdot> ?\<sigma>(Suc k)) \<in> SN_rel ?Rs ?rstep" unfolding SN_rel_def by simp
        thus ?thesis using Suc unfolding SN_rel_def by auto
      qed
    qed
    from steps have conn: "?conn (?entry k) (?entry (Suc k))" using gcMatchCapRMSound gcMatchCapRRMSound by auto
    from inP obtain stri nstri where graph: "Scg (?entry k) (?entry k) stri nstri \<in> set Gs \<and> 
         (\<forall>(i,j)\<in>set stri. i \<le> length(args(?s k)) \<and> j \<le> length(args(?t k)) \<and> (get_arg(?s k) i,get_arg(?t k) j) \<in> supt) \<and> 
         (\<forall>(i,j)\<in>set nstri. i \<le> length(args(?s k)) \<and> j \<le> length(args(?t k)) \<and> (get_arg(?s k) i,get_arg(?t k) j) \<in> supteq)" using graphs by force
    let ?g = "Scg (?entry k) (?entry k) stri nstri"
    from graph have gGs: "?g \<in> set Gs" ..
    let ?check = "\<lambda>g. ((?entry k,?sk),(?entry (Suc k),?sk1))
           \<in> sct_semantics.steps ?supt ?supteq ?conn g"
    have "?check ?g"
    proof (simp add: sct_semantics.steps.simps sct_semantics conn[simplified], rule conjI)
      {
        fix i j
        assume "(i,j) \<in> set nstri"
        with graph have i: "i \<le> length(args(?s k))" and j: "j \<le> length(args(?t k))"
          and NS: "(get_arg(?s k) i,get_arg(?t k) j) \<in> supteq" by auto
        from j stepsNS have astepsNS: "(?tk j,?sk1 j) \<in> ?NS" by auto
        from SN_ski and i have snski: "SN_elt ?rstep (?sk i)" by simp
        from NS supteq_sub_closed have "(?sk i,?tk j) \<in> supteq"
          unfolding sub_closed_def using csI by blast
        hence non_strict: "(?sk i,?tk j) \<in> ?NS" using snski
          unfolding SN_rel_def supteq_supt_eq_conv by force
        from rtrancl_trans[OF non_strict astepsNS] have "(?sk i,?sk1 j) \<in> ?NS" by auto
      }
      thus "\<forall>ij\<in>set nstri.(\<lambda>(i,j). (?sk i,?sk1 j) \<in> ?NS) ij" by auto
    next
      {
        fix i j
        assume "(i,j) \<in> set stri"
        with graph have i: "i \<le> length (args(?s k))" and j: "j \<le> length(args(?t k))"
          and S: "(get_arg(?s k) i, get_arg(?t k) j) \<in> supt" by auto
        from j stepsNS have astepsNS: "(?tk j,?sk1 j) \<in> ?NS" by auto
        from SN_ski and i have snski: "SN_elt ?rstep (?sk i)" by simp
        from S supt_sub_closed have "(?sk i,?tk j) \<in> supt" unfolding sub_closed_def using csI by blast
        hence strict: "(?sk i,?tk j) \<in> ?supt" using snski
          unfolding SN_rel_def by force
        have "(?sk i,?sk1 j) \<in> ?S" using strict astepsNS by auto
      }
      thus "\<forall>ij\<in>set stri.(\<lambda>(i,j). (?sk i,?sk1 j) \<in> ?S) ij" by auto
    qed
    with gGs have "\<exists>g\<in>set Gs. ?check g" by blast
  }
  hence main: "?R \<subseteq> (\<Union>G \<in> set Gs. sct_semantics.steps ?supt ?supteq ?conn G)" by auto
  have "SN ?R"
    by (rule sct_semantics.SCT_correctness2[where S = ?supt and NS = ?supteq],rule sct_semantics,rule main,rule check)
  obtain f where "f = ?pairs" by auto
  hence id: "?R = {(f k, f (Suc k)) | k. True}" by auto
  have "\<not> SN ?R" unfolding SN_defs by (simp only: id, blast)
  with `SN ?R` show False by auto
qed

text {* towards automation *}

(* use rulemaps to check tuple and no_left_var condition (as they are precomputed in anyway).
  this might also be useful for subterm-criterion implementation *)
lemma ruleMap_not_None_eq_no_vars_left:
  shows "(computeRuleMap R \<noteq> None) = (\<forall> lr \<in> set R. \<forall> x. fst lr \<noteq> Var x)"
proof (cases "computeRuleMap R")
  case None
  from None computeRuleMapSound2 have "\<exists> (l,r) \<in> set R. root l = None" by auto
  from this obtain l r where "(l,r) \<in> set R \<and> root l = None" by auto
  from this have "(l,r) \<in> set R \<and> \<not> (\<forall> x. fst (l,r) \<noteq> Var x)" by (cases l, auto)
  with None show ?thesis by blast
next
  case (Some rm)
  hence left: "computeRuleMap R \<noteq> None" by auto
  from Some computeRuleMapSound2 have "\<forall> (l,r) \<in> set R. root l \<noteq> None" by force
  hence "\<forall> lr \<in> set R. \<forall> x. fst lr \<noteq> Var x" by auto
  with left show ?thesis by blast
qed

lemma ruleMap_empty_eq_nondef: fixes R :: "('f,'v)trsL"
  assumes "computeRuleMap R = Some rm"
  shows "(rm f = []) = (f \<notin> defs (set R))"
proof -
  from assms computeRuleMapSound have rm: "\<And> f :: 'f. set (rm f) = {(l,r) | l r. (l, r) \<in> set R \<and> root l = Some f}" by force
  show ?thesis
  proof (cases "rm f")
    case Nil
    with rm have "f \<notin> defs (set R)" unfolding defs_def defined_def by force
    with Nil show ?thesis by blast
  next
    case (Cons lr RR)
    hence left: "rm f \<noteq> []" by auto
    from Cons rm[where f = f] have "f \<in> defs (set R)" unfolding defs_def defined_def by (cases lr, force)
    with left show ?thesis by blast
  qed
qed

definition check_supt :: "('f::shows,'v::shows)rule \<Rightarrow> shows monad"
where "check_supt lr \<equiv> checkB ((fst lr) \<rhd> (snd lr))
                              (''the rhs is not a proper subterm of the lhs'' +#+ shows_nl)
                           \<leftarrow> (\<lambda>s. ''Could not orient rule '' +#+ shows_rule lr +@+ shows_nl +@+ s)"

definition check_supteq :: "('f::shows,'v::shows)rule \<Rightarrow> shows monad"
where "check_supteq lr \<equiv> checkB ((fst lr) \<unrhd> (snd lr))
                              (''the rhs is not a subterm of the lhs'' +#+ shows_nl)
                           \<leftarrow> (\<lambda>s. ''Could not orient rule '' +#+ shows_rule lr +@+ shows_nl +@+ s)"

lemma check_supt_sound[simp]:
  "isOK(check_supt r) = (r \<in> supt)"
  by (cases r, simp add: check_supt_def supt_def)

lemma check_supteq_sound[simp]:
  "isOK(check_supteq r) = (r \<in> supteq)"
  by (cases r, auto simp: check_supteq_def supteq_def)

definition check_sct_subterm_proc :: "
  ('f::shows,'v::shows)trsL \<Rightarrow>
  (('f,'v)rule \<times> ((nat\<times>nat)list \<times> (nat\<times>nat)list))list \<Rightarrow>
  ('f \<Rightarrow> ('f,'v)trsL)option \<Rightarrow>
  ('f \<Rightarrow> ('f,'v)trsL)option \<Rightarrow> shows monad"
where "check_sct_subterm_proc P Gs rm rrm = (
    case rm of None \<Rightarrow> error(''TRS has variables as left-hand sides'' +#+ shows_nl)
             | Some irm \<Rightarrow> (
               checkSubset P (map fst Gs)
               \<leftarrow> (\<lambda>s. ''there is no size-change graph for DP '' +#+ shows_rule s +@+ shows_nl)
               \<guillemotright> checkAllM (\<lambda>((s,t),(stri,nstri)). ((
                 checkB (\<not> is_var s \<and> \<not> is_var t \<and> irm(the(root t)) = [])
                        (''roots of DP must be constructors'' +#+ shows_nl)
               \<guillemotright> (let n  = length (args s) in
                  let m  = length (args t) in (
                    checkAllM (\<lambda>(i,j). checkB (i \<le> n \<and> j \<le> m \<and> isOK(check_supt (get_arg s i,get_arg t j)))
                                              (''problem with edge '' +#+ shows i +@+ '' |> '' +#+ shows j +@+ shows_nl)) stri
                  \<guillemotright> checkAllM (\<lambda>(i,j). checkB (i \<le> n \<and> j \<le> m \<and> isOK(check_supteq (get_arg s i,get_arg t j)))
                                              (''problem with edge '' +#+ shows i +@+ '' |>= '' +#+ shows j +@+ shows_nl)) nstri
                  ))
               )) \<leftarrow> (\<lambda>m. ''problems with DP '' +#+ shows_rule(s,t) +@+ shows_nl +@+ m)) Gs
             \<guillemotright> checkB (check_SCT (\<lambda>(s,t,cs,ct) (u,v,cu,cv). gcMatchI ct u \<and> gcMatchI cu t) (map (\<lambda>((s,t),(stri,nstri)). let e = (s,t,capRM rrm s,capRM rm t) in Scg e e stri nstri) Gs))
                      (''size-change analysis failed'' +#+ shows_nl)
    ) \<leftarrow> (\<lambda>s. ''could not apply the size-change processor with subterm-relation for the following reason'' +#+ shows_nl +@+ s)
)"

lemma check_sct_subterm_proc_sound:
  "chain_identifying_onlyDP_proc_impl (\<lambda>P R rm rrm rP. isOK(check_sct_subterm_proc P Gs rm rrm))"
unfolding chain_identifying_onlyDP_proc_impl_def
proof (intro allI impI)
  fix P R rP s t \<sigma>
  assume check: "isOK(check_sct_subterm_proc P Gs 
                   (computeRuleMap R) (computeRuleMap(reverse_trs R)))"
    and chain: "min_ichain (set P,set R) s t \<sigma>"
  let ?cap = "capRM (computeRuleMap R)"
  let ?capRM = "capRM (computeRuleMap (reverse_trs R))"
  let ?entry = "\<lambda> s t. (s,t,?capRM s, ?cap t)"
  from check obtain rm where rm: "computeRuleMap R = Some rm" by (cases "computeRuleMap R") (simp add: check_sct_subterm_proc_def)
  let ?Gs = "map (\<lambda>((s,t),(stri,nstri)). Scg (?entry s t) (?entry s t) stri nstri) Gs"
  let ?conn = "\<lambda>(s,t,cs,ct) (u,v,cu,cv). gcMatch ct u \<and> gcMatch cu t"
  from check rm have check1: "set P \<subseteq> fst ` set Gs \<and> check_SCT ?conn ?Gs" by (simp add: Let_def check_sct_subterm_proc_def)
  let ?check2'' = "\<lambda>f g s t stri nstri. \<not> is_var s \<and> \<not> is_var t \<and> rm(the(root t)) = [] \<and>
    (\<forall>(i,j) \<in> set stri. i \<le> length(args s) \<and> j \<le> length(args t) \<and> f(get_arg s i,get_arg t j)) \<and>
    (\<forall>(i,j) \<in> set nstri. i \<le> length(args s) \<and> j \<le> length(args t) \<and> g(get_arg s i,get_arg t j))"
  let ?check2'  = "?check2'' (\<lambda>r. isOK(check_supt r)) (\<lambda>r. isOK(check_supteq r))"
  let ?check2    = "?check2'' (\<lambda>r. r \<in> supt) (\<lambda>r. r \<in> supteq)"
  from check have "\<forall>((s,t),(stri,nstri)) \<in> set Gs. ?check2' s t stri nstri"
    unfolding rm by (force simp:Let_def check_sct_subterm_proc_def)
  hence check2: "\<forall>((s,t),(stri,nstri)) \<in> set Gs. ?check2 s t stri nstri"
    using check_supt_sound check_supteq_sound by fast
  let ?tuple = "\<lambda>s t. \<not> is_var s \<and> \<not> is_var t \<and> the(root t) \<notin> defs(set R)"
  let ?graph = "\<lambda>s t.\<exists>stri nstri. (
    Scg (?entry s t) (?entry s t) stri nstri \<in> set ?Gs \<and>
    (\<forall>(i,j)\<in>set stri. i \<le> length(args s) \<and> j \<le> length(args t) \<and> (get_arg s i,get_arg t j) \<in> supt) \<and>
    (\<forall>(i,j)\<in>set nstri. i \<le> length(args s) \<and> j \<le> length(args t) \<and> (get_arg s i,get_arg t j) \<in> supteq)
  )"
  {
    fix s t
    assume "(s,t) \<in> set P"
    with check1 obtain stri nstri where mem: "((s,t),(stri,nstri)) \<in> set Gs" by auto
    hence memGs: "Scg (?entry s t) (?entry s t) stri nstri \<in> set ?Gs" by force
    from mem check2 have ch: "?check2 s t stri nstri" by auto
    from rm ch ruleMap_empty_eq_nondef have tuple: "?tuple s t" by force
    from ch memGs have "?graph s t" unfolding split_def by auto
    with tuple have "?tuple s t \<and> ?graph s t" by auto
  }
  hence tuple_graph: "\<forall>(s,t)\<in>set P. ?tuple s t \<and> ?graph s t" by auto
  from tuple_graph have tuple: "\<forall>(s,t)\<in>set P. ?tuple s t" by auto
  from tuple_graph have graph: "\<forall>(s,t)\<in>set P. ?graph s t" by auto
  have False by (rule sct_with_subterm[OF chain tuple graph], rule conjE[OF check1], simp)
  thus "\<exists>i. ichain (set rP,set R) (shift s i) (shift t i) (shift \<sigma> i)" by simp
qed

definition check_sct_proc :: "
  (('f::shows,'v::shows)rule \<Rightarrow> shows monad) \<Rightarrow>
  (('f,'v)rule \<Rightarrow> shows monad) \<Rightarrow>
   shows \<Rightarrow>
   ('f,'v)trsL \<Rightarrow> (('f,'v)rule \<times> ((nat \<times> nat)list \<times> (nat \<times> nat)list))list \<Rightarrow> ('f,'v)trsL \<Rightarrow>
  ('f \<Rightarrow> ('f,'v)trsL)option \<Rightarrow>
  ('f \<Rightarrow> ('f,'v)trsL)option \<Rightarrow> shows monad"
where "check_sct_proc checkSS checkNSS representation P Gs R rm rrm = (
    case rm of None \<Rightarrow> error(''TRS has variables as left-hand sides'' +#+ shows_nl)
      | Some RM \<Rightarrow> (
        checkSubset P (map fst Gs)
           \<leftarrow> (\<lambda>s. ''there is no size-change graph for DP '' +#+ shows_rule s)
      \<guillemotright> checkAllM (\<lambda> ((s,t),(stri,nstri)). ((
              checkB (\<not> is_var s \<and> \<not> is_var t \<and> RM (the (root t)) = []) (''roots of DP must be constructors'' +#+ shows_nl)
         \<guillemotright> (let n = length (args s) in let m = length (args t) in (
                checkAllM (\<lambda> (i,j). checkB (i \<le> n \<and> j \<le> m \<and> isOK(checkSS (get_arg s i, get_arg t j))) (''problem with edge '' +#+ shows i +@+ '' -S-> '' +#+ shows j)) stri
              \<guillemotright> checkAllM (\<lambda> (i,j). checkB (i \<le> n \<and> j \<le> m \<and> isOK(checkNSS (get_arg s i, get_arg t j))) (''problem with edge '' +#+ shows i +@+ '' -NS-> '' +#+ shows j)) nstri
              ))
           )) \<leftarrow> (\<lambda>m. ''problems with DP '' +#+ shows_rule(s,t) +@+ shows_nl +@+ m)
        ) Gs
      \<guillemotright> checkAllM checkNSS R
              \<leftarrow> (\<lambda>s. ''problem when orienting rules'' +#+ shows_nl +@+ s)
      \<guillemotright> checkB (check_SCT (\<lambda>(s,t,cs,ct) (u,v,cu,cv). gcMatchI ct u \<and> gcMatchI cu t) (map (\<lambda>((s,t),(stri,nstri)). let e = (s,t,capRM rrm s,capRM rm t) in Scg e e stri nstri) Gs))
           (''size-change analysis failed'' +#+ shows_nl)
      )
  \<leftarrow> (\<lambda>s. ''could not apply the size-change processor with the following '' +#+ shows_nl +@+
         representation +@+ shows_nl +@+ ''for the following reason'' +#+ shows_nl +@+ s)
)"

context redpair_impl
begin

lemma check_sct_proc_sound_1:
   assumes chain: "ichain (set P,set R) s t \<sigma>"
   and check: "isOK(check_sct_proc checkS checkNS repr P Gs R (computeRuleMap R) (computeRuleMap (reverse_trs R)))"
   shows False
proof -
  from check obtain rm where rm: "computeRuleMap R = Some rm" by (cases "computeRuleMap R", auto simp: check_sct_proc_def)
  let ?cap = "capRM (computeRuleMap R)"
  let ?capRM = "capRM (computeRuleMap (reverse_trs R))"
  let ?entry = "\<lambda> s t. (s,t,?capRM s, ?cap t)"
  let ?Gs = "(map (\<lambda> ((s,t),(stri,nstri)). Scg (?entry s t) (?entry s t) stri nstri) Gs)"
  let ?conn = "\<lambda>(s,t,cs,ct) (u,v,cu,cv). gcMatch ct u \<and> gcMatch cu t"
  from check rm have check1: "set P \<subseteq> fst ` set Gs \<and>
     set R \<subseteq> NS \<and>
     check_SCT ?conn ?Gs"
    by (auto simp: checkNSSound Let_def check_sct_proc_def)
  let ?check2 = "\<lambda> s t stri nstri. \<not> is_var s \<and> \<not> is_var t \<and> rm (the (root t)) = [] \<and> 
    (\<forall> (i,j) \<in> set stri. i \<le> length (args s) \<and> j \<le> length (args t) \<and> (get_arg s i, get_arg t j) \<in> S) \<and> 
    (\<forall> (i,j) \<in> set nstri. i \<le> length (args s) \<and> j \<le> length (args t) \<and> (get_arg s i, get_arg t j) \<in> NS)"
  from check rm have check2: "\<forall> ((s,t),(stri,nstri)) \<in> set Gs. ?check2 s t stri nstri"
    by (force simp: Let_def checkNSSound checkSSound check_sct_proc_def)
  from rm have no_left_vars: "\<forall> lr \<in> set R. \<forall> x. fst lr \<noteq> Var x" using ruleMap_not_None_eq_no_vars_left by force
  let ?tuple = "\<lambda> s t. \<not> is_var s \<and> \<not> is_var t \<and> the (root t) \<notin> defs (set R)"
  let ?graph = "\<lambda> s t. \<exists> stri non_stri. (
         Scg (?entry s t) (?entry s t) stri non_stri \<in> set ?Gs \<and>
         (\<forall> (i,j) \<in> set stri. i \<le> length (args s) \<and> j \<le> length (args t) \<and> (get_arg s i, get_arg t j) \<in> S) \<and> 
         (\<forall> (i,j) \<in> set non_stri. i \<le> length (args s) \<and> j \<le> length (args t) \<and> (get_arg s i, get_arg t j) \<in> NS))"
  {
    fix s t
    assume "(s,t) \<in> set P"
    with check1 obtain stri nstri where mem: "((s,t),(stri,nstri)) \<in> set Gs" by auto
    hence memGs: "Scg (?entry s t) (?entry s t) stri nstri \<in> set ?Gs" by force
    from mem check2 have ch: "?check2 s t stri nstri" by auto
    from rm ch ruleMap_empty_eq_nondef have tuple: "?tuple s t" by force
    from ch memGs have "?graph s t" by blast
    with tuple have "?tuple s t \<and> ?graph s t" by auto
  }
  hence  tuple_graph:  "\<forall> (s,t) \<in> set P. ?tuple s t \<and> ?graph s t" by auto
  from tuple_graph have tuple: "\<forall> (s,t) \<in> set P. ?tuple s t" by blast
  from tuple_graph have graph: "\<forall> (s,t) \<in> set P. ?graph s t" by blast
  show False
    by (rule sct_with_redpair[OF chain no_left_vars tuple _ graph], simp add: check1, rule conjE[OF check1], simp)
qed

lemma check_sct_proc_sound:
  "chain_identifying_onlyDP_proc_impl (\<lambda> P R rm rrm rP. isOK(check_sct_proc checkS checkNS repr P Gs R rm rrm))"
unfolding chain_identifying_onlyDP_proc_impl_def
proof (intro allI, intro impI, intro allI, intro impI)
  fix P R rP s t \<sigma>
  assume check: "isOK(check_sct_proc checkS checkNS repr P Gs R (computeRuleMap R) (computeRuleMap (reverse_trs R)))"
  and chain: "min_ichain (set P, set R) s t \<sigma>"
  hence ichain: "ichain (set P, set R) s t \<sigma>" by auto
  have False by (rule check_sct_proc_sound_1, rule ichain, rule check)
  thus "\<exists> i. ichain (set rP, set R) (shift s i) (shift t i) (shift \<sigma> i)" ..
qed
end

definition check_sct_ur_af_proc :: "
  (('f::shows,'v::shows)rule \<Rightarrow> shows monad) \<Rightarrow>
  (('f,'v)rule \<Rightarrow> shows monad) \<Rightarrow>
   shows \<Rightarrow>
   ('f,'v)trsL \<Rightarrow> (('f,'v)rule \<times> ((nat \<times> nat)list \<times> (nat \<times> nat)list))list  \<Rightarrow> ('f \<Rightarrow> ('f,'v)trsL)option \<Rightarrow>
  ('f \<Rightarrow> ('f,'v)trsL)option \<Rightarrow>
  ('f,'v)trsL \<Rightarrow> 'f af \<Rightarrow>
  shows monad"
where "check_sct_ur_af_proc checkSS checkNSS representation P Gs rm rrm ur \<pi> = (
    case rm of None \<Rightarrow> error(''TRS has variables as left-hand sides'' +#+ shows_nl)
      | Some RM \<Rightarrow> (let Ggs = filter (\<lambda> g. fst g mem P) Gs in (
        checkAllM (\<lambda> lr. check_ur_closed_term_rm_af RM ur \<pi> (snd lr)) ur
           \<leftarrow> (\<lambda>s. ''error when checking closure properties of rhs of usable rules'' +#+ shows_nl +@+ s)
      \<guillemotright> checkSubset P (map fst Ggs)
           \<leftarrow> (\<lambda>s. ''there is no size-change graph for DP '' +#+ shows_rule s)
      \<guillemotright> checkAllM (\<lambda>((s,t),(stri,nstri)). ((
              checkB (\<not> is_var s \<and> \<not> is_var t \<and> RM (the (root t)) = []) (''roots of DP must be constructors'' +#+ shows_nl)
         \<guillemotright> checkAllM (\<lambda> i. checkB ( i \<le> length (args s)) (''left-index to large'' +#+ shows i +@+ shows_nl)) (remdups (map fst (List.append stri nstri)))
         \<guillemotright> checkAllM (\<lambda> j. checkB ( j \<le> length (args t) \<and> isOK(check_ur_closed_term_rm_af RM ur \<pi> (get_arg t j))) (''right-index to large or argument violates usable-rules condition'' +#+ shows j)) (remdups (map snd (List.append stri nstri)))
         \<guillemotright> (let ss = args s in let ts = args t in  (
                checkAllM (\<lambda> (i,j). checkB (isOK(checkSS (get_arg s i, get_arg t j))) (''problem with edge '' +#+ shows i +@+ '' -S-> '' +#+ shows j)) stri
              \<guillemotright> checkAllM (\<lambda> (i,j). checkB (isOK(checkNSS (get_arg s i, get_arg t j))) (''problem with edge '' +#+ shows i +@+ '' -NS-> '' +#+ shows j)) nstri
              ))
           )) \<leftarrow> (\<lambda> m. ''problems with DP '' +#+ shows_rule(s,t) +@+ shows_nl +@+ m) ) Ggs
      \<guillemotright> checkAllM checkNSS ur
              \<leftarrow> (\<lambda>s. ''problem when orienting usable rules'' +#+ shows_nl +@+ s)
      \<guillemotright> checkB (check_SCT (\<lambda>(s,t,cs,ct) (u,v,cu,cv). gcMatchI ct u \<and> gcMatchI cu t) (map (\<lambda>((s,t),(stri,nstri)). let e = (s,t,capRM rrm s,capRM rm t) in Scg e e stri nstri) Ggs))
          (''size-change analysis failed'' +#+ shows_nl)
      ))
  \<leftarrow> (\<lambda>s. ''could not apply the size-change processor with the following '' +#+ shows_nl +@+
          representation +@+ shows_nl +@+ ''for the following reason'' +#+ shows_nl +@+ s)
)"

context ce_af_redpair_impl
begin
lemma check_sct_ur_af_proc_sound_1:
   assumes chain: "min_ichain (set P,set R) s t \<sigma>"
   and check: "isOK(check_sct_ur_af_proc checkS checkNS repr P Gs (computeRuleMap R) (computeRuleMap (reverse_trs R)) ur \<pi>)"
   shows False
proof -
  from check obtain rm where rm: "computeRuleMap R = Some rm" by (cases "computeRuleMap R", auto simp: check_sct_ur_af_proc_def)
  let ?cap = "capRM (computeRuleMap R)"
  let ?capRM = "capRM (computeRuleMap (reverse_trs R))"
  let ?entry = "\<lambda> s t. (s,t,?capRM s, ?cap t)"
  let ?us = "(\<Union> ( (funas_term o snd) ` (set ur \<union> set P)))"
  let ?GGs = "filter (\<lambda> g. fst g \<in> set P) Gs"
  let ?Gs = "(map (\<lambda> ((s,t),(stri,nstri)). Scg (?entry s t) (?entry s t) stri nstri) ?GGs)"
  let ?conn = "\<lambda>(s,t,cs,ct) (u,v,cu,cv). gcMatch ct u \<and> gcMatch cu t"
  {
    fix l r
    assume lr: "(l,r) \<in> set ur"
    with check rm have ch: "isOK(check_ur_closed_term_rm_af rm ur \<pi> r)" by (auto simp: check_sct_ur_af_proc_def Let_def)
    from lr have rus: "funas_term r \<subseteq> ?us" by force
    from check_ur_closed_term_rm_af_sound[OF ch rus rm]  lr
    have "ur_closed_term_af r R (set ur) ?us \<pi>" by force
  } 
  with check rm have check1: "set P \<subseteq> fst ` set ?GGs \<and>
     set ur \<subseteq> NS \<and>
     ur_closed_af (set ur) ?us R \<pi> \<and> 
     check_SCT ?conn ?Gs"
    by (auto simp: mem_iff checkNSSound Let_def check_sct_ur_af_proc_def)
  let ?check2a = "\<lambda> s t stri nstri. \<not> is_var s \<and> \<not> is_var t \<and> rm (the (root t)) = [] \<and> 
    (\<forall> i \<in> fst ` set (stri @ nstri).  i \<le> length (args s)) \<and>
    (\<forall> j \<in> snd ` set (stri @ nstri).  j \<le> length (args t) \<and> isOK(check_ur_closed_term_rm_af rm ur \<pi> (get_arg t j))) \<and> 
    (\<forall> (i,j) \<in> set stri. (get_arg s i, get_arg t j) \<in> S) \<and>
    (\<forall> (i,j) \<in> set nstri. (get_arg s i, get_arg t j) \<in> NS)"
  let ?check2 = "\<lambda> s t stri nstri. \<not> is_var s \<and> \<not> is_var t \<and> rm (the (root t)) = [] \<and> 
    (\<forall> i \<in> fst ` set (stri @ nstri).  i \<le> length (args s)) \<and>
    (\<forall> j \<in> snd ` set (stri @ nstri).  j \<le> length (args t) \<and> ur_closed_term_af (get_arg t j) R (set ur) ?us \<pi>) \<and> 
    (\<forall> (i,j) \<in> set stri. (get_arg s i, get_arg t j) \<in> S) \<and>
    (\<forall> (i,j) \<in> set nstri. (get_arg s i, get_arg t j) \<in> NS)"
  from check rm have check2a: "\<forall> ((s,t),(stri,nstri)) \<in> set ?GGs. ?check2a s t stri nstri"
    by (force simp: mem_iff Let_def checkNSSound checkSSound check_sct_ur_af_proc_def simp del: set_append map_append)
  {
    fix s t stri nstri
    assume g: "((s,t),(stri,nstri)) \<in> set ?GGs" (is "?g \<in> _")
    from bspec[OF check2a g] have t: "\<not> is_var t" and other: "\<forall> j \<in> snd ` set (stri @ nstri). j \<le> length (args t) \<and> isOK(check_ur_closed_term_rm_af rm ur \<pi> (get_arg t j))"
      by auto
    from g have st: "(s,t) \<in> set P" by auto
    hence tus: "funas_term t \<subseteq> ?us" by auto
    have "(\<forall> j \<in> snd ` set (stri @ nstri).  j \<le> length (args t) \<and> ur_closed_term_af (get_arg t j) R (set ur) ?us \<pi>)"
    proof
      fix j
      assume j: "j \<in> snd ` set (stri @ nstri)"
      with bspec[OF check2a g]
      have jl: "j \<le> length (args t)" and ch: "isOK(check_ur_closed_term_rm_af rm ur \<pi> (get_arg t j))" by auto
      have "funas_term (get_arg t j) \<subseteq> funas_term t" 
      proof (cases j, simp)
        case (Suc i)
        from t obtain f ts where t: "t = Fun f ts" by (cases t, auto)
        with jl Suc have "i < length ts" by auto
        with t Suc have arg: "get_arg t j \<in> set ts" by auto
        with t show "funas_term (get_arg t j) \<subseteq> funas_term t" by auto
      qed
      with tus have tjus: "funas_term (get_arg t j) \<subseteq> ?us" by auto
      from check_ur_closed_term_rm_af_sound[OF ch tjus rm] jl
      show "j \<le> length (args t) \<and> ur_closed_term_af (get_arg t j) R (set ur) ?us \<pi>" by simp
    qed
  }
  with check2a have check2: "\<forall> ((s,t),(stri,nstri)) \<in> set ?GGs. ?check2 s t stri nstri" by best
  let ?tuple = "\<lambda> s t. \<not> is_var s \<and> \<not> is_var t \<and> the (root t) \<notin> defs (set R)"
  let ?pgraph = "\<lambda> s t stri non_stri. (
         Scg (?entry s t) (?entry s t) stri non_stri \<in> set ?Gs \<and>
         (\<forall> (i,j) \<in> set stri. i \<le> length (args s) \<and> j \<le> length (args t) \<and> ur_closed_term_af (get_arg t j) R (set ur) ?us \<pi> \<and> (get_arg s i, get_arg t j) \<in> S) \<and> 
         (\<forall> (i,j) \<in> set non_stri. i \<le> length (args s) \<and> j \<le> length (args t) \<and> ur_closed_term_af (get_arg t j) R (set ur) ?us \<pi> \<and> (get_arg s i, get_arg t j) \<in> NS))"
  let ?graph = "\<lambda> s t. \<exists> stri non_stri. ?pgraph s t stri non_stri"
  {
    fix s t
    assume "(s,t) \<in> set P"
    with check1 obtain stri nstri where mem: "((s,t),(stri,nstri)) \<in> set ?GGs" by auto
    hence memGs: "Scg (?entry s t) (?entry s t) stri nstri \<in> set ?Gs" by force
    from mem check2 have ch: "?check2 s t stri nstri" by (blast del: not_is_var_imp_Fun)
    from ch have stri_ur: "\<forall> (i,j) \<in> set stri. i \<le> length (args s) \<and> j \<le> length (args t) \<and> ur_closed_term_af (get_arg t j) R (set ur) ?us \<pi>" by (force del: not_is_var_imp_Fun)
    from ch have nstri_ur: "\<forall> (i,j) \<in> set nstri. i \<le> length (args s) \<and> j \<le> length (args t) \<and> ur_closed_term_af (get_arg t j) R (set ur) ?us \<pi>" by (force del: not_is_var_imp_Fun)
    from rm ch ruleMap_empty_eq_nondef have tuple: "?tuple s t" by force
    from ch memGs stri_ur nstri_ur have "?pgraph s t stri nstri" by auto
    hence "?graph s t" by blast
    with tuple have "?tuple s t \<and> ?graph s t" by auto
  }
  hence  tuple_graph:  "\<forall> (s,t) \<in> set P. ?tuple s t \<and> ?graph s t" by auto
  from tuple_graph have tuple: "\<forall> (s,t) \<in> set P. ?tuple s t" by blast
  from tuple_graph have graph: "\<forall> (s,t) \<in> set P. ?graph s t" by blast
  show False
    by (rule sct_ur_af_redpair[OF chain tuple _ _ graph], (simp add: check1)+, (rule conjE[OF check1], auto)+)
qed

lemma check_sct_ur_af_proc_sound:
  "chain_identifying_onlyDP_proc_impl (\<lambda> P R rm rrm rP. isOK(check_sct_ur_af_proc checkS checkNS repr P Gs rm rrm ur \<pi>))"
unfolding chain_identifying_onlyDP_proc_impl_def
proof (intro allI, intro impI, intro allI, intro impI)
  fix P R rP s t \<sigma>
  assume check: "isOK(check_sct_ur_af_proc checkS checkNS repr P Gs (computeRuleMap R) (computeRuleMap (reverse_trs R)) ur \<pi>)"
  and chain: "min_ichain (set P, set R) s t \<sigma>"
  have False by (rule check_sct_ur_af_proc_sound_1, rule chain, rule check)
  thus "\<exists> i. ichain (set rP, set R) (shift s i) (shift t i) (shift \<sigma> i)" ..
qed
end

end
