(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Show imports Main Util begin

types "shows" = "string \<Rightarrow> string"

definition shows_string :: "string \<Rightarrow> shows" where "shows_string \<equiv> op @"

class "shows" =
  fixes "shows" :: "'a \<Rightarrow> shows"
    and shows_list :: "'a list \<Rightarrow> shows"
  assumes assoc[simp]: "shows x r @ s = shows x (r@s)"
    and assoc_list[simp]: "shows_list xs r @ s = shows_list xs (r@s)"

definition "show" :: "'a::shows \<Rightarrow> string" where "show x \<equiv> shows x ''''"

abbreviation shows_cons :: "string \<Rightarrow> shows \<Rightarrow> shows" (infixr "+#+" 10)
where "s +#+ p \<equiv> shows_string s \<circ> p"
abbreviation shows_append :: "shows \<Rightarrow> shows \<Rightarrow> shows" (infixr "+@+" 10)
where "s +@+ p \<equiv> s \<circ> p"

definition newline :: char where "newline \<equiv> Char Nibble0 NibbleA"
definition shows_between :: "shows \<Rightarrow> shows \<Rightarrow> shows \<Rightarrow> shows"
where "shows_between l p r \<equiv> l +@+ p +@+ r"

(*
fun showl :: "('a \<Rightarrow> shows) \<Rightarrow> string \<Rightarrow> 'a list \<Rightarrow> string"
where "showl showx s [] = CHR '']'' # s"
    | "showl showx s (x#xs) = CHR '','' # showx x (showl showx s xs)"

fun shows_list_aux :: "('a \<Rightarrow> shows) \<Rightarrow> 'a list \<Rightarrow> shows"
where "shows_list_aux _ [] s = ''[]'' @ s"
    | "shows_list_aux showx (x#xs) s = (CHR ''['') # showx x (showl showx s xs)"

hide const showl
*)

(*
definition unquote :: "shows" where "unquote x \<equiv> filter (\<lambda> c. \<not> c = CHR ''\"'') x"
definition showF :: "'a::show \<Rightarrow> string"
  where "showF x \<equiv> unquote(show x)"

abbreviation sAppend :: "sstring \<Rightarrow> sstring \<Rightarrow> sstring" (infixl "+++" 10)
  where "sAppend s t \<equiv> (s @ t)"

abbreviation sConcat :: "string list \<Rightarrow> sstring" where "sConcat \<equiv> concat"

definition sEmpty :: "sstring" where "sEmpty = []"
*)

fun shows_sep :: "('a \<Rightarrow> shows) \<Rightarrow> shows \<Rightarrow> 'a list \<Rightarrow> shows"
where "shows_sep s sep [] = shows_string ''''"
    | "shows_sep s sep [x] = s x"
    | "shows_sep s sep (x#xs) = s x \<circ> sep \<circ> shows_sep s sep xs"

lemma shows_sep_assoc[simp]:
  assumes elt: "\<And>r s. \<forall>x\<in>set xs. elt x r @ s = elt x (r@s)"
    and "\<And>r s. sep r @ s = sep(r@s)"
  shows "shows_sep elt sep xs r @ s = shows_sep elt sep xs (r@s)"
using elt proof (induct xs)
  case Nil show ?case by (simp add: shows_string_def)
next
  case (Cons x xs) note c = this thus ?case proof (cases xs)
    case Nil show ?thesis by (simp add: assms c Nil)
  next
    case (Cons y ys) show ?thesis
      using c unfolding Cons by (simp add: assms)
  qed
qed

lemma shows_string_assoc[simp]: "shows_string x r @ s = shows_string x (r@s)"
unfolding shows_string_def by simp

lemma shows_between_assoc[simp]:
  assumes l: "\<And>s t. l s @ t = l(s@t)"
    and m: "\<And>s t. m s @ t = m(s@t)"
    and r: "\<And>s t. r s @ t = r(s@t)"
  shows "shows_between l m r s @ t = shows_between l m r (s@t)"
unfolding shows_between_def o_def l r[symmetric] m by simp

definition shows_list_gen
  :: "('a \<Rightarrow> shows) \<Rightarrow> string \<Rightarrow> string \<Rightarrow> string \<Rightarrow> string \<Rightarrow> 'a list \<Rightarrow> shows"
where "shows_list_gen elt e l s r xs \<equiv> (if xs = [] then (
  shows_string e
) else (
  shows_between (shows_string l) (shows_sep elt (shows_string s) xs) (shows_string r)
))"

lemma shows_list_gen_assoc[simp]:
  assumes "\<And>r s. \<forall>x \<in> set xs. elt x r @ s = elt x (r@s)"
  shows "shows_list_gen elt e l sep r xs s @ t = shows_list_gen elt e l sep r xs (s@t)"
proof (cases "xs")
  case Nil show ?thesis unfolding shows_list_gen_def if_P[OF Nil] shows_string_def by simp
next
  case (Cons y ys)
  hence ne: "\<not> xs = []" by simp
  have "\<And>s t. shows_string l s @ t = shows_string l (s@t)" by (rule shows_string_assoc)
  moreover have "\<And>s t. shows_sep elt (shows_string sep) xs s @ t =
    shows_sep elt (shows_string sep) xs (s@t)"
  proof -
    fix s t
    have "\<And>s t. shows_string sep s @ t = shows_string sep (s@t)" by (rule shows_string_assoc)
    thus "?thesis s t" by (rule shows_sep_assoc[OF assms])
  qed
  moreover have "\<And>s t. shows_string r s @ t = shows_string r (s@t)"
    by (rule shows_string_assoc)
  ultimately show ?thesis unfolding shows_list_gen_def if_not_P[OF ne]
    by (rule shows_between_assoc)
qed

definition shows_list_aux :: "('a \<Rightarrow> shows) \<Rightarrow> 'a list \<Rightarrow> shows"
where "shows_list_aux s xs \<equiv> shows_list_gen s ''[]'' ''['' '','' '']'' xs"

lemma (in "shows") assoc_elt: "\<forall>x\<in>set xs. shows x r @ s = shows x (r @ s)"
using assoc[of _ r s] by simp

lemma shows_list_aux_assoc[simp]:
  assumes "\<And>r s. \<forall>x\<in>set xs. elt x r @ s = elt x (r@s)"
  shows "shows_list_aux elt xs r @ s = shows_list_aux elt xs (r@s)"
unfolding shows_list_aux_def by (rule shows_list_gen_assoc[OF assms])

instantiation unit :: "shows" begin
  definition "shows(x::unit) \<equiv> shows_string ''()''"
  definition "shows_list(xs::unit list) \<equiv> shows_list_aux shows xs"

  lemma assoc_unit: "shows (x::unit) r @ s = shows x (r@s)"
    by (unfold shows_unit_def shows_string_def,simp)
  lemma assoc_unit_elt: "\<forall>(x::unit)\<in>set xs. shows x r @ s = shows x (r@s)"
    using assoc_unit by auto
  instance by
    (intro_classes,rule assoc_unit,unfold shows_list_unit_def)
    (rule shows_list_aux_assoc[OF assoc_unit_elt])
end

instantiation bool :: "shows" begin
  definition "shows(b::bool) \<equiv> (if b then shows_string ''True'' else shows_string ''False'')"
  definition "shows_list(bs::bool list) \<equiv> shows_list_aux shows bs"

  lemma assoc_bool: "shows (b::bool) r @ s = shows b (r@s)"
    by (unfold shows_bool_def shows_string_def,simp)
  lemma assoc_bool_elt: "\<forall>(b::bool)\<in>set bs. shows b r @ s = shows b (r@s)"
    using assoc_bool by auto
  instance by 
    (intro_classes,rule assoc_bool,unfold shows_list_bool_def)
    (rule shows_list_aux_assoc[OF assoc_bool_elt])
end

instantiation char :: "shows" begin
  definition "shows(c::char) \<equiv> op # c"
  definition "shows_list(cs::string) \<equiv> shows_string cs"
  instance
    by (intro_classes,unfold shows_char_def shows_list_char_def shows_string_def,simp) auto
end

definition shows_nl :: "shows" where "shows_nl \<equiv> shows newline"
definition shows_space :: "shows" where "shows_space \<equiv> shows(CHR '' '')"

definition shows_paren :: "shows \<Rightarrow> shows"
where "shows_paren p \<equiv> shows(CHR ''('') +@+ p +@+ shows(CHR '')'')"

lemmas show_defs =
  shows_char_def shows_nl_def shows_space_def shows_string_def shows_between_def shows_paren_def

instantiation * :: ("shows","shows")"shows" begin
  definition "shows(p::('a::shows\<times>'b::shows)) \<equiv>
    shows_paren (shows(fst p) +@+ shows(CHR '','') +@+ shows(snd p))"
  definition "shows_list(ps::('a::shows\<times>'b::shows)list) \<equiv> shows_list_aux shows ps"

  lemma assoc_prod: "shows (p::('a::shows\<times>'b::shows)) r @ s = shows p (r@s)"
  proof (cases p)
    case (Pair a b) show ?thesis unfolding Pair shows_prod_def show_defs o_def by simp
  qed
  lemma assoc_prod_elt: "\<forall>(p::('a::shows\<times>'b::shows))\<in>set ps. shows p r @ s = shows p (r@s)"
    using assoc_prod by auto
  instance by
    (intro_classes,rule assoc_prod,unfold shows_list_prod_def)
    (rule shows_list_aux_assoc[OF assoc_prod_elt])
end

instantiation list :: ("shows")"shows" begin
  definition "shows(l::'a::shows list) \<equiv> shows_list l"
  definition "shows_list(ls::'a::shows list list) \<equiv> shows_list_aux shows ls"

  lemma assoc_list_elt: "\<forall>(l::('a::shows list))\<in>set ls. shows l r @ s = shows l (r@s)"
    unfolding shows_list_def by simp
  instance by
    (intro_classes,unfold shows_list_def,simp,unfold shows_list_list_def)
    (rule shows_list_aux_assoc[OF assoc_list_elt])
end

definition shows_lines :: "'a::shows list \<Rightarrow> shows"
where "shows_lines \<equiv> shows_sep shows shows_nl"
definition shows_many :: "'a::shows list \<Rightarrow> shows"
where "shows_many \<equiv> shows_sep shows id"
definition shows_words :: "'a::shows list \<Rightarrow> shows"
where "shows_words \<equiv> shows_sep shows shows_space"

fun shows_map :: "('a \<Rightarrow> shows) \<Rightarrow> 'a list \<Rightarrow> shows"
where "shows_map s [] = id"
    | "shows_map s (x#xs) = (s x +@+ shows_map s xs)"

lemma shows_nl_assoc[simp]: "shows_nl r @ s = shows_nl(r@s)" unfolding show_defs by simp
lemma shows_id_assoc[simp]: "id r @ s = id(r@s)" by simp
lemma shows_space_assoc[simp]: "shows_space r @ s = shows_space(r@s)"
  unfolding show_defs by simp

lemma shows_lines_assoc[simp]: "shows_lines xs r @ s = shows_lines xs (r@s)"
unfolding shows_lines_def by (rule shows_sep_assoc[OF assoc_elt shows_nl_assoc])

lemma shows_many_assoc[simp]: "shows_many xs r @ s = shows_many xs (r@s)"
unfolding shows_many_def by (rule shows_sep_assoc[OF assoc_elt shows_id_assoc])

lemma shows_words_assoc[simp]: "shows_words xs r @ s = shows_words xs (r@s)"
unfolding shows_words_def by (rule shows_sep_assoc[OF assoc_elt shows_space_assoc])

lemma shows_map_assoc[simp]:
  assumes "\<And>r s.\<forall>x\<in>set xs. elt x r @ s = elt x (r@s)"
  shows "shows_map elt xs r @ s = shows_map elt xs (r@s)"
using assms by (induct xs arbitrary: r s) auto

fun shows_concat :: "shows list \<Rightarrow> shows"
where "shows_concat [] = id"
    | "shows_concat(s#ss) = (s +@+ shows_concat ss)"

lemma shows_map_cong[fundef_cong,recdef_cong]:
  assumes "xs = ys" and "\<And>x. x \<in> set ys \<Longrightarrow> f x = g x" shows "shows_map f xs = shows_map g ys"
proof
  fix s show "shows_map f xs s = shows_map g ys s" using assms
  proof (induct xs arbitrary: ys)
    case Nil thus ?case by (induct ys) auto
  next
    case (Cons x xs) thus ?case by (induct ys) auto
  qed
qed

lemma shows_sep_cong[fundef_cong,recdef_cong]:
  assumes "xs = ys" and "\<And>x. x \<in> set ys \<Longrightarrow> f x = g x" shows "shows_sep f sep xs = shows_sep g sep ys"
proof
  fix s show "shows_sep f sep xs s = shows_sep g sep ys s" using assms
  proof (induct xs arbitrary: ys)
    case Nil thus ?case by (induct ys) auto
  next
    case (Cons x xs) note c = this thus ?case
    proof (cases xs)
      case Nil show ?thesis using c unfolding Nil by (induct ys) auto
    next
      case (Cons x' xs') show ?thesis using c unfolding Cons by (induct ys) auto
    qed
  qed
qed

lemma shows_list_gen_cong[fundef_cong,recdef_cong]:
  assumes "xs = ys" and "\<And>x. x \<in> set ys \<Longrightarrow> f x = g x" shows "shows_list_gen f e l sep r xs = shows_list_gen g e l sep r ys"
using shows_sep_cong[of xs ys f g] assms unfolding shows_list_gen_def by (cases "xs = []") auto

(*
(* empty, left paren, separator, right paren, list *)
fun show_list_gen :: "string \<Rightarrow> string \<Rightarrow> string \<Rightarrow> string \<Rightarrow> string list \<Rightarrow> string"
where "show_list_gen e lp co rp [] = e"
 |  "show_list_gen e lp co rp (x # xs) = (lp +++ x +++ sConcat (map (\<lambda> y. co +++ y) xs) +++ rp)"

fun showList  :: "sstring list \<Rightarrow> string"
 where "showList xs = concat (map (\<lambda> x. x @ newline) xs)"
*)

instantiation nat :: "shows" begin
  fun digit2string :: "nat \<Rightarrow> string"
  where "digit2string n = 
    (if n = 0 then ''0'' else
    (if n = 1 then ''1'' else
    (if n = 2 then ''2'' else
    (if n = 3 then ''3'' else
    (if n = 4 then ''4'' else
    (if n = 5 then ''5'' else
    (if n = 6 then ''6'' else
    (if n = 7 then ''7'' else
    (if n = 8 then ''8'' else
    (if n = 9 then ''9'' else ''not a digit''))))))))))"
(*
  fun digit2string :: "nat \<Rightarrow> string"
  where "digit2string n = (case n of
    0 \<Rightarrow> ''0''
  | Suc 0 \<Rightarrow> ''1''
  | Suc(Suc 0) \<Rightarrow> ''2''
  | Suc(Suc(Suc 0)) \<Rightarrow> ''3''
  | Suc(Suc(Suc(Suc 0))) \<Rightarrow> ''4''
  | Suc(Suc(Suc(Suc(Suc 0)))) \<Rightarrow> ''5''
  | Suc(Suc(Suc(Suc(Suc(Suc 0))))) \<Rightarrow> ''6''
  | Suc(Suc(Suc(Suc(Suc(Suc(Suc 0)))))) \<Rightarrow> ''7''
  | Suc(Suc(Suc(Suc(Suc(Suc(Suc(Suc 0))))))) \<Rightarrow> ''8''
  | Suc(Suc(Suc(Suc(Suc(Suc(Suc(Suc(Suc 0)))))))) \<Rightarrow> ''9''
  | _ \<Rightarrow> ''not a digit''
  )"
*)

  fun shows_nat where "shows_nat(n::nat) = (
    if n < 10 then shows_string(digit2string n)
              else shows_nat(n div 10) \<circ> shows_string(digit2string(n mod 10))
  )"
  definition "shows_list(ns::nat list) \<equiv> shows_list_aux shows ns"

  lemma assoc_nat: "shows (n::nat) r @ s = shows n (r@s)"
  proof (induct n arbitrary: r s rule: nat_less_induct)
    case (1 n) show ?case
    proof (cases "n < 10")
      case True thus ?thesis by simp
    next
      let ?m = "n div 10"
      case False
      hence "?m < n" by simp
      with 1 have "\<And>s t. shows ?m s @ t = shows ?m (s@t)" by simp
      thus ?thesis by auto
    qed
  qed
  lemma assoc_nat_elt: "\<forall>(n::nat)\<in>set ns. shows n r @ s = shows n (r@s)"
    using assoc_nat by auto
  instance by
    (intro_classes,rule assoc_nat,unfold shows_list_nat_def)
    (rule shows_list_aux_assoc[OF assoc_nat_elt])
end

instantiation int :: "shows" begin
  definition "shows(i::int) \<equiv> (if i < 0 then shows(CHR ''-'') \<circ> shows(nat(0 - i))
                                        else shows(nat i))"
  definition "shows_list(is::int list) \<equiv> shows_list_aux shows is"

  lemma assoc_int: "shows (i::int) r @ s = shows i (r@s)"
  proof (cases "i < 0")
    case True thus ?thesis
      unfolding shows_int_def o_def if_P[OF True] assoc by simp
  next
    case False thus ?thesis unfolding shows_int_def o_def if_not_P[OF False] assoc by simp
  qed
  lemma assoc_int_elt: "\<forall>(i::int)\<in>set is. shows i s @ r = shows i (s@r)"
  using assoc_int by auto
  instance by
    (intro_classes,rule assoc_int,unfold shows_list_int_def)
    (rule shows_list_aux_assoc[OF assoc_int_elt])
end

instantiation rat :: "shows" begin
  definition show_rat :: "rat \<Rightarrow> string"
  where "show_rat x \<equiv> case (normalize x) of (Pair den num) \<Rightarrow> if num = 1 then shows den '''' else ((shows den \<circ> shows(CHR ''/'') \<circ> shows num) '''')"
  definition shows_rat :: "rat \<Rightarrow> shows"
  where "shows_rat x \<equiv> shows_string (show_rat x)"
  definition shows_list_rat where "shows_list_rat(rs::rat list) \<equiv> shows_list_aux shows rs"

  lemma assoc_rat: "shows (x::rat) r @ s = shows x (r@s)" unfolding shows_rat_def by  auto
  lemma assoc_rat_elt: "\<forall>(x::rat)\<in>set rs. shows x r @ s = shows x (r@s)"
    using assoc_rat by auto
  instance by
    (intro_classes,rule assoc_rat,unfold shows_list_rat_def)
    (rule shows_list_aux_assoc[OF assoc_rat_elt])
end


end
