(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
header {* Trs *}
theory Trs imports Term Infinite_Set begin

text {* A theory on first-order term rewrite systems (TRSs). *}

text {* A rewrite rule is a pair of terms. A TRS is a set of
rewrite rules. *}
types 
 ('f,'v)rule = "('f,'v)term \<times> ('f,'v)term"
 ('f,'v)trs  = "('f,'v)rule set"

subsection {* Variables of Rules *}

definition vars_rule :: "('f,'v)rule \<Rightarrow> 'v set"
where "vars_rule r = vars_term(fst r) \<union> vars_term(snd r)"

definition vars_rule_list' :: "('f,'v)rule \<Rightarrow> 'v list"
where "vars_rule_list' r \<equiv> vars_term_list'(fst r) @ vars_term_list'(snd r)"

lemma vars_rule_list'_sound[simp]:
  fixes r::"('f,'v)rule" shows "set(vars_rule_list' r) = vars_rule r"
  unfolding vars_rule_list'_def vars_rule_def by (cases r) simp

definition vars_rule_list :: "('f,'v)rule \<Rightarrow> 'v list"
where "vars_rule_list r \<equiv> remdups(vars_rule_list'(r::('f,'v)rule))"

lemma vars_rule_list_sound[simp]:
  fixes r::"('f,'v)rule" shows "set(vars_rule_list r) = vars_rule r"
  unfolding vars_rule_list_def by simp


subsection {* Variables of TRSs *}

definition vars_trs :: "('f,'v)trs \<Rightarrow> 'v set"
where "vars_trs(R) \<equiv> \<Union>r\<in>R. vars_rule r"

lemma finite_trs_has_finite_vars:
  fixes R::"('f,'v)trs"
  assumes "finite(R)" shows "finite(vars_trs R)"
using assms unfolding vars_trs_def vars_rule_def by simp

definition vars_trs_list :: "('f,'v)rule list \<Rightarrow> 'v list"
where "vars_trs_list trs \<equiv> remdups(flat_map vars_rule_list' trs)"

lemma vars_trs_list_Nil[simp]: "vars_trs_list [] = []" unfolding vars_trs_list_def by simp

lemma vars_trs_list_sound[simp]:
  fixes trs::"('f,'v)rule list" shows "set(vars_trs_list trs) = vars_trs(set trs)"
  by (induct trs) (auto simp: vars_trs_def vars_trs_list_def)

lemmas vars_defs = vars_trs_def vars_rule_def vars_trs_list_def vars_rule_list_def

subsection {* Function Symbols of Rules *}

definition funs_rule :: "('f,'v)rule \<Rightarrow> 'f set"
where "funs_rule r \<equiv> funs_term(fst r) \<union> funs_term(snd r)"

definition funs_rule_list' :: "('f,'v)rule \<Rightarrow> 'f list"
where "funs_rule_list' r \<equiv> funs_term_list'(fst r) @ funs_term_list'(snd r)"

lemma funs_rule_list'_sound[simp]:
  fixes r::"('f,'v)rule" shows "set(funs_rule_list' r) = funs_rule r"
  unfolding funs_rule_list'_def funs_rule_def by (cases r) simp

definition funs_rule_list :: "('f,'v)rule \<Rightarrow> 'f list"
where "funs_rule_list r \<equiv> remdups(funs_rule_list'(r::('f,'v)rule))"

lemma funs_rule_list_sound[simp]:
  fixes r::"('f,'v)rule" shows "set(funs_rule_list r) = funs_rule r"
  unfolding funs_rule_list_def by simp

text {* The same including arities. *}
definition funas_rule :: "('f,'v)rule \<Rightarrow> ('f\<times>nat)set"
where "funas_rule r \<equiv> funas_term(fst r) \<union> funas_term(snd r)"

definition funas_rule_list' :: "('f,'v)rule \<Rightarrow> ('f\<times>nat)list"
where "funas_rule_list' r \<equiv> funas_term_list'(fst r) @ funas_term_list'(snd r)"

lemma funas_rule_list'_sound[simp]:
  fixes r::"('f,'v)rule" shows "set(funas_rule_list' r) = funas_rule r"
  unfolding funas_rule_list'_def funas_rule_def by (cases r) simp

definition funas_rule_list :: "('f,'v)rule \<Rightarrow> ('f\<times>nat)list"
where "funas_rule_list r \<equiv> remdups(funas_rule_list'(r::('f,'v)rule))"

lemma funas_rule_list_sound[simp]:
  fixes r::"('f,'v)rule" shows "set(funas_rule_list r) = funas_rule r"
  unfolding funas_rule_list_def by simp

subsection {* Function Symbols of TRSs *}

definition funs_trs :: "('f,'v)trs \<Rightarrow> 'f set"
where "funs_trs(R) \<equiv> \<Union>r\<in>R. funs_rule r"

definition funs_trs_list :: "('f,'v)rule list \<Rightarrow> 'f list"
where "funs_trs_list trs \<equiv> remdups(flat_map funs_rule_list' trs)"

lemma funs_trs_list_sound[simp]:
  fixes trs::"('f,'v)rule list" shows "set(funs_trs_list trs) = funs_trs(set trs)"
  by (induct trs) (auto simp: funs_trs_def funs_trs_list_def)

definition funas_trs :: "('f,'v)trs \<Rightarrow> ('f\<times>nat)set"
where "funas_trs(R) \<equiv> \<Union>r\<in>R. funas_rule r"

lemma funas_empty[simp]: "funas_trs {} = {}" unfolding funas_trs_def by simp

definition funas_trs_list :: "('f,'v)rule list \<Rightarrow> ('f\<times>nat)list"
where "funas_trs_list trs \<equiv> remdups(flat_map funas_rule_list' trs)"

lemma funas_trs_list_sound[simp]:
  fixes trs::"('f,'v)rule list" shows "set(funas_trs_list trs) = funas_trs(set trs)"
  by (induct trs) (auto simp: funas_trs_def funas_trs_list_def)

lemmas funs_defs = funs_trs_def funs_trs_list_def funs_rule_def funs_rule_list_def
lemmas funas_defs = funas_trs_def funas_trs_list_def funas_rule_def funas_rule_list_def

text {* A function symbol is said to be \emph{defined} (w.r.t.\ to a given
TRS) if it occurs as root of some left-hand side. *}
definition defined :: "('f,'v)trs \<Rightarrow> 'f \<Rightarrow> bool" where
  "defined R f \<equiv> (\<exists>l r. (l,r) \<in> R \<and> root l = Some f)"

text {* The set of all defined symbols. *}
definition "defs" :: "('f,'v)trs \<Rightarrow> 'f set" where
  "defs R \<equiv> {f. defined R f}"

fun defs_list :: "('f,'v)rule list \<Rightarrow> 'f list"
where "defs_list [] = []"
    | "defs_list ((Fun f _,_) # ts) = (let fs = defs_list ts in if f mem fs then fs else f # fs)"
    | "defs_list ((Var _, _) # ts) = defs_list ts"

lemma defs_list_sound:
  fixes R::"('f,'v)rule list" 
  shows "defined (set R) f = f mem defs_list R"
proof (induct R, simp add: defined_def)
  case (Cons lr R)
  show ?case
  proof (cases lr)
    case (Pair l r)
    show ?thesis
    proof (cases l)
      case (Var x)
      with Pair Cons show ?thesis by (auto simp: defined_def mem_iff)
    next
      case (Fun g ss)
      with Pair
      have "defined (set (lr # R)) f = defined (set ((Fun g ss,r) # R)) f" by simp
      also have "\<dots> =  (\<exists>l ra. (l, ra) \<in> set ((Fun g ss, r) # R) \<and> root l = Some f)" (is "_ = ?P (set ((Fun g ss,r) # R))") by (simp add: defined_def)
      also have "\<dots> = (?P (set R) \<or> ?P {(Fun g ss, r)})" by auto
      also have "\<dots> = (f mem defs_list R \<or> ?P {(Fun g ss, r)})"  using Cons by (unfold defined_def, auto)
      also have "\<dots> = (f mem defs_list ((Fun g ss,r) # R))" (is "?l = ?r")
      proof (cases "g mem defs_list R")
	case False
	thus "?l = ?r" by (auto simp: Let_def)
      next
	case True
	thus "?l = ?r" by (simp add: Let_def, blast)
      qed		
      finally show ?thesis using Fun Pair Cons by simp
    qed
  qed
qed

text {* The remaining symbols are called \emph{constructor symbols}, since
they are only used to construct data. *}
definition cons :: "('f,'v)trs \<Rightarrow> 'f set" where "cons R \<equiv> funs_trs R - defs R"

text {* Closure under contexts of a binary relation on terms. *}
inductive_set
 closure_ctxt :: "('f,'v)trs \<Rightarrow> ('f,'v)trs" ("CC")
 for r :: "('f,'v)trs"
where
 ccI[intro]: "(s,t) \<in> r \<Longrightarrow> (C\<langle>s\<rangle>,C\<langle>t\<rangle>) \<in> CC r"

lemma clctxt_superset_rel: "r \<subseteq> CC r"
proof (rule subsetI2)
 fix s t assume "(s,t) \<in> r" hence "(\<box>\<langle>s\<rangle>,\<box>\<langle>t\<rangle>) \<in> CC r" using ccI[where C = "\<box>"] by simp
 thus "(s,t) \<in> CC r" by simp
qed

definition ctxt_closed :: "('f,'v)trs \<Rightarrow> bool"
where
 "ctxt_closed r = (CC r \<subseteq> r)"

lemma ctxt_closed_correctness: "ctxt_closed r \<Longrightarrow> r = CC r"
proof -
 assume "ctxt_closed r"
 hence a: "CC r \<subseteq> r" by (unfold ctxt_closed_def)
 from clctxt_superset_rel have b: "r \<subseteq> CC r" .
 from a b show "r = CC r" by simp
qed

text {* For well-formed terms, having the same root symbol implies having the same number of arguments. *}
lemma wft_num_args_eq: 
 assumes "wf_term \<Sigma> (Fun f ts)" and "wf_term \<Sigma> (Fun f ss)"
 shows "length ts = length ss"
proof -
 from `wf_term \<Sigma> (Fun f ss)` have ss: "length ss = arity \<Sigma> f" by (cases rule: wf_term.cases) auto
 from `wf_term \<Sigma> (Fun f ts)` have ts: "length ts = arity \<Sigma> f" by (cases rule: wf_term.cases) auto
 show ?thesis unfolding ss ts ..
qed

lemma ctxt_closed_imp_ctxt:
 fixes R :: "('f,'v)trs"
 assumes "ctxt_closed R" and "(s,t) \<in> R" shows "(C\<langle>s\<rangle>,C\<langle>t\<rangle>) \<in> R"
proof -
 from `ctxt_closed R` have "R = CC R" by (rule ctxt_closed_correctness)
 from `(s,t) \<in> R` have "(C\<langle>s\<rangle>,C\<langle>t\<rangle>) \<in> CC R" by (rule ccI)
 with `R = CC R` show ?thesis by simp
qed

lemma ctxt_closed_R_imp_suptp_R_distr:
 fixes R :: "('f,'v)trs"
 assumes "ctxt_closed R" and "s \<rhd> t" and "(t,u) \<in> R" shows "\<exists>t. (s,t) \<in> R \<and> t \<rhd> u"
proof -
 from `s \<rhd> t` obtain C where "C \<noteq> \<box>" and "C\<langle>t\<rangle> = s" using suptp_imp_nectxt by best 
 from `ctxt_closed R` and `(t,u) \<in> R` have RCtCu: "(C\<langle>t\<rangle>,C\<langle>u\<rangle>) \<in> R" by (rule ctxt_closed_imp_ctxt)
 from `C \<noteq> \<box>` have "C\<langle>u\<rangle> \<rhd> u" using nectxt_imp_suptp by auto
 from RCtCu have "(s,C\<langle>u\<rangle>) \<in> R" unfolding `C\<langle>t\<rangle> = s` .
 from this and `C\<langle>u\<rangle> \<rhd> u` show ?thesis by auto
qed

lemma ctxt_closed_suptp_R_imp_R_suptp:
 fixes R :: "('f,'v)trs"
 assumes "ctxt_closed R" and "s \<rhd> t" and "(t,u) \<in> R" shows "\<exists>t. (s,t) \<in> R \<and> t \<rhd> u"
proof -
 from `s \<rhd> t` obtain "C" where "C \<noteq> \<box>" and "C\<langle>t\<rangle> = s" using suptp_imp_nectxt by best
 from `ctxt_closed R` and `(t,u) \<in> R` have RCtCu: "(C\<langle>t\<rangle>,C\<langle>u\<rangle>) \<in> R" by (rule ctxt_closed_imp_ctxt)
 from `C \<noteq> \<box>` have "C\<langle>u\<rangle> \<rhd> u" by (rule nectxt_imp_suptp_ctxt)
 from `(C\<langle>t\<rangle>,C\<langle>u\<rangle>) \<in> R` have "(s,C\<langle>u\<rangle>) \<in> R" unfolding `C\<langle>t\<rangle> = s` by simp
 from this and `C\<langle>u\<rangle> \<rhd> u` show ?thesis by auto
qed

lemma ctxt_closed_R_imp_R_qc_suptp:
 assumes "ctxt_closed R" shows "quasi_commute R {(x,y). x \<rhd> y}" (is "quasi_commute ?r ?s")
proof -
 have "?s O ?r \<subseteq> ?r O (?r \<union> ?s)^*"
 proof (rule subsetI2)
  fix x y assume "(x,y) \<in> ?s O ?r"
  then obtain z where "x \<rhd> z" and "(z,y) \<in> R" by best
  with `ctxt_closed R` obtain w where "(x,w) \<in> R" and "w \<rhd> y" using ctxt_closed_suptp_R_imp_R_suptp[where R = R and s = x and t = z and u = y] by auto
  hence "(w,y) \<in> ?s" by simp
  hence "(w,y) \<in> ?r \<union> ?s" by auto
  hence "(w,y) \<in> (?r \<union> ?s)^*" by auto
  with `(x,w) \<in> R` show "(x,y) \<in> ?r O (?r \<union> ?s)^*" by auto
 qed
 thus ?thesis unfolding quasi_commute_def by force
qed

lemma closure_ctxt_idemp[simp]: "(CC(CC R)) = (CC R)"
proof
 show "CC(CC R) \<subseteq> CC R"
 proof (rule subsetI2)
  fix s t assume "(s,t) \<in> CC(CC R)"
  thus "(s,t) \<in> CC R"
  proof
   fix s' t' C assume s:"s = C\<langle>s'\<rangle>" and t: "t = C\<langle>t'\<rangle>" and "(s',t') \<in> CC R"
   from `(s',t') \<in> CC R` show "(s,t) \<in> CC R"
   proof
    fix s'' t'' D assume s': "s' = D\<langle>s''\<rangle>" and t': "t' = D\<langle>t''\<rangle>" and "(s'',t'') \<in> R"
    hence Rst'': "(s'',t'') \<in> R" by auto
    with s t s' t' have st: "s = (C \<circ>c D)\<langle>s''\<rangle>" "t = (C \<circ>c D)\<langle>t''\<rangle>" by auto
    from Rst'' show ?thesis unfolding st by (rule ccI)
   qed
  qed
 qed
next
 show "CC R \<subseteq> CC(CC R)"
 proof (rule subsetI2)
  fix s t assume "(s,t) \<in> CC R"
  hence "(\<box>\<langle>s\<rangle>,\<box>\<langle>t\<rangle>) \<in> CC(CC R)" by (rule ccI)
  thus "(s,t) \<in> CC(CC R)" by auto
 qed
qed

text {* Let~$R$ be a relation on terms that is closed under contexts. If~$R$
is well-founded then $R \cup \rhd$ is well-founed. *}
lemma SN_imp_SN_union_supt:
 assumes "SN R" and "ctxt_closed R"
 shows "SN(R \<union> supt)" (is "SN(?r \<union> ?s)")
proof -
 from `ctxt_closed R` have "quasi_commute ?r ?s" by (rule ctxt_closed_R_imp_R_qc_suptp[to_set])
 have "SN ?s" by (rule SN_supt)
 from `SN ?r` and `SN ?s` and `quasi_commute ?r ?s`
 show ?thesis by (rule quasi_commute_imp_SN)
qed

inductive_set
 closure_sub :: "('f,'v)trs \<Rightarrow> ('f,'v)trs" ("CS")
 for r :: "('f,'v)trs"
where
 csI[intro]: "((s::('f,'v)term),t) \<in> r \<Longrightarrow> (s\<cdot>(\<sigma>::('f,'v)sub),t\<cdot>\<sigma>) \<in> CS r"

lemma clsub_superset_rel: "R \<subseteq> CS R" 
proof (rule subsetI2)
 fix s t assume "(s,t) \<in> R"
 thus "(s,t) \<in> CS R" using csI[where \<sigma>="\<iota>"] by simp
qed

definition sub_closed :: "('f,'v) trs \<Rightarrow> bool"
where 
 "sub_closed r = (CS r \<subseteq> r)"

lemma sub_closed_correctness: assumes "sub_closed R" shows "R = CS R"
 using assms clsub_superset_rel unfolding sub_closed_def by auto


lemma sub_closed_union: 
  assumes r: "sub_closed r"
  and s: "sub_closed s"
  shows "sub_closed (r \<union> s)"
unfolding sub_closed_def
proof (rule subsetI2)
  fix x y	
  assume "(x,y) \<in> CS (r \<union> s)"
  thus "(x,y) \<in> r \<union> s"
  proof
    fix xx yy sig
    assume x: "x = xx \<cdot> sig" and y: "y = yy \<cdot> sig" and mem: "(xx,yy) \<in> r \<union> s"
    from mem show ?thesis
    proof
      assume "(xx,yy) \<in> r"
      hence "(x,y) \<in> CS r" 
        by (simp add: x y, rule csI)
      thus ?thesis using r unfolding sub_closed_def by auto
    next
      assume "(xx,yy) \<in> s"
      hence "(x,y) \<in> CS s" 
        by (simp add: x y, rule csI)
      thus ?thesis using s unfolding sub_closed_def by auto
    qed
  qed
qed

lemma sub_closed_comp: 
  assumes r: "sub_closed r"
  and s: "sub_closed s"
  shows "sub_closed (r O s)"
unfolding sub_closed_def
proof (rule subsetI2)
  fix x y	
  assume "(x,y) \<in> CS (r O s)"
  thus "(x,y) \<in> r O s"
  proof
    fix xx yy sig
    assume x: "x = xx \<cdot> sig" and y: "y = yy \<cdot> sig" and mem: "(xx,yy) \<in> r O s"    
    from mem obtain zz where xx: "(xx,zz) \<in> r" and yy: "(zz,yy) \<in> s" by auto
    let ?z = "zz \<cdot> sig"
    from xx have "(x,?z) \<in> CS r" 
      by (simp add: x, rule csI)
    hence x: "(x,?z) \<in> r" using r unfolding sub_closed_def by auto
    from yy have "(?z,y) \<in> CS s" 
      by (simp add: y, rule csI)
    hence y: "(?z,y) \<in> s" using s unfolding sub_closed_def by auto
    from x y show "(x,y) \<in> r O s" by auto
  qed
qed

lemma sub_closed_rtrancl: 
  assumes r: "sub_closed r"
  shows "sub_closed (r^*)"
unfolding sub_closed_def
proof (rule subsetI2)
  fix x y	
  assume "(x,y) \<in> CS (r^*)"
  thus "(x,y) \<in> r^*"
  proof
    fix xx yy sig
    assume x: "x = xx \<cdot> sig" and y: "y = yy \<cdot> sig" and mem: "(xx,yy) \<in> r^*"
    from mem have "(xx \<cdot> sig, yy \<cdot> sig) \<in> r^*"
    proof (induct, simp)
      case (step yy zz)
      hence "(yy \<cdot> sig, zz \<cdot> sig) \<in> CS r" by auto
      hence "(yy \<cdot> sig, zz \<cdot> sig) \<in> r" using r unfolding sub_closed_def by auto
      with step show ?case by auto
    qed
    with x y show "(x,y) \<in> r^*" by simp
  qed
qed

lemma sub_closed_decompose: assumes "(s,t) \<in> (CS R)" shows "\<exists>\<sigma> u v. (s = u\<cdot>\<sigma>) \<and> (t = v\<cdot>\<sigma>) \<and> (u,v) \<in> R"
 using assms by (cases rule:closure_sub.cases) force

lemma sub_sub_closed_subseteq_sub_closed: "CS (CS R) \<subseteq> CS R"
proof (rule subsetI2)
 fix s t assume a: "(s,t) \<in> CS (CS R)"
 hence "\<exists>\<sigma> u v. (s = u\<cdot>\<sigma>) \<and> (t = v\<cdot>\<sigma>) \<and> (u,v) \<in> CS R" using sub_closed_decompose by auto
 hence "\<exists>\<sigma> u v. (s = u\<cdot>\<sigma>) \<and> (t = v\<cdot>\<sigma>) \<and> (\<exists> \<tau> l r. (u = l\<cdot>\<tau>) \<and> (v = r\<cdot>\<tau>) \<and> (l,r) \<in> R)" using sub_closed_decompose by best
 hence "\<exists>\<sigma> \<tau> l r. (s = l\<cdot>(\<tau>\<circ>s\<sigma>)) \<and> (t = r\<cdot>(\<tau>\<circ>s\<sigma>)) \<and> (l,r) \<in> R" by auto
 thus "(s,t) \<in> (CS R)" by blast
qed

lemma sub_sub_closed_equal_sub_closed: "CS (CS R) = CS R" using sub_sub_closed_subseteq_sub_closed clsub_superset_rel by auto

lemma sub_closure_is_closed_under_subs [simp]: "sub_closed (CS R)" unfolding sub_closed_def sub_sub_closed_equal_sub_closed by simp

lemma supteq_sub_closed: "sub_closed supteq" unfolding sub_closed_def
proof (rule subsetI2)
  fix s t assume "(s,t) \<in> CS supteq"
  then obtain u v \<sigma> where "(u,v) \<in> supteq" and "u\<cdot>\<sigma> = s" and "v\<cdot>\<sigma> = t"
    by (cases rule: closure_sub.cases) auto
  with supteqp_stable[to_set,of u v \<sigma>] show "(s,t) \<in> supteq" by simp
qed

lemma supt_sub_closed: "sub_closed supt" unfolding sub_closed_def
proof (rule subsetI2)
 fix s t assume "(s,t) \<in> CS supt"
 then obtain u v \<sigma> where s: "s = u \<cdot> \<sigma>" and t: "t = v \<cdot> \<sigma>" and "(u,v) \<in> supt" by (cases rule: closure_sub.cases) auto
 thus "(s,t) \<in> supt" unfolding s t using suptp_stable[to_set] by simp
qed


text {* The single-step relation induced by a TRS~@{text R} is the relation @{text R} closed under
substitutions and contexts. *}
abbreviation rstep :: "('f,'v)trs \<Rightarrow> ('f,'v)trs" where "rstep R \<equiv> CC (CS R)"

definition wfrstep :: "('f,'v)trs \<Rightarrow> ('f,'v)trs"
where "wfrstep R \<equiv> {(s,t). (s,t) \<in> rstep R \<and> s \<in> Term(funas_trs R) \<and> t \<in> Term(funas_trs R)}"

lemma ctxt_closed_rstep: "ctxt_closed(rstep R)" unfolding ctxt_closed_def by simp

lemma ctxt_closed_rstep_seq: "ctxt_closed((rstep R)^*)" unfolding ctxt_closed_def
proof (rule subsetI2)
 fix s t assume "(s,t) \<in> CC((rstep R)^*)" thus "(s,t) \<in> (rstep R)^*"
 proof (induct)
  case (ccI s t C) thus ?case
  proof (induct)
   case base thus ?case by simp
  next
   case (step y z)
   hence "(C\<langle>y\<rangle>,C\<langle>z\<rangle>) \<in> rstep R" using ctxt_closed_rstep unfolding ctxt_closed_def by blast
   with `(C\<langle>s\<rangle>,C\<langle>y\<rangle>) \<in> (rstep R)^*` show ?case by simp
  qed
 qed
qed

lemma rstep_contains_r : "R \<subseteq> rstep R"
proof
  fix lr
  assume "lr \<in> R"
  show "lr \<in> rstep R" 
  proof (cases lr)
    case (Pair l r)
    with `lr \<in> R` have "(l \<cdot> \<iota>,r \<cdot> \<iota>) \<in> CS R"  using csI by blast
    hence "(l,r) \<in> CS R" by simp
    hence "(Hole \<langle>l\<rangle>, Hole \<langle>r\<rangle>) \<in> rstep R" using ccI by blast
    with Pair show ?thesis by simp 
  qed
qed      


lemma sub_closed_rstep_subseteq_rstep: "CS(rstep R) \<subseteq> rstep R"
proof (rule subsetI2)
  fix s t assume assm: "(s, t) \<in> CS (rstep R)" 
  from assm obtain \<sigma> u v where a:"s = u \<cdot> \<sigma>" and b:"t = v \<cdot> \<sigma>"
    and c:"(u, v) \<in> CC (CS R)" by (cases rule:closure_sub.cases) simp
  from c obtain C l r where d:"u = C\<langle>l\<rangle>" and e: "v = C\<langle>r\<rangle>" and f: "(l, r) \<in> CS R"
    by (cases rule:closure_ctxt.cases) simp
  from f obtain \<tau> m n where g:"l= m \<cdot> \<tau>" and h:"r = n \<cdot> \<tau>" and i:"(m, n) \<in> R"
    by (cases rule:closure_sub.cases) simp
  hence n: "s = (C \<cdot>c \<sigma>)\<langle>m \<cdot> (\<tau> \<circ>s \<sigma>)\<rangle>"
    and o:"t = (C \<cdot>c \<sigma>)\<langle>n \<cdot> (\<tau> \<circ>s \<sigma>)\<rangle>" using a b d e g h ctxt_sub_distr by auto
 thus "(s, t) \<in> (CC (CS R))" using i csI unfolding n o by best
qed

lemma rstep_sub_closed: "sub_closed(rstep R)" using sub_closed_rstep_subseteq_rstep unfolding sub_closed_def by auto

lemma rstep_seq_sub_closed: "sub_closed((rstep R)^*)" unfolding sub_closed_def
proof (rule subsetI2)
 fix s t assume "(s,t) \<in> CS((rstep R)^*)" thus "(s,t) \<in> (rstep R)^*"
 proof (induct)
  case (csI s t \<sigma>) thus ?case
  proof (induct)
   case base thus ?case by simp
  next
   case (step y z)
   hence "(y \<cdot> \<sigma>,z \<cdot> \<sigma>) \<in> rstep R" using rstep_sub_closed unfolding sub_closed_def by blast
   with `(s \<cdot> \<sigma>,y \<cdot> \<sigma>) \<in> (rstep R)^*` show ?case by simp
  qed
 qed
qed

lemma comp_stable:
 assumes R: "\<And>x y \<sigma>. (x,y) \<in> R \<Longrightarrow> (x \<cdot> \<sigma>,y \<cdot> \<sigma>) \<in> R" and S: "\<And>x y \<sigma>. (x,y) \<in> S \<Longrightarrow> (x \<cdot> \<sigma>,y \<cdot> \<sigma>) \<in> S"
 shows "\<And>x y \<sigma>. (x,y) \<in> R O S \<Longrightarrow> (x \<cdot> \<sigma>,y \<cdot> \<sigma>) \<in> R O S"
proof -
 fix s t \<sigma> assume "(s,t) \<in> R O S"
 then obtain u where "(s,u) \<in> R" and "(u,t) \<in> S" by auto
 from R[OF `(s,u) \<in> R`] have "(s \<cdot> \<sigma>,u \<cdot> \<sigma>) \<in> R" by simp
 from S[OF `(u,t) \<in> S`] have "(u \<cdot> \<sigma>,t \<cdot> \<sigma>) \<in> S" by simp
 with `(s \<cdot> \<sigma>,u \<cdot> \<sigma>) \<in> R`
 show "(s \<cdot> \<sigma>,t \<cdot> \<sigma>) \<in> R O S" by auto
qed

lemma comp_sub_closed: assumes "sub_closed R" and "sub_closed S" shows "sub_closed(R O S)" unfolding sub_closed_def
proof (rule subsetI2)
 fix s t assume "(s,t) \<in> CS(R O S)"
 then obtain u v \<sigma> where s: "s = u \<cdot> \<sigma>" and t: "t = v \<cdot> \<sigma>" and "(u,v) \<in> R O S" by (cases rule: closure_sub.cases) auto
 from `(u,v) \<in> R O S` obtain w where "(u,w) \<in> R" and "(w,v) \<in> S" by auto
 from `sub_closed R` and `(u,w) \<in> R` have "(s,w \<cdot> \<sigma>) \<in> R" unfolding sub_closed_def s by auto
 from `sub_closed S` and `(w,v) \<in> S` have "(w \<cdot> \<sigma>,t) \<in> S" unfolding sub_closed_def t by auto
 with `(s,w \<cdot> \<sigma>) \<in> R` show "(s,t) \<in> R O S" by auto
qed

text {* The set of minimally non-terminating Terms, i.e., non-terminating terms whose
proper subterms are terminating w.r.t.\ a given TRS. *}
definition Tinf :: "('f, 'v) trs \<Rightarrow> ('f, 'v) term set"
where "Tinf R \<equiv> {t. \<not> SN_elt (rstep R) t \<and> (\<forall>s\<lhd>t. SN_elt (rstep R) s)}"

lemma [elim!]: fixes t :: "('f, 'v) term" assumes "Var x \<rhd> t" shows P
using assms suptp_var[of x t] by simp

lemma [elim]: assumes "Fun f ts \<rhd> s" obtains t where "t \<in> set ts" and "t \<unrhd> s"
using assms by (cases) auto

lemma not_SN_imp_subt_Tinf:
  assumes "\<not> SN_elt (rstep R) t" shows "\<exists>s\<unlhd>t. s \<in> Tinf R"
using assms proof (induct rule: subterm_induct)
  case (Var x) thus ?case by (auto simp: Tinf_def)
next
  case (Fun f ts)
  let ?t = "Fun f ts"
  show ?case
  proof (cases "\<forall>s\<lhd>?t. SN_elt (rstep R) s")
    case True
    with `\<not> SN_elt (rstep R) ?t` show ?thesis by (auto simp: Tinf_def)
  next
    case False
    then obtain s where "?t \<rhd> s" and "\<not> SN_elt (rstep R) s" by best
    from `?t \<rhd> s` obtain t where "t \<in> set ts" and "t \<unrhd> s" ..
    with Fun and `\<not> SN_elt (rstep R) s` obtain u where "u \<unlhd> s" and "u \<in> Tinf R" by auto
    from `t \<in> set ts` have "?t \<rhd> t" by auto
    also have "\<dots> \<unrhd> s" by (rule `t \<unrhd> s`)
    also have "\<dots> \<unrhd> u" by (rule `u \<unlhd> s`)
    finally have "?t \<rhd> u" .
    with `u \<in> Tinf R` show ?thesis by best
  qed
qed

lemma not_SN_imp_Tinf:
  assumes "\<not> SN (rstep R)" shows "\<exists>t. t \<in> Tinf R"
using assms not_SN_imp_subt_Tinf unfolding SN_def by auto

(* @{term "Tinf"} characterizes termination. *)
lemma Tinf_SN_conv: "(Tinf R = {}) = (SN (rstep R))"
proof
  assume "Tinf R = {}" thus "SN (rstep R)" using not_SN_imp_Tinf by blast
next
  assume "SN (rstep R)" thus "Tinf R = {}" unfolding Tinf_def SN_defs by blast
qed

definition rstep_r_p_s ::
 "('f,'v)trs \<Rightarrow> ('f,'v)rule \<Rightarrow> pos \<Rightarrow> ('f,'v)sub \<Rightarrow> ('f,'v)trs"
where
 "rstep_r_p_s R r p \<sigma> \<equiv> {(s,t).
 let C = ctxt_of_pos_term p s in p \<in> poss s \<and> r \<in> R \<and> (C\<langle>fst r\<cdot>\<sigma>\<rangle> = s) \<and> (C\<langle>snd r\<cdot>\<sigma>\<rangle> = t)}"

lemma "p \<in> poss t \<Longrightarrow> hole_pos(ctxt_of_pos_term p t) = p" by (induct p t rule: ctxt_of_pos_term.induct) auto

lemma ctxt_of_pos_of_ctxt: "ctxt_of_pos_term (hole_pos C) (C\<langle>s\<rangle>) = C" by (induct C) auto

lemma ep_pos_term: "\<epsilon> \<in> poss t" by (induct rule: term_induct) auto

(* The position of the hole in a context C is contained in the set
of positions of any term obtained by applying C. *)
lemma hole_pos_in_poss: "hole_pos C \<in> poss (C\<langle>t\<rangle>)" by (induct C arbitrary: t) (auto simp: ep_pos_term)

lemma rstep_iff_rstep_r_p_s:
 fixes s t :: "('f,'v)term"
 shows "(s,t) \<in> rstep R = (\<exists>l r p \<sigma>. (s,t) \<in> rstep_r_p_s R (l,r) p \<sigma>)"
proof
 assume "(s,t) \<in> CC (CS R)"
 then obtain C s' t' where A:"(s',t') \<in> CS R"
 and B[simp]: "s = C\<langle>s'\<rangle>" and C[simp]: "t = C\<langle>t'\<rangle>" by (auto elim: closure_ctxt.cases)
 from A obtain \<sigma> :: "('f,'v)sub" and l r :: "('f,'v)term" where [simp]:"s' = l\<cdot>\<sigma>" "t' = r\<cdot>\<sigma>" and rule: "(l,r) \<in> R"
   by (auto elim: closure_sub.cases)
 from `t = C\<langle>t'\<rangle>` `s = C\<langle>s'\<rangle>` obtain p :: pos where p: "p = hole_pos C" by auto
 hence pos: "p \<in> poss s" by (auto simp: hole_pos_in_poss)
 from p have "ctxt_of_pos_term p s = C" by (simp add: ctxt_of_pos_of_ctxt)
 with pos rule have
 "let C = ctxt_of_pos_term p s in p \<in> poss s \<and> (l,r) \<in> R \<and> C\<langle>l\<cdot>\<sigma>\<rangle> = s \<and> C\<langle>r\<cdot>\<sigma>\<rangle> = t" by auto
 thus "\<exists>l r p \<sigma>. (s,t) \<in> rstep_r_p_s R (l,r) p \<sigma>" unfolding rstep_r_p_s_def by auto
next
 assume "\<exists>l r p \<sigma>. (s,t) \<in> rstep_r_p_s R (l,r) p \<sigma>"
 then obtain l r p \<sigma> where rstep: "(s,t) \<in> rstep_r_p_s R (l,r) p \<sigma>" by auto
 let ?C = "(ctxt_of_pos_term p s)"
 from rstep have "(l,r) \<in> R" and eqs:"s = ?C\<langle>l\<cdot>\<sigma>\<rangle>" "t = ?C\<langle>r\<cdot>\<sigma>\<rangle> "
  unfolding rstep_r_p_s_def by (auto simp: Let_def)
 hence "(l\<cdot>\<sigma>,r\<cdot>\<sigma>) \<in> CS R" by auto
 hence "(?C\<langle>l\<cdot>\<sigma>\<rangle>,?C\<langle>r\<cdot>\<sigma>\<rangle>) \<in> CC (CS R)" by auto
 with eqs show  "(s,t) \<in> CC (CS R)" by auto
qed

text {* Rewriting steps below the root position. *}
definition nrrstep :: "('f,'v)trs \<Rightarrow> ('f,'v)trs"
where "nrrstep R \<equiv> {(s,t). \<exists>r i ps \<sigma>. (s,t) \<in> rstep_r_p_s R r (i<#ps) \<sigma>}"

text {* An alternative characterisation of non-root rewrite steps. *}
lemma nrrstep_def':
  "nrrstep R = {(s, t). \<exists>l r C \<sigma>. (l, r) \<in> R \<and> C \<noteq> \<box> \<and> s = C\<langle>l\<cdot>\<sigma>\<rangle> \<and> t = C\<langle>r\<cdot>\<sigma>\<rangle>}"
  (is "?lhs = ?rhs")
proof
  show "?lhs \<subseteq> ?rhs"
  proof (rule subsetI2)
    fix s t assume "(s, t) \<in> nrrstep R"
    then obtain l r i ps \<sigma> where step: "(s, t) \<in> rstep_r_p_s R (l, r) (i <# ps) \<sigma>"
      unfolding nrrstep_def by best
    let ?C = "ctxt_of_pos_term (i <# ps) s"
    from step have"i <# ps \<in> poss s" and "(l, r) \<in> R" and "s = ?C\<langle>l\<cdot>\<sigma>\<rangle>" and "t = ?C\<langle>r\<cdot>\<sigma>\<rangle>"
      unfolding rstep_r_p_s_def Let_def by auto
    moreover from `i <# ps \<in> poss s` have "?C \<noteq> \<box>" by (induct s) auto
    ultimately show "(s, t) \<in> ?rhs" by auto
  qed
next
  show "?rhs \<subseteq> ?lhs"
  proof (rule subsetI2)
    fix s t assume "(s, t) \<in> ?rhs"
    then obtain l r C \<sigma> where in_R: "(l, r) \<in> R" and "C \<noteq> \<box>"
      and s: "s = C\<langle>l\<cdot>\<sigma>\<rangle>" and t: "t = C\<langle>r\<cdot>\<sigma>\<rangle>" by auto
    from `C \<noteq> \<box>` obtain i p where ip: "hole_pos C = i <# p" by (induct C) auto
    have "i <# p \<in> poss s" using hole_pos_poss[of C] unfolding s ip by simp
    hence C: "C = ctxt_of_pos_term (i <# p) s" unfolding s ip[symmetric]
      by (simp add: ctxt_of_pos_of_ctxt)
    from `i <# p \<in> poss s` in_R s t have "(s, t) \<in> rstep_r_p_s R (l, r) (i <# p) \<sigma>"
      unfolding rstep_r_p_s_def C by simp
    thus "(s, t) \<in> nrrstep R" unfolding nrrstep_def by best
  qed
qed

text {* Rewriting step at the root position. *}
definition rrstep :: "('f, 'v) trs \<Rightarrow> ('f, 'v) trs"
where "rrstep R \<equiv> {(s,t). \<exists>r \<sigma>. (s,t) \<in> rstep_r_p_s R r \<epsilon> \<sigma>}"

text {* An alternative characterisation of root rewrite steps. *}
lemma rrstep_def': "rrstep R = {(s, t). \<exists>l r \<sigma>. (l, r) \<in> R \<and> s = l\<cdot>\<sigma> \<and> t = r\<cdot>\<sigma>}"
  (is "_ = ?rhs")
by (auto simp: rrstep_def rstep_r_p_s_def)

lemma rstep_iff_rrstep_or_nrrstep: "rstep R = (rrstep R \<union> nrrstep R)"
proof
 show "rstep R \<subseteq> rrstep R \<union> nrrstep R"
 proof (rule subsetI2)
  fix s t assume "(s,t) \<in> rstep R"
  then obtain l r p \<sigma> where rstep_rps: "(s,t) \<in> rstep_r_p_s R (l,r) p \<sigma>"
   by (auto simp: rstep_iff_rstep_r_p_s)
  thus "(s,t) \<in> rrstep R \<union> nrrstep R" unfolding rrstep_def nrrstep_def by (cases p) auto
 qed
next
 show "rrstep R \<union> nrrstep R \<subseteq> rstep R"
 proof (rule subsetI2)
  fix s t assume "(s,t) \<in> rrstep R \<union> nrrstep R"
  thus "(s,t) \<in> rstep R" by (auto simp: rrstep_def nrrstep_def rstep_iff_rstep_r_p_s)
 qed
qed

lemma rstep_i_pos_imp_rstep_arg_i_pos:
 assumes nrrstep: "(Fun f ss,t) \<in> rstep_r_p_s R (l,r) (i<#ps) \<sigma>"
 shows "(ss!i,t|_<i>) \<in> rstep_r_p_s R (l,r) ps \<sigma>"
proof -
 from nrrstep obtain C where C:"C = ctxt_of_pos_term (i<#ps) (Fun f ss)"
  and pos: "(i<#ps) \<in> poss (Fun f ss)"
  and Rlr: "(l,r) \<in> R"
  and Fun: "C\<langle>l\<cdot>\<sigma>\<rangle> = Fun f ss"
  and t: "C\<langle>r\<cdot>\<sigma>\<rangle> = t" unfolding rstep_r_p_s_def Let_def by auto
 then obtain D where C':"C = More f (take i ss) D (drop (Suc i) ss)" by auto
 hence CFun: "C\<langle>l\<cdot>\<sigma>\<rangle> = Fun f (take i ss @ (D\<langle>l\<cdot>\<sigma>\<rangle>) # drop (Suc i) ss)" by auto
 from pos have len: "i < length ss" by auto
 from len
 have "(take i ss @ (D\<langle>l\<cdot>\<sigma>\<rangle>) # drop (Suc i) ss)!i = D\<langle>l\<cdot>\<sigma>\<rangle>" by (auto simp: nth_append)
 with C Fun CFun have ssi: "ss!i = D\<langle>l\<cdot>\<sigma>\<rangle>" by auto
 from C' t have t': "t = Fun f (take i ss @ (D\<langle>r\<cdot>\<sigma>\<rangle>) # drop (Suc i) ss)" by auto
 from len
 have "(take i ss @ (D\<langle>r\<cdot>\<sigma>\<rangle>) # drop (Suc i) ss)!i = D\<langle>r\<cdot>\<sigma>\<rangle>" by (auto simp: nth_append)
 with t' have "t|_<i> = (D\<langle>r\<cdot>\<sigma>\<rangle>)|_\<epsilon>" by auto
 hence subt_at: "t|_<i> = D\<langle>r\<cdot>\<sigma>\<rangle>" by simp
 from C C' have "D = ctxt_of_pos_term ps (ss!i)" by auto
 with pos Rlr ssi[symmetric] subt_at[symmetric]
 show ?thesis unfolding rstep_r_p_s_def Let_def by auto
qed

lemma ctxt_closure_is_closed_under_ctxts: "ctxt_closed (CC R)" unfolding ctxt_closed_def by auto

theorem ctxt_closed_rstep_equal_rstep:  "CC (rstep R) = rstep R" by simp

theorem sub_closed_rstep_equal_rstep: "CS (rstep R) = rstep R" 
 using sub_closed_rstep_subseteq_rstep sub_closed_correctness unfolding sub_closed_def by blast

(* Rewriting steps are closed under contexts. *)
lemma rstep_imp_ctxt_rstep: "(s,t) \<in> rstep R \<Longrightarrow> (C\<langle>s\<rangle>,C\<langle>t\<rangle>) \<in> rstep R"
proof -
 assume "(s,t) \<in> rstep R"
 hence "(C\<langle>s\<rangle>,C\<langle>t\<rangle>) \<in> CC(CC(CS R))" by (rule ccI)
 thus "(C\<langle>s\<rangle>,C\<langle>t\<rangle>) \<in> CC(CS R)" by simp
qed

lemma rstep_supt_commute: "supt O rstep R \<subseteq> rstep R O supt"
proof (rule subsetI2)
  fix s t assume "(s,t) \<in> supt O rstep R" thus "(s,t) \<in> rstep R O supt"
  proof (induct s)
    case (Var x)
    hence "\<exists>u. (Var x, u) \<in> supt \<and> (u,t) \<in> rstep R" by auto
    then obtain u where "(Var x,u) \<in> supt" and "(u,t) \<in> rstep R" by auto
    from `(Var x,u) \<in> supt`[to_pred] show ?case by (cases rule: suptp.cases) auto
  next
    case (Fun f ss)
    then obtain u where "(Fun f ss,u) \<in> supt" and "(u,t) \<in> rstep R" by auto
    from `(Fun f ss,u) \<in> supt`[to_pred] obtain C where "C \<noteq> \<box>" and "C\<langle>u\<rangle> = Fun f ss" using suptp_imp_nectxt by auto
    from `C \<noteq> \<box>` have "C\<langle>t\<rangle> \<rhd> t" by (rule nectxt_imp_suptp_ctxt)
    from `(u,t) \<in> rstep R` have "(C\<langle>u\<rangle>,C\<langle>t\<rangle>) \<in> rstep R" by (rule rstep_imp_ctxt_rstep)
    hence "(Fun f ss,C\<langle>t\<rangle>) \<in> rstep R" unfolding `C\<langle>u\<rangle> = Fun f ss` .
    with `C\<langle>t\<rangle> \<rhd> t`[to_set] show ?case by auto
  qed
qed

lemma rstep_seq_supt_commute: "supt O (rstep R)^* \<subseteq> (rstep R)^* O supt"
proof (rule subsetI2)
  fix s t assume "(s,t) \<in> supt O (rstep R)^*"
  then obtain u where "(u,t) \<in> (rstep R)^*" and "(s,u) \<in> supt" by auto
  thus "(s,t) \<in> (rstep R)^* O supt"
  proof (induct rule: rtrancl.induct)
    case (rtrancl_refl u) thus ?case by auto
  next
    case (rtrancl_into_rtrancl x y z)
    hence IH: "(s,y) \<in> (rstep R)^* O supt" by simp
    then obtain u where "(s,u) \<in> (rstep R)^*" and "(u,y) \<in> supt" by auto
    with `(y,z) \<in> rstep R` have "(u,z) \<in> supt O rstep R" by auto
    hence "(u,z) \<in> rstep R O supt" using rstep_supt_commute by auto
    then obtain v where "(u,v) \<in> rstep R" and "(v,z) \<in> supt" by auto
    from `(s,u) \<in> (rstep R)^*` and `(u,v) \<in> rstep R` have "(s,v) \<in> (rstep R)^*" by auto
    with `(v,z) \<in> supt` show ?case by auto
  qed
qed

lemma ne_rstep_seq_imp_list_of_terms:
  assumes "(s,t) \<in> (rstep R)^+"
  shows "\<exists>ts. length ts > 1 \<and> ts!0 = s \<and> ts!(length ts - 1) = t \<and>
         (\<forall>i<length ts - 1. (ts!i,ts!(Suc i)) \<in> (rstep R))" (is "\<exists>ts. _ \<and> _ \<and> _ \<and> ?P ts")
using assms proof (induct rule: trancl.induct)
  case (r_into_trancl x y)
  let ?ts = "[x,y]"
  have "length ?ts > 1" by simp
  moreover have "?ts!0 = x" by simp
  moreover have "?ts!(length ?ts - 1) = y" by simp
  moreover from r_into_rtrancl r_into_trancl have "?P ?ts" by auto
  ultimately show ?case by fast
next
  case (trancl_into_trancl x y z)
  then obtain ts where "length ts > 1" and "ts!0 = x"
    and last1: "ts!(length ts - 1) = y" and "?P ts" by auto
  let ?ts = "ts@[z]"
  have len: "length ?ts = length ts + 1" by simp
  from `length ts > 1` have "length ?ts > 1" by auto
  moreover with `ts!0 = x` have "?ts!0 = x" by (induct ts) auto
  moreover have last2: "?ts!(length ?ts - 1) = z" by simp
  moreover have "?P ?ts"
  proof (intro allI impI)
    fix i assume A: "i < length ?ts - 1"
    show "(?ts!i,?ts!(Suc i)) \<in> rstep R"
    proof (cases "i < length ts - 1")
      case True with `?P ts` A show ?thesis unfolding len unfolding nth_append by auto
    next
      case False with A have "i = length ts - 1" by simp
      with last1 `length ts > 1` have "?ts!i = y" unfolding nth_append by auto
      have "Suc i = length ?ts - 1" using `i = length ts - 1` using `length ts > 1` by auto
      with last2 have "?ts!(Suc i) = z" by auto
      from `(y,z) \<in> rstep R` show ?thesis unfolding `?ts!(Suc i) = z` `?ts!i = y` .
    qed
  qed
  ultimately show ?case by fast
qed

definition SN_rel_supt :: "('f,'v)trs \<Rightarrow> ('f,'v)trs"
where "SN_rel_supt R \<equiv> (SN_rel R R) \<union> (SN_rel supt R)"

lemma SN_rel_imp_SN_subt:
 fixes R::"('f,'v)trs" and t::"('f,'v)term" 
 assumes "ctxt_closed R" and "SN_elt R s" and "s \<unrhd> t"
 shows "SN_elt R t"
proof (rule ccontr)
 assume "\<not> SN_elt R t"
 then obtain A::"nat \<Rightarrow> ('f,'v)term" where "A 0 = t" and "\<forall>i. (A i,A (Suc i)) \<in> R"
  unfolding SN_elt_def[where r = "R" and x = "t"] by best
 from `s \<unrhd> t` obtain C where "s = C\<langle>t\<rangle>" using supteqp_imp_ctxt by force
 let ?B = "\<lambda>i. C\<langle>A i\<rangle>"
 have "\<forall>i. (?B i,?B(Suc i)) \<in> R"
 proof
  fix i
  from `\<forall>i. (A i,A(Suc i)) \<in> R` have "(?B i,?B(Suc i)) \<in> CC(R)" using ccI by fast
  thus "(?B i,?B(Suc i)) \<in> R" using `ctxt_closed R` unfolding ctxt_closed_def by auto
 qed
 with `A 0 = t` have "?B 0 = s \<and> (\<forall>i. (?B i,?B(Suc i)) \<in> R)" unfolding `s = C\<langle>t\<rangle>` by auto
 hence "\<not> SN_elt R s" unfolding SN_elt_def by auto
 with assms show "False" by simp
qed

lemma ctxt_closed_R_imp_suptp_SN_rel_distr:
 fixes R :: "('f,'v)trs"
 assumes "ctxt_closed R" and "(s,t) \<in> (SN_rel supt R)" and "(t,u) \<in> SN_rel R R" shows "(\<exists>t. (s,t) \<in> SN_rel R R \<and> (t,u) \<in> SN_rel supt R)" (is "\<exists> t. _ \<and> (t,u) \<in> ?snSub")
proof -
 from `(s,t) \<in> ?snSub` have "SN_elt R s" and "s \<rhd> t" unfolding SN_rel_def supt_def by auto
 from `s \<rhd> t` obtain C where "C \<noteq> Hole" and "C\<langle>t\<rangle> = s" using suptp_imp_nectxt by best
 from `(t,u) \<in> SN_rel R R` have "(t,u) \<in> R" unfolding SN_rel_def by auto
 from `ctxt_closed R` and this have RCtCu: "(C\<langle>t\<rangle>,C\<langle>u\<rangle>) \<in> R" by (rule ctxt_closed_imp_ctxt)
 with `C\<langle>t\<rangle> = s` `SN_elt R s` have one: "SN_elt R C\<langle>u\<rangle>" using step_preserves_SN by force
 from `C \<noteq> \<box>` have "C\<langle>u\<rangle> \<rhd> u" using nectxt_imp_suptp by auto
 with one have "(C\<langle>u\<rangle>,u) \<in> ?snSub" unfolding SN_rel_def supt_def by auto
 from RCtCu have "(s,C\<langle>u\<rangle>) \<in> R" unfolding `C\<langle>t\<rangle> = s` .
 from this and `SN_elt R s` and `(C\<langle>u\<rangle>,u) \<in> ?snSub` show ?thesis unfolding SN_rel_def by auto
qed

lemma ctxt_closed_R_imp_SN_rel_qc_suptp:
 assumes "ctxt_closed R" shows "quasi_commute (SN_rel R R) (SN_rel supt R)" (is "quasi_commute ?r ?s")
proof -
  have "?s O ?r \<subseteq> ?r O (?r \<union> ?s)^*" 
  proof (rule subsetI2)
    fix x y 
    assume "(x,y) \<in> ?s O ?r"
    from this obtain z where "(x,z) \<in> ?s" and "(z,y) \<in> ?r" by best
    hence "x \<rhd> z" and "(z,y) \<in> R" and "SN_elt R x" unfolding SN_rel_def supt_def by auto
    with `ctxt_closed R` obtain w where "(x,w) \<in> R" and "w \<rhd> y" using ctxt_closed_suptp_R_imp_R_suptp[where R = R and s = x and t = z and u = y] by auto
    from this and `SN_elt R x` have "SN_elt R w" using step_preserves_SN by auto
    with `w \<rhd> y` have "(w,y) \<in> ?s" unfolding SN_rel_def supt_def by auto
    with `(x,w) \<in> R` `SN_elt R x` show "(x,y) \<in> ?r O (?r \<union> ?s)^*" unfolding SN_rel_def by auto
  qed
  thus ?thesis unfolding quasi_commute_def .
qed

lemma ctxt_closed_imp_SN_SN_rel_supt:
 assumes "ctxt_closed R"
 shows "SN (SN_rel_supt R)"
proof -
  let ?r = "SN_rel R R"
  let ?s = "SN_rel supt R"
  from `ctxt_closed R` have "quasi_commute ?r ?s" by (rule ctxt_closed_R_imp_SN_rel_qc_suptp[to_set])
  have "SN supt" by (rule SN_supt)
  hence "SN ?s" unfolding SN_rel_def SN_def SN_elt_def by auto
  have "SN ?r" by (rule SN_SN_rel)
  from `SN ?r` and `SN ?s` and `quasi_commute ?r ?s`
  show ?thesis unfolding SN_rel_supt_def by  (rule quasi_commute_imp_SN)
qed

lemma SN_SN_rel_supt_rstep: 
  shows "SN (SN_rel_supt (rstep R))"
proof -
  have "ctxt_closed (rstep R)" by (rule ctxt_closed_rstep)
  thus ?thesis by (rule ctxt_closed_imp_SN_SN_rel_supt)
qed

lemma nrrstep_imp_pos_term:
 fixes t :: "('f,'v)term" and R :: "('f,'v)trs"
 shows "(Fun f ss,t) \<in> nrrstep R \<Longrightarrow> \<exists>i s. t = Fun f (ss[i:=s]) \<and> (ss!i,s) \<in> rstep R \<and> i < length ss"
proof -
 assume "(Fun f ss,t) \<in> nrrstep R"
 then obtain l r i ps \<sigma> where rstep_rps:"(Fun f ss,t) \<in> rstep_r_p_s R (l,r) (i<#ps) \<sigma>"
  unfolding nrrstep_def by auto
 then obtain C
 where "(l,r) \<in> R"
  and pos: "(i<#ps) \<in> poss (Fun f ss)"
  and C: "C = ctxt_of_pos_term (i<#ps) (Fun f ss)"
  and "C\<langle>l\<cdot>\<sigma>\<rangle> = Fun f ss"
  and t: "C\<langle>r\<cdot>\<sigma>\<rangle> = t"
 unfolding rstep_r_p_s_def Let_def by auto
 then obtain D where "C = More f (take i ss) D (drop (Suc i) ss)" by auto
 with t have t': "t = Fun f (take i ss @ (D\<langle>r\<cdot>\<sigma>\<rangle>) # drop (Suc i) ss)" by auto
 from rstep_rps have "(ss!i,t|_<i>) \<in> rstep_r_p_s R (l,r) ps \<sigma>"
  by (rule rstep_i_pos_imp_rstep_arg_i_pos)
 hence rstep:"(ss!i,t|_<i>) \<in> rstep R" by (auto simp: rstep_iff_rstep_r_p_s)
 hence "(C\<langle>ss!i\<rangle>,C\<langle>t|_<i>\<rangle>) \<in> rstep R" by (rule rstep_imp_ctxt_rstep)
 from pos have len: "i < length ss" by auto
 from len
 have "(take i ss @ (D\<langle>r\<cdot>\<sigma>\<rangle>) # drop (Suc i) ss)!i = D\<langle>r\<cdot>\<sigma>\<rangle>" by (auto simp: nth_append)
 with C t' have "t|_<i> = D\<langle>r\<cdot>\<sigma>\<rangle>" by auto
 with t' have "t = Fun f (take i ss @ (t|_<i>) # drop (Suc i) ss)" by auto
 with len have "t = Fun f (ss[i:=t|_<i>])" by (auto simp: upd_conv_take_nth_drop)
 with rstep len show "\<exists>i s. t = Fun f (ss[i:=s]) \<and> (ss!i,s) \<in> rstep R \<and> i < length ss" by auto
qed

lemma rstep_cases[consumes 1]:
 "\<lbrakk>(s,t) \<in> rstep R; (s,t) \<in> rrstep R \<Longrightarrow> P; (s,t) \<in> nrrstep R \<Longrightarrow> P\<rbrakk> \<Longrightarrow> P"
by (auto simp: rstep_iff_rrstep_or_nrrstep)

lemma nrrstep_imp_rstep: "(s,t) \<in> nrrstep R \<Longrightarrow> (s,t) \<in> rstep R"
by (auto simp: rstep_iff_rrstep_or_nrrstep)

lemma nrrstep_imp_Fun: "(s,t) \<in> nrrstep R \<Longrightarrow> \<exists>f ss. s = Fun f ss"
proof -
 assume "(s,t) \<in> nrrstep R"
 then obtain i ps where "i<#ps \<in> poss s"
 unfolding nrrstep_def rstep_r_p_s_def Let_def by auto
 thus "\<exists>f ss. s = Fun f ss" by (cases s) auto
qed

fun num_args :: "('f,'v)term \<Rightarrow> nat"
where "num_args(Var x) = 0"
    | "num_args(Fun f ts) = length ts"

lemma nrrstep_imp_subt_rstep:
 assumes "(s,t) \<in> nrrstep R"
 shows "\<exists>i. i < num_args s \<and> num_args s = num_args t \<and> (s|_<i>,t|_<i>) \<in> rstep R \<and> (\<forall>j. i \<noteq> j \<longrightarrow> s|_<j> = t|_<j>)"
proof -
 from `(s,t) \<in> nrrstep R` obtain f ss where "s = Fun f ss" using nrrstep_imp_Fun by blast
 with `(s,t) \<in> nrrstep R` have "(Fun f ss,t) \<in> nrrstep R" by simp
 then obtain i u where "t = Fun f (ss[i := u])" and "(ss!i,u) \<in> rstep R" and "i < length ss"
  using nrrstep_imp_pos_term by best
 from `s = Fun f ss` and `t = Fun f (ss[i := u])` have "num_args s = num_args t" by auto
 from `i < length ss` and `s = Fun f ss` have "i < num_args s" by auto
 from `s = Fun f ss` have "s|_<i> = (ss!i)" by auto
 from `t = Fun f (ss[i := u])` and `i < length ss` have "t|_<i> = u" by auto
 from `s = Fun f ss` and `t = Fun f (ss[i := u])`
 have  "\<forall>j. j \<noteq> i \<longrightarrow> s|_<j> = t|_<j>" by auto
 with `(ss!i,u) \<in> rstep R`
  and `i < num_args s`
  and `num_args s = num_args t`
  and `s|_<i> = (ss!i)`[symmetric] and `t|_<i> = u`[symmetric]
 show ?thesis by auto
qed

lemma finite_range:
  fixes f :: "nat \<Rightarrow> 'a"
  assumes fin: "finite (range f)"
  shows "\<exists>x. \<exists>\<^sub>\<infinity>i. f i = x"
proof (rule classical)
  assume "\<not>(\<exists>x. \<exists>\<^sub>\<infinity>i. f i = x)"
  hence "\<forall>x. \<exists>j. \<forall>i>j. f i \<noteq> x" unfolding INFM_nat by blast
  with choice have "\<exists>j. \<forall>x. \<forall>i>(j x). f i \<noteq> x" .
  then obtain j where neq: "\<And>x i. j x < i \<Longrightarrow> f i \<noteq> x" by blast
  from fin have "finite (range (j o f))" unfolding comp_def using finite_range_imageI by blast
  with finite_nat_bounded obtain m where "range (j o f) \<subseteq> {..<m}" by blast
  hence "j (f m) < m" unfolding comp_def by auto
  with neq[of "f m" m] show ?thesis by blast
qed

lemma rstep_stable[intro]: assumes "(s, t) \<in> R" shows "(s\<cdot>\<sigma>, t\<cdot>\<sigma>) \<in> rstep R"
using ccI[OF csI[OF assms], of "\<box>"] by simp

lemma nrrstep_subt: assumes "(s, t) \<in> nrrstep R" shows "\<exists>u\<lhd>s. \<exists>v\<lhd>t. (u, v) \<in> rstep R"
proof -
  from assms obtain l r C \<sigma> where "(l, r) \<in> R" and "C \<noteq> \<box>"
    and s: "s = C\<langle>l\<cdot>\<sigma>\<rangle>" and t: "t = C\<langle>r\<cdot>\<sigma>\<rangle>" unfolding nrrstep_def' by best
  from `C \<noteq> \<box>` s have "s \<rhd> l\<cdot>\<sigma>" by (rule ctxt_suptp_I)
  moreover from `C \<noteq> \<box>` t have "t \<rhd> r\<cdot>\<sigma>" by (rule ctxt_suptp_I)
  moreover from `(l, r) \<in> R` have "(l\<cdot>\<sigma>, r\<cdot>\<sigma>) \<in> rstep R" ..
  ultimately show ?thesis by auto
qed

lemma nrrstep_args:
  assumes "(s, t) \<in> nrrstep R"
  shows "\<exists>f ss ts. s = Fun f ss \<and> t = Fun f ts \<and> length ss = length ts
    \<and> (\<exists>j<length ss. (ss!j, ts!j) \<in> rstep R \<and> (\<forall>i<length ss. i \<noteq> j \<longrightarrow> ss!i = ts!i))"
proof -
  from assms obtain l r C \<sigma> where "(l, r) \<in> R" and "C \<noteq> \<box>"
    and s: "s = C\<langle>l\<cdot>\<sigma>\<rangle>" and t: "t = C\<langle>r\<cdot>\<sigma>\<rangle>" unfolding nrrstep_def' by best
  from `C \<noteq> \<box>` obtain f ss1 D ss2 where C: "C = More f ss1 D ss2" by (induct C) auto
  have "s = Fun f (ss1 @ D\<langle>l\<cdot>\<sigma>\<rangle> # ss2)" (is "_ = Fun f ?ss") by (simp add: s C)
  moreover have "t = Fun f (ss1 @ D\<langle>r\<cdot>\<sigma>\<rangle> # ss2)" (is "_ = Fun f ?ts") by (simp add: t C)
  moreover have "length ?ss = length ?ts" by simp
  moreover have "\<exists>j<length ?ss.
    (?ss!j, ?ts!j) \<in> rstep R \<and> (\<forall>i<length ?ss. i \<noteq> j \<longrightarrow> ?ss!i = ?ts!i)"
  proof -
    let ?j = "length ss1"
    have "?j < length ?ss" by simp
    moreover have "(?ss!?j, ?ts!?j) \<in> rstep R"
    proof -
      from `(l, r) \<in> R` have "(D\<langle>l\<cdot>\<sigma>\<rangle>, D\<langle>r\<cdot>\<sigma>\<rangle>) \<in> rstep R" by auto
      thus ?thesis by auto
    qed
    moreover have "\<forall>i<length ?ss. i \<noteq> ?j \<longrightarrow> ?ss!i = ?ts!i" (is "\<forall>i<length ?ss. _ \<longrightarrow> ?P i")
    proof (intro allI impI)
      fix i assume "i < length ?ss" and "i \<noteq> ?j"
      hence "i < length ss1 \<or> i > length ss1" by auto
      thus "?P i"
      proof
        assume "i < length ss1" thus "?P i" by (auto simp: nth_append)
      next
        assume "i > length ss1" thus "?P i"
          using `i < length ?ss` by (auto simp: nth_Cons' nth_append)
      qed
    qed
    ultimately show ?thesis by best
  qed
  ultimately show ?thesis by auto
qed

fun idx
where "idx f 0 = 0"
    | "idx f (Suc i) = Suc(f(idx f i))"

lemma Tinf_imp_SN_elt_nrrstep:
  assumes "t \<in> Tinf R" shows "SN_elt (nrrstep R) t"
proof (rule ccontr)
  assume "\<not> SN_elt (nrrstep R) t"
  then obtain S where "t = S 0" and nrseq: "\<forall>i. (S i, S (Suc i)) \<in> nrrstep R" by auto
  hence "\<forall>i. \<exists>f ss ts. S i = Fun f ss \<and> S (Suc i) = Fun f ts \<and> length ss = length ts
    \<and> (\<exists>j<length ss. (ss!j, ts!j) \<in> rstep R \<and> (\<forall>k<length ss. k \<noteq> j \<longrightarrow> ss!k = ts!k))"
    (is "\<forall>i. \<exists>f ss ts. ?P i f ss ts")
    using nrrstep_args by fast
  from choice[OF this] obtain f where "\<forall>i. \<exists>ss ts. ?P i (f i) ss ts" by best
  from choice[OF this] obtain ss where "\<forall>i. \<exists>ts. ?P i (f i) (ss i) ts" by best
  from choice[OF this] obtain ts where P: "\<forall>i. ?P i (f i) (ss i) (ts i)" by best
  moreover have "\<forall>i. f i = f (Suc i) \<and> ts i = ss (Suc i)"
  proof
    fix i
    from P have "S (Suc i) = Fun (f (Suc i)) (ss (Suc i))" by simp
    moreover from P have "S (Suc i) = Fun (f i) (ts i)" by blast+
    ultimately show "f i = f (Suc i) \<and> ts i = ss (Suc i)" by simp
  qed
  ultimately have "\<forall>i. ?P i (f i) (ss i) (ss (Suc i))" by auto
  hence "\<forall>i. length (ss i) = length (ss (Suc i))"
    and "\<forall>i. \<exists>j<length (ss i). (ss i ! j, ss (Suc i) ! j) \<in> rstep R
    \<and> (\<forall>k<length (ss i). k \<noteq> j \<longrightarrow> ss i ! k = ss (Suc i) ! k)" by blast+
  from choice[OF this(2)] obtain \<pi> where pi: "\<forall>i. \<pi> i < length (ss i)
    \<and> (ss i ! \<pi> i, ss (Suc i) ! \<pi> i) \<in> rstep R
    \<and> (\<forall>k<length (ss i). k \<noteq> \<pi> i \<longrightarrow> ss i ! k = ss (Suc i) ! k)" by blast+
  hence seq: "\<forall>i. (ss i ! \<pi> i, ss (Suc i) ! \<pi> i) \<in> rstep R" by simp
  have len: "\<forall>i. length (ss i) = length (ss 0)"
  proof
    fix i show "length (ss i) = length (ss 0)"
    proof (induct i)
      case 0 show ?case ..
    next
      case (Suc i) thus ?case using `\<forall>i. length (ss i) = length (ss (Suc i))` by simp
    qed
  qed
  have "\<forall>i. \<pi> i < length (ss 0)"
  proof
    fix i
    have "length (ss i) = length (ss 0)" using len by simp
    moreover from pi have "\<pi> i < length (ss i)" by simp
    ultimately show "\<pi> i < length (ss 0)" by simp
  qed
  hence "range \<pi> \<subseteq> {i. i < length (ss 0)}" by auto
  moreover have "finite {i. i < length (ss 0)}" by simp
  ultimately have "finite (range \<pi>)" by (rule finite_subset)
  from finite_range[OF this] obtain j where j: "\<forall>i. \<exists>k\<ge>i. \<pi> k = j" by (auto simp: INFM_nat_le)
  from choice[OF this] obtain q where q: "\<forall>i. q i \<ge> i \<and> \<pi> (q i) = j" by auto
  have j_len: "\<forall>i. j < length (ss i)"
  proof
    fix i
    from j obtain k where "k \<ge> i" and "\<pi> k = j" by best
    moreover from pi have "\<pi> k < length (ss k)" by simp
    ultimately have "j < length (ss k)" by simp
    moreover from len have "length (ss k) = length (ss 0)" by simp
    moreover from len have "length (ss i) = length (ss 0)" by simp
    ultimately show "j < length (ss i)" by simp
  qed
  from seq have step: "\<forall>i. (ss (q i) ! \<pi> (q i), ss (Suc (q i)) ! \<pi> (q i)) \<in> rstep R" by simp
  have refl_seq: "\<forall>i. (ss i ! j, ss (Suc i) ! j) \<in> (rstep R)^="
  proof
    fix i show "(ss i ! j, ss (Suc i) ! j) \<in> (rstep R)^="
    proof (cases "\<pi> i = j")
      assume "\<pi> i = j"
      moreover from pi have "(ss i ! \<pi> i, ss (Suc i) ! \<pi> i) \<in> rstep R" by simp
      ultimately show ?thesis by simp
    next
      assume "\<pi> i \<noteq> j"
      moreover from `\<forall>i. j < length (ss i)` have "j < length (ss i)" ..
      moreover from pi have "j < length (ss i) \<and> j \<noteq> \<pi> i \<longrightarrow> ss i ! j = ss (Suc i) ! j" by simp
      ultimately show ?thesis by simp
    qed
  qed
  let ?s = "\<lambda>i. ss (idx q i) ! j"
  have "\<forall>i. (ss i ! j, ss (Suc (q i)) ! j) \<in> (rstep R)^+"
  proof
    fix i
    from q have "q i \<ge> i" by simp
    from steps_imp_seq[OF refl_seq this] have "(ss i ! j, ss (q i) ! j) \<in> (rstep R)^*" by simp
    moreover from step have "(ss (q i) ! j, ss (Suc (q i)) ! j) \<in> rstep R" by (simp add: q)
    ultimately show "(ss i ! j, ss (Suc (q i)) ! j) \<in> (rstep R)^+" by simp
  qed
  hence "\<forall>i. (?s i, ?s (Suc i)) \<in> (rstep R)^+" by simp
  hence "\<not> SN_elt (rstep R) (?s 0)" using SN_elt_imp_SN_elt_trancl by best
  moreover have "?s 0 \<lhd> t"
  proof -
    from P have t: "t = Fun (f 0) (ss 0)" by (simp add: `t = S 0`)
    have "j < length (ss 0)" using j_len by simp
    hence "?s 0 \<in> set (ss 0)" by simp
    thus ?thesis unfolding t ..
  qed
  moreover from assms have "\<forall>s\<lhd>t. SN_elt (rstep R) s" unfolding Tinf_def by simp
  ultimately show False by simp
qed

hide const idx

lemma rrstep_imp_rule_subs:
 fixes R :: "('f,'v)trs" and s t :: "('f,'v)term"
 assumes "(s,t) \<in> rrstep R"
 shows "\<exists>l r (\<sigma>::('f,'v)sub). (l,r) \<in> R \<and> (l\<cdot>\<sigma>) = s \<and> (r\<cdot>\<sigma>) = t"
proof -
 have "ctxt_of_pos_term \<epsilon> s = Hole" by auto
 obtain l r \<sigma> where "(s,t) \<in> rstep_r_p_s R (l,r) \<epsilon> \<sigma>" using assms unfolding rrstep_def by best
 hence "let C = ctxt_of_pos_term \<epsilon> s in \<epsilon> \<in> poss s \<and> (l,r) \<in> R \<and> C\<langle>l\<cdot>\<sigma>\<rangle> = s \<and> C\<langle>r\<cdot>\<sigma>\<rangle> = t" unfolding rstep_r_p_s_def by simp
 with `ctxt_of_pos_term \<epsilon> s = Hole` have "(l,r) \<in> R" and "l\<cdot>\<sigma> = s" and "r\<cdot>\<sigma> = t" unfolding Let_def by auto
 thus ?thesis by auto
qed

lemma nrrstep_preserves_root:
 assumes "(Fun f ss,t) \<in> nrrstep R" (is "(?s,t) \<in> nrrstep R") shows "\<exists>ts. t = (Fun f ts)"
using assms unfolding nrrstep_def rstep_r_p_s_def Let_def by auto

lemma nrrstep_equiv_root: assumes "(s,t) \<in> nrrstep R" shows "\<exists>f ss ts. s = Fun f ss \<and> t = Fun f ts"
proof -
 from assms obtain f ss where "s = Fun f ss" using nrrstep_imp_Fun by best
 with assms obtain ts where "t = Fun f ts" using nrrstep_preserves_root by best
 from `s = Fun f ss` and `t = Fun f ts` show ?thesis by best
qed

lemma nrrstep_reflects_root:
 assumes "(s,Fun g ts) \<in> nrrstep R" (is "(s,?t) \<in> nrrstep R")
 shows "\<exists>ss. s = (Fun g ss)"
proof -
 from assms obtain f ss ts' where "s = Fun f ss" and "Fun g ts = Fun f ts'" using nrrstep_equiv_root by best
 hence "f = g" by simp
 with `s = Fun f ss` have "s = Fun g ss" by simp
 thus ?thesis by auto
qed

lemma nrrstep_seq_preserves_root:
 assumes "(Fun f ss,t) \<in> (nrrstep R)^*"
 shows "\<exists>ts. t = (Fun f ts)"
using assms by induct (auto simp: nrrstep_preserves_root)

lemma nrrstep_non_var_imp_arg_seq:
 fixes ss :: "('f,'v) term list"
 assumes "(Fun f ss,Fun f ts) \<in> nrrstep R" (is "(?s,?t) \<in> nrrstep R") and "i < length ss"
 shows "(ss!i,ts!i) \<in> (rstep R)^*"
proof -
 from assms have "<i> \<in> poss ?s" using ep_pos_term by simp
 from `(?s,?t) \<in> nrrstep R`
 obtain l r :: "('f,'v)term" and  j :: nat and ps :: pos and \<sigma> :: "('f,'v)sub"
 where A: "let C = ctxt_of_pos_term (j<#ps) ?s in (j<#ps) \<in> poss ?s \<and> (l,r) \<in> R \<and> C\<langle>l\<cdot>\<sigma>\<rangle> = ?s \<and> C\<langle>r\<cdot>\<sigma>\<rangle> = ?t" unfolding nrrstep_def rstep_r_p_s_def by force
 let ?C = "ctxt_of_pos_term (j<#ps) ?s"
 from A have "(j<#ps) \<in> poss ?s" and "(l,r) \<in> R" and "?C\<langle>l\<cdot>\<sigma>\<rangle> = ?s" and "?C\<langle>r\<cdot>\<sigma>\<rangle> = ?t" using Let_def by auto
 have C: "?C = More f (take j ss) (ctxt_of_pos_term ps (ss!j)) (drop (Suc j) ss)" (is "?C = More f ?ss1 ?D ?ss2") by auto
 from `?C\<langle>l\<cdot>\<sigma>\<rangle> = ?s` have "?D\<langle>l\<cdot>\<sigma>\<rangle> = (ss!j)" by (auto simp: take_drop_imp_nth)
 from `(l,r) \<in> R` have "(l\<cdot>\<sigma>,r\<cdot>\<sigma>) \<in> (CS R)" by auto
 hence "(?D\<langle>l\<cdot>\<sigma>\<rangle>,?D\<langle>r\<cdot>\<sigma>\<rangle>) \<in> (CC(CS R))" and "(?C\<langle>l\<cdot>\<sigma>\<rangle>,?C\<langle>r\<cdot>\<sigma>\<rangle>) \<in> (CC(CS R))" by (auto simp only: closure_ctxt.ccI)
 hence D_rstep: "(?D\<langle>l\<cdot>\<sigma>\<rangle>,?D\<langle>r\<cdot>\<sigma>\<rangle>) \<in> rstep R" and "(?C\<langle>l\<cdot>\<sigma>\<rangle>,?C\<langle>r\<cdot>\<sigma>\<rangle>) \<in> rstep R" by auto
 from `?C\<langle>r\<cdot>\<sigma>\<rangle> = ?t` and C have "?t = Fun f (take j ss @ ?D\<langle>r\<cdot>\<sigma>\<rangle> # drop (Suc j) ss)" by auto
 hence ts: "ts = (take j ss @ ?D\<langle>r\<cdot>\<sigma>\<rangle> # drop (Suc j) ss)" by auto
 have "j = i \<or> j \<noteq> i" by simp
 from `j<#ps \<in> poss ?s` have "j < length ss" by simp
 hence "(take j ss @ ?D\<langle>r\<cdot>\<sigma>\<rangle> # drop (Suc j) ss) ! j = ?D\<langle>r\<cdot>\<sigma>\<rangle>" by (auto simp: nth_append_take)
 with ts have "ts!j = ?D\<langle>r\<cdot>\<sigma>\<rangle>" by auto
 have "j = i \<or> j \<noteq> i" by simp
 thus "(ss!i,ts!i) \<in> (rstep R)^*"
 proof
  assume "j = i"
  with `ts!j = ?D\<langle>r\<cdot>\<sigma>\<rangle>` and `?D\<langle>l\<cdot>\<sigma>\<rangle> = ss!j` and D_rstep show ?thesis by auto
 next
  assume "j \<noteq> i"
  with `i < length ss` and `j < length ss` have "(take j ss @ ?D\<langle>r\<cdot>\<sigma>\<rangle> # drop (Suc j) ss) ! i = ss!i" by (auto simp: nth_append_take_drop_is_nth_conv)
  with ts have "ts!i = ss!i" by auto
  thus ?thesis by auto
 qed
qed

lemma nrrstep_imp_arg_rstep:
 assumes "(s,t) \<in> nrrstep R" and "i < num_args s" shows "(args s!i,args t!i) \<in> (rstep R)^*"
proof (cases s)
 fix x assume "s = Var x" thus ?thesis using assms by auto
next
 fix f ss assume "s = Fun f ss"
 hence "(Fun f ss,t) \<in> nrrstep R" using assms by simp
 then obtain ts where "t = Fun f ts" using nrrstep_preserves_root by best
 with `(s,t) \<in> nrrstep R` and `s = Fun f ss` have "(Fun f ss,Fun f ts) \<in> nrrstep R" by simp
 from  `s = Fun f ss` and `i < num_args s` have "i < length ss" by simp
 with `(Fun f ss,Fun f ts) \<in> nrrstep R` have "(ss!i,ts!i) \<in> (rstep R)^*" by (rule nrrstep_non_var_imp_arg_seq)
 with `s = Fun f ss` and `t = Fun f ts` show ?thesis by simp
qed

lemma nrrstep_rtrancl_imp_rstep_rtrancl: "(s,t) \<in> (nrrstep R)^* \<Longrightarrow> (s,t) \<in> (rstep R)^*"
proof (induct rule: rtrancl.induct)
 case (rtrancl_refl a) thus ?case by simp
next
 case (rtrancl_into_rtrancl a b c)
 hence IH: "(a,b) \<in> (rstep R)^*" and nrrstep: "(b,c) \<in> nrrstep R" by auto
 from nrrstep have "(b,c) \<in> rstep R" using nrrstep_imp_rstep by auto
 with IH show ?case by auto
qed

lemma nrrstep_non_var_preserves_num_args:
 fixes R :: "('f,'v)trs"
 assumes "(Fun f ss,Fun f ts) \<in> nrrstep R" (is "(?s,?t) \<in> nrrstep R")
 shows "length ss = length ts"
proof -
 from assms obtain l r :: "('f,'v)term" and i and ps :: pos and \<sigma> :: "('f,'v)sub" where "let C = ctxt_of_pos_term (i<#ps) ?s in (i<#ps) \<in> poss ?s \<and> (l,r) \<in> R \<and> C\<langle>l\<cdot>\<sigma>\<rangle> = ?s \<and> C\<langle>r\<cdot>\<sigma>\<rangle> = ?t" (is "let C = ?C in _")
 unfolding nrrstep_def rstep_r_p_s_def by force
 hence "(l,r) \<in> R" and Cl: "?C\<langle>l\<cdot>\<sigma>\<rangle> = ?s" and Cr: "?C\<langle>r\<cdot>\<sigma>\<rangle> = ?t" unfolding Let_def by auto
 have C: "?C = More f (take i ss) (ctxt_of_pos_term ps (ss!i)) (drop (Suc i) ss)" (is "?C = More f ?ss1 ?D ?ss2") by simp
 from C and Cl have s: "?s = Fun f (take i ss @ ?D\<langle>l\<cdot>\<sigma>\<rangle> # drop (Suc i) ss)" (is "?s = Fun f ?ss") by simp
 from C and Cr have t: "?t = Fun f (take i ss @ ?D\<langle>r\<cdot>\<sigma>\<rangle> # drop (Suc i) ss)" (is "?t = Fun f ?ts") by simp
 from s and t have ss: "ss = ?ss" and ts: "ts = ?ts" by auto
 have "length ?ss = length ?ts" by auto
 with ss and ts show ?thesis by simp
qed

lemma nrrstep_equiv_num_args:
 assumes "(s,t) \<in> nrrstep R" shows "num_args s = num_args t"
proof -
 from assms obtain f ss ts where s:"s = Fun f ss" and t:"t = Fun f ts" using nrrstep_equiv_root by best
 with assms have "(Fun f ss,Fun f ts) \<in> nrrstep R" by simp
 hence "length ss = length ts" by (rule nrrstep_non_var_preserves_num_args)
 with s and t show ?thesis by auto
qed
 
lemma nrrstep_preserves_num_args:
 assumes "(s,t) \<in> nrrstep R" and "i < num_args s" shows "i <  num_args t"
proof (cases s)
 fix x assume "s = Var x" thus ?thesis using assms by auto
next
 fix f ss assume "s = Fun f ss"
 with assms obtain ts where "t = Fun f ts" using nrrstep_preserves_root by best
 from `(s,t) \<in> nrrstep R` have "length ss = length ts" unfolding `s = Fun f ss` and `t = Fun f ts` by (rule nrrstep_non_var_preserves_num_args)
 with assms and `s = Fun f ss` and `t = Fun f ts` show ?thesis by auto
qed

lemma nrrstep_reflects_num_args:
 assumes "(s,t) \<in> nrrstep R" and "i < num_args t" shows "i < num_args s"
proof (cases t)
 fix x assume "t = Var x" thus ?thesis using assms by auto
next
 fix g ts assume "t = Fun g ts"
 with assms obtain ss where "s = Fun g ss" using nrrstep_reflects_root by best
 from `(s,t) \<in> nrrstep R` have "length ss = length ts" unfolding `s = Fun g ss` and `t = Fun g ts` by (rule nrrstep_non_var_preserves_num_args)
 with assms and `s = Fun g ss` and `t = Fun g ts` show ?thesis by auto
qed

lemma rtrancl_nrrstep_imp_arg_rstep: 
 assumes "(s,t) \<in> (nrrstep R)^*" and "i < num_args s"
 shows "(args s!i,args t!i) \<in> (rstep R)^*"
using assms
proof (induct rule: rtrancl.induct)
 case (rtrancl_refl a) thus ?case by auto
next
 case (rtrancl_into_rtrancl a b c)
 hence IH: "(args a!i,args b!i) \<in> (rstep R)^*" by auto
 from `(a,b) \<in> (nrrstep R)^*` and `i < num_args a` have "i < num_args b" by induct (auto simp: nrrstep_preserves_num_args)
 with `(b,c) \<in> nrrstep R` have "(args b!i,args c!i) \<in> (rstep R)^*" by (rule nrrstep_imp_arg_rstep)
 with IH show ?case by simp
qed

definition wf_trs :: "('f,'v)trs \<Rightarrow> bool"
where
 "wf_trs R = (\<forall>l r. (l,r) \<in> R \<longrightarrow> (\<exists>f ts. l = Fun f ts) \<and> vars_term r \<subseteq> vars_term l)"

lemma wf_trs_imp_lhs_Fun:"wf_trs R \<Longrightarrow> (l,r) \<in> R \<Longrightarrow> \<exists>f ts. l = Fun f ts" unfolding wf_trs_def by blast

lemma rstep_imp_Fun:
 assumes "wf_trs R"
 shows "(s, t) \<in> rstep R \<Longrightarrow> \<exists>f ss. s = Fun f ss"
proof -
  assume "(s, t) \<in> rstep R"
  then obtain l r p \<sigma> where "(s, t) \<in> rstep_r_p_s R (l, r) p \<sigma>"
    using rstep_iff_rstep_r_p_s by best
  hence "(ctxt_of_pos_term p s)\<langle>l \<cdot> \<sigma>\<rangle> = s" (is "?C\<langle>l \<cdot> \<sigma>\<rangle> = s") and "(l, r) \<in> R"
    unfolding rstep_r_p_s_def Let_def by auto
  with `wf_trs R` obtain f ss where "l = Fun f ss" unfolding wf_trs_def by best
  from `?C\<langle>l \<cdot> \<sigma>\<rangle> = s` obtain C where "C\<langle>l \<cdot> \<sigma>\<rangle> = s" by best
  hence "C\<langle>Fun f (map (\<lambda>x. x \<cdot> \<sigma>) ss)\<rangle> = s" unfolding `l = Fun f ss` by simp
  thus "\<exists>f ss. s = Fun f ss" by (induct C) auto
qed

lemma suptp_imp_arg_or_suptp_of_arg:
 assumes "Fun f ss \<rhd> t"
 shows "t \<in> set ss \<or> (\<exists>s \<in> set ss. s \<rhd> t)"
using assms by (rule suptp.cases) auto

lemma var_imp_supteqp:
 "x \<in> vars_term(t::('f,'v)term) \<Longrightarrow> t \<unrhd> (Var x)" and "x \<in> vars_term((Fun f ss)::('f,'v)term) \<Longrightarrow> Fun f ss \<unrhd> (Var x)"
proof (induct t and ss rule: term.inducts)
 case (Var x) thus ?case using assms by auto
next
 case (Fun g ts)
 from `x \<in> vars_term(Fun g ts)` have "x \<in> vars_term(Fun f ts)" by simp
 with Fun have "Fun f ts \<unrhd> Var x" by simp
 have "Fun f ts \<noteq> Var x" by auto
 with `Fun f ts \<unrhd> Var x` have "Fun f ts \<rhd> Var x" by (cases rule: supteqp_cases) auto
 hence "Fun g ts \<rhd> Var x" by (cases rule: suptp.cases) auto
 thus ?case unfolding supt_iff_supteq_neq by auto
next
 case Nil_term thus ?case by auto
next
 case (Cons_term u us)
 hence "x \<in> vars_term u \<or>  x \<in> \<Union>(set (map vars_term us))" by auto
 thus ?case
 proof
  assume "x \<in> vars_term u" with Cons_term show ?case by auto
 next
  assume "x \<in> \<Union>(set (map vars_term us))" 
  hence "x \<in> vars_term(Fun f us)" by auto
  with Cons_term have "Fun f us \<unrhd> Var x" by auto
  have "Fun f us \<noteq> Var x" by auto
  with `Fun f us \<unrhd> Var x` have "Fun f us \<rhd> Var x" by (cases rule: supteqp_cases) auto
  hence "Var x \<in> set us \<or> (\<exists>u \<in> set us. u \<rhd> Var x)" by (rule suptp.cases) auto
  thus ?case
  proof
   assume "Var x \<in> set us" hence "Var x \<in> set (u#us)" by simp
   thus ?thesis by auto
  next
   assume "\<exists>u \<in> set us. u \<rhd> Var x"
   then obtain v where "v \<in> set us" and "v \<rhd> Var x" by auto
   hence "v \<in> set (u#us)" by auto
   from `v \<rhd> Var x` have "v \<unrhd> Var x" using supteqp_iff_suptp_or_eq by auto
   with `v \<in> set (u#us)` show ?thesis by auto
  qed
 qed
qed

lemma suptp_Fun_imp_arg_supteqp: assumes "Fun f ss \<rhd> t" shows "\<exists>s \<in> set ss. s \<unrhd> t"
using assms
proof (cases rule: suptp.cases)
 fix s ts g assume "Fun f ss = Fun g ts" and "t = s" and "(s :: ('a, 'b) term) \<in> set ts"
 thus ?thesis by auto
next
 fix s u ts g assume "Fun f ss = Fun g ts" and "t = u" and "s \<rhd> u" and "(s :: ('a, 'b) term) \<in> set ts"
 thus ?thesis unfolding supt_iff_supteq_neq by auto
qed

lemma not_SN_subt_imp_not_SN:
 fixes R :: "('f,'v)trs" and t :: "('f,'v)term"
 assumes "\<not> SN_elt (rstep R) t" and "s \<unrhd> t"
 shows "\<not> SN_elt (rstep R) s"
proof -
 from assms obtain A :: "nat \<Rightarrow> ('f,'v)term"  where "A 0 = t" and "\<forall>i. (A i,A(Suc i)) \<in> rstep R" unfolding SN_elt_def by best
 from `s \<unrhd> t` obtain C where "s = C\<langle>t\<rangle>" using supteqp_imp_ctxt by best
 let ?B = "\<lambda>i. C\<langle>A i\<rangle>"
 from `A 0 = t` and `s = C\<langle>t\<rangle>` have "?B 0 = s" by simp
 have "\<forall>i. (?B i,?B(Suc i)) \<in> rstep R"
 proof
  fix i
  from `\<forall>i. (A i,A(Suc i)) \<in> rstep R` have "(A i,A(Suc i)) \<in> rstep R" by simp
  hence "(?B i,?B(Suc i)) \<in> CC(rstep R)" by (rule ccI)
  thus "(?B i,?B(Suc i)) \<in> rstep R" by auto
 qed
 with `?B 0 = s` show ?thesis unfolding SN_elt_def by best
qed

lemma SN_instance_imp_SN_var: 
 fixes \<sigma>::"('f,'v)sub" and R::"('f,'v)trs" and t :: "('f,'v)term"
 assumes "SN_elt (rstep R) (t\<cdot>\<sigma>)" and "x \<in> vars_term t"
 shows "SN_elt (rstep R) (Var x\<cdot>\<sigma>)"
proof (rule ccontr)
 assume "\<not> SN_elt (rstep R) (Var x\<cdot>\<sigma>)"
 then obtain A :: "nat \<Rightarrow> ('f,'v)term"
  where "A 0 = (Var x\<cdot>\<sigma>)" and "\<forall>i. (A i,A(Suc i)) \<in> rstep R"
  using SN_elt_def[where r = R and x = "Var x\<cdot>\<sigma>"] by fast
 from `x \<in> vars_term t` have "t \<unrhd> Var x" by (rule var_imp_supteqp)
 hence "t\<cdot>\<sigma> \<unrhd> Var x\<cdot>\<sigma>" by (rule supteqp_stable)
 then obtain C where "C\<langle>Var x\<cdot>\<sigma>\<rangle> = t\<cdot>\<sigma>" using supteqp_imp_ctxt by best
 with `A 0 = (Var x\<cdot>\<sigma>)`[symmetric] have "t\<cdot>\<sigma> = C\<langle>A 0\<rangle>" by simp
 have "\<forall>i. (C\<langle>A i\<rangle>,C\<langle>A (Suc i)\<rangle>) \<in> rstep R"
 proof
  fix i
  from `\<forall>i. (A i,A(Suc i)) \<in> rstep R` have "(A i,A(Suc i)) \<in> rstep R" by simp
  hence "(C\<langle>A i\<rangle>,C\<langle>A (Suc i)\<rangle>) \<in> CC(rstep R)" by (rule ccI)
  thus "(C\<langle>A i\<rangle>,C\<langle>A (Suc i)\<rangle>) \<in> rstep R" by auto
 qed
 with `t\<cdot>\<sigma> = C\<langle>A 0\<rangle>` have "C\<langle>A 0\<rangle> = t\<cdot>\<sigma> \<and> (\<forall>i. (C\<langle>A i\<rangle>,C\<langle>A (Suc i)\<rangle>) \<in> rstep R)" by auto
 hence "\<exists>B. B 0 = t\<cdot>\<sigma> \<and> (\<forall>i. (B i,B(Suc i)) \<in> rstep R)" by auto
 hence "\<not> SN_elt (rstep R) (t\<cdot>\<sigma>)" unfolding SN_elt_def by best
 with `SN_elt (rstep R) (t\<cdot>\<sigma>)` show False by simp
qed

lemma var_imp_var_of_arg:
 fixes ss::"('f,'v)term list"
 assumes "x \<in> vars_term(Fun f ss)" (is "x \<in> vars_term(?s::('f,'v)term)")
 shows "\<exists>i < num_args(Fun f ss). x \<in> vars_term(ss!i)"
proof -
 from assms have "x \<in> \<Union>set (map vars_term ss)" by simp
 hence "x \<in> (\<Union>i<length ss. vars_term(ss!i))" unfolding set_conv_nth by auto
 hence "\<exists>i<length ss. x \<in> vars_term(ss!i)" using UN_iff by best
 then obtain i where "i < length ss" and "x \<in> vars_term(ss!i)" by auto
 hence "i < num_args ?s" by simp
 with `x \<in> vars_term(ss!i)` show ?thesis by auto
qed

lemma subt_instance_and_not_subs_imp_subt:
 fixes \<sigma> :: "('f,'v)sub"
 shows "s\<cdot>\<sigma> \<unrhd> t \<Longrightarrow> \<forall>x \<in> vars_term s. \<not>((Var x)\<cdot>\<sigma> \<unrhd> t) \<Longrightarrow> \<exists>u. s \<unrhd> u \<and> t = u\<cdot>\<sigma>"
proof (induct s arbitrary: t rule: term_induct)
 case (Var x) thus ?case by auto
next
 case (Fun f ss)
 from `Fun f ss\<cdot>\<sigma> \<unrhd> t` have"(Fun f ss\<cdot>\<sigma> = t) \<or> (Fun f ss\<cdot>\<sigma> \<rhd> t)" using supteqp_iff_suptp_or_eq by auto
 thus ?case
 proof
  assume "Fun f ss\<cdot>\<sigma> = t" with Fun show ?thesis by auto
 next
  assume "Fun f ss\<cdot>\<sigma> \<rhd> t"
  hence "Fun f (map (\<lambda>t. t\<cdot>\<sigma>) ss) \<rhd> t" by simp
  hence "\<exists>s \<in> set (map (\<lambda>t. t\<cdot>\<sigma>) ss). s \<unrhd> t" (is "\<exists>s \<in> set ?ss'. s \<unrhd> t") by (rule suptp_Fun_imp_arg_supteqp)
  then obtain s' where "s' \<in> set ?ss'" and "s' \<unrhd> t" by best
  hence "\<exists>i < length ?ss'. ?ss'!i = s'" using in_set_conv_nth[where ?x = "s'"] by best
  then obtain i where "i < length ?ss'" and "?ss'!i = s'" by best
  hence "?ss'!i = (ss!i)\<cdot>\<sigma>" by auto
  from `?ss'!i = s'` have "s' = (ss!i)\<cdot>\<sigma>" unfolding `?ss'!i = (ss!i)\<cdot>\<sigma>` by simp
  from `s' \<unrhd> t` have "(ss!i)\<cdot>\<sigma> \<unrhd> t" unfolding `s' = (ss!i)\<cdot>\<sigma>` by simp
  with `i < length ?ss'` have "(ss!i) \<in> set ss" by auto
  with `(ss!i)\<cdot>\<sigma> \<unrhd> t` have "\<exists>s \<in> set ss. s\<cdot>\<sigma> \<unrhd> t" by best
  then obtain s where "s \<in> set ss" and "s\<cdot>\<sigma> \<unrhd> t" by best
  with Fun have "\<forall>x \<in> vars_term s. \<not>((Var x)\<cdot>\<sigma> \<unrhd> t)" by force
  from Fun 
  have IH: "s \<in> set ss \<longrightarrow> (\<forall>v. s\<cdot>\<sigma> \<unrhd> v \<longrightarrow> (\<forall>x \<in> vars_term s. \<not> Var x\<cdot>\<sigma> \<unrhd> v) \<longrightarrow> (\<exists>u. s \<unrhd> u \<and> v = u\<cdot>\<sigma>))"
  by auto
  with `s \<in> set ss` have "!!v. s\<cdot>\<sigma> \<unrhd> v \<Longrightarrow> (\<forall>x \<in> vars_term s. \<not> Var x\<cdot>\<sigma> \<unrhd> v) \<longrightarrow> (\<exists>u. s \<unrhd> u \<and> v = u\<cdot>\<sigma>)"
  by simp
  with `s\<cdot>\<sigma> \<unrhd> t` have "(\<forall>x \<in> vars_term s. \<not> Var x\<cdot>\<sigma> \<unrhd> t) \<longrightarrow> (\<exists>u. s \<unrhd> u \<and> t = u\<cdot>\<sigma>)" by simp
  with `\<forall>x \<in> vars_term s. \<not> Var x\<cdot>\<sigma> \<unrhd> t` obtain u where "s \<unrhd> u" and "t = u\<cdot>\<sigma>" by best
  with `s \<in> set ss` have "Fun f ss \<unrhd> u" by auto
  with `t = u\<cdot>\<sigma>` show ?thesis by best
 qed
qed

lemma SN_imp_SN_subt:
 fixes R::"('f,'v)trs" and t::"('f,'v)term"
 assumes "SN_elt (rstep R) s" and "s \<unrhd> t"
 shows "SN_elt (rstep R) t"
proof (rule ccontr)
 assume "\<not> SN_elt (rstep R) t"
 then obtain A::"nat \<Rightarrow> ('f,'v)term" where "A 0 = t" and "\<forall>i. (A i,A (Suc i)) \<in> rstep R"
  unfolding SN_elt_def[where r = "rstep R" and x = "t"] by best
 from `s \<unrhd> t` obtain C where "s = C\<langle>t\<rangle>" using supteqp_imp_ctxt by force
 let ?B = "\<lambda>i. C\<langle>A i\<rangle>"
 have "\<forall>i. (?B i,?B(Suc i)) \<in> rstep R"
 proof
  fix i
  from `\<forall>i. (A i,A(Suc i)) \<in> rstep R` have "(?B i,?B(Suc i)) \<in> CC(rstep R)" using ccI by fast
  thus "(?B i,?B(Suc i)) \<in> rstep R" by auto
 qed
 with `A 0 = t` have "?B 0 = s \<and> (\<forall>i. (?B i,?B(Suc i)) \<in> rstep R)" unfolding `s = C\<langle>t\<rangle>` by auto
 hence "\<not> SN_elt (rstep R) s" unfolding SN_elt_def by auto
 with assms show "False" by simp
qed

lemma var_subt_imp_subt_in_vars: "r \<unrhd> Var x \<Longrightarrow> x \<in> vars_term r"
proof (induct r rule: term_induct)
 case (Var y) hence "x = y" by (cases rule: supteqp.cases) auto
 thus ?case by simp
next
 case (Fun f ss)
 from `Fun f ss \<unrhd> Var x` have "(Fun f ss = Var x) \<or> (Fun f ss \<rhd> Var x)"
 using supteqp_iff_suptp_or_eq by auto
 thus ?case
 proof
  assume "Fun f ss = Var x" thus ?thesis by auto
 next
  assume "Fun f ss \<rhd> Var x"
  then obtain s where "s \<in> set ss" and "s \<unrhd> Var x" using suptp_Fun_imp_arg_supteqp by best
  with Fun have "s \<unrhd> Var x \<Longrightarrow> x \<in> vars_term s" by auto
  with `s \<unrhd> Var x` have "x \<in> vars_term s" by simp
  with `s \<in> set ss` show ?thesis by auto
 qed
qed

lemma SN_args_imp_SN_subt:
 fixes R::"('f,'v)trs" and ss::"('f,'v)term list"
 assumes "!!i. i < num_args (Fun f ss) \<Longrightarrow> SN_elt (rstep R) (ss!i)"
 shows "\<forall>s. (Fun f ss) \<rhd> s \<longrightarrow> SN_elt (rstep R) s"
proof
 fix s show "(Fun f ss) \<rhd> s \<longrightarrow> SN_elt (rstep R) s"
 proof
  assume "Fun f ss \<rhd> s" (is "?t \<rhd> s")
  hence "\<exists>u \<in> set ss. u \<unrhd> s" using suptp_Fun_imp_arg_supteqp by best
  then obtain u where "u \<in> set ss" and "u \<unrhd> s" by best
  hence "\<exists>i < length ss. ss!i = u" unfolding in_set_conv_nth by simp
  then obtain i where "i < length ss" and "ss!i = u" by best
  hence "i < num_args ?t" by simp
  with assms have "SN_elt (rstep R) (ss!i)" by simp
  with `u \<unrhd> s` show "SN_elt (rstep R) s" unfolding `ss!i = u`[symmetric]
   using SN_imp_SN_subt[where R = R and s = "ss!i"] by auto
 qed
qed

lemma SN_imp_SN_arg : fixes R :: "('f,'v)trs" and t :: "('f,'v) term"
  assumes sn: "SN_elt (rstep R) (Fun f ts)" and arg: "t \<in> set ts" shows "SN_elt (rstep R) t"
proof -
  from arg have "Fun f ts \<unrhd> t" by auto
  with sn show ?thesis by (rule SN_imp_SN_subt)
qed

lemma SNinstance_imp_SN : fixes R :: "('f,'v)trs" and t :: "('f,'v)term"
  assumes "SN_elt (rstep R) (t \<cdot> \<sigma>)" 
  shows "SN_elt (rstep R) t"
proof(rule ccontr)
  assume "\<not> SN_elt (rstep R) t"
  hence "\<not> SN_elt (rstep R) (t \<cdot> \<sigma>)" 
  proof (simp add: SN_elt_def)
    assume "\<exists> f. f 0 = t \<and> (\<forall> i. (f i, f (Suc i)) \<in> rstep R)"
    from this obtain f where prem: "f 0 = t \<and> (\<forall> i. (f i, f (Suc i)) \<in> rstep R)" ..
    let ?g = "\<lambda> i. (f i) \<cdot> \<sigma>"
    from prem have "?g 0 = t \<cdot> \<sigma> \<and> (\<forall> i. (?g i, ?g (Suc i)) \<in> rstep R)" using rstep_sub_closed 
      unfolding sub_closed_def by auto 
    thus "\<exists> f. f 0 = t \<cdot> \<sigma> \<and> (\<forall> i. (f i, f (Suc i)) \<in> rstep R)" by auto
  qed
  with assms show "False" by simp
qed


lemma first_R_step:
 fixes S::"'a brel" and R::"'a brel"
 assumes step: "\<And>i. (t i, t (Suc i)) \<in> (R \<union> S)" and "SN_elt S (t 0)"
 shows "\<exists>i. (t i, t (Suc i)) \<in> R \<and> (\<forall>j<i. (t j, t (Suc j)) \<in> S \<and> \<not>((t j, t (Suc j)) \<in> R))"
proof -
  from `SN_elt S (t 0)` have "\<And>s. s 0 \<noteq> t 0 \<or> (\<exists>i. \<not>(s i,s (Suc i)) \<in> S)"
    unfolding SN_elt_def by simp
  hence "\<exists>i. \<not>(t i,t(Suc i)) \<in> S" by auto
  then obtain i where "\<not>(t i,t(Suc i)) \<in> S" by auto
  with step have "(t i,t(Suc i)) \<in> R" (is "?P i") by auto
  hence "?P (Least ?P)" by (rule LeastI)
  let ?i = "Least ?P"
  have no_R_step_before: "\<forall>j<?i. \<not>(t j,t(Suc j)) \<in> R" using not_less_Least by auto
  with step have "\<forall>j<?i. (t j,t(Suc j)) \<in> S" by best
  with no_R_step_before have "\<forall>j<?i. (t j,t(Suc j)) \<in> S \<and> \<not>(t j,t(Suc j)) \<in> R" by best
  with `?P ?i` show ?thesis by best
qed

lemma rstep_imp_C_s_r:
 fixes R :: "('f,'v)trs"
 assumes "(s,t) \<in> rstep R"
 shows "\<exists>C \<sigma> l r. (l,r) \<in> R \<and> s = C\<langle>l\<cdot>\<sigma>\<rangle> \<and> t = C\<langle>r\<cdot>\<sigma>\<rangle>"
proof -
 from assms obtain l r p \<sigma> where step:"(s,t) \<in> rstep_r_p_s R (l,r) p \<sigma>"
  unfolding rstep_iff_rstep_r_p_s by best
 let ?C = "ctxt_of_pos_term p s"
 from step have "p \<in> poss s" and "(l,r) \<in> R" and "?C\<langle>l\<cdot>\<sigma>\<rangle> = s" and "?C\<langle>r\<cdot>\<sigma>\<rangle> = t"
  unfolding rstep_r_p_s_def Let_def by auto
 hence "(l,r) \<in> R \<and> s = ?C\<langle>l\<cdot>\<sigma>\<rangle> \<and> t = ?C\<langle>r\<cdot>\<sigma>\<rangle>" by auto
 thus ?thesis by force
qed

lemma rstep_D[elim]:
  assumes "(s, t) \<in> rstep R"
  obtains l r C \<sigma>
    where "(l, r) \<in> R" and "s = C\<langle>l\<cdot>\<sigma>\<rangle>" and "t = C\<langle>r\<cdot>\<sigma>\<rangle>"
using rstep_imp_C_s_r[OF assms] by auto

lemma SN_Var:
 assumes "wf_trs R" shows "SN_elt (rstep R) (Var x)"
proof (rule ccontr)
  assume "\<not> ?thesis"
  then obtain S where [symmetric]: "S 0 = Var x"
    and seq: "\<forall>i. (S i, S (Suc i)) \<in> rstep R" by auto
  hence "(Var x, S (Suc 0)) \<in> rstep R" by simp
  then obtain C l r \<sigma> where "(l, r) \<in> R" and x: "Var x = C\<langle>l\<cdot>\<sigma>\<rangle>" by best
  hence "Var x = l\<cdot>\<sigma>" by (induct C) simp_all
  then obtain y where "l = Var y" by (induct l) simp_all
  with assms and `(l, r) \<in> R` show False unfolding wf_trs_def by blast
qed

lemma Tinf_lemma:
  assumes "wf_trs R" and "t \<in> Tinf R"
  shows "\<exists>l r u \<sigma>.
    (l, r) \<in> R \<and> 
    r \<unrhd> u \<and> 
    \<not> is_var u \<and> 
    (t, l\<cdot>\<sigma>) \<in> (nrrstep R)^* \<and> 
    (l\<cdot>\<sigma>, r\<cdot>\<sigma>) \<in> rrstep R \<and> 
    r\<cdot>\<sigma> \<unrhd> u\<cdot>\<sigma> \<and>
    l\<cdot>\<sigma> \<in> Tinf R \<and>
    u\<cdot>\<sigma> \<in> Tinf R \<and>
    \<not>(l\<cdot>\<sigma> \<rhd> u\<cdot>\<sigma>)"
proof (cases t)
  case (Var x)
  from `t \<in> Tinf R` have "\<not> SN_elt (rstep R) (Var x)" unfolding Var Tinf_def by simp
  with SN_Var[OF `wf_trs R`] show ?thesis by simp
next
  case (Fun f ts)
  let ?t = "Fun f ts"
  from `t \<in> Tinf R` have "\<forall>s\<lhd>?t. SN_elt (rstep R) s"
    and "\<not> SN_elt (rstep R) ?t" by (auto simp: Tinf_def Fun)
  then obtain S where "S 0 = ?t" and seq: "\<forall>i. (S i, S (Suc i)) \<in> rstep R" by best
  have "\<exists>i. (S 0, S i) \<in> (nrrstep R)^* \<and> (S i, S (Suc i)) \<in> rrstep R"
  proof -
    from `t \<in> Tinf R` have "SN_elt (nrrstep R) (S 0)"
      by (simp add: Tinf_imp_SN_elt_nrrstep `S 0 = ?t` Fun)
    with seq first_R_step[of S "rrstep R" "nrrstep R"]
      obtain j where first_step: "(S j, S (Suc j)) \<in> rrstep R"
      and nrsteps: "\<forall>i<j. (S i, S (Suc i)) \<in> nrrstep R"
      by (auto simp: rstep_iff_rrstep_or_nrrstep)
    from nrsteps have "(S 0, S j) \<in> (nrrstep R)^*" by (induct j) auto
    with first_step show ?thesis by auto
  qed
  then obtain j where nrseq: "(S 0, S j) \<in> (nrrstep R)^*"
    and rrstep: "(S j, S (Suc j)) \<in> rrstep R" by auto

  from rrstep_imp_rule_subs[OF rrstep]
    obtain l r \<sigma> where "(l, r) \<in> R" and "l\<cdot>\<sigma> = S j" and "r\<cdot>\<sigma> = S (Suc j)" by auto

  have "\<not> SN_elt (rstep R) (r\<cdot>\<sigma>)"
  proof -
    let ?T = "\<lambda>i. S (Suc j + i)"
    from `r\<cdot>\<sigma> = S (Suc j)` have "?T 0 = r\<cdot>\<sigma>" by simp
    moreover from `\<forall>i. (S i, S (Suc i)) \<in> rstep R` have "\<forall>i. (?T i, ?T (Suc i)) \<in> rstep R"
      by simp
    ultimately show ?thesis unfolding SN_defs by best
  qed
  then obtain t' where "r\<cdot>\<sigma> \<unrhd> t'" and "t' \<in> Tinf R" using not_SN_imp_subt_Tinf by auto

  from nrseq have "(?t, S j) \<in> (nrrstep R)^*" by (simp add: `S 0 = ?t`)
  moreover from nrrstep_seq_preserves_root[OF this] obtain ls'
    where "S j = Fun f ls'" by auto
  ultimately have "l\<cdot>\<sigma> = Fun f ls'" and "(?t, Fun f ls') \<in> (nrrstep R)^*"
    by (auto simp: `l\<cdot>\<sigma> = S j`)
  hence "(?t, l\<cdot>\<sigma>) \<in> (nrrstep R)^*" by simp

  from wf_trs_imp_lhs_Fun[OF `wf_trs R` `(l, r) \<in> R`] obtain g ls
    where "l = Fun g ls" by best
  with `l\<cdot>\<sigma> = Fun f ls'` have "\<forall>i<length ls'. ls'!i = (ls!i)\<cdot>\<sigma>" by auto

  have "\<forall>i<num_args ?t. (ts!i, (ls!i)\<cdot>\<sigma>) \<in> (rstep R)^*"
  proof (intro allI impI)
    fix i assume "i < num_args ?t"
    with `(?t, S j) \<in> (nrrstep R)^*` have "(args ?t!i, args (Fun f ls')!i) \<in> (rstep R)^*"
      unfolding `S j = Fun f ls'` by (rule rtrancl_nrrstep_imp_arg_rstep)
    hence "(ts!i, ls'!i) \<in> (rstep R)^*" by simp
    from `(?t, Fun f ls') \<in> (nrrstep R)^*` and `i < num_args ?t`
      have "i < num_args (Fun f ls')" by induct (auto simp: nrrstep_preserves_num_args)
    hence "i < length ls'" by simp
    with `\<forall>i<length ls'. ls'!i = (ls!i)\<cdot>\<sigma>` and `(ts!i, ls'!i) \<in> (rstep R)^*`
    show "(ts!i, (ls!i)\<cdot>\<sigma>) \<in> (rstep R)^*" by simp
  qed

  from `(?t, l\<cdot>\<sigma>) \<in> (nrrstep R)^*` have "num_args ?t = num_args (l\<cdot>\<sigma>)"
    by induct (auto simp: nrrstep_equiv_num_args)
  with `l\<cdot>\<sigma> = Fun f ls'` and `\<forall>i<length ls'. ls'!i = (ls!i)\<cdot>\<sigma>`
    have "\<forall>i<num_args ?t. (ls!i)\<cdot>\<sigma> = ls'!i" by simp
  moreover with `\<forall>s\<lhd>?t. SN_elt (rstep R) s` have "\<forall>i<num_args ?t. SN_elt (rstep R) (ts!i)"
    by auto
  ultimately have "\<forall>i<num_args ?t. SN_elt (rstep R) ((ls!i)\<cdot>\<sigma>)"
    using `\<forall>i<num_args ?t. (ts!i, (ls!i)\<cdot>\<sigma>) \<in> (rstep R)^*`
    using rtrancl_step_preserves_SN[of _ _ "rstep R"] by blast
  with `\<forall>i<num_args ?t. (ls!i)\<cdot>\<sigma> = ls'!i` and `num_args ?t = num_args (l\<cdot>\<sigma>)`
    and `l\<cdot>\<sigma> = Fun f ls'` and `\<forall>i<num_args ?t. SN_elt (rstep R) ((ls!i)\<cdot>\<sigma>)`
    have "\<forall>i<num_args (Fun f ls'). SN_elt (rstep R) (ls'!i)" by auto
  hence "\<forall>s\<lhd>Fun f ls'. SN_elt (rstep R) s" using SN_args_imp_SN_subt[of f ls' R] by blast
  hence "\<forall>s\<lhd>l\<cdot>\<sigma>. SN_elt (rstep R) s" by (simp add: `l\<cdot>\<sigma> = Fun f ls'`)

  from `(l, r) \<in> R` have "(l, r) \<in> R" .
  
  moreover have "l\<cdot>\<sigma> \<in> Tinf R"
  proof -
    let ?T = "\<lambda>i. S (j + i)"
    have "?T 0 = l\<cdot>\<sigma>" by (simp add: `l\<cdot>\<sigma> = S j`)
    moreover from `\<forall>i. (S i, S (Suc i)) \<in> rstep R`
      have "\<forall>i. (?T i, ?T (Suc i)) \<in> rstep R" by simp
    ultimately have "\<not> SN_elt (rstep R) (l\<cdot>\<sigma>)" unfolding SN_defs by best
    with `\<forall>s\<lhd>l\<cdot>\<sigma>. SN_elt (rstep R) s` show ?thesis by (simp add: Tinf_def)
  qed

  moreover have "\<exists>u. r \<unrhd> u \<and> \<not> is_var u \<and> r\<cdot>\<sigma> \<unrhd> u\<cdot>\<sigma> \<and> u\<cdot>\<sigma> \<in> Tinf R \<and> \<not>(l\<cdot>\<sigma> \<rhd> u\<cdot>\<sigma>)"
  proof -
    have "\<forall>x\<in>vars_term r. SN_elt (rstep R) (Var x \<cdot> \<sigma>)"
    proof
      fix x
      assume "x \<in> vars_term r"
      hence "x \<in> vars_term l" using `(l, r) \<in> R` and `wf_trs R` by (auto simp: wf_trs_def)
      then obtain j where "j < num_args l" and "x \<in> vars_term (ls!j)"
        using var_imp_var_of_arg unfolding `l = Fun g ls` by best
      hence "j < num_args (l\<cdot>\<sigma>)" unfolding `l = Fun g ls` by simp
      with `(?t, l\<cdot>\<sigma>) \<in> (nrrstep R)^*` have "j < num_args ?t"
        by (induct rule: converse_rtrancl_induct) (auto simp: nrrstep_reflects_num_args)
      with `\<forall>i<num_args ?t. SN_elt (rstep R) ((ls!i)\<cdot>\<sigma>)`
        have "SN_elt (rstep R) ((ls!j)\<cdot>\<sigma>)" by simp
      with `x \<in> vars_term (ls!j)` show "SN_elt (rstep R) (Var x \<cdot> \<sigma>)"
        using SN_instance_imp_SN_var[of R "ls!j"] by simp
    qed

    have "\<forall>x\<in>vars_term r. \<not>(Var x\<cdot>\<sigma> \<unrhd> t')"
    proof
      fix x
      assume "x \<in> vars_term r"
      hence "SN_elt (rstep R) (Var x \<cdot> \<sigma>)"
        using `\<forall>x\<in>vars_term r. SN_elt (rstep R) (Var x \<cdot> \<sigma>)` by simp
      show "\<not> Var x \<cdot> \<sigma>  \<unrhd> t'"
      proof
        assume "Var x \<cdot> \<sigma> \<unrhd> t'"
        with `SN_elt (rstep R) (Var x \<cdot> \<sigma>)` have "SN_elt (rstep R) t'"
          using SN_imp_SN_subt by auto

        with `t' \<in> Tinf R` show False unfolding Tinf_def by auto
      qed
    qed
    from subt_instance_and_not_subs_imp_subt[OF `r\<cdot>\<sigma> \<unrhd> t'` this]
      obtain u where "r \<unrhd> u" and "t' = u\<cdot>\<sigma>" by best
    
    moreover have "\<not> is_var u"
    proof
      assume "is_var u"
      then obtain x where "u = Var x" by (cases u) auto
      with `r \<unrhd> u` have "r \<unrhd> Var x" by simp
      from var_subt_imp_subt_in_vars[OF this] have "x \<in> vars_term r" .
      with `\<forall>x\<in>vars_term r. SN_elt (rstep R) (Var x \<cdot> \<sigma>)` and `t' \<in> Tinf R`
      show False by (simp add: Tinf_def `u = Var x` `t' = u\<cdot>\<sigma>`)
    qed

    moreover from `r \<unrhd> u` have "r\<cdot>\<sigma> \<unrhd> u\<cdot>\<sigma>" by (rule supteqp_stable)
    moreover from `t' \<in> Tinf R` have "u\<cdot>\<sigma> \<in> Tinf R" by (simp add: `t' = u\<cdot>\<sigma>`)
    moreover have "\<not>(l\<cdot>\<sigma> \<rhd> u\<cdot>\<sigma>)"
    proof
      assume "l\<cdot>\<sigma> \<rhd> u\<cdot>\<sigma>"
      with `l\<cdot>\<sigma> \<in> Tinf R` have "SN_elt (rstep R) (u\<cdot>\<sigma>)" unfolding Tinf_def by force
      with `u\<cdot>\<sigma> \<in> Tinf R` show False by (simp add: Tinf_def)
    qed

    ultimately show ?thesis by auto
  qed

  moreover from `(?t, l\<cdot>\<sigma>) \<in> (nrrstep R)^*` have "(?t, l\<cdot>\<sigma>) \<in> (nrrstep R)^*" .

  moreover from `(l, r) \<in> R` have "(l\<cdot>\<sigma>, r\<cdot>\<sigma>) \<in> rrstep R" unfolding rrstep_def' by best
  ultimately show ?thesis by (auto simp: Fun)
qed
 
lemma Tinf_imp_def_root:
  assumes "wf_trs R" and "t \<in> Tinf R"
  shows "\<exists>f ss. t = Fun f ss \<and> defined R f"
proof -
  from assms obtain l r \<sigma> where "(l, r) \<in> R"
    and "(t, l\<cdot>\<sigma>) \<in> (nrrstep R)^*" using Tinf_lemma by blast
  with `wf_trs R` obtain f ls where l: "l = Fun f ls" unfolding wf_trs_def by best
  hence "l\<cdot>\<sigma> = Fun f (map (\<lambda>t. t\<cdot>\<sigma>) ls)" (is "_ = Fun f ?ls'") by simp
  with `(t, l\<cdot>\<sigma>) \<in> (nrrstep R)^*` have "\<exists>ts. t = Fun f ts"
    by (induct rule: converse_rtrancl_induct) (auto simp: nrrstep_reflects_root)
  then obtain ts where t: "t = Fun f ts" by best
  have "root l = Some f" by (simp add: l)
  with `(l, r) \<in> R` have "defined R f" unfolding defined_def by force
  thus ?thesis by (simp add: t)
qed

datatype 'f shp = Sharp 'f ("_\<sharp>") | Plain 'f ("_@")

instantiation shp :: ("shows")"shows" begin
  fun shows_shp where "shows_shp(Sharp s) = (shows s +@+ shows_string ''#'')"
                    | "shows_shp(Plain s) = shows s"
  definition shows_list_shp where "shows_list_shp(fs::('f::shows)shp list) \<equiv> shows_list_aux shows fs"

  lemma assoc_shp: "shows (f::('f::shows)shp) r @ s = shows f (r@s)"
    by (cases f) (auto simp: shows_string_def)
  lemma assoc_shp_elt: "\<forall>(f::('f::shows)shp)\<in>set fs. shows f r @ s = shows f (r@s)"
    using assoc_shp by auto
  instance by
    (intro_classes,rule assoc_shp,unfold shows_list_shp_def)
    (rule shows_list_aux_assoc[OF assoc_shp_elt])
end

definition sharp_sig :: "'f sig \<Rightarrow> ('f shp)sig" ("_^\<sharp>")
where
 "\<Sigma>^\<sharp> = ({f\<sharp> | f. f :s \<Sigma>} \<union> {f@ | f. f :s \<Sigma>}, (\<lambda>f. (case f of f\<sharp> \<Rightarrow> arity \<Sigma> f | f@ \<Rightarrow> arity \<Sigma> f)))"

definition unsharp_sig :: "('f shp)sig \<Rightarrow> 'f sig"
where
 "unsharp_sig \<Sigma> = ({f | f. f\<sharp> :s \<Sigma> \<or> f@ :s \<Sigma>}, (\<lambda>f. arity \<Sigma> f@))"

fun plain :: "('f,'v)term \<Rightarrow> ('f shp,'v)term"
where "plain (Var x) = Var x"
    | "plain (Fun f ss) = Fun f@ (map plain ss)"

fun sharp :: "('f,'v)term \<Rightarrow> ('f shp,'v)term"
where "sharp (Var x) = Var x"
    | "sharp (Fun f ss) = Fun f\<sharp> (map plain ss)"

fun unsharp_fun :: "'f shp \<Rightarrow> 'f" ("!_!")
where "!f@! = f"
    | "!f\<sharp>! = f"

fun unsharp :: "('f shp,'v) term \<Rightarrow> ('f,'v) term"
where "unsharp(Var x) = Var x"
    | "unsharp(Fun f ss) = Fun !f! (map unsharp ss)"

fun plain_ctxt :: "('f,'v)ctxt \<Rightarrow> ('f shp,'v)ctxt"
where "plain_ctxt \<box> = \<box>"
    | "plain_ctxt (More f ss1 D ss2) = More f@ (map plain ss1) (plain_ctxt D) (map plain ss2)"

fun unsharp_ctxt :: "('f shp,'v)ctxt \<Rightarrow> ('f,'v)ctxt"
where "unsharp_ctxt \<box> = \<box>"
    | "unsharp_ctxt (More f ss1 D ss2) = More !f! (map unsharp ss1) (unsharp_ctxt D) (map unsharp ss2)"

fun sharp_ctxt :: "('f,'v)ctxt \<Rightarrow> ('f shp,'v)ctxt"
where "sharp_ctxt \<box> = \<box>"
    | "sharp_ctxt (More f ss1 D ss2) = More f\<sharp> (map plain ss1) (plain_ctxt D) (map plain ss2)"

definition DP :: "('f,'v)trs \<Rightarrow> ('f shp,'v)trs"
where
 "DP R = {(s,t). \<exists>l r h us. s = sharp l \<and> t = sharp(Fun h us) \<and> (l,r) \<in> R \<and> r \<unrhd> (Fun h us) \<and> defined R h \<and> \<not> l \<rhd> Fun h us}"

definition ID :: "('f,'v)trs \<Rightarrow> ('f shp,'v)trs"
where
 "ID R = {(s,t). \<exists>l r. s = plain l \<and> t = plain r \<and> (l,r) \<in> R}"

lemma plain_sub_distr':
 fixes \<sigma> :: "'v \<Rightarrow> ('f,'v)term" and t :: "('f,'v)term"
 shows "((\<lambda>t. t\<cdot>(Sub(plain \<circ> \<sigma>))) \<circ> plain) t = (plain \<circ> (\<lambda>t. t\<cdot>(Sub \<sigma>))) t" (is "?f t = ?g t")
proof (induct t)
 case (Var x) show ?case by simp
next
 case (Fun f ss)
 hence IH: "\<And>s. s \<in> set ss \<longrightarrow> ?f s = ?g s" by simp
 have "?f(Fun f ss) = (Fun f@ (map plain ss))\<cdot>(Sub(plain \<circ> \<sigma>))" unfolding o_def by simp
 also have "\<dots> = Fun f@ (map (\<lambda>t. t\<cdot>(Sub(plain \<circ> \<sigma>))) (map plain ss))" by simp
 also have "\<dots> = Fun f@ (map ?f ss)" unfolding o_def by simp
 also have "\<dots> = Fun f@ (map ?g ss)" unfolding map_ext[OF IH] by simp
 also have "\<dots> = Fun f@ (map plain (map (\<lambda>t. t\<cdot>(Sub \<sigma>)) ss))" by simp
 also have "\<dots> = plain(Fun f (map (\<lambda>t. t\<cdot>(Sub \<sigma>)) ss))" by simp
 also have "\<dots> = ?g(Fun f ss)" by simp
 finally show ?case by force
qed

lemma plain_sub_distr'': "(\<lambda>t. t\<cdot>(Sub(plain \<circ> \<sigma>))) \<circ> plain = plain \<circ> (\<lambda>t. t\<cdot>(Sub \<sigma>))" (is "?f = ?g")
proof (rule ext)
 fix t show "?f t = ?g t" using plain_sub_distr'[where t = t] unfolding o_def by simp
qed

lemma plain_sub_distr[simp]: "plain(t\<cdot>\<sigma>) = (plain t)\<cdot>(Sub(plain \<circ> (case \<sigma> of Sub s \<Rightarrow> s)))" 
proof -
 obtain s where "\<sigma> = Sub s" by (cases \<sigma>) auto
 hence "plain(t\<cdot>\<sigma>) = plain(t\<cdot>(Sub s))" by simp
 also have "\<dots> = (plain \<circ> (\<lambda>t. t\<cdot>(Sub s))) t" unfolding o_def by simp
 also have "\<dots> = ((\<lambda>t. t\<cdot>(Sub(plain \<circ> s))) \<circ> plain) t" unfolding plain_sub_distr'' by simp
 also have "\<dots> = (plain t)\<cdot>(Sub(plain \<circ> s))" by simp
 also have "\<dots> = (plain t)\<cdot>(Sub(plain \<circ> (case \<sigma> of Sub s \<Rightarrow> s)))" unfolding `\<sigma> = Sub s` by simp
 finally show ?thesis .
qed

(* In the case of sharp the lemma only holds for non-variable terms. *)
lemma sharp_sub_distr: "(sharp(Fun f ss))\<cdot>(Sub(plain \<circ> \<sigma>)) = sharp((Fun f ss)\<cdot>(Sub \<sigma>))"
proof -
 have A: "\<And>s. s \<in> set ss \<Longrightarrow> (plain s)\<cdot>(Sub(plain \<circ> \<sigma>)) = plain(s\<cdot>(Sub \<sigma>))" using plain_sub_distr'[where \<sigma>= \<sigma>] unfolding o_def by auto
 have "(sharp(Fun f ss))\<cdot>(Sub(plain \<circ> \<sigma>)) = Fun f\<sharp> (map plain ss)\<cdot>(Sub(plain \<circ> \<sigma>))" by simp
 also have "\<dots> = Fun f\<sharp> (map (\<lambda>t. t\<cdot>(Sub(plain \<circ> \<sigma>))) (map plain ss))" by simp
 also have "\<dots> = Fun f\<sharp> (map ((\<lambda>t. t\<cdot>(Sub(plain \<circ> \<sigma>))) \<circ> plain) ss)" by (simp add: o_def)
 also have "\<dots> = Fun f\<sharp> (map (plain \<circ> (\<lambda>t. t\<cdot>(Sub \<sigma>))) ss)" unfolding plain_sub_distr'' by simp
 also have "\<dots> = Fun f\<sharp> (map plain (map (\<lambda>t. t\<cdot>(Sub \<sigma>)) ss))" by simp
 finally show ?thesis by simp
qed

lemma plain_ctxt_distr[simp]: "plain(C\<langle>t\<rangle>) = (plain_ctxt C)\<langle>plain t\<rangle>" by (induct C) auto
lemma sharp_ctxt_distr[simp]:
 "sharp ((More f ss1 D ss2)\<langle>t\<rangle>) = (sharp_ctxt (More f ss1 D ss2))\<langle>plain t\<rangle>" by (induct D) auto

lemma unsharp_sub_distr': "((\<lambda>t. t\<cdot>(Sub(unsharp \<circ> \<sigma>))) \<circ> unsharp) t = (unsharp \<circ> (\<lambda>t. t\<cdot>(Sub \<sigma>))) t" (is "?f t = ?g t")
proof (induct t)
 case (Var x) show ?case by simp
next
 case (Fun f ss)
 hence IH: "\<And>s. s \<in> set ss \<longrightarrow> ?f s = ?g s" by simp
 obtain g where g:"(f = g@) \<or> (f = g\<sharp>)" by (induct rule: shp.induct) auto
 have "?f(Fun f ss) = (Fun g (map unsharp ss))\<cdot>(Sub(unsharp \<circ> \<sigma>))" unfolding o_def using g by cases auto
 also have "\<dots> = Fun g (map (\<lambda>t. t\<cdot>(Sub(unsharp \<circ> \<sigma>))) (map unsharp ss))" by simp
 also have "\<dots> = Fun g (map ?f ss)" by (simp add: o_def)
 also have "\<dots> = Fun g (map  ?g ss)" unfolding map_ext[OF IH] by simp
 also have "\<dots> = Fun g (map unsharp (map (\<lambda>t. t\<cdot>(Sub \<sigma>)) ss))" by simp
 also have "\<dots> = unsharp(Fun f (map (\<lambda>t. t\<cdot>(Sub \<sigma>)) ss))" using g by cases auto
 also have "\<dots> = ?g(Fun f ss)" by simp
 finally show ?case by force
qed

lemma unsharp_sub_distr'': "(\<lambda>t. t\<cdot>(Sub(unsharp \<circ> \<sigma>))) \<circ> unsharp = unsharp \<circ> (\<lambda>t. t\<cdot>(Sub \<sigma>))" (is "?f = ?g")
proof (rule ext)
 fix t show "?f t = ?g t" using unsharp_sub_distr'[where t = t] unfolding o_def by simp
qed

lemma unsharp_sub_distr[simp]: "unsharp(t\<cdot>\<sigma>) = (unsharp t)\<cdot>(Sub(unsharp \<circ> (case \<sigma> of Sub s \<Rightarrow> s)))"
proof -
 obtain s where "\<sigma> = Sub s" by (cases \<sigma>) auto
 hence "unsharp(t\<cdot>\<sigma>) = unsharp(t\<cdot>(Sub s))" by simp
 also have "\<dots> = (unsharp \<circ> (\<lambda>t. t\<cdot>(Sub s))) t" unfolding o_def by simp
 also have "\<dots> = ((\<lambda>t. t\<cdot>(Sub(unsharp \<circ> s))) \<circ> unsharp) t" unfolding unsharp_sub_distr'' by simp
 also have "\<dots> = (unsharp t)\<cdot>(Sub(unsharp \<circ> s))" by simp
 also have "\<dots> = (unsharp t)\<cdot>(Sub(unsharp \<circ> (case \<sigma> of Sub s \<Rightarrow> s)))" unfolding `\<sigma> = Sub s` by simp
 finally show ?thesis .
qed

lemma unsharp_plain_id[simp]: "unsharp(plain t) = t"
proof (induct t rule: term_induct)
 case (Var x) thus ?case by simp
next
 case (Fun f ss)
 hence "!!s. s \<in> set ss \<Longrightarrow> (unsharp \<circ> plain) s = s" by simp
 hence a: "map (unsharp \<circ> plain) ss = ss" by (rule map_idI)
 have "unsharp (plain (Fun f ss)) = unsharp (Fun f@ (map plain ss))" by simp
 also have "\<dots> = Fun f (map unsharp (map plain ss))" by simp
 also have "\<dots> = Fun f (map (unsharp \<circ> plain) ss)" by simp
 finally show ?case unfolding a .
qed

lemma unsharp_ctxt_distr[simp]: "unsharp C\<langle>t\<rangle> = (unsharp_ctxt C)\<langle>unsharp t\<rangle>"
proof (induct C)
 case Hole thus ?case by simp
next
 case (More f ss1 C ss2)
 have "unsharp ((More f ss1 C ss2)\<langle>t\<rangle>) = unsharp (Fun f (ss1@C\<langle>t\<rangle>#ss2))" by simp
 then obtain g where "unsharp ((More f ss1 C ss2)\<langle>t\<rangle>) = Fun g (map unsharp (ss1@C\<langle>t\<rangle>#ss2))"
  and "f = g\<sharp> \<or> f = g@"
  by (induct rule: shp.induct) auto
 hence "unsharp ((More f ss1 C ss2)\<langle>t\<rangle>) =
        Fun g (map unsharp ss1@((unsharp_ctxt C)\<langle>unsharp t\<rangle>)#(map unsharp ss2))" using More by simp
 hence a: "unsharp ((More f ss1 C ss2)\<langle>t\<rangle>) =
               Fun g (map unsharp ss1@((unsharp_ctxt C)\<langle>unsharp t\<rangle>)#(map unsharp ss2))" by simp
 from `f = g\<sharp> \<or> f = g@` show ?case
 proof
  assume "f = g\<sharp>"
  have "(unsharp_ctxt (More f ss1 C ss2))\<langle>unsharp t\<rangle> =
   (More g (map unsharp ss1) (unsharp_ctxt C) (map unsharp ss2))\<langle>unsharp t\<rangle>" unfolding `f = g\<sharp>` by simp
  also have "\<dots> = Fun g ((map unsharp ss1)@(unsharp (C\<langle>t\<rangle>))#(map unsharp ss2))"
   unfolding `f = g\<sharp>` using More by simp
  also have "\<dots> = Fun g (map unsharp (ss1@C\<langle>t\<rangle>#ss2))" by simp
  also have "\<dots> = unsharp (Fun f (ss1@C\<langle>t\<rangle>#ss2))" unfolding `f = g\<sharp>` by simp
  finally have "(unsharp_ctxt (More f ss1 C ss2))\<langle>unsharp t\<rangle> = unsharp ((More f ss1 C ss2)\<langle>t\<rangle>)" by simp
  thus ?case by simp
 next
  assume "f = g@"
  have "(unsharp_ctxt (More f ss1 C ss2))\<langle>unsharp t\<rangle> =
   (More g (map unsharp ss1) (unsharp_ctxt C) (map unsharp ss2))\<langle>unsharp t\<rangle>" unfolding `f = g@` by simp
  also have "\<dots> = Fun g ((map unsharp ss1)@(unsharp (C\<langle>t\<rangle>))#(map unsharp ss2))"
   unfolding `f = g@` using More by simp
  also have "\<dots> = Fun g (map unsharp (ss1@C\<langle>t\<rangle>#ss2))" by simp
  also have "\<dots> = unsharp (Fun f (ss1@C\<langle>t\<rangle>#ss2))" unfolding `f = g@` by simp
  finally have "(unsharp_ctxt (More f ss1 C ss2))\<langle>unsharp t\<rangle> = unsharp ((More f ss1 C ss2)\<langle>t\<rangle>)" by simp
  thus ?case by simp
 qed
qed

lemma rstep_iff_plain_rstep:
 fixes R :: "('f,'v)trs"
 shows "((s,t) \<in> rstep R) = ((plain s,plain t) \<in> rstep(ID R))"
proof
 let ?ps = "map (\<lambda>(x,t). (x, plain t))"
 assume "(s,t) \<in> rstep R" hence "(s,t) \<in> CC(CS R)" .
 then obtain C u v where "(u,v) \<in> (CS R)" and "C\<langle>u\<rangle> = s" and "C\<langle>v\<rangle> = t"
  by (cases rule: closure_ctxt.cases) blast
 then obtain \<sigma> :: "('f,'v)sub" and w x where "(w,x) \<in> R" and "w\<cdot>\<sigma> = u" and "x\<cdot>\<sigma> = v"
  by (cases rule: closure_sub.cases) best
 from `(w,x) \<in> R` have "(plain w,plain x) \<in> ID R" unfolding ID_def by best
 hence "(plain(w\<cdot>\<sigma>),plain(x\<cdot>\<sigma>)) \<in> CS(ID R)" unfolding plain_sub_distr by (rule csI)
 hence "(plain(C\<langle>w\<cdot>\<sigma>\<rangle>),plain(C\<langle>x\<cdot>\<sigma>\<rangle>)) \<in> CC(CS(ID R))" unfolding plain_ctxt_distr by (rule ccI)
 thus "(plain s,plain t) \<in> rstep(ID R)" unfolding `w\<cdot>\<sigma> = u` `x\<cdot>\<sigma> = v` `C\<langle>u\<rangle> = s` `C\<langle>v\<rangle> = t` .
next
 assume "(plain s,plain t) \<in> rstep(ID R)" hence "(plain s,plain t) \<in> CC(CS(ID R))" .
 then obtain C u v where "(u,v) \<in> CS(ID R)" and "C\<langle>u\<rangle> = plain s" and "C\<langle>v\<rangle> = plain t"
  by (cases rule: closure_ctxt.cases) force
 then obtain \<sigma> w x where "(w,x) \<in> ID R" and "w\<cdot>\<sigma> = u" and "x\<cdot>\<sigma> = v"
  by (cases rule: closure_sub.cases) force
 then obtain l r where "w = plain l" and "x = plain r" and "(l,r) \<in> R" unfolding ID_def by best
 have ps: "C\<langle>plain l\<cdot>\<sigma>\<rangle> = plain s" and pt: "C\<langle>plain r\<cdot>\<sigma>\<rangle> = plain t"
  unfolding `w = plain l`[symmetric] `x = plain r`[symmetric] `w\<cdot>\<sigma> = u` `x\<cdot>\<sigma> = v`
  `C\<langle>u\<rangle> = plain s` `C\<langle>v\<rangle> = plain t` by auto
 let ?C = "unsharp_ctxt C"
 let ?\<sigma> = "Sub(unsharp \<circ> (case \<sigma> of Sub s \<Rightarrow> s))"
 from ps and pt have "s = unsharp C\<langle>(plain l)\<cdot>\<sigma>\<rangle>" and "t = unsharp C\<langle>(plain r)\<cdot>\<sigma>\<rangle>" by auto
 hence s: "s = ?C\<langle>unsharp (plain l\<cdot>\<sigma>)\<rangle>" and t: "t = ?C\<langle>unsharp (plain r\<cdot>\<sigma>)\<rangle>" using unsharp_ctxt_distr by auto
 from s have "s = ?C\<langle>l\<cdot>?\<sigma>\<rangle>" unfolding unsharp_sub_distr by simp
 from t have "t = ?C\<langle>r\<cdot>?\<sigma>\<rangle>" unfolding unsharp_sub_distr by simp
 from `(l,r) \<in> R` have "(l\<cdot>?\<sigma>,r\<cdot>?\<sigma>) \<in> CS R" by (rule csI)
 hence "(?C\<langle>l\<cdot>?\<sigma>\<rangle>,?C\<langle>r\<cdot>?\<sigma>\<rangle>) \<in> CC(CS R)" by (rule ccI)
 thus "(s,t) \<in> rstep R" unfolding `s = ?C\<langle>l\<cdot>?\<sigma>\<rangle>` `t = ?C\<langle>r\<cdot>?\<sigma>\<rangle>` .
qed

lemma nrrstep_imp_sharp_rstep:
 assumes "(s,t) \<in> nrrstep R"
 shows "(sharp s,sharp t) \<in> rstep(ID R)"
proof -
 from assms obtain l r i ps \<sigma> where "(s,t) \<in> rstep_r_p_s R (l,r) (i<#ps) \<sigma>" unfolding nrrstep_def by best
 hence "let C = ctxt_of_pos_term (i<#ps) s in (i<#ps) \<in> poss s \<and> (l,r) \<in> R \<and> C\<langle>l\<cdot>\<sigma>\<rangle> = s \<and> C\<langle>r\<cdot>\<sigma>\<rangle> = t"
  (is "let C = ?C in _") unfolding rstep_r_p_s_def by simp
 hence "(l,r) \<in> R" and s: "?C\<langle>l\<cdot>\<sigma>\<rangle> = s" and t: "?C\<langle>r\<cdot>\<sigma>\<rangle> = t" unfolding Let_def by auto
 from `(s,t) \<in> nrrstep R` obtain f ss where "s = Fun f ss" using nrrstep_imp_Fun by best
 hence C: "?C = More f (take i ss) (ctxt_of_pos_term ps (ss!i)) (drop (Suc i) ss)"
  (is "?C = More f ?ss1 ?D ?ss2") by simp
 let ?ps = "\<lambda>\<sigma>. Sub(plain \<circ> (case \<sigma> of Sub s \<Rightarrow> s))"
 let ?C' = "sharp_ctxt (More f ?ss1 ?D ?ss2)"
 from `(l,r) \<in> R` have "(plain l,plain r) \<in> ID R" unfolding ID_def by auto
 hence "(plain l\<cdot>(?ps \<sigma>),plain r\<cdot>(?ps \<sigma>)) \<in> CS(ID R)" by (rule csI)
 hence "(?C'\<langle>plain l\<cdot>(?ps \<sigma>)\<rangle>,?C'\<langle>plain r\<cdot>(?ps \<sigma>)\<rangle>) \<in> CC(CS(ID R))" by (rule ccI)
 hence ID_R_step: "(?C'\<langle>plain l\<cdot>(?ps \<sigma>)\<rangle>,?C'\<langle>plain r\<cdot>(?ps \<sigma>)\<rangle>) \<in> rstep(ID R)" by simp
 from `?C\<langle>l\<cdot>\<sigma>\<rangle> = s` have "sharp s = sharp ?C\<langle>l\<cdot>\<sigma>\<rangle>" by simp
 hence "sharp s = ?C'\<langle>plain (l\<cdot>\<sigma>)\<rangle>" using C sharp_ctxt_distr[symmetric] by simp
 hence s: "sharp s = ?C'\<langle>plain l\<cdot>(?ps \<sigma>)\<rangle>" unfolding plain_sub_distr[symmetric] by simp
 from `?C\<langle>r\<cdot>\<sigma>\<rangle> = t` have "sharp t = sharp ?C\<langle>r\<cdot>\<sigma>\<rangle>" by simp
 hence "sharp t = ?C'\<langle>plain (r\<cdot>\<sigma>)\<rangle>" using C sharp_ctxt_distr[symmetric] by simp
 hence t: "sharp t = ?C'\<langle>plain r\<cdot>(?ps \<sigma>)\<rangle>" unfolding plain_sub_distr[symmetric] by simp
 from ID_R_step show ?thesis unfolding s t .
qed

lemma nrrstep_seq_imp_sharp_rstep_seq:
 "(s,t) \<in> (nrrstep R)^* \<Longrightarrow> (sharp s,sharp t) \<in> (rstep (ID R))^*"
proof (induct rule: rtrancl_induct)
 case base thus ?case by simp
next
 case (step a b)
 from `(a,b) \<in> nrrstep R` have "(sharp a,sharp b) \<in> rstep(ID R)" by (rule nrrstep_imp_sharp_rstep)
 with step show ?case by auto
qed

lemma instance_no_subt_imp_no_subt:
 assumes "\<not> s\<cdot>\<sigma> \<rhd> (t\<cdot>\<sigma>)"
 shows "\<not> s \<rhd> t"
proof (rule ccontr)
 assume "\<not> \<not> s \<rhd> t"
 have "s \<rhd> t \<Longrightarrow> s\<cdot>\<sigma> \<rhd> (t\<cdot>\<sigma>)" by (rule suptp_stable)
 with `\<not> \<not> s \<rhd> t` have "s\<cdot>\<sigma> \<rhd> (t\<cdot>\<sigma>)" by simp
 with assms show "False" by simp
qed

lemma Tinf_imp_Fun: 
 assumes "wf_trs R" and "t \<in> Tinf R"
 shows "\<exists>f ts. t = Fun f ts"
proof (cases t)
 fix x assume "t = Var x"
 from `wf_trs R` have "SN_elt (rstep R) t" unfolding `t = Var x` using SN_Var by best
 with `t \<in> Tinf R` show ?thesis unfolding Tinf_def by auto
next
 fix f ts assume "t = Fun f ts" thus ?thesis by auto
qed

lemma sub_case_simp[simp]: "Sub(case \<sigma> of Sub s \<Rightarrow> s) = \<sigma>" by (cases \<sigma>) simp

lemma sub_plain_distr: "(plain t)\<cdot>Sub(plain \<circ> s) = plain(t \<cdot> Sub s)" by (induct t) simp_all

lemma sub_sharp_distr: "(sharp(Fun f ss))\<cdot>Sub(plain \<circ> s) = sharp(Fun f ss \<cdot> Sub s)"
proof -
 have A: "\<forall>t\<in>set ss. (plain t)\<cdot>Sub(\<lambda>t. plain(s t)) = plain(t \<cdot> Sub s)" using sub_plain_distr unfolding o_def by auto
 show ?thesis by simp
qed

lemma cs_subseteq_cs_ext1:
 fixes R::"('f,'v)trs"
 assumes "(s,t) \<in> CS R"
 shows "(s,t) \<in> CS(R \<union> S)"
proof -
 from assms obtain \<sigma> u v 
 where R:"(u,v) \<in> R" and s: "s = u\<cdot>\<sigma>" and t:"t = v\<cdot>\<sigma>" by (cases rule: closure_sub.cases) auto
 from R have "(u,v) \<in> (R \<union> S)" ..
 thus ?thesis unfolding s t by (rule csI)
qed

lemma cs_subseteq_cs_ext2:
 assumes "(s,t) \<in> CS S"
 shows "(s,t) \<in> CS(R \<union> S)"
proof -
 have comm:"(S \<union> R) = (R \<union> S)" by auto
 from assms have "(s,t) \<in> CS(S \<union> R)" by (rule cs_subseteq_cs_ext1)
 thus ?thesis unfolding comm .
qed

lemma cs_subseteq_cs_ext1': "CS R \<subseteq> CS (R \<union> S)"
by (intro subsetI2) (rule cs_subseteq_cs_ext1)

lemma cs_union_distr: "CS(R \<union> S) = (CS R) \<union> (CS S)"
proof
 show "CS (R \<union> S) \<subseteq> ((CS R) \<union> (CS S))"
 proof (intro subsetI2)
  fix s t assume "(s,t) \<in> CS(R \<union> S)"
  then obtain \<sigma> u v where RS:"(u,v) \<in> (R \<union> S)" and s:"s = u\<cdot>\<sigma>" and t:"t = v\<cdot>\<sigma>"
   by (cases rule: closure_sub.cases) auto
  from RS have "(u,v) \<in> R \<or> (u,v) \<in> S" by simp
  hence "(s,t) \<in> CS R \<or> (s,t) \<in> CS S"
  proof
   assume "(u,v) \<in> R" thus ?thesis unfolding s t using csI by auto
  next
   assume "(u,v) \<in> S" thus ?thesis unfolding s t using csI by auto
  qed
  thus "(s,t) \<in> CS R \<union> CS S" by simp
 qed
next
 show "(CS R) \<union> (CS S) \<subseteq> CS (R \<union> S)"
 proof (intro subsetI2)
  fix s t assume "(s,t) \<in> CS R \<union> CS S"
  hence "(s,t) \<in> CS R \<or> (s,t) \<in> CS S" by simp
  thus "(s,t) \<in> CS(R \<union> S)"
  proof (rule disjE)
   assume "(s,t) \<in> CS R" thus ?thesis by (rule cs_subseteq_cs_ext1)
  next
   assume "(s,t) \<in> CS S" thus ?thesis by (rule cs_subseteq_cs_ext2)
  qed
 qed
qed

lemma cc_subseteq_cc_ext1:
 assumes "(s,t) \<in> CC R"
 shows "(s,t) \<in> CC(R \<union> S)"
proof -
 from assms obtain C u v
  where R:"(u,v) \<in> R" and s:"s = C\<langle>u\<rangle>" and t:"t = C\<langle>v\<rangle>" by (cases rule: closure_ctxt.cases) auto
 from R have "(u,v) \<in> (R \<union> S)" ..
 thus ?thesis unfolding s t by (rule ccI)
qed

lemma cc_subseteq_cc_ext2:
 assumes "(s,t) \<in> CC S"
 shows "(s,t) \<in> CC(R \<union> S)"
proof -
 have comm:"(S \<union> R) = (R \<union> S)" by auto
 from assms have "(s,t) \<in> CC(S \<union> R)" by (rule cc_subseteq_cc_ext1)
 thus ?thesis unfolding comm .
qed

lemma cc_union_distr: "CC(R \<union> S) = (CC R) \<union> (CC S)"
proof
 show "CC (R \<union> S) \<subseteq> (CC R \<union> CC S)"
 proof (rule subsetI2)
  fix s t assume "(s,t) \<in> CC(R \<union> S)"
  then obtain C u v where RS:"(u,v) \<in> (R \<union> S)" and s:"s = C\<langle>u\<rangle>" and t:"t = C\<langle>v\<rangle>"
   by (cases rule: closure_ctxt.cases) auto
  from RS have "(u,v) \<in> R \<or> (u,v) \<in> S" by simp
  thus "(s,t) \<in> CC R \<union> CC S"
  proof
   assume "(u,v) \<in> R" thus ?thesis unfolding s t using ccI by auto
  next
   assume "(u,v) \<in> S" thus ?thesis unfolding s t using ccI by auto
  qed
 qed
next
 show "(CC R) \<union> (CC S) \<subseteq> CC (R \<union> S)"
 proof (rule subsetI2)
  fix s t assume "(s,t) \<in> CC R \<union> CC S"
  hence "(s,t) \<in> CC R \<or> (s,t) \<in> CC S" by simp
  thus "(s,t) \<in> CC(R \<union> S)"
  proof (rule disjE)
   assume "(s,t) \<in> CC R" thus ?thesis by (rule cc_subseteq_cc_ext1)
  next
   assume "(s,t) \<in> CC S" thus ?thesis by (rule cc_subseteq_cc_ext2)
  qed
 qed
qed

lemma CC_monotone : assumes "r \<subseteq> s" shows "CC r \<subseteq> CC s"
using assms
proof auto
  fix u v
  assume "(u,v) \<in> CC r"
  thus "(u,v) \<in> CC s"
  proof
    fix uu vv C
    assume "u = C\<langle>uu\<rangle>" and "v = C\<langle>vv\<rangle>" and "(uu,vv) \<in> r"
    with assms show ?thesis by (auto simp: ccI)
  qed
qed

lemma CS_monotone : assumes "r \<subseteq> s" shows "CS r \<subseteq> CS s"
using assms
proof auto
  fix u v
  assume "(u,v) \<in> CS r"
  thus "(u,v) \<in> CS s"
  proof
    fix uu vv \<sigma>
    assume "u = uu \<cdot> \<sigma>" and "v = vv \<cdot> \<sigma>" and "(uu,vv) \<in> r"
    with assms show ?thesis by (auto simp: csI)
  qed
qed

lemma rstep_mono:
 assumes "rR \<subseteq> R"
 shows "rstep rR \<subseteq> rstep R"
proof-
 from assms have "CS rR \<subseteq> CS R" using CS_monotone by auto
 thus "CC (CS rR) \<subseteq> CC (CS R)" using CC_monotone by auto
qed

lemma cs_diff_distr[simp]: "CS A - CS B  \<subseteq> CS (A - B)" (* SW *)
proof-
 have "A = (A - B) \<union> (B \<inter> A)" by auto
 hence "CS A = CS ((A - B) \<union> (B \<inter> A))" by auto
 hence "CS A = CS (A - B) \<union> CS (B \<inter> A)" using cs_union_distr by auto
 hence "CS A - CS (B \<inter> A) = CS (A - B) \<union> CS (B \<inter> A) - CS (B \<inter> A)" by auto
 hence one:"CS A - CS (B \<inter> A) \<subseteq> CS (A - B)" by auto
 have "B \<inter> A \<subseteq> B" by auto
 from CS_monotone[OF this] have "CS (B \<inter> A) \<subseteq>  CS B" .
 hence "CS A - CS B \<subseteq> CS A - CS (B \<inter> A)" by auto
 with one show "CS A - CS B \<subseteq> CS (A - B)" by auto
qed

lemma cc_diff_distr[simp]: "CC A - CC B \<subseteq> CC (A - B)"  (* SW *)
proof-
 have "A = (A - B) \<union> (B \<inter> A)" by auto
 hence "CC A = CC ((A - B) \<union> (B \<inter> A))" by auto
 hence "CC A = CC (A - B) \<union> CC (B \<inter> A)" using cc_union_distr by auto
 hence one:"CC A - CC (B \<inter> A) = CC (A - B) \<union> CC (B \<inter> A) - CC (B \<inter> A)" by auto
 hence one:"CC A - CC (B \<inter> A) \<subseteq> CC (A - B)" by auto
 have "B \<inter> A \<subseteq> B" by auto
 from CC_monotone[OF this] have "CC (B \<inter> A) \<subseteq>  CC B" .
 hence "CC A - CC B \<subseteq> CC A - CC (B \<inter> A)" by auto
 with one show "CC A - CC B \<subseteq> CC (A - B)" by auto
qed

lemma rstep_diff:  (* SW *)
 shows "rstep R - rstep R' \<subseteq> rstep (R - R')"
proof-
 have "CC (CS R) - CC (CS R') \<subseteq> CC (CS R - CS R')" using cc_diff_distr by auto
 also have "\<dots> \<subseteq> CC (CS (R - R'))" using cs_diff_distr CC_monotone[where ?r="CS R - CS R'"] by auto
 finally show ?thesis by auto
qed

lemma sub_ctxt_distr: "s = C\<langle>t\<rangle>\<cdot>\<sigma> \<Longrightarrow> \<exists>D. s = D\<langle>t\<cdot>\<sigma>\<rangle>"
proof (induct C arbitrary: s)
 case Hole
 have "\<box>\<langle>t\<rangle>\<cdot>\<sigma> = \<box>\<langle>t\<cdot>\<sigma>\<rangle>" by simp
 hence "s = \<box>\<langle>t\<cdot>\<sigma>\<rangle>" unfolding Hole .
 thus ?case by best
next
 case (More f ss1 C ss2)
 hence "s = Fun f (ss1@C\<langle>t\<rangle>#ss2)\<cdot>\<sigma>" by simp
 hence "s = Fun f (map (\<lambda>t. t\<cdot>\<sigma>) (ss1@C\<langle>t\<rangle>#ss2))" by simp
 hence "s = Fun f ((map (\<lambda>t. t\<cdot>\<sigma>) ss1)@(C\<langle>t\<rangle>\<cdot>\<sigma>)#(map (\<lambda>t. t\<cdot>\<sigma>) ss2))"
  (is "s = Fun f (?ss1@(C\<langle>t\<rangle>\<cdot>\<sigma>)#?ss2)") by simp
 with More obtain D where "s = Fun f (?ss1@D\<langle>t\<cdot>\<sigma>\<rangle>#?ss2)" by force
 hence "s = (More f ?ss1 D ?ss2)\<langle>t\<cdot>\<sigma>\<rangle>" by simp
 thus ?case by best
qed

lemma Tinf_lemma_DP:
  assumes "wf_trs R" "s \<in> Tinf R"
  shows "\<exists>t u. t \<in> Tinf R \<and> u \<in> Tinf R \<and>
    (sharp s, sharp t) \<in> (rstep (ID R))^* \<and> (sharp t, sharp u) \<in> rstep (DP R)"
proof -
  obtain l r u \<sigma> where "(l, r) \<in> R" and "r \<unrhd> u" and "\<not> is_var u"
    and nrseq: "(s,l\<cdot>\<sigma>) \<in> (nrrstep R)^*"
    and "(l\<cdot>\<sigma>, r\<cdot>\<sigma>) \<in> rrstep R" and "l\<cdot>\<sigma> \<in> Tinf R" and "u\<cdot>\<sigma> \<in> Tinf R"
    and "\<not> l\<cdot>\<sigma> \<rhd> (u\<cdot>\<sigma>)"
    using Tinf_lemma[where ?R = "R" and ?t = "s",OF `wf_trs R` `s \<in> Tinf R`] by auto
  let ?l = "sharp l"
  let ?r = "sharp u"
  let ?ps = "\<lambda>\<sigma>. Sub(plain \<circ> (case \<sigma> of Sub s \<Rightarrow> s))"
  let ?\<sigma> = "?ps \<sigma>"
  from `\<not> is_var u` obtain f us where "u = Fun f us" by (cases u) auto
  hence "u\<cdot>\<sigma> = Fun f (map (\<lambda>t. t\<cdot>\<sigma>) us)" (is "_ = Fun f ?us'") by simp
  hence "root u = root (u\<cdot>\<sigma>)" unfolding `u = Fun f us` by simp
  from `wf_trs R` and `u\<cdot>\<sigma> \<in> Tinf R` obtain h vs where "(u\<cdot>\<sigma>) = Fun h vs" and "defined R h"
    using Tinf_imp_def_root by best
  from `\<not> l\<cdot>\<sigma> \<rhd> (u\<cdot>\<sigma>)` have "\<not> l \<rhd> u" by (rule instance_no_subt_imp_no_subt)
  from nrseq have "(sharp s, sharp (l\<cdot>\<sigma>)) \<in> (rstep (ID R))^*"
    by (rule nrrstep_seq_imp_sharp_rstep_seq)
  from `(l,r) \<in> R` and `wf_trs R` obtain g ts
    where "l = Fun g ts" unfolding wf_trs_def by blast
  from `u\<cdot>\<sigma> = Fun h vs` and `u = Fun f us` have "f = h" by auto
  hence "defined R f" using `defined R h` by simp
  from `(l, r) \<in> R` and `r \<unrhd> u` and `defined R f` and `\<not> l \<rhd> u`
    have "(?l, ?r) \<in> DP R" unfolding DP_def `u = Fun f us` by auto
  hence "(?l\<cdot>?\<sigma>, ?r\<cdot>?\<sigma>) \<in> CS (DP R)" by auto
  hence "(?l\<cdot>?\<sigma>, ?r\<cdot>?\<sigma>) \<in> rstep (DP R)" using ccI[where C = "\<box>"] by simp
  hence "(sharp (l\<cdot>\<sigma>), sharp (u\<cdot>\<sigma>)) \<in> rstep (DP R)"
    unfolding `l = Fun g ts` `u = Fun f us` sharp_sub_distr by simp
  with `l\<cdot>\<sigma> \<in> Tinf R` and `u\<cdot>\<sigma> \<in> Tinf R` and `(sharp s,sharp(l\<cdot>\<sigma>)) \<in> (rstep (ID R))^*`
    show ?thesis by auto
qed

lemma Tinf_lemma_DP_explicit_sub:
  assumes "wf_trs R" and "s \<in> Tinf R"
  shows "\<exists>l r u \<sigma>. (l, r) \<in> DP R \<and> r\<cdot>\<sigma> = sharp u \<and> u \<in> Tinf R
    \<and> (sharp s, l\<cdot>\<sigma>) \<in> (rstep (ID R))^*"
proof -
  obtain l r u \<sigma> where "(l, r) \<in> R" and "r \<unrhd> u" and "\<not> is_var u"
    and "(s, l\<cdot>\<sigma>) \<in> (nrrstep R)^*" and "(l\<cdot>\<sigma>, r\<cdot>\<sigma>) \<in> rrstep R"
    and "l\<cdot>\<sigma> \<in> Tinf R" and "u\<cdot>\<sigma> \<in> Tinf R" and "\<not> l\<cdot>\<sigma> \<rhd> u\<cdot>\<sigma>"
    using Tinf_lemma[of R s,OF assms] by auto
  let ?u = "u\<cdot>\<sigma>"
  let ?l = "sharp l"
  let ?r = "sharp u"
  let ?\<sigma> = "Sub (plain \<circ> (case \<sigma> of Sub s \<Rightarrow> s))"

  from `\<not> is_var u` obtain f us where u: "u = Fun f us" by (cases u) auto

  have "(?l, ?r) \<in> DP R"
  proof -
    obtain h vs where "?u = Fun h vs" and "defined R h"
      using Tinf_imp_def_root[OF `wf_trs R` `?u \<in> Tinf R`] by auto
    from `?u = Fun h vs` and `u = Fun f us` have "f = h" by simp
    with `defined R h` have "defined R f" by simp
    moreover from `\<not> (l\<cdot>\<sigma> \<rhd> ?u)` have "\<not> (l \<rhd> u)" by (rule instance_no_subt_imp_no_subt)
    ultimately show "(?l, ?r) \<in> DP R" using `(l, r) \<in> R` and `r \<unrhd> u` by (auto simp: DP_def u)
  qed

  moreover have "?r\<cdot>?\<sigma> = sharp ?u" by (simp add: u sharp_sub_distr)

  moreover have "(sharp s, ?l\<cdot>?\<sigma>) \<in> (rstep (ID R))^*"
  proof -
    from `(l, r) \<in> R` and `wf_trs R` obtain g ts where l: "l = Fun g ts"
      unfolding wf_trs_def by best
        from `(s, l\<cdot>\<sigma>) \<in> (nrrstep R)^*`
      have "(sharp s, sharp(l\<cdot>\<sigma>)) \<in> (rstep(ID R))^*"
      by (rule nrrstep_seq_imp_sharp_rstep_seq)
    moreover have "sharp (l\<cdot>\<sigma>) = ?l\<cdot>?\<sigma>" by (simp add: l)
    ultimately show ?thesis by simp
  qed

  ultimately show ?thesis using `?u \<in> Tinf R` by auto
qed

fun next2 :: "('f,'v)trs \<Rightarrow> ('f,'v)term \<Rightarrow> (('f,'v)term \<times> ('f,'v)term)"
where "next2 R s = (SOME (t, u).
 s \<in> Tinf R \<and> t \<in> Tinf R \<and> u \<in> Tinf R \<and>
 (sharp s,sharp t) \<in> (rstep(ID R))^* \<and> (sharp t,sharp u) \<in> rstep(DP R))"

fun concrete_seq :: "('f,'v)trs \<Rightarrow> ('f,'v)term \<Rightarrow> nat \<Rightarrow> (('f,'v)term \<times> ('f,'v)term \<times> ('f,'v)term)"
where "concrete_seq R s 0       = (s, next2 R s)"
    | "concrete_seq R r (Suc n) = (let (_,_,s) = concrete_seq R r n in (let (t,u) = next2 R s in (s,t,u)))"
   
lemma someI_ex2: "\<exists>x y. P (x,y) \<Longrightarrow> P (SOME (x,y). P (x,y))" using some_eq_ex by auto

definition minimal_R_pred :: 
"('f,'v)trs \<Rightarrow> (nat \<Rightarrow> ('f,'v)term) \<Rightarrow> (nat \<Rightarrow> ('f,'v)term) \<Rightarrow> nat \<Rightarrow> bool"
where "minimal_R_pred R A B i = (
 (A i) \<in> Tinf R \<and> B i \<in> Tinf R \<and> A (Suc i) \<in> Tinf R \<and> 
 (sharp(A i),sharp(B i)) \<in> (rstep(ID R))^* \<and>
 (sharp(B i),sharp(A(Suc i))) \<in> rstep(DP R))"

definition minimal_R_seq :: "('f,'v)trs \<Rightarrow> (nat \<Rightarrow> ('f,'v)term) \<Rightarrow> (nat \<Rightarrow> ('f,'v)term) \<Rightarrow> bool"
where
 "minimal_R_seq R A B = (\<forall>i. minimal_R_pred R A B i)"

theorem not_SN_imp_minimal_R_seq:
 fixes R::"('f,'v)trs"
 assumes "wf_trs R" and "\<not> SN (rstep R)"
 shows "\<exists>A B. minimal_R_seq R A B"
proof -
 from `\<not> SN (rstep R)` obtain s'::"('f,'v)term" where "\<not> SN_elt (rstep R) s'" unfolding SN_def
  by auto
 then obtain s where "s' \<unrhd> s" and "s \<in> Tinf R" using not_SN_imp_subt_Tinf by auto
 let ?A = "\<lambda>i. fst (concrete_seq R s i)"
 let ?B = "\<lambda>i. fst (snd (concrete_seq R s i))"
 let ?P = "\<lambda>s t u. s \<in> Tinf R \<and> t \<in> Tinf R \<and> u \<in> Tinf R \<and> (sharp s,sharp t) \<in> (rstep(ID R))^* \<and> (sharp t,sharp u) \<in> rstep(DP R)"
 let ?P' = "\<lambda>(x,y). ?P s x y"
 have "!!i. ?P (?A i) (?B i) (?A (Suc i))"
 proof -
  fix i show "?P (?A i) (?B i) (?A (Suc i))"
  proof (induct i)
   case 0
   from `wf_trs R` and `s \<in> Tinf R` have "\<exists>t u. ?P' (t,u)" using Tinf_lemma_DP by best
   hence "?P' (next2 R s)" unfolding next2.simps using someI_ex2[where P = "?P'"] by best
   have "?A 0 = s" unfolding concrete_seq.simps by simp
   have "?B 0 = fst (next2 R s)" unfolding concrete_seq.simps by simp
   have "?A (Suc 0) = snd (next2 R s)" unfolding concrete_seq.simps Let_def split_def by simp
   show ?case unfolding `?A 0 = s` `?B 0 = fst (next2 R s)` `?A (Suc 0) = snd (next2 R s)`
   using `?P' (next2 R s)` by auto
  next
   case (Suc i)
   let ?P'' = "\<lambda>(x,y). ?P (?A (Suc i)) x y"
   from Suc have "?A (Suc i) \<in> Tinf R" by simp
   with `wf_trs R` have "\<exists>t u. ?P'' (t,u)" using Tinf_lemma_DP[where s = "?A (Suc i)" and R = "R"] by best
   hence "?P'' (next2 R (?A (Suc i)))" unfolding next2.simps using someI_ex2[where P = "?P''"] by simp
   have "?B (Suc i) = fst (next2 R (?A (Suc i)))" unfolding concrete_seq.simps Let_def split_def by simp
   have "?A (Suc (Suc i)) = snd (next2 R (?A (Suc i)))" unfolding concrete_seq.simps Let_def split_def by simp
   show ?case unfolding `?B (Suc i) = fst (next2 R (?A (Suc i)))` `?A (Suc (Suc i)) = snd (next2 R (?A (Suc i)))`
   using `?P'' (next2 R (?A (Suc i)))` by auto
  qed
 qed
 hence "\<forall>i. ?P (?A i) (?B i) (?A (Suc i))" by (rule allI)
 hence "\<exists>B. \<forall>i. ?P (?A i) (B i) (?A (Suc i))" by (rule exI[where x = "?B"])
 hence "\<exists>A. \<exists>B. \<forall>i. ?P (A i) (B i) (A (Suc i))" by (rule exI[where x = "?A"])
 hence "\<exists>A B. \<forall>i. minimal_R_pred R A B i" unfolding minimal_R_pred_def .
 thus ?thesis unfolding minimal_R_seq_def .
qed

hide const concrete_seq next2

lemma suptp_imp_supteqp_of_arg:
 assumes "Fun f ss \<rhd> t"
 shows "\<exists>s\<in>set ss. s \<unrhd> t"
proof -
 from assms have "Fun f ss \<unrhd> t" and "Fun f ss \<noteq> t" using supt_iff_supteq_neq by auto
 from `Fun f ss \<unrhd> t` show ?thesis
 proof (cases rule: supteqp.cases)
  fix s assume "Fun f ss = s" and "t = s" thus ?thesis using `Fun f ss \<noteq> t` by auto
 next
  fix u ts v g assume "Fun f ss = Fun g ts" and "t = v" and "u \<in> set ts" and "u \<unrhd> v"
  hence "ss = ts" by simp
  from `u \<in> set ts` and `u \<unrhd> v ` have "u \<in> set ss \<and> u \<unrhd> t" unfolding `t = v` and `ss = ts` by auto
  thus ?thesis by auto
 qed
qed

lemma subt_iff_eq_or_subt_of_arg: 
 assumes "s = Fun f ss"
 shows "{t. s \<unrhd> t} = ((\<Union>u \<in> set ss. {t. u \<unrhd> t})\<union>{s})"
using assms proof (induct s)
 case (Var x) thus ?case by auto
next
 case (Fun g ts)
 hence "g = f" and "ts = ss" by auto
 show ?case
 proof
  show "{a. Fun g ts \<unrhd> a} \<subseteq> (\<Union>u\<in>set ss. {a. u \<unrhd> a}) \<union> {Fun g ts}"
  proof
   fix x
   assume "x \<in> {a. Fun g ts \<unrhd> a}"
   hence "Fun g ts \<unrhd> x" by simp
   hence "Fun g ts \<rhd> x \<or> Fun g ts = x" using supteqp_iff_suptp_or_eq by auto
   thus "x \<in> (\<Union>u\<in>set ss. {a. u \<unrhd>a}) \<union> {Fun g ts}"
   proof
    assume "Fun g ts \<rhd> x"
    then obtain u where "u \<in> set ts" and "u \<unrhd> x" using suptp_imp_supteqp_of_arg by best
    hence "x \<in> {a. u \<unrhd> a}" by simp
    with `u \<in> set ts` have "x \<in> (\<Union>u\<in>set ts. {a. u \<unrhd> a})" by auto
    thus ?thesis unfolding `ts = ss` by simp
   next
    assume "Fun g ts = x" thus ?thesis by simp
   qed
  qed
 next
  show "(\<Union>u\<in>set ss. {a. u \<unrhd> a}) \<union> {Fun g ts} \<subseteq> {a. Fun g ts \<unrhd> a}"
  proof
   fix x
   assume "x \<in> (\<Union>u\<in>set ss. {a. u \<unrhd> a}) \<union> {Fun g ts}"
   hence "x \<in> (\<Union>u\<in>set ss. {a. u \<unrhd> a}) \<or> x = Fun g ts" by auto
   thus "x \<in> {a. Fun g ts \<unrhd> a}"
   proof
    assume "x \<in> (\<Union>u\<in>set ss. {a. u \<unrhd> a})"
    then obtain u where "u \<in> set ss" and "u \<unrhd> x" by auto
    thus ?thesis unfolding `ts = ss` by auto
   next
    assume "x = Fun g ts" thus ?thesis by auto
   qed
  qed
 qed
qed

text {* The set of subterms of a term is finite. *}
lemma finite_subterms: "finite{s. t \<unrhd> s}"
proof (induct t)
 case (Var x)
 hence "!!s. (Var x \<unrhd> s) = (Var x = s)" using supteqp.cases by best
 thus ?case unfolding `!!s. (Var x \<unrhd> s) = (Var x = s)` by simp
next
 case (Fun f ss)
 have "Fun f ss = Fun f ss" by simp
 from Fun show ?case
 unfolding subt_iff_eq_or_subt_of_arg[OF `Fun f ss = Fun f ss`] by auto
qed

lemma subsetI3:"(\<And>x y z. (x,y,z) \<in> A \<Longrightarrow> (x,y,z) \<in> B) \<Longrightarrow> A \<subseteq> B" by auto

lemma aux: "(\<Union>a\<in>P. {(x,y,z). x = fst a \<and> y = snd a \<and> Q a z}) = {(x,y,z). (x,y) \<in> P \<and> Q (x,y) z}" (is "?P = ?Q")
proof
 show "?P \<subseteq> ?Q"
 proof
  fix x assume "x \<in> ?P"
  then obtain a where "a \<in> P" and "x \<in> {(x,y,z). x = fst a \<and> y = snd a \<and> Q a z}" by auto
  then obtain b where "Q (fst x, fst (snd x)) (snd (snd x))" and "(fst x, fst (snd x)) = a" and "snd (snd x) = b" by force
  from `a \<in> P` have "(fst a,snd a) \<in> P" unfolding split_def by simp
  from `Q (fst x, fst (snd x)) (snd (snd x))` have "Q a b" unfolding `(fst x, fst (snd x)) = a` `snd (snd x) = b` .
  from `(fst a,snd a) \<in> P` and `Q a b` show "x \<in> ?Q" unfolding split_def `(fst x, fst (snd x)) = a`[symmetric] `snd (snd x) = b`[symmetric] by simp
 qed
next
 show "?Q \<subseteq> ?P"
 proof (rule subsetI3)
  fix x y z assume "(x,y,z) \<in> ?Q"
  hence "(x,y) \<in> P" and "Q (x,y) z" by auto
  hence "x = fst(x,y) \<and> y = snd(x,y) \<and> Q (x,y) z" by auto
  hence "(x,y,z) \<in> {(x,y,z). x = fst(x,y) \<and> y = snd(x,y) \<and> Q (x,y) z}" by auto
  hence "\<exists>p\<in>P. (x,y,z) \<in> {(x,y,z). x = fst p \<and> y = snd p \<and> Q p z}" using `(x,y) \<in> P` by blast
  thus "(x,y,z) \<in> ?P" unfolding UN_iff[symmetric] by simp
 qed
qed
 
lemma finiteR_imp_finiteDP': 
 assumes "finite R"
 shows "finite{(l,r,u). (\<exists>h us. u = Fun h us \<and> (l,r) \<in> R \<and> r \<unrhd> u \<and> defined R h \<and> \<not>(l \<rhd> u))}"
proof -
 have "!!l r. (l,r) \<in> R \<Longrightarrow> finite{u. r \<unrhd> u}"
 proof -
  fix l r
  assume "(l,r) \<in> R"
  show "finite{u. r \<unrhd> u}" by (rule finite_subterms)
 qed
 with `finite R` have "finite(\<Union>(l,r) \<in> R. {u. r \<unrhd> u})" by auto
 have "finite(\<Union>lr\<in>R. {(l,r,u). l = fst lr \<and> r = snd lr \<and> snd lr \<unrhd> u})"
 proof (rule finite_UN_I)
  show "finite R" by (rule `finite R`)
 next
  fix lr
  assume "lr \<in> R"
  have "finite {u. snd lr \<unrhd> u}" by (rule finite_subterms)
  thus "finite {(l,r,u). l = fst lr \<and> r = snd lr \<and> snd lr \<unrhd> u}" by auto
 qed
 hence "finite {(l,r,u). (l,r) \<in> R \<and> r \<unrhd> u}" unfolding aux by auto
 have "{(l,r,u). (l,r) \<in> R \<and> r \<unrhd> u} \<supseteq> {(l,r,u). (\<exists>h us. u = Fun h us \<and> (l,r) \<in> R \<and> r \<unrhd> u \<and> defined R h \<and> \<not>(l \<rhd> u))}" by auto
 with `finite {(l,r,u). (l,r) \<in> R \<and> r \<unrhd> u}` show ?thesis using finite_subset by fast
qed

lemma card_image_le':
 assumes "finite S"
 shows "card (\<Union>y\<in>S.{x. x = f y}) \<le> card S"
proof -
 have A:"(\<Union>y\<in>S. {x. x = f y}) = f ` S" by auto
 from assms show ?thesis unfolding A using card_image_le by auto
qed

lemma finiteR_imp_finiteDP:
 assumes "finite R"
 shows "finite(DP R)"
proof -
 from assms have fS:
  "finite{(l,r,u). (\<exists>h us. u = Fun h us \<and> (l,r) \<in> R \<and> r \<unrhd> u \<and> defined R h \<and> \<not>(l \<rhd> u))}"
  (is "finite ?S")
  by (rule finiteR_imp_finiteDP')
 let ?f = "\<lambda>(x,y,z). (sharp x, sharp z)"
 have eq1: "(\<Union>y\<in>?S. {x. x = ?f y}) = ?f ` ?S" by blast
 with fS have "finite(?f ` ?S)" by auto
 have "DP R = ?f ` ?S" (is "?DP = ?T")
 proof
  show "?DP \<subseteq> ?T"
  proof
   fix x assume "x \<in> ?DP"
   then obtain l r h us where "fst x = sharp l" "snd x = sharp(Fun h us)" "(l,r) \<in> R" "r \<unrhd> Fun h us" "defined R h" " \<not>(l \<rhd> Fun h us)" unfolding DP_def split_def by auto
   hence "(l,r,Fun h us) \<in> ?S" by auto
   thus "x \<in> ?T" unfolding eq1[symmetric]
   proof (rule UN_I)
    have "x = (fst x, snd x)" by simp
    hence "x = (sharp l, sharp(Fun h us))" unfolding `fst x = sharp l` `snd x = sharp(Fun h us)` .
    thus "x \<in> {x. x = ?f (l,r,Fun h us)}" by auto
   qed
  qed
 next
  show "?T \<subseteq> ?DP"
  proof
   fix x assume "x \<in> ?T"
   hence "x \<in> (\<Union>y\<in>?S. {x. x = ?f y})" unfolding eq1 .
   then obtain y where "y \<in> ?S" and "x = ?f y" by fast
   then obtain l r h us where "y = (l,r,Fun h us)" and dp: "(l,r) \<in> R" "r \<unrhd> Fun h us" "defined R h" "\<not>(l \<rhd> Fun h us)" by blast
   with `x = ?f y` have "x = (sharp l, sharp(Fun h us))" by auto
   thus "x \<in> ?DP" using dp unfolding DP_def by auto
  qed
 qed
 with fS show ?thesis by auto
qed

lemma wf_term_sharp_iff_wf_term_plain: "wf_term \<Sigma>^\<sharp> (sharp s) = wf_term \<Sigma>^\<sharp> (plain s)"
proof (induct s)
 case (Var x) thus ?case by simp
next
 case (Fun f ss)
 have symb: "(f@ :s \<Sigma>^\<sharp>) = (f\<sharp> :s \<Sigma>^\<sharp>)" unfolding sharp_sig_def by (cases "f :s \<Sigma>") auto
 hence ar: "arity \<Sigma>^\<sharp> f@ = arity \<Sigma>^\<sharp> f\<sharp>" unfolding sharp_sig_def by auto
 have "wf_term \<Sigma>^\<sharp> (plain (Fun f ss)) = wf_term \<Sigma>^\<sharp> (Fun f@ (map plain ss))" by simp
 also have "\<dots> = wf_term \<Sigma>^\<sharp> (Fun f\<sharp> (map plain ss))" unfolding wf_term.simps symb ar by simp
 finally show ?case by simp
qed

lemma vars_sharp_eq_vars_plain: "vars_term(sharp s) = vars_term(plain s)" by (induct s) auto

lemma vars_sharp_eq_vars: "vars_term(sharp t) = vars_term t"
proof (induct t)
 case (Var x) thus ?case by auto
next
 case (Fun f ss)
 hence IH: "!!s. s \<in> set ss \<Longrightarrow> vars_term(plain s) = vars_term s" unfolding vars_sharp_eq_vars_plain .
 have "vars_term(sharp (Fun f ss)) = \<Union>set (map vars_term (map plain ss))" by simp
 thus ?case using IH by auto
qed

lemma vars_plain_eq_vars: "vars_term(plain s) = vars_term s" by (induct s) auto

lemma wf_term_sharp_eq_wf_term: "wf_term \<Sigma>^\<sharp> (sharp s) = wf_term \<Sigma> s"
proof -
 have A: "wf_term \<Sigma>^\<sharp> (sharp s) = wf_term \<Sigma>^\<sharp> (plain s)" by (rule wf_term_sharp_iff_wf_term_plain)
 show ?thesis unfolding A
 proof (induct s)
  case (Var x) thus ?case by simp
 next
  case (Fun f ss)
  have "(\<forall>s \<in> set (map plain ss). wf_term \<Sigma>^\<sharp> s) = (\<forall>s\<in>set ss. wf_term \<Sigma>^\<sharp> (plain s))" by simp
  hence A: "(\<forall>s \<in> set (map plain ss). wf_term \<Sigma>^\<sharp> s) = (\<forall>s\<in>set ss. wf_term \<Sigma> s)" using Fun by auto
  have "wf_term \<Sigma>^\<sharp> (plain (Fun f ss)) = wf_term \<Sigma>^\<sharp> (Fun f@ (map plain ss))" by simp
  also have "\<dots> = (f@ :s \<Sigma>^\<sharp> \<and> arity \<Sigma>^\<sharp> f@ = length (map plain ss) \<and> (\<forall>s\<in>set (map plain ss). wf_term \<Sigma>^\<sharp> s))" by auto
  also have "\<dots> = (f :s \<Sigma> \<and> arity \<Sigma> f = length ss \<and> (\<forall>s\<in>set (map plain ss). wf_term \<Sigma>^\<sharp> s))"
   unfolding sharp_sig_def by (cases "f :s \<Sigma>") auto
  also have "\<dots> = wf_term \<Sigma> (Fun f ss)" unfolding wf_term.simps A by simp
  finally show ?case .
 qed
qed

lemma wf_term_plain_eq_wf_term: "wf_term \<Sigma>^\<sharp> (plain s) = wf_term \<Sigma> s"
using wf_term_sharp_eq_wf_term unfolding wf_term_sharp_iff_wf_term_plain .

lemma wf_trs_imp_wf_DP:
 assumes "wf_trs R"
 shows "wf_trs (DP R)"
unfolding wf_trs_def
proof (intro allI impI)
 fix s t
 assume "(s,t) \<in> DP R"
 then obtain l r h us where "s = sharp l" "t = sharp(Fun h us)" "(l,r) \<in> R" "r \<unrhd> (Fun h us)" "defined R h" "\<not>(l \<rhd> Fun h us)"
  unfolding DP_def by auto
 from `wf_trs R` and `(l,r) \<in> R` have "\<exists>f ss. l = Fun f ss"
  "vars_term r \<subseteq> vars_term l" unfolding wf_trs_def by auto
 from `\<exists>f ss. l = Fun f ss` obtain f ss where "l = Fun f ss" by auto
 hence "s = Fun f\<sharp> (map plain ss)" unfolding `s = sharp l` by simp
 hence "\<exists>f ss. s = Fun f ss" by auto
 from `r \<unrhd> Fun h us` have "vars_term(Fun h us) \<le> vars_term r" by (induct rule: supteqp.induct) auto
 hence "vars_term t \<subseteq> vars_term s" unfolding `s = sharp l` `t = sharp(Fun h us)` vars_sharp_eq_vars using `vars_term r \<le> vars_term l` by simp
 from `\<exists>f ss. s = Fun f ss` `vars_term t \<subseteq> vars_term s`
 show "(\<exists>f ss. s = Fun f ss) \<and> vars_term t \<subseteq> vars_term s" by simp
qed


lemma wf_trs_imp_wf_ID:
 assumes "wf_trs R"
 shows "wf_trs (ID R)"
unfolding wf_trs_def
proof (intro allI impI)
 fix s t
 assume "(s,t) \<in> ID R"
 then obtain l r where "s = plain l" "t = plain r" "(l,r) \<in> R" unfolding ID_def by best
 from `wf_trs R` and `(l,r) \<in> R` have "\<exists>f ss. l = Fun f ss"
  "vars_term r \<subseteq> vars_term l" unfolding wf_trs_def by auto
 from `\<exists>f ss. l = Fun f ss` obtain f ss where "l = Fun f ss" by auto
 hence "s = Fun f@ (map plain ss)" unfolding `s = plain l` by simp
 hence "\<exists>f ss. s = Fun f ss" by auto
 from `vars_term r \<subseteq> vars_term l` have "vars_term t \<subseteq> vars_term s" unfolding `s = plain l` `t = plain r` vars_plain_eq_vars .
 from `\<exists>f ss. s = Fun f ss` `vars_term t \<subseteq> vars_term s`
 show "(\<exists>f ss. s = Fun f ss) \<and> vars_term t \<subseteq> vars_term s" by simp
qed


lemma subteq_of_plain_imp_plain: "plain s \<unrhd> t \<Longrightarrow> \<exists>u. t = plain u"
proof (induct s arbitrary: t)
 case (Var x)
 hence "plain (Var x) \<rhd> t \<or> plain (Var x) = t" unfolding supteqp_iff_suptp_or_eq .
 thus ?case
 proof
  assume "plain (Var x) \<rhd> t" thus ?thesis by (cases rule: suptp.cases) auto
 next
  assume "plain (Var x) = t" thus ?thesis by best
 qed
next
 case (Fun f ss)
 hence "plain (Fun f ss) \<rhd> t \<or> plain (Fun f ss) = t" unfolding supteqp_iff_suptp_or_eq by simp
 thus ?case
 proof
  assume "plain (Fun f ss) \<rhd> t"
  thus ?case using Fun by (cases rule: suptp.cases) (auto simp: suptp_iff_supteqp_neq)
 next
  assume "plain (Fun f ss) = t" thus ?thesis by best
 qed
qed

lemma subt_of_sharp_imp_plain: "sharp s \<rhd> t \<Longrightarrow> \<exists>u. t = plain u"
proof (induct s)
 case (Var x) thus ?case by (cases rule: suptp.cases) auto
next
 case (Fun f ss)
 hence "Fun f\<sharp> (map plain ss) \<rhd> t" by simp
 then obtain u where "u\<in>set (map plain ss)" and "u \<unrhd> t" using suptp_imp_supteqp_of_arg[where f = "f\<sharp>" and ss = "map plain ss" and t = t] by fast
 from `u\<in>set(map plain ss)` obtain v where "u = plain v" by auto
 with `u \<unrhd> t` have "plain v \<unrhd> t" by simp
 thus ?case by (rule subteq_of_plain_imp_plain)
qed

lemma rstepDP_imp_rrstepDP: 
 assumes "(sharp s,sharp t) \<in> rstep(DP R)" and "wf_trs R"
 shows "(sharp s,sharp t) \<in> rrstep(DP R)" 
using assms proof (cases rule:rstep_cases)
 assume "(sharp s,sharp t) \<in> rrstep(DP R)" thus ?thesis .
next
 assume "(sharp s,sharp t) \<in> nrrstep(DP R)"
 then obtain l r i ps \<sigma> where "(sharp s,sharp t) \<in> rstep_r_p_s (DP R) (l,r) (i<#ps) \<sigma>" unfolding nrrstep_def by auto
 hence "(l,r) \<in> DP R" and "(ctxt_of_pos_term (i<#ps) (sharp s))\<langle>l\<cdot>\<sigma>\<rangle> = (sharp s)"
   (is "?C\<langle>l\<cdot>\<sigma>\<rangle> = sharp s") unfolding rstep_r_p_s_def Let_def by auto
 from `wf_trs R` have "wf_trs (DP R)" by (rule wf_trs_imp_wf_DP)
 with `(l,r) \<in> DP R` obtain f ss where "l = Fun f ss" unfolding wf_trs_def by best
 from `(sharp s,sharp t) \<in> nrrstep(DP R)` obtain g ts where "(sharp s) = Fun g ts"
  using nrrstep_imp_Fun by best
 have "?C \<noteq> \<box>" unfolding `(sharp s) = Fun g ts` by auto
 hence "(sharp s) \<rhd> (l\<cdot>\<sigma>)" using nectxt_imp_suptp_ctxt[where C = ?C and t = "l\<cdot>\<sigma>"] unfolding `?C\<langle>l\<cdot>\<sigma>\<rangle> = sharp s` by simp
 then obtain u where "(l\<cdot>\<sigma>) = plain u" using subt_of_sharp_imp_plain[where s = s and t = "l\<cdot>\<sigma>"] by auto
 hence "Fun f ss\<cdot>\<sigma> = plain u" unfolding `l = Fun f ss` by simp
 hence "Fun f (map (\<lambda>x. x\<cdot>\<sigma>) ss) = plain u" by simp
 hence "\<exists>h us. plain u = plain (Fun h us)" by (induct u) auto
 then obtain h us where "plain u = plain (Fun h us)" by auto
 hence "root(plain u) = Some h@" by simp
 from `(l,r) \<in> DP R` obtain lhs rhs where "l = sharp lhs" and "(lhs,rhs) \<in> R" unfolding DP_def by auto
 with `wf_trs R` obtain e vs where "lhs = Fun e vs" unfolding wf_trs_def by best
 hence "l = sharp (Fun e vs)" unfolding `l = sharp lhs` by simp
 hence "root l = Some e\<sharp>" by auto
 from `root (plain u) = Some h@` have "root(Fun f ss\<cdot>\<sigma>) = Some h@" unfolding `Fun f ss\<cdot>\<sigma> = plain u` by simp
 hence "root (Fun f ss) = Some h@" by simp
 hence "root l = Some h@" unfolding `l = Fun f ss` by simp
 with `root l = Some e\<sharp>` show ?thesis by auto
qed

lemma rstepID_imp_nrrstepID:
 assumes "(sharp s,sharp t) \<in> rstep(ID R)" and "wf_trs R"
 shows "(sharp s,sharp t) \<in> nrrstep(ID R)"
proof (rule ccontr)
 assume "(sharp s,sharp t) \<notin> nrrstep(ID R)"
 from assms have "(sharp s,sharp t) \<in> rrstep(ID R) \<or> (sharp s,sharp t) \<in> nrrstep(ID R)" using rstep_iff_rrstep_or_nrrstep by auto
 with `(sharp s,sharp t) \<notin> nrrstep(ID R)` have "(sharp s,sharp t) \<in> rrstep(ID R)" by simp
 then obtain l r \<sigma> where "(sharp s,sharp t) \<in> rstep_r_p_s (ID R) (l,r) \<epsilon> \<sigma>" unfolding rrstep_def by best
 hence "(l,r) \<in> ID R" and "(ctxt_of_pos_term \<epsilon> (sharp s))\<langle>l\<cdot>\<sigma>\<rangle> = (sharp s)" (is "?C\<langle>l\<cdot>\<sigma>\<rangle> = sharp s") unfolding rstep_r_p_s_def Let_def by auto
 hence "?C = \<box>" by auto
 with `?C\<langle>l\<cdot>\<sigma>\<rangle> = sharp s` have "l\<cdot>\<sigma> = sharp s" by simp
 from `wf_trs R` have "wf_trs (ID R)" by (rule wf_trs_imp_wf_ID)
 with `(l,r) \<in> ID R` obtain f ss where "l = Fun f ss" unfolding wf_trs_def by best
 from `(l,r) \<in> ID R` obtain u where "l = plain u" unfolding ID_def by best
 hence "Fun f ss = plain u" unfolding `l = Fun f ss` .
 then obtain h us where "plain u = plain (Fun h us)" by (induct u) auto
 hence "l = plain (Fun h us)" unfolding `l = plain u` .
 hence "root l = Some h@" by simp
 from `l\<cdot>\<sigma> = sharp s` have "Fun f (map (\<lambda>x. x\<cdot>\<sigma>) ss) = sharp s" unfolding `l = Fun f ss` by simp
 hence "\<exists>f. root (sharp s) = Some f\<sharp>" by (induct s) auto
 then obtain g where "root (sharp s) = Some g\<sharp>" by best
 have "root (Fun f (map (\<lambda>x. x\<cdot>\<sigma>) ss)) = root (Fun f ss)" by simp
 hence "root (l\<cdot>\<sigma>) = root l" unfolding `l = Fun f ss` by simp
 with `root (sharp s) = Some g\<sharp>` have "root l = Some g\<sharp>" unfolding `l\<cdot>\<sigma> = sharp s` by simp
 with `root l = Some h@` show False by auto
qed

lemma rstepDP_imp_ruleDP: 
 assumes "(sharp s,sharp t) \<in> rstep(DP R)" and "wf_trs R"
 shows "\<exists>l r \<sigma>. (l,r) \<in> DP R \<and> l\<cdot>\<sigma> = sharp s \<and> r\<cdot>\<sigma> = sharp t"
proof -
 from assms have "(sharp s,sharp t) \<in> rrstep(DP R)" by (rule rstepDP_imp_rrstepDP)
 thus ?thesis unfolding rrstep_def rstep_r_p_s_def Let_def by auto
qed

lemma rstep_seq_imp_plain_rstep_seq:
 shows "(s,t) \<in> (rstep R)^* \<Longrightarrow> (plain s,plain t) \<in> (rstep(ID R))^*"
proof (induct rule: rtrancl_induct)
 case base thus ?case by simp
next
 case (step a b)
 from `(a,b) \<in> rstep R` have "(plain a,plain b) \<in> rstep(ID R)" using rstep_iff_plain_rstep by auto
 with step show ?case by auto
qed

lemma plain_eq_imp_eq:
 assumes "plain s = plain t"
 shows "s = t"
using assms proof (induct s arbitrary: t)
 case (Var x) with assms show ?case by (induct t) auto
next
 case (Fun f ss) thus ?case
 proof (induct t)
  case (Var y) thus ?case by auto
 next
  case (Fun g ts)
  hence A: "map plain ss = map plain ts" by simp
  hence len_eq:"length ss = length ts" by (rule map_eq_imp_length_eq)
  from A have "!!i. i < length ss \<Longrightarrow> (map plain ss)!i = (map plain ts)!i" by auto
  with len_eq have eq: "!!i. i < length ss \<Longrightarrow> plain (ss!i) = plain (ts!i)" using nth_map by auto
  have in_set: "!!i. i < length ss \<Longrightarrow> (ss!i) \<in> set ss" by auto
  from Fun have "\<forall>i < length ss. (ss!i) = (ts!i)" using in_set eq by auto
  with len_eq have "ss = ts" using nth_equalityI[where xs = ss and ys = ts] by simp
  have "f = g" using Fun by auto
  with `ss = ts` show ?case by simp
 qed
qed

lemma plain_unsharp_plain_eq_plain[simp]:"plain(unsharp(plain t)) = plain t"
proof (induct t)
 case (Var x) thus ?case by simp
next
 case (Fun f ss)
 from Fun have IH: "\<forall>s \<in> set ss. (plain \<circ> unsharp \<circ> plain) s = plain s" by simp
 have "plain (unsharp (plain (Fun f ss))) = plain (unsharp (Fun f@ (map plain ss)))" by simp
 also have "\<dots> = plain (Fun f (map unsharp (map plain ss)))" by simp
 also have "\<dots> = Fun f@ (map plain (map unsharp (map plain ss)))" by simp
 also have "\<dots> = Fun f@ (map (plain \<circ> unsharp) (map plain ss))" by simp
 also have "\<dots> = Fun f@ (map (plain \<circ> unsharp \<circ> plain) ss)" by simp
 finally show ?case using IH by auto
qed

lemma map_plain_unsharp_plain_eq_map_plain[simp]: "map plain (map unsharp (map plain ts)) = map plain ts"
proof -
 have "map plain (map unsharp (map plain ts)) = map (plain \<circ> unsharp) (map plain ts)" by simp
 also have "\<dots> = map (plain \<circ> unsharp \<circ> plain) ts" by simp
 finally show ?thesis by simp
qed

lemma ctxt_plain_unsharp_plain_eq_plain[simp]:"plain_ctxt(unsharp_ctxt(plain_ctxt E)) = (plain_ctxt E)"
proof (induct E)
 case Hole thus ?case by simp
next
 case (More f ss1 D ss2)
 have "plain_ctxt (unsharp_ctxt (plain_ctxt (More f ss1 D ss2))) = plain_ctxt (unsharp_ctxt (More f@ (map plain ss1) (plain_ctxt D) (map plain ss2)))" by simp
 also have "\<dots> = plain_ctxt (More f (map unsharp (map plain ss1)) (unsharp_ctxt (plain_ctxt D)) (map unsharp (map plain ss2)))" by simp
 also have "\<dots> = More f@ (map plain (map unsharp (map plain ss1))) (plain_ctxt D) (map plain (map unsharp (map plain ss2)))" using More by simp
 also have "\<dots> = More f@ (map (plain \<circ> unsharp) (map plain ss1)) (plain_ctxt D) (map (plain \<circ> unsharp) (map plain ss2))" by simp
 also have "\<dots> = More f@ (map (plain \<circ> unsharp \<circ> plain) ss1) (plain_ctxt D) (map (plain \<circ> unsharp \<circ> plain) ss2)" by simp
 also have "\<dots> = More f@ (map plain ss1) (plain_ctxt D) (map plain ss2)" by auto
 finally show ?case by auto
qed

lemma nth_take_prefix:
 "length ys \<le> length xs \<Longrightarrow> \<forall>i < length ys. xs!i = ys!i \<Longrightarrow> take (length ys) xs = ys"
proof (induct ys arbitrary: xs)
 case Nil thus ?case by auto
next
 case (Cons y ys) note IH = this
 show ?case
 proof (cases xs)
  case Nil thus ?thesis using Cons by simp
 next
  case (Cons z zs)
  hence "xs!0 = z" by simp
  with IH have "z = y" by auto
  from IH have "\<forall>i < length ys. xs!(Suc i) = ys!i" by auto
  hence A:"\<forall>i < length ys. zs!i = ys!i" using Cons by auto
  have B:"length ys \<le> length zs" using IH Cons by auto
  have C:"take (length ys) zs = ys" using IH A B by simp
  have "take (length (y#ys)) (z#zs) = z#(take (length ys) zs)" by simp
  also have "\<dots> = z#ys" using C by simp
  finally show ?thesis unfolding `z = y` `xs = z#zs` .
 qed
qed

lemma unsharp_plain_eq[simp]: "unsharp(plain s) = s"
proof (induct s)
 case (Var x) thus ?case by simp
next
 case (Fun f ss)
 hence "\<forall>i<length ss. (unsharp\<circ>plain) s = s" by simp
 hence IH: "map (unsharp\<circ>plain) ss = ss" by (auto simp: list_eq_iff_nth_eq)
 have "unsharp(plain(Fun f ss)) = Fun f (map (unsharp\<circ>plain) ss)" by simp
 thus ?case using IH by simp
qed

lemma drop_length: "i < length xs \<Longrightarrow> xs!i#drop (Suc i) xs = drop i xs"
proof (induct xs arbitrary: i)
 case Nil thus ?case by simp
next
 case (Cons x xs) thus ?case by (induct i) auto
qed

lemma plain_plain_ctxt: "plain(C\<langle>s\<rangle>) = (plain_ctxt C)\<langle>plain s\<rangle>" by (induct C) auto

lemma unsharp_plain: "map unsharp (map plain ss) = ss" by (induct ss) auto

lemma plain_ctxt_term: "plain s = C\<langle>t\<rangle> \<Longrightarrow> t = plain(unsharp t) \<and> C = plain_ctxt(unsharp_ctxt C)"
proof (induct C arbitrary: s)
 case Hole 
 hence "t = plain s" by simp
 show ?case unfolding `t = plain s` by simp
next
 case (More f ss1 D ss2)
 let ?C = "More f ss1 D ss2"
 from More show ?case
 proof (induct s)
  case (Var x) thus ?case by simp
 next
  case (Fun g ss)
  hence "f = g@" and args:"map plain ss = ss1@D\<langle>t\<rangle>#ss2" (is "_ = ?ss") by auto
  hence lsum:"length ss = length ss1 + 1 + length ss2" using length_map[where f = plain and xs = ss] by simp
  with args have "\<forall>i<length ss. (map plain ss)!i = ?ss!i" by simp
  hence "plain(ss!(length ss1)) = D\<langle>t\<rangle>" (is "plain(ss!?i) = _") using nth_map lsum by simp
  with Fun have one:"t = plain(unsharp t)" and "plain_ctxt(unsharp_ctxt D) = D" by auto
  from args have A:"take ?i (map plain ss) = ss1" by simp
  from args have B:"drop (Suc ?i) (map plain ss) = ss2" by simp
  from args have C:"map unsharp ?ss = ss" using unsharp_plain[where ss = ss] by simp
  have "take ?i ss = take ?i (map unsharp ?ss)" unfolding C by simp
  also have "\<dots> = map unsharp (take ?i ?ss)" unfolding take_map by simp
  finally have A':"take ?i ss = map unsharp ss1" by simp
  have "drop (Suc ?i) ss = drop (Suc ?i) (map unsharp ?ss)" unfolding C by simp
  also have "\<dots> = map unsharp (drop (Suc ?i) ?ss)" unfolding drop_map by simp
  finally have B':"drop (Suc ?i) ss = map unsharp ss2" by simp
  have "More f ss1 D ss2 = More f (take ?i (map plain ss)) D (drop (Suc ?i) (map plain ss))" unfolding A B by simp
  also have "\<dots> = More f (map plain (take ?i ss)) (plain_ctxt(unsharp_ctxt D)) (map plain (drop (Suc ?i) ss))" unfolding `plain_ctxt(unsharp_ctxt D) = D` take_map drop_map by simp
  also have "\<dots> = plain_ctxt(More g (take ?i ss) (unsharp_ctxt D) (drop (Suc ?i) ss))" unfolding `f = g@` by simp
  also have "\<dots> = plain_ctxt(More g (map unsharp ss1) (unsharp_ctxt D) (map unsharp ss2))" unfolding A' B' by simp
  finally show ?case using one by simp
 qed
qed
  
lemma plain_ctxt_term': assumes "(plain s) = (C\<langle>t\<rangle>)" shows "\<exists>D u. t = plain u \<and> C = plain_ctxt D"
using plain_ctxt_term[OF assms] by auto

lemma pair_fun_eq[simp]:
 fixes f :: "'a \<Rightarrow> 'b" and g :: "'b \<Rightarrow> 'a"
 shows "((\<lambda>(x,y). (x,f y)) \<circ> (\<lambda>(x,y). (x,g y))) = (\<lambda>(x,y). (x,(f \<circ> g) y))" (is "?f = ?g")
proof (rule ext)
 fix ab :: "'c * 'b"
 obtain a b where "ab = (a,b)" by force
 have "((\<lambda>(x,y). (x,f y))\<circ>(\<lambda>(x,y). (x,g y))) (a,b) = (\<lambda>(x,y). (x,(f\<circ>g) y)) (a,b)" by simp
 thus "?f ab = ?g ab" unfolding `ab = (a,b)` by simp
qed

lemma sub_ext: "\<sigma> = \<tau> \<Longrightarrow> \<forall>s. s\<cdot>\<sigma> = s\<cdot>\<tau>" by auto

lemma restrict_singleton: assumes "x \<in> dom \<sigma>" shows "\<exists>t. \<sigma>|_{x} = (Sub(\<lambda>y. if y = x then t else Var y))"
proof (cases \<sigma>)
 case (Sub s)
 have "(Sub s)|_{x} = Sub(\<lambda>y. if y = x then s y else Var y)" by simp
 have "(Sub s)|_{x} = Sub(\<lambda>y. if y = x then s x else Var y)" by (simp cong: if_cong)
 thus ?thesis unfolding Sub by (rule exI[where x = "s x"])
qed

lemma sub_case_aux: "(case (Sub t) of Sub s \<Rightarrow> s) = t" by simp

lemma sub_restrict_f: 
 assumes "\<And>x. f(Var x) = Var x"
 shows "Sub(f \<circ> (case \<sigma> of Sub s \<Rightarrow> s))|_V = Sub(f \<circ> (case \<sigma>|_V of Sub s \<Rightarrow> s))"
proof (cases \<sigma>)
 case (Sub s)
 have "Sub(f \<circ> (case \<sigma>|_V of Sub s \<Rightarrow> s)) = Sub(f \<circ> (case (Sub s)|_V of Sub s \<Rightarrow> s))" unfolding Sub by simp
 also have "\<dots> = Sub(f \<circ> (case Sub(\<lambda>z. if z \<in> V then s z else Var z) of Sub s \<Rightarrow> s))" by simp
 also have "\<dots> = Sub(f \<circ> (\<lambda>z. if z \<in> V then s z else Var z))" by simp
 also have "\<dots> = Sub(\<lambda>z. f(if z \<in> V then s z else Var z))" unfolding o_def by simp
 also have "\<dots> = Sub(\<lambda>z. if z \<in> V then f(s z) else f(Var z))" unfolding if_distrib by simp
 also have "\<dots> = Sub(\<lambda>z. if z \<in> V then f(s z) else Var z)" using assms by (auto cong: if_cong)
 also have "\<dots> = Sub(\<lambda>z. if z \<in> V then (f \<circ> s) z else Var z)" unfolding o_def by simp
 also have "\<dots> = Sub(f \<circ> (case Sub s of Sub s \<Rightarrow> s))|_V" by simp
 finally have "Sub(f \<circ> (case \<sigma>|_V of Sub s \<Rightarrow> s)) = Sub(f \<circ> (case \<sigma> of Subs \<Rightarrow> s))|_V" by simp
 thus ?thesis unfolding `\<sigma> = Sub s` unfolding sub_case_aux by simp
qed

lemma unsharp_Var_id: "\<And>x. unsharp(Var x) = Var x" by simp
lemma plain_Var_id: "\<And>x. plain(Var x) = Var x" by simp

lemma plain_sub_term: assumes "plain s = t\<cdot>\<sigma>"
shows "t = plain(unsharp t) \<and> t\<cdot>(\<sigma>|_(vars_term t)) = t\<cdot>(Sub(plain \<circ> unsharp \<circ> (case \<sigma>|_(vars_term t) of Sub s \<Rightarrow> s)))"
using assms proof (induct t arbitrary: s)
 case (Var x)
 obtain s' where "\<sigma> = Sub s'" by (cases \<sigma>) simp
 have "\<sigma>|_(vars_term(Var x)) = \<sigma>|_{x}" by simp
 have "Var x = plain(unsharp(Var x))" by simp
 show ?case
 proof (cases "x \<in> dom \<sigma>")
 assume "x \<in> dom \<sigma>"
 then obtain t where singleton:"\<sigma>|_{x} = Sub(\<lambda>z. if z = x then t else Var z)" (is "\<sigma>|_{x} = ?\<sigma> t") using restrict_singleton by best
 from `plain s = (Var x)\<cdot>\<sigma>` have "plain s = (Var x)\<cdot>(\<sigma>|_vars_term(Var x))" unfolding coincidence_lemma[where t = "Var x" and \<sigma> = \<sigma>] by simp
  hence "plain s = (Var x)\<cdot>(\<sigma>|_{x})" by simp
  hence "plain s = (Var x)\<cdot>(?\<sigma> t)" unfolding singleton by simp
  hence "plain s = t" by simp
  let ?\<tau> = "Sub(unsharp \<circ> (case \<sigma>|_{x} of Sub s \<Rightarrow> s))"
  have "\<sigma>|_vars_term(Var x) = ?\<sigma> t" unfolding `\<sigma>|_vars_term(Var x) = \<sigma>|_{x}` using singleton by simp
  also have "\<dots> = ?\<sigma> t" unfolding `\<sigma>|_{x} = ?\<sigma> t` by simp
  also have "\<dots> = ?\<sigma>(plain s)" unfolding `plain s = t` by simp
  also have "\<dots> = Sub(\<lambda>z. if z = x then plain s else plain(Var z))" by (simp cong: if_cong)
  also have "\<dots> = Sub(\<lambda>z. plain(if z = x then s else Var z))" unfolding if_distrib by simp
  also have "\<dots> = Sub(\<lambda>z. plain(unsharp(plain(if z = x then s else Var z))))" by simp
  also have "\<dots> = Sub(plain \<circ> unsharp \<circ> (\<lambda>z. plain(if z = x then s else Var z)))" unfolding o_def by simp
  also have "\<dots> = Sub(plain \<circ> unsharp \<circ> (\<lambda>z. if z = x then plain s else Var z))" unfolding if_distrib by (simp cong: if_cong)
  also have "\<dots> = Sub(plain \<circ> unsharp \<circ> (\<lambda>z. if z = x then t else Var z))" unfolding `plain s = t` by simp
  also have "\<dots> = Sub(plain \<circ> unsharp \<circ> (case Sub(\<lambda>z. if z = x then t else Var z) of Sub t \<Rightarrow> t))" unfolding sub_case_aux by simp
  also have "\<dots> = Sub(plain \<circ> unsharp \<circ> (case \<sigma>|_{x} of Sub t \<Rightarrow> t))" unfolding singleton by simp
  also have "\<dots> = Sub(plain \<circ> (case ?\<tau> of Sub t \<Rightarrow> t))" unfolding sub_case_aux unfolding o_assoc by simp
  finally have "\<sigma>|_vars_term(Var x) = Sub(plain \<circ> (case ?\<tau> of Sub t \<Rightarrow> t))" by simp
  hence "(Var x)\<cdot>(\<sigma>|_vars_term(Var x)) = (Var x)\<cdot>(Sub(plain \<circ> (case ?\<tau> of Sub t \<Rightarrow> t)))" unfolding `\<sigma>|_vars_term(Var x) = \<sigma>|_{x}` by simp
  with `Var x = plain(unsharp(Var x))` show ?thesis unfolding `\<sigma>|_vars_term(Var x) = \<sigma>|_{x}` unfolding sub_case_aux unfolding o_assoc by (rule conjI)
 next
  assume "x \<notin> dom \<sigma>"
  have "\<sigma>|_vars_term(Var x) = \<sigma>|_{x}" by simp
  also have "\<dots> = \<iota>" using `x \<notin> dom \<sigma>` restrict_empty by best
  finally have "\<sigma>|_vars_term(Var x) = Sub(plain \<circ> unsharp \<circ> (case \<sigma>|_vars_term(Var x) of Sub t \<Rightarrow> t))" unfolding `\<sigma>|_vars_term(Var x) = \<sigma>|_{x}` unfolding `\<sigma> = Sub s'` unfolding o_def by simp
  hence "(Var x)\<cdot>(\<sigma>|_vars_term(Var x)) = (Var x)\<cdot>(Sub(plain \<circ> unsharp \<circ> (case \<sigma>|_vars_term(Var x) of Sub t \<Rightarrow> t)))" by simp
  thus ?thesis using `Var x = plain(unsharp(Var x))` by simp
 qed
next
 case (Fun g ts)
 hence "plain s = Fun g (map (\<lambda>x. x\<cdot>\<sigma>) ts)" (is "_ = Fun g ?ts") by simp
 then obtain f ss where "s = Fun f ss" and "g = f@" by (induct s) auto
 hence "plain s = Fun f@ (map plain ss)" (is "_ = Fun _ ?ss") by simp
 with `plain s = Fun g ?ts` have "?ss = ?ts" by simp
 hence "length ?ss = length ?ts" by simp
 hence "length ts = length ss" by simp
 obtain s' where "\<sigma> = Sub s'" by (cases \<sigma>) auto
 have ss:"\<forall>i<length ts. (map plain ss)!i = plain(ss!i)" unfolding `length ts = length ss` using nth_map by simp
 have ts:"\<forall>i<length ts. ?ts!i = (ts!i)\<cdot>\<sigma>" using nth_map by simp
 hence eq:"\<forall>i<length ts. plain(ss!i) = (ts!i)\<cdot>\<sigma>" using ss ts `?ss = ?ts` by simp
 have set:"\<forall>i<length ts. (ts!i) \<in> set ts" by simp
 have "length ts = length(map plain (map unsharp ts))" by simp
 have IH:"\<forall>i<length ts. (ts!i) = plain(unsharp(ts!i)) \<and>
          (ts!i)\<cdot>(\<sigma>|_vars_term(ts!i)) = (ts!i)\<cdot>(Sub(plain \<circ> unsharp \<circ> (case \<sigma>|_vars_term(ts!i) of Sub s \<Rightarrow> s)))" using Fun eq set by best
 hence "\<forall>i<length ts. (ts!i) = plain(unsharp(ts!i))" by simp
 hence "\<forall>i<length ts. (ts!i) = plain(map unsharp ts!i)" by simp
 hence "\<forall>i<length ts. (ts!i) = (map plain (map unsharp ts))!i" by simp
 hence "ts = map plain (map unsharp ts)" using list_eq_iff_nth_eq[where xs = ts and ys = "map plain (map unsharp ts)"] `length ts = length(map plain (map unsharp ts))` by best
 have "plain(unsharp(Fun g ts)) = Fun g (map plain (map unsharp ts))" unfolding `g = f@` by simp
 also have "\<dots> = Fun g ts" unfolding `ts = map plain (map unsharp ts)`[symmetric] by simp
 finally have one:"Fun g ts = plain(unsharp(Fun g ts))" by simp
 have "\<forall>i < length ts. vars_term(ts!i) \<subseteq> \<Union>set(map vars_term ts)"
 proof (induct ts)
  case Nil thus ?case by simp
 next
  case (Cons t ts)
  have zero:"vars_term((t#ts)!0) \<subseteq> \<Union>set(map vars_term (t#ts))" by auto
  show ?case
  proof
   fix i have "i = (0::nat) \<or> i \<noteq> 0" by auto
   thus "i < length(t#ts) \<longrightarrow> vars_term((t#ts)!i) \<subseteq> \<Union>set(map vars_term (t#ts))"
   proof
    assume "i = 0" thus ?thesis using zero by auto
   next
    assume "i \<noteq> 0" show ?thesis
    proof
     assume "i < length(t#ts)"
     with `i \<noteq> 0` have "((t#ts)!i) \<in> set ts" by (induct i) auto
     with Cons show "vars_term((t#ts)!i) \<subseteq> \<Union>set(map vars_term(t#ts))" by auto
    qed
   qed
  qed
 qed
 hence a:"\<And>i. i < length ts \<Longrightarrow> vars_term(ts!i) \<subseteq> vars_term(Fun g ts)" by simp
 hence b:"!!i. i < length ts \<Longrightarrow> (ts!i)\<cdot>(\<sigma>|_vars_term(ts!i)) = (ts!i)\<cdot>(\<sigma>|_vars_term(Fun g ts))" using coincidence_lemma' by best
 let ?un = "\<lambda>\<sigma>. Sub(unsharp \<circ> (case \<sigma> of Sub s \<Rightarrow> s))"
 let ?pl = "\<lambda>\<sigma>. Sub(plain \<circ> (case \<sigma> of Sub s \<Rightarrow> s))"
 from IH have A:"!!i. i < length ts \<Longrightarrow> (ts!i)\<cdot>(\<sigma>|_vars_term(ts!i)) = (ts!i)\<cdot>(?pl (?un (\<sigma>|_vars_term(ts!i))))" unfolding sub_case_aux unfolding o_assoc by simp
 have "Var x = unsharp(Var x)" by simp
 have c:"!!i. i < length ts \<Longrightarrow> (ts!i)\<cdot>(?pl (?un (\<sigma>|_vars_term(ts!i)))) = (ts!i)\<cdot>(?pl (?un (\<sigma>|_vars_term(Fun g ts))))" 
 proof -
  fix i assume "i < length ts"
  have "(ts!i)\<cdot>(?pl (?un (\<sigma>|_vars_term(ts!i)))) = (ts!i)\<cdot>(?pl ((?un \<sigma>)|_vars_term(ts!i)))" unfolding sub_restrict_f[OF unsharp_Var_id] by simp
  also have "\<dots> = (ts!i)\<cdot>((?pl (?un \<sigma>))|_vars_term(ts!i))" unfolding sub_restrict_f[OF plain_Var_id] by simp
  also have "\<dots> = (ts!i)\<cdot>((?pl (?un \<sigma>))|_vars_term(Fun g ts))" using a[OF `i < length ts`] coincidence_lemma'[where t = "ts!i" and V = "vars_term(Fun g ts)"] by best
  also have "\<dots> = (ts!i)\<cdot>(?pl ((?un \<sigma>)|_vars_term(Fun g ts)))" unfolding sub_restrict_f[OF plain_Var_id] by simp
  finally show "(ts!i)\<cdot>(?pl (?un (\<sigma>|_vars_term(ts!i)))) = (ts!i)\<cdot>(?pl (?un (\<sigma>|_vars_term(Fun g ts))))" unfolding sub_restrict_f[OF unsharp_Var_id] by simp
 qed
 from A b c have "\<forall>i < length ts. (ts!i)\<cdot>(\<sigma>|_(vars_term(Fun g ts))) = (ts!i)\<cdot>(?pl (?un (\<sigma>|_vars_term(Fun g ts))))" by simp
 hence d: "\<forall>i < length ts. map (\<lambda>x. x\<cdot>(\<sigma>|_(vars_term(Fun g ts)))) ts!i = map (\<lambda>x. x\<cdot>(?pl(?un(\<sigma>|_vars_term(Fun g ts))))) ts!i" by simp
 have e: "length(map (\<lambda>x. x\<cdot>(?pl (?un (\<sigma>|_vars_term(Fun g ts))))) ts) = length ts" by simp
 from d and e have f:"map (\<lambda>x. x\<cdot>(\<sigma>|_vars_term(Fun g ts))) ts = map (\<lambda>x. x\<cdot>(?pl(?un(\<sigma>|_vars_term(Fun g ts))))) ts"
  using list_eq_iff_nth_eq[where xs = "map (\<lambda>x. x\<cdot>(\<sigma>|_vars_term(Fun g ts))) ts"
                             and ys = "map (\<lambda>x. x\<cdot>(?pl(?un(\<sigma>|_vars_term(Fun g ts))))) ts"] by simp
 have "Fun g ts\<cdot>(\<sigma>|_vars_term(Fun g ts)) = Fun g (map (\<lambda>x. x\<cdot>(\<sigma>|_vars_term(Fun g ts))) ts)" by simp
 also have "\<dots> = Fun g (map (\<lambda>x. x\<cdot>(?pl(?un(\<sigma>|_vars_term(Fun g ts))))) ts)" unfolding f by simp
 finally have two:"Fun g ts\<cdot>(\<sigma>|_vars_term(Fun g ts)) = Fun g ts\<cdot>(?pl(?un(\<sigma>|_vars_term(Fun g ts))))" by simp
 from one and two show ?case unfolding sub_case_aux unfolding o_assoc by fast
qed
 
lemma plain_sub_term':
 assumes "plain s = t\<cdot>\<sigma>"
 shows "\<exists>u \<tau>. t = plain u \<and> t\<cdot>(\<sigma>|_vars_term t) = t\<cdot>(Sub(plain \<circ> \<tau>))"
proof -
 from plain_sub_term[OF assms] have "t = plain(unsharp t)"
  and "t\<cdot>(\<sigma>|_vars_term t) = t\<cdot>(Sub(plain \<circ> (unsharp \<circ> (case \<sigma>|_vars_term t of Sub s \<Rightarrow> s))))" unfolding o_assoc by auto
 hence "\<exists>\<tau>. t = plain(unsharp t) \<and> t\<cdot>(\<sigma>|_vars_term t) = t\<cdot>(Sub(plain \<circ> \<tau>))" using exI[where x = "unsharp \<circ> (case \<sigma>|_vars_term t of Sub s \<Rightarrow> s)"] by auto
 thus ?thesis by auto
qed

lemma sub_dom: "s\<cdot>\<sigma> = s\<cdot>\<tau> \<Longrightarrow> vars_term t \<subseteq> vars_term s \<Longrightarrow> t\<cdot>\<sigma> = t\<cdot>\<tau>" by (induct t) (induct s,auto)

lemma plain_rstep_imp_plain_rhs:
 assumes "wf_trs (ID R)" and "(plain s,t) \<in> rstep(ID R)"
 shows "\<exists>u. t = plain u"
proof -
 from assms obtain l r p \<sigma> where "(plain s,t) \<in> rstep_r_p_s (ID R) (l,r) p \<sigma>"
  using rstep_iff_rstep_r_p_s by best
 hence "(l,r) \<in> ID R" and "p \<in> poss(plain s)"
  and "(ctxt_of_pos_term p (plain s))\<langle>l\<cdot>\<sigma>\<rangle> = (plain s)" (is "?C\<langle>l\<cdot>\<sigma>\<rangle> = ?s")
  and "(ctxt_of_pos_term p (plain s))\<langle>r\<cdot>\<sigma>\<rangle> = t" unfolding rstep_r_p_s_def Let_def by auto
 from `(l,r) \<in> ID R` obtain l' r' where "(l',r') \<in> R" and "l = plain l'" and "r = plain r'" unfolding ID_def by auto
 from `?C\<langle>l\<cdot>\<sigma>\<rangle> = plain s` have "plain s = (?C\<langle>l\<cdot>\<sigma>\<rangle>)" by simp
 then obtain D u where "l\<cdot>\<sigma> = plain u" and "?C = plain_ctxt D" using plain_ctxt_term by best
 hence "plain u = l\<cdot>\<sigma>" by simp
 let ?un = "\<lambda>\<sigma>. Sub(unsharp \<circ> (case \<sigma> of Sub s \<Rightarrow> s))"
 let ?pl = "\<lambda>\<sigma>. Sub(plain \<circ> (case \<sigma> of Sub s \<Rightarrow> s))"
 obtain \<sigma>' where "\<forall>s. s\<cdot>\<sigma> = s\<cdot>\<sigma>'" by auto
 hence "plain u = l\<cdot>\<sigma>'" unfolding `plain u = l\<cdot>\<sigma>` by simp
 have "l = plain(unsharp l)" (is "l = plain ?l") and "l\<cdot>(\<sigma>'|_vars_term l) = l\<cdot>(?pl(?un(\<sigma>'|_vars_term l)))" (is "_ = l\<cdot>?\<tau>") using plain_sub_term[where \<sigma> = "\<sigma>'" and s = u and t = l, OF `plain u = l\<cdot>\<sigma>'`] unfolding o_assoc[symmetric] unfolding sub_case_aux by auto
 have "plain(?l\<cdot>(?un(\<sigma>'|_vars_term l))) = (plain ?l)\<cdot>(?pl(?un(\<sigma>'|_vars_term l)))" unfolding plain_sub_distr by simp
 also have "\<dots> = l\<cdot>?\<tau>" unfolding `l = plain ?l`[symmetric] by simp
 also have "\<dots> = l\<cdot>(\<sigma>'|_vars_term l)" unfolding `l\<cdot>(\<sigma>'|_vars_term l) = l\<cdot>?\<tau>` by simp
 also have "\<dots> = l\<cdot>\<sigma>'" unfolding coincidence_lemma[symmetric] by simp
 finally have "plain(?l\<cdot>(?un(\<sigma>'|_vars_term l))) = l\<cdot>\<sigma>" unfolding `plain u = l\<cdot>\<sigma>`[symmetric] using `plain u = l\<cdot>\<sigma>'` by simp
 from `\<forall>s. s\<cdot>\<sigma> = s\<cdot>\<sigma>'` have "r\<cdot>\<sigma> = r\<cdot>\<sigma>'" by simp
 hence "r\<cdot>\<sigma>' = r\<cdot>(\<sigma>'|_vars_term r)" using coincidence_lemma[where t = r] by simp
 from `wf_trs (ID R)` and `(l,r) \<in> ID R` have "vars_term r \<subseteq> vars_term l" unfolding wf_trs_def by simp
 with `r\<cdot>\<sigma>' = r\<cdot>(\<sigma>'|_vars_term r)` have "r\<cdot>\<sigma> = r\<cdot>(\<sigma>'|_vars_term l)" unfolding `r\<cdot>\<sigma> = r\<cdot>\<sigma>'` using coincidence_lemma' by auto
 have A:"?C\<langle>r\<cdot>(\<sigma>'|_vars_term l)\<rangle> = t" using `?C\<langle>r\<cdot>\<sigma>\<rangle> = t` unfolding `r\<cdot>\<sigma> = r\<cdot>(\<sigma>'|_vars_term l)` by simp
 with `l\<cdot>(\<sigma>'|_vars_term l) = l\<cdot>(?pl(?un(\<sigma>'|_vars_term l)))` and `vars_term r \<subseteq> vars_term l` have "r\<cdot>(\<sigma>'|_vars_term l) = r\<cdot>(?pl(?un(\<sigma>'|_vars_term l)))" using sub_dom by auto
 also have "\<dots> = r\<cdot>(?pl((?un \<sigma>')|_vars_term l))" unfolding sub_restrict_f[OF unsharp_Var_id] by simp
 also have "\<dots> = r\<cdot>((?pl(?un \<sigma>'))|_vars_term l)" unfolding sub_restrict_f[OF plain_Var_id] by simp
 also have "\<dots> = r\<cdot>((?pl(?un \<sigma>'))|_vars_term r)" unfolding coincidence_lemma'[where t = r and V = "vars_term l", OF `vars_term r \<subseteq> vars_term l`] by simp
 also have "\<dots> = r\<cdot>(?pl((?un \<sigma>')|_vars_term r))" unfolding sub_restrict_f[OF plain_Var_id] by simp
 finally have "r\<cdot>(\<sigma>'|_vars_term l) = r\<cdot>(?pl(?un(\<sigma>'|_vars_term r)))" unfolding sub_restrict_f[OF unsharp_Var_id] by simp
 with A have "?C\<langle>r\<cdot>(?pl(?un(\<sigma>'|_vars_term r)))\<rangle> = t" by simp
 hence "?C\<langle>plain r'\<cdot>(?pl(?un(\<sigma>'|_vars_term r)))\<rangle> = t" unfolding `r = plain r'` by simp
 hence "?C\<langle>plain(r'\<cdot>(?un(\<sigma>'|_vars_term r)))\<rangle> = t" using plain_sub_distr[where t = r' and \<sigma> = "?un(\<sigma>'|_vars_term r)"] by simp
 hence "(plain_ctxt D)\<langle>plain(r'\<cdot>(?un(\<sigma>'|_vars_term r)))\<rangle> = t" unfolding `?C = plain_ctxt D` by simp
 hence "plain(D\<langle>r'\<cdot>(?un(\<sigma>'|_vars_term r))\<rangle>) = t" unfolding plain_ctxt_distr by simp
 thus ?thesis by best
qed

lemma plain_rstep_seq_imp_rstep_seq': 
 assumes "wf_trs (ID R)"
 shows "(plain s,plain t) \<in> rstep(ID R)^^n \<Longrightarrow> (s,t) \<in> (rstep R ^^n)"
proof (induct n arbitrary: s t)
 case 0 thus ?case using plain_eq_imp_eq by simp
next
 case (Suc n s t)
 from rel_pow_Suc_D2[OF this(2)]
   obtain u where "(plain s,u) \<in> rstep(ID R)" and "(u,plain t) \<in> (rstep (ID R) ^^ n)" by best
 from `wf_trs (ID R)` and `(plain s,u) \<in> rstep(ID R)` obtain v where "u = plain v" using plain_rstep_imp_plain_rhs by best
 with `(u,plain t) \<in> (rstep (ID R) ^^ n)` have "(plain v,plain t) \<in> (rstep (ID R) ^^ n)" by simp
 with Suc have "(v,t) \<in> (rstep R ^^ n)" by simp
 from `(plain s,u) \<in> rstep(ID R)` have "(plain s,plain v) \<in> rstep(ID R)" unfolding `u = plain v` by simp
 with rstep_iff_plain_rstep have "(s,v) \<in> rstep R" by best
 with `(v,t) \<in> (rstep R ^^ n)` show ?case using rel_pow_Suc_I2[where R = "rstep R" and x = s] by simp
qed
  
lemma plain_rstep_seq_imp_rstep_seq:
 assumes "wf_trs (ID R)"
 shows "(plain s,plain t) \<in> (rstep(ID R))^* \<Longrightarrow> (s,t) \<in> (rstep R)^*"
proof -
 assume "(plain s,plain t) \<in> (rstep (ID R))^*"
 then obtain n where "(plain s,plain t) \<in> (rstep (ID R) ^^ n)" using rtrancl_imp_rel_pow' by best
 hence "(s,t) \<in> (rstep R ^^ n)" using plain_rstep_seq_imp_rstep_seq'[OF assms] by simp
 thus "(s,t) \<in> (rstep R)^*" using rel_pow_imp_rtrancl by best
qed

lemma sharp_eq_imp_eq: "sharp s = sharp t \<Longrightarrow> s = t"
proof (cases s)
 fix x assume "sharp s = sharp t" and "s = Var x" thus ?thesis by (induct t) auto
next
 fix f ss assume "sharp s = sharp t" and "s = Fun f ss"
 hence "sharp(Fun f ss) = sharp t" by simp
 then obtain g ts where "t = Fun g ts" by (induct t) auto
 with `sharp s = sharp t` have "sharp(Fun f ss) = sharp(Fun g ts)" unfolding `s = Fun f ss` `t = Fun g ts`
  by simp
 hence "f\<sharp> = g\<sharp>" and a: "map plain ss = map plain ts" by auto
 hence "f = g" by auto
 from a have "length ss = length ts" using map_eq_imp_length_eq by auto
 from a have "\<forall>i < length ss. (map plain ss)!i = (map plain ts)!i" by auto
 hence "\<forall>i < length ss. plain(ss!i) = (map plain ts)!i" using nth_map by auto
 hence b:"\<forall>i < length ss. plain(ss!i) = plain(ts!i)" unfolding `length ss = length ts` using nth_map by auto
 have "\<forall>i < length ss. ss!i = ts!i"
 proof (intro allI impI)
  fix i assume "i < length ss"
  with b have "plain(ss!i) = plain(ts!i)" by auto 
  thus "ss!i = ts!i" by (rule plain_eq_imp_eq)
 qed
 with `length ss = length ts` have "(length ss = length ts \<and> (\<forall>i<length ss. ss!i = ts!i))" by simp
 hence "ss = ts" using list_eq_iff_nth_eq by auto
 with `f = g` have "Fun f ss = Fun g ts" by simp
 thus ?thesis unfolding `s = Fun f ss` and `t = Fun g ts` by simp
qed

lemma rstep_seq_closed_ctxt: assumes "(s,t) \<in> (rstep R)^*" shows "(C\<langle>s\<rangle>,C\<langle>t\<rangle>) \<in> (rstep R)^*"
proof -
 from assms obtain n where "(s,t) \<in> (rstep R)^^n" using rtrancl_is_UN_rel_pow by auto
 thus ?thesis
 proof (induct n arbitrary: s)
  case 0 thus ?case by auto
 next
  case (Suc n)
  from rel_pow_Suc_D2[OF `(s,t) \<in> (rstep R)^^Suc n`] obtain u
    where "(s,u) \<in> (rstep R)" and "(u,t) \<in> (rstep R)^^n" by auto
  from `(s,u) \<in> (rstep R)` have "(C\<langle>s\<rangle>,C\<langle>u\<rangle>) \<in> (rstep R)" by (rule rstep_imp_ctxt_rstep)
  from Suc and `(u,t) \<in> (rstep R)^^n` have "(C\<langle>u\<rangle>,C\<langle>t\<rangle>) \<in> (rstep R)^*" by simp
  with `(C\<langle>s\<rangle>,C\<langle>u\<rangle>) \<in> (rstep R)` show ?case by auto
 qed
qed

lemma nth_drop_0: "0 < length ss \<Longrightarrow> (ss!0)#drop (Suc 0) ss = ss" by (induct ss) auto

lemma not_Nil_imp_last: "xs \<noteq> [] \<Longrightarrow> \<exists>ys y. xs = ys@[y]"
proof (induct xs)
 case Nil thus ?case by simp
next
 case (Cons x xs) show ?case
 proof (cases xs)
  assume "xs = []" with Cons show ?thesis by simp
 next
  fix x' xs' assume "xs = x'#xs'"
  hence "xs \<noteq> []" by simp
  with Cons obtain ys y where "xs = ys@[y]" by auto
  hence "x#xs = x#(ys@[y])" by simp
  hence "x#xs = (x#ys)@[y]" by simp
  thus ?thesis by auto
 qed
qed

lemma Nil_or_last: "xs = [] \<or> (\<exists>ys y. xs = ys@[y])"
proof (cases xs)
 assume "xs = []" thus ?thesis by simp
next
 fix x' xs' assume "xs = x'#xs'"
 hence "xs \<noteq> []" by simp
 thus ?thesis using not_Nil_imp_last by auto
qed

lemma args_seq_imp_seq: assumes "length ss = length ts" and args:"\<forall>i<length ss. (ss!i,ts!i) \<in> (rstep R)^*"
shows "(Fun f ss,Fun f ts) \<in> (rstep R)^*"
proof -
 let ?C = "\<lambda>i. More f (take i ts) \<box> (drop (Suc i) ss)"
 have a:"\<forall>i<length ss. (Fun f ss,(?C i)\<langle>ts!i\<rangle>) \<in> (rstep R)^*"
 proof (intro allI impI)
  fix i assume "i < length ss" thus "(Fun f ss,(?C i)\<langle>ts!i\<rangle>) \<in> (rstep R)^*"
  proof (induct i)
   case 0
   with args have "(ss!0,ts!0) \<in> (rstep R)^*" by simp
   hence "((?C 0)\<langle>ss!0\<rangle>,(?C 0)\<langle>ts!0\<rangle>) \<in> (rstep R)^*" by (rule rstep_seq_closed_ctxt)
   hence "((More f [] \<box> (drop (Suc 0) ss))\<langle>ss!0\<rangle>,(?C 0)\<langle>ts!0\<rangle>) \<in> (rstep R)^*" by simp
   hence "(Fun f ((ss!0)#drop (Suc 0) ss),(?C 0)\<langle>ts!0\<rangle>) \<in> (rstep R)^*" by simp
   thus "(Fun f ss,(?C 0)\<langle>ts!0\<rangle>) \<in> (rstep R)^*" using `0 < length ss`
     using nth_drop_0[where ss = ss] by simp
  next
   case (Suc i)
   hence IH: "(Fun f ss,(?C i)\<langle>ts!i\<rangle>) \<in> (rstep R)^*" by auto
   from Suc have "i < length ss" by simp
   with `length ss = length ts` have "i < length ts" by simp
   have "(?C(Suc i))\<langle>ss!(Suc i)\<rangle> = Fun f (take (Suc i) ts @ (ss!(Suc i)) # drop (Suc(Suc i)) ss)" by simp
   also have "\<dots> = Fun f ((take i ts @ [ts!i]) @ (ss!(Suc i)) # drop (Suc(Suc i)) ss)"
    using take_Suc_conv_app_nth[OF `i < length ts`] by simp
   also have "\<dots> = Fun f ((take i ts @ [ts!i]) @ drop (Suc i) ss)"
    using drop_length[OF `Suc i < length ss`] by simp
   also have "\<dots> = Fun f (take i ts @ (ts!i) # drop (Suc i) ss)" by simp
   also have "\<dots> = (More f (take i ts) \<box> (drop (Suc i) ss))\<langle>ts!i\<rangle>" by simp
   finally have b:"(?C(Suc i))\<langle>ss!(Suc i)\<rangle> = (?C i)\<langle>ts!i\<rangle>" by auto
   from IH have c: "(Fun f ss,(?C(Suc i))\<langle>ss!(Suc i)\<rangle>) \<in> (rstep R)^*" unfolding b[symmetric] by simp
   from Suc args have "(ss!(Suc i),ts!(Suc i)) \<in> (rstep R)^*" by simp
   hence "((?C(Suc i))\<langle>ss!(Suc i)\<rangle>,(?C(Suc i))\<langle>ts!(Suc i)\<rangle>) \<in> (rstep R)^*" by (rule rstep_seq_closed_ctxt)
   with c show ?case by auto
  qed
 qed
 have "ss = [] \<or> (\<exists>ss' s'. ss = ss'@[s'])" by (rule Nil_or_last)
 thus ?thesis
 proof
  assume "ss = []"
  hence "ts = []" using `length ss = length ts` by simp
  show ?thesis unfolding `ss = []` `ts = []` by auto
 next
  assume "\<exists>ss' s'. ss = ss' @ [s']"
  then obtain s' ss' where "ss = ss'@[s']" by auto
  let ?i = "length ss'"
  have "?i < length ss" using `ss = ss'@[s']` by simp
  have "ts = [] \<or> (\<exists>ts' t'. ts = ts'@[t'])" by (rule Nil_or_last)
  thus ?thesis
  proof
   assume "ts = []" with `ss = ss'@[s']` and `length ss = length ts` show ?thesis by simp
  next
   assume "\<exists>ts' t'. ts = ts'@[t']"
   then obtain t' ts' where "ts = ts'@[t']" by auto
   with a and `?i < length ss` have b:"(Fun f ss,(?C ?i)\<langle>(ts'@[t'])!?i\<rangle>) \<in> (rstep R)^*" by simp
   from `length ss = length ts` have "length ss' = length ts'"
    unfolding `ss = ss'@[s']` `ts = ts'@[t']` by simp
   hence "?i = length ts'" by simp
   have "(?C ?i)\<langle>(ts'@[t'])!?i\<rangle> = Fun f (take ?i ts @ ((ts'@[t'])!?i) # drop (Suc ?i) ss)" by simp
   also have "\<dots> = Fun f (take ?i (ts'@[t']) @ ((ts'@[t'])!?i) # drop (Suc ?i) (ss'@[s']))"
    unfolding `ss = ss'@[s']` and `ts = ts'@[t']` by simp
   also have "\<dots> = Fun f (ts' @ ((ts'@[t'])!?i) # drop (Suc ?i) (ss'@[s']))"
    unfolding `length ss' = length ts'` by auto
   also have "\<dots> = Fun f (ts' @ t' # drop (Suc ?i) (ss'@[s']))"
    unfolding `length ss' = length ts'` by auto
   also have "\<dots> = Fun f (ts' @ t' # [])" by auto
   also have "\<dots> = Fun f (ts @ [])" using `ts = ts'@[t']` by simp
   finally have "(?C ?i)\<langle>(ts'@[t'])!?i\<rangle> = Fun f ts" by simp
   with b show ?thesis by auto
  qed
 qed
qed

lemma nrrstep_non_var_imp_arg_rstep:
 fixes ss :: "('f,'v)term list"
 assumes "(Fun f ss,Fun f ts) \<in> nrrstep R" (is "(?s,?t) \<in> nrrstep R")
 shows "\<exists>C i. i < length ss \<and> (ss!i,ts!i) \<in> rstep R \<and> C\<langle>ss!i\<rangle> = Fun f ss \<and> C\<langle>ts!i\<rangle> = Fun f ts"
proof -
 from `(?s,?t) \<in> nrrstep R`
 obtain l r j ps \<sigma> where A: "let C = ctxt_of_pos_term (j<#ps) ?s in (j<#ps) \<in> poss ?s \<and> (l,r) \<in> R \<and> C\<langle>l\<cdot>\<sigma>\<rangle> = ?s \<and> C\<langle>r\<cdot>\<sigma>\<rangle> = ?t" unfolding nrrstep_def rstep_r_p_s_def by force
 let ?C = "ctxt_of_pos_term (j<#ps) ?s"
 from A have "(j<#ps) \<in> poss ?s" and "(l,r) \<in> R" and "?C\<langle>l\<cdot>\<sigma>\<rangle> = ?s" and "?C\<langle>r\<cdot>\<sigma>\<rangle> = ?t" using Let_def by auto
 have C: "?C = More f (take j ss) (ctxt_of_pos_term ps (ss!j)) (drop (Suc j) ss)" (is "?C = More f ?ss1 ?D ?ss2") by auto
 from `?C\<langle>l\<cdot>\<sigma>\<rangle> = ?s` have "?D\<langle>l\<cdot>\<sigma>\<rangle> = (ss!j)" by (auto simp: take_drop_imp_nth)
 from `(l,r) \<in> R` have "(l\<cdot>\<sigma>,r\<cdot>\<sigma>) \<in> (CS R)" by auto
 hence "(?D\<langle>l\<cdot>\<sigma>\<rangle>,?D\<langle>r\<cdot>\<sigma>\<rangle>) \<in> (CC(CS R))" and "(?C\<langle>l\<cdot>\<sigma>\<rangle>,?C\<langle>r\<cdot>\<sigma>\<rangle>) \<in> (CC(CS R))" by (auto simp only: closure_ctxt.ccI)
 hence D_rstep: "(?D\<langle>l\<cdot>\<sigma>\<rangle>,?D\<langle>r\<cdot>\<sigma>\<rangle>) \<in> rstep R" and "(?C\<langle>l\<cdot>\<sigma>\<rangle>,?C\<langle>r\<cdot>\<sigma>\<rangle>) \<in> rstep R" by auto
 from `?C\<langle>r\<cdot>\<sigma>\<rangle> = ?t` and C have "?t = Fun f (take j ss @ ?D\<langle>r\<cdot>\<sigma>\<rangle> # drop (Suc j) ss)" by auto
 hence ts: "ts = (take j ss @ ?D\<langle>r\<cdot>\<sigma>\<rangle> # drop (Suc j) ss)" by auto
 from `j<#ps \<in> poss ?s` have r0: "j < length ss" by simp
 hence "(take j ss @ ?D\<langle>r\<cdot>\<sigma>\<rangle> # drop (Suc j) ss) ! j = ?D\<langle>r\<cdot>\<sigma>\<rangle>" by (auto simp: nth_append_take)
 with ts have "ts!j = ?D\<langle>r\<cdot>\<sigma>\<rangle>" by auto
 let ?C' = "More f (take j ss) \<box> (drop (Suc j) ss)"
 from D_rstep have r1: "(ss!j,ts!j) \<in> rstep R" unfolding `ts!j = ?D\<langle>r\<cdot>\<sigma>\<rangle>` `?D\<langle>l\<cdot>\<sigma>\<rangle> = ss!j` by simp
 have "?s = ?C\<langle>l\<cdot>\<sigma>\<rangle>" unfolding `?C\<langle>l\<cdot>\<sigma>\<rangle> = ?s` by simp
 also have "\<dots> = ?C'\<langle>?D\<langle>l\<cdot>\<sigma>\<rangle>\<rangle>" unfolding C by simp
 finally have r2:"?C'\<langle>ss!j\<rangle> = ?s" unfolding `?D\<langle>l\<cdot>\<sigma>\<rangle> = ss!j` by simp
 have "?t = ?C\<langle>r\<cdot>\<sigma>\<rangle>" unfolding `?C\<langle>r\<cdot>\<sigma>\<rangle> = ?t` by simp
 also have "\<dots> = ?C'\<langle>?D\<langle>r\<cdot>\<sigma>\<rangle>\<rangle>" unfolding C by simp
 finally have r3:"?C'\<langle>ts!j\<rangle> = ?t" unfolding `ts!j = ?D\<langle>r\<cdot>\<sigma>\<rangle>` by simp
 from r0 r1 r2 r3 show ?thesis by best
qed

lemma sharp_nrrstep_imp_rstep:
 assumes "wf_trs (ID R)" and "(sharp s,sharp t) \<in> nrrstep(ID R)"
 shows "(s,t) \<in> rstep R"
proof -
 from assms obtain f ss where "sharp s = Fun f ss" using nrrstep_imp_Fun by best
 with assms have "(Fun f ss,sharp t) \<in> nrrstep(ID R)" by simp
 then obtain ts where "sharp t = Fun f ts" using nrrstep_preserves_root by best
 from `sharp s = Fun f ss` obtain f' ss' where "sharp s = sharp(Fun f' ss')" by (induct s) auto
 from `sharp t = Fun f ts` obtain g' ts' where "sharp t = sharp(Fun g' ts')" by (induct t) auto
 have "root(sharp s) = root(sharp t)" unfolding `sharp s = Fun f ss` `sharp t = Fun f ts` by simp
 hence "f'\<sharp> = g'\<sharp>" unfolding `sharp s = sharp(Fun f' ss')` `sharp t = sharp(Fun g' ts')` by simp
 hence "f' = g'" by simp
 from `(sharp s,sharp t) \<in> nrrstep(ID R)`
  have a:"(Fun f'\<sharp> (map plain ss'),Fun g'\<sharp> (map plain ts')) \<in> nrrstep(ID R)"
  unfolding `sharp s = sharp(Fun f' ss')` and `sharp t = sharp(Fun g' ts')` by simp
 hence "length(map plain ss') = length(map plain ts')" unfolding `f' = g'`
  using nrrstep_non_var_preserves_num_args[where ss = "map plain ss'"] by auto
 hence "length ss' = length ts'" by simp
 from a have "(Fun f'\<sharp> (map plain ss'),Fun f'\<sharp> (map plain ts')) \<in> nrrstep(ID R)"
  unfolding `f'\<sharp> = g'\<sharp>` by auto
 then obtain C i where "i < length(map plain ss')"
  and b:"(map plain ss'!i,map plain ts'!i) \<in> rstep(ID R)"
  and c:"C\<langle>map plain ss'!i\<rangle> = Fun f'\<sharp> (map plain ss')"
  and d:"C\<langle>map plain ts'!i\<rangle> = Fun g'\<sharp> (map plain ts')"
  unfolding `f'\<sharp> = g'\<sharp>` using nrrstep_non_var_imp_arg_rstep[where f = "g'\<sharp>" and ss = "map plain ss'"] 
  by best
 from `i < length(map plain ss')` have "map plain ss'!i = plain(ss'!i)" by simp
 from `i < length(map plain ss')` have "map plain ts'!i = plain(ts'!i)" unfolding `length(map plain ss') = length(map plain ts')` by simp
 from b have "(plain(ss'!i),plain(ts'!i)) \<in> rstep(ID R)"
  unfolding `map plain ss'!i = plain(ss'!i)` `map plain ts'!i = plain(ts'!i)` by simp
 hence "(ss'!i,ts'!i) \<in> rstep R" unfolding rstep_iff_plain_rstep[symmetric] by simp
 from `sharp s = sharp(Fun f' ss')` have "s = Fun f' ss'" by (rule sharp_eq_imp_eq)
 from `sharp t = sharp(Fun g' ts')` have "t = Fun g' ts'" by (rule sharp_eq_imp_eq)
 let ?C = "unsharp_ctxt C"
 have "?C\<langle>ss'!i\<rangle> = ?C\<langle>unsharp(plain(ss'!i))\<rangle>" by simp
 also have "\<dots> = unsharp(C\<langle>plain(ss'!i)\<rangle>)" unfolding unsharp_ctxt_distr by simp
 also have "\<dots> = unsharp(Fun f'\<sharp> (map plain ss'))"
  using c unfolding `map plain ss'!i = plain(ss'!i)` by simp
 also have "\<dots> = Fun f' (map unsharp (map plain ss'))" by simp
 finally have e:"?C\<langle>ss'!i\<rangle> = Fun f' ss'" unfolding unsharp_plain by simp
 have "?C\<langle>ts'!i\<rangle> = ?C\<langle>unsharp(plain(ts'!i))\<rangle>" by simp
 also have "\<dots> = unsharp(C\<langle>plain(ts'!i)\<rangle>)" unfolding unsharp_ctxt_distr by simp
 also have "\<dots> = unsharp(Fun g'\<sharp> (map plain ts'))"
  using d unfolding `map plain ts'!i = plain(ts'!i)` by simp
 also have "\<dots> = Fun g' (map unsharp (map plain ts'))" by simp
 finally have f:"?C\<langle>ts'!i\<rangle> = Fun g' ts'" unfolding unsharp_plain by simp
 from `(ss'!i,ts'!i) \<in> rstep R` have "(?C\<langle>ss'!i\<rangle>,?C\<langle>ts'!i\<rangle>) \<in> CC(rstep R)" by (rule ccI)
 hence "(?C\<langle>ss'!i\<rangle>,?C\<langle>ts'!i\<rangle>) \<in> rstep R" by auto
 hence "(Fun f' ss',Fun g' ts') \<in> rstep R" unfolding e f by simp
 thus ?thesis unfolding `s = Fun f' ss'` `t = Fun g' ts'` by simp
qed

lemma sub_Var_imp_Var: assumes "s\<cdot>\<sigma> = Var x" shows "\<exists>y. s = Var y" using assms by (induct s) auto 

lemma unsharp_plain_id': "unsharp\<circ>plain = (\<lambda>x. x)"
proof (rule ext)
 fix x :: "('f,'v)term"
 have "unsharp(plain x) = x" by (rule unsharp_plain_eq)
 thus "(unsharp\<circ>plain) x = x" by simp
qed

lemma sharp_Fun_sub_imp_unsharp_sub:
 assumes "sharp(Fun f ss) = sharp(Fun g ts)\<cdot>\<sigma>"
 shows "Fun f ss = (Fun g ts)\<cdot>(Sub(unsharp \<circ> (case \<sigma> of Sub s \<Rightarrow> s)))" (is "?s = ?t\<cdot>?\<sigma>")
proof -
 from assms have "Fun f\<sharp> (map plain ss) = (Fun g\<sharp> (map plain ts))\<cdot>\<sigma>" (is "Fun f\<sharp> ?ss = (Fun g\<sharp> ?ts)\<cdot>\<sigma>")
  by simp
 hence "Fun f\<sharp> ?ss = Fun g\<sharp> (map (\<lambda>t. t\<cdot>\<sigma>) ?ts)" by simp
 hence "f = g" and "?ss = map (\<lambda>t. t\<cdot>\<sigma>) ?ts" by auto
 hence "?ss = map (\<lambda>t. (plain t)\<cdot>\<sigma>) ts" by simp
 hence "map unsharp ?ss = map unsharp (map (\<lambda>t. (plain t)\<cdot>\<sigma>) ts)" by auto
 hence "map (unsharp\<circ>plain) ss = map unsharp (map (\<lambda>t. (plain t)\<cdot>\<sigma>) ts)" by simp
 hence "map (\<lambda>x. x) ss = map unsharp (map (\<lambda>t. (plain t)\<cdot>\<sigma>) ts)" unfolding unsharp_plain_id' by simp
 hence "ss = map unsharp (map (\<lambda>t. (plain t)\<cdot>\<sigma>) ts)" unfolding map_ident by simp
 hence "ss = map (\<lambda>t. unsharp((plain t)\<cdot>\<sigma>)) ts" by simp
 hence "ss = map (\<lambda>t. (unsharp(plain t))\<cdot>?\<sigma>) ts"
  unfolding unsharp_sub_distr by simp
 hence "ss = map (\<lambda>t. t\<cdot>?\<sigma>) ts" by auto
 hence "Fun f ss = Fun g (map (\<lambda>t. t\<cdot>?\<sigma>) ts)" unfolding `f = g` by simp
 thus ?thesis by auto
qed

lemma sharp_Fun_sub_imp_unsharp_sub':
 assumes "sharp s = sharp(Fun g ts)\<cdot>\<sigma>"
 shows "s = Fun g ts\<cdot>Sub(unsharp \<circ> (case \<sigma> of Sub s \<Rightarrow> s))"
proof (cases s)
 fix x assume "s = Var x" thus ?thesis using assms by auto
next
 fix f ss assume "s = Fun f ss"
 show ?thesis using assms unfolding `s = Fun f ss` by (rule sharp_Fun_sub_imp_unsharp_sub)
qed

lemma sharp_Var_sub_imp_unsharp_sub:
 assumes "sharp(Var x) = sharp(Var y)\<cdot>\<sigma>"
 shows "Var x = (Var y)\<cdot>Sub(unsharp \<circ> (case \<sigma> of Sub s \<Rightarrow> s))" (is "?s = ?t\<cdot>?\<sigma>")
proof (cases \<sigma>)
 case (Sub s)
 from assms have "s y = Var x" unfolding Sub by simp
 have "(Var y)\<cdot>?\<sigma> = Var y \<cdot> Sub(unsharp \<circ> s)" unfolding Sub by simp
 also have "\<dots> = (unsharp \<circ> s) y" by simp
 also have "\<dots> = unsharp(s y)" unfolding o_def by simp
 also have "\<dots> = unsharp(Var x)" unfolding `s y = Var x` by simp
 finally have "?t\<cdot>?\<sigma> = Var x" by simp
 thus ?thesis unfolding Sub by simp
qed

lemma Some_root_imp_Fun: "root t = Some f \<Longrightarrow> \<exists>f ss. t = Fun f ss" by (induct t) auto

lemma defined_root_imp_Fun: 
 assumes "wf_trs R" and "root s = Some f" and "defined R f"
 shows "\<exists>ss. s = Fun f ss" using assms by (induct s) auto

lemma sharp_rrstep_imp_rstep:
 assumes "(sharp s,sharp t) \<in> rstep(DP R)" and "wf_trs R" shows "\<exists>C. (s,C\<langle>t\<rangle>) \<in> rstep R"
proof -
 from `wf_trs R` have "wf_trs (DP R)" by (rule wf_trs_imp_wf_DP)
 from assms have rrstep:"(sharp s,sharp t) \<in> rrstep(DP R)" by (rule rstepDP_imp_rrstepDP)
 then obtain l r \<sigma> where "(l,r) \<in> DP R" and "l\<cdot>\<sigma> = sharp s" and "r\<cdot>\<sigma> = sharp t"
  using rrstep_imp_rule_subs by best
 let ?\<sigma> = "Sub(unsharp \<circ> (case \<sigma> of Sub s \<Rightarrow> s))"
 from `(l,r) \<in> DP R` obtain l' r' h' us' where "l = sharp l'" and "r = sharp(Fun h' us')" (is "r = sharp ?u'")
  and "(l',r') \<in> R" and "r' \<unrhd> (Fun h' us')" and "defined R h'" and "\<not> l' \<rhd> Fun h' us'"
  unfolding DP_def by auto
 from `wf_trs R` and `(l',r') \<in> R` obtain f' ss' where "l' = Fun f' ss'" using wf_trs_imp_lhs_Fun by best
 from `r' \<unrhd> ?u'` obtain C where "C\<langle>?u'\<rangle> = r'" using supteqp_imp_ctxt by best
 from `wf_trs (DP R)` `(sharp s,sharp t) \<in> rstep(DP R)` obtain f ss where "sharp s = Fun f ss"
  using rstep_imp_Fun[where R = "DP R"] by best
 then obtain f'' ss'' where "sharp s = sharp(Fun f'' ss'')" by (induct s) auto
 hence "s = Fun f'' ss''" by (rule sharp_eq_imp_eq)
 hence "sharp(Fun f'' ss'') = l\<cdot>\<sigma>" unfolding `l\<cdot>\<sigma> = sharp s` by simp
 hence "sharp(Fun f'' ss'') = (sharp l')\<cdot>\<sigma>" unfolding `l = sharp l'` by simp
 hence "sharp(Fun f'' ss'') = sharp(Fun f' ss')\<cdot>\<sigma>" unfolding `l' = Fun f' ss'` by simp
 hence "Fun f'' ss'' = (Fun f' ss')\<cdot>?\<sigma>" using sharp_Fun_sub_imp_unsharp_sub by best
 hence "s = l'\<cdot>?\<sigma>" unfolding `s = Fun f'' ss''` `l' = Fun f' ss'` by simp
 have "r'\<cdot>?\<sigma> = C\<langle>?u'\<rangle>\<cdot>?\<sigma>" unfolding `C\<langle>?u'\<rangle> = r'` by simp
 then obtain D where "r'\<cdot>?\<sigma> = D\<langle>?u'\<cdot>?\<sigma>\<rangle>" using sub_ctxt_distr by best
 have "sharp t = r\<cdot>\<sigma>" unfolding `r\<cdot>\<sigma> = sharp t` by simp
 hence "sharp t = sharp ?u'\<cdot>\<sigma>" unfolding `r = sharp ?u'` by simp
 hence "t = ?u'\<cdot>?\<sigma>" using sharp_Fun_sub_imp_unsharp_sub' by best
 from `(l',r') \<in> R` have "(l'\<cdot>?\<sigma>,r'\<cdot>?\<sigma>) \<in> CS R" by (rule csI)
 hence "(\<box>\<langle>l'\<cdot>?\<sigma>\<rangle>,\<box>\<langle>r'\<cdot>?\<sigma>\<rangle>) \<in> rstep R" by (rule ccI[where C = "\<box>"])
 hence "(l'\<cdot>?\<sigma>,r'\<cdot>?\<sigma>) \<in> rstep R" by simp
 hence "(l'\<cdot>?\<sigma>,D\<langle>?u'\<cdot>?\<sigma>\<rangle>) \<in> rstep R" unfolding `r'\<cdot>?\<sigma> = D\<langle>?u'\<cdot>?\<sigma>\<rangle>` by simp
 thus ?thesis unfolding `s = l'\<cdot>?\<sigma>` `t = ?u'\<cdot>?\<sigma>` by best
qed

lemma rstepID_imp_rstep:
 assumes a:"wf_trs R" and b: "(sharp s,sharp t) \<in> rstep(ID R)"
 shows "(s,t) \<in> rstep R"
proof -
 from a have c:"wf_trs (ID R)" by (rule wf_trs_imp_wf_ID)
 from b and a have "(sharp s,sharp t) \<in> nrrstep(ID R)" by (rule rstepID_imp_nrrstepID)
 with c show ?thesis by (rule sharp_nrrstep_imp_rstep)
qed

definition rstep_r_c_s :: "('f,'v)rule \<Rightarrow> ('f,'v)ctxt \<Rightarrow> ('f,'v)sub \<Rightarrow> ('f,'v)term brel"
where "rstep_r_c_s r C \<sigma> = {(s,t) | s t. s = C\<langle>fst r \<cdot> \<sigma>\<rangle> \<and> t = C\<langle>snd r \<cdot> \<sigma>\<rangle>}"

lemma rstep_iff_rstep_r_c_s : "((s,t) \<in> rstep R) = (\<exists> l r C \<sigma>. (l,r) \<in> R \<and> (s,t) \<in> rstep_r_c_s (l,r) C \<sigma>)" (is "?left = ?right")
proof
  assume ?left
  with rstep_iff_rstep_r_p_s obtain l r p \<sigma> where step: "(s,t) \<in> rstep_r_p_s R (l,r) p \<sigma>" by blast
  obtain D where D: "D = ctxt_of_pos_term p s" by auto
  with step have Rrule: "(l,r) \<in> R" and s: "s = D\<langle>l \<cdot> \<sigma>\<rangle>" and t: "t = D\<langle>r \<cdot> \<sigma>\<rangle>" 
    unfolding rstep_r_p_s_def by (force simp: Let_def)+
  thus ?right unfolding rstep_r_c_s_def by auto
next
  assume ?right
  from this obtain l r C \<sigma> where "(l,r) \<in> R \<and> (s,t) \<in> rstep_r_c_s (l,r) C \<sigma>" by auto
  hence rule: "(l,r) \<in> R" and s: "s = C\<langle>l \<cdot> \<sigma>\<rangle>" and t: "t = C\<langle>r \<cdot> \<sigma>\<rangle>" unfolding rstep_r_c_s_def by auto
  show ?left by (simp add: s t, rule ccI, rule csI, rule rule)
qed

lemma rstep_subset_characterization:
  "(rstep R \<subseteq> rstep S) = (\<forall> l r. (l,r) \<in> R \<longrightarrow> (\<exists> l' r' C \<sigma> . (l',r') \<in> S \<and> l = C\<langle>l' \<cdot> \<sigma>\<rangle> \<and> r = C\<langle>r' \<cdot> \<sigma>\<rangle>))" (is "?left = ?right")
proof
  assume ?right 
  show ?left
  proof (rule subsetI, simp add: split_paired_all)
    fix s t
    assume "(s,t) \<in> rstep R"
    then obtain l r C \<sigma> where step: "(l,r) \<in> R \<and> (s,t) \<in> rstep_r_c_s (l,r) C \<sigma>"
      unfolding rstep_iff_rstep_r_c_s by best
    hence Rrule: "(l,r) \<in> R" and s: "s = C\<langle>l \<cdot> \<sigma>\<rangle>" and t: "t = C\<langle>r \<cdot> \<sigma>\<rangle>" 
      unfolding rstep_r_c_s_def by (force)+
    from Rrule `?right` obtain l' r' C' \<sigma>' where Srule: "(l',r') \<in> S" and l: "l = C'\<langle>l' \<cdot> \<sigma>'\<rangle>" and r: "r = C'\<langle>r' \<cdot> \<sigma>'\<rangle>"
      by (force simp: Let_def)+
    let ?D = "C \<circ>c (C' \<cdot>c \<sigma>)"
    let ?sig = "\<sigma>' \<circ>s \<sigma>"
    have s2: "s = ?D\<langle>l' \<cdot> ?sig\<rangle>" by (simp add: s l)
    have t2: "t = ?D\<langle>r' \<cdot> ?sig\<rangle>" by (simp add: t r)
    from s2 t2 have sStep: "(s,t) \<in> rstep_r_c_s (l',r') ?D ?sig" unfolding rstep_r_c_s_def by force
    with Srule show "(s,t) \<in> rstep S" by (simp only: rstep_iff_rstep_r_c_s, blast)
  qed
next
  assume ?left
  show ?right
  proof (rule ccontr)
    assume "\<not> ?right"
    from this obtain l r where "(l,r) \<in> R" and cond: "\<forall> l' r' C \<sigma>. (l',r') \<in> S \<longrightarrow> (l \<noteq> C\<langle>l' \<cdot> \<sigma>\<rangle> \<or> r \<noteq> C\<langle>r' \<cdot> \<sigma>\<rangle>)" by blast
    hence "(l,r) \<in> rstep R" using rstep_contains_r by blast
    with `?left` have "(l,r) \<in> rstep S" by auto
    with cond show False by (simp only: rstep_iff_rstep_r_c_s, unfold rstep_r_c_s_def, force)
  qed
qed

lemma rstep_preserves_wf_fun_terms:
  fixes R::"('f,'v)trs"
  assumes "funas_trs R \<subseteq> F" and "s \<in> Term(F)" and "(s,t) \<in> rstep R" and "wf_trs R"
  shows "t \<in> Term(F)"
proof -
  from `(s,t) \<in> rstep R` obtain l r C \<sigma> where R: "(l,r) \<in> R"
    and s: "s = C\<langle>l\<cdot>\<sigma>\<rangle>" and t: "t = C\<langle>r\<cdot>\<sigma>\<rangle>" using rstep_imp_C_s_r[of s t R] by fast
  from `s \<in> Term(F)` have "funas_term s \<subseteq> F" by (induct s) auto
  from `funas_trs R \<subseteq> F` and R have "funas_term r \<subseteq> F" unfolding funas_defs by force
  from `wf_trs R` and R have "vars_term r \<subseteq> vars_term l" unfolding wf_trs_def by simp
  have "t \<in> Term(funas_term t)" by (rule funas_term_id)
  have "funas_term s = funas_ctxt C \<union> funas_term(l\<cdot>\<sigma>)" unfolding s by simp
  hence funas_s: "funas_term s = funas_ctxt C \<union> funas_term l \<union> (\<Union>x\<in>vars_term l. funas_term(Var x \<cdot> \<sigma>))" using funas_term_sub_apply[of l \<sigma>] by auto
  have "funas_term t = funas_ctxt C \<union> funas_term(r\<cdot>\<sigma>)" unfolding t by simp
  hence "funas_term t = funas_ctxt C \<union> funas_term r \<union> (\<Union>x\<in>vars_term r. funas_term(Var x \<cdot> \<sigma>))" using funas_term_sub_apply[of r \<sigma>] by auto
  hence "funas_term t \<subseteq> funas_ctxt C \<union> funas_term r \<union> (\<Union>x\<in>vars_term l. funas_term(Var x \<cdot> \<sigma>))" using `vars_term r \<subseteq> vars_term l` by auto
  with `funas_term r \<subseteq> F` `funas_term s \<subseteq> F` have "funas_term t \<subseteq> F" unfolding funas_s by blast
  from wf_fun_terms_subset[OF `funas_term t \<subseteq> F`] and `t \<in> Term(funas_term t)` show ?thesis by auto
qed

lemma no_Var_rstep[simp]: fixes R::"('f,'v)trs" assumes "wf_trs R" and "(Var x,t) \<in> rstep R" shows "False"
using assms unfolding wf_trs_def rstep_iff_rstep_r_p_s rstep_r_p_s_def Let_def by force

lemma lhs_wf:
  fixes R::"('f,'v)trs"
  assumes R: "(l,r) \<in> R" and subset: "funas_trs R \<subseteq> F" shows "l \<in> Term(F)"
proof -
  have "funas_term l \<subseteq> funas_trs R" using R unfolding funas_trs_def funas_rule_def by auto
  hence "Term(funas_term l) \<subseteq> Term(funas_trs R)" by (rule wf_fun_terms_subset)
  moreover have "l \<in> Term(funas_term l)" by simp
  ultimately have "l \<in> Term(funas_trs R)" by auto
  moreover from subset have "Term(funas_trs R) \<subseteq> Term(F)" by (rule wf_fun_terms_subset)
  ultimately show ?thesis by auto
qed

lemma rhs_wf:
  fixes R::"('f,'v)trs"
  assumes R: "(l,r) \<in> R" and subset: "funas_trs R \<subseteq> F" shows "r \<in> Term(F)"
proof -
  have "funas_term r \<subseteq> funas_trs R" using R unfolding funas_trs_def funas_rule_def by auto
  hence "Term(funas_term r) \<subseteq> Term(funas_trs R)" by (rule wf_fun_terms_subset)
  moreover have "r \<in> Term(funas_term r)" by simp
  ultimately have "r \<in> Term(funas_trs R)" by auto
  moreover from subset have "Term(funas_trs R) \<subseteq> Term(F)" by (rule wf_fun_terms_subset)
  ultimately show ?thesis by auto
qed

lemma poss_plain_unsharp_eq[simp]: "poss(plain(unsharp s)) = poss s"
proof (induct s)
 case (Var x) show ?case by simp
next
 case (Fun f ss)
 hence IH: "\<And>i. i < length ss \<longrightarrow> poss(map (plain \<circ> unsharp) ss!i) = poss(ss!i)" by simp
 have "poss(plain(unsharp(Fun f ss))) = ({\<epsilon>} \<union> {i<#p | i p. i < length ss \<and> p \<in> poss((map (plain \<circ> unsharp) ss)!i)})" by simp
 also have "\<dots> = {\<epsilon>} \<union> {i<#p | i p. i < length ss \<and> p \<in> poss(ss!i)}" using IH by auto
 finally show ?case by simp
qed

lemma rstepID_imp_plain_rstep:
 assumes "(s,t) \<in> rstep(ID R)"
 shows "(plain(unsharp s),plain(unsharp t)) \<in> rstep(ID R)"
proof -
 from assms obtain C l r p \<sigma> where "C = ctxt_of_pos_term p s" and "p \<in> poss s" and"(l,r) \<in> ID R" and "C\<langle>l\<cdot>\<sigma>\<rangle> = s" and "C\<langle>r\<cdot>\<sigma>\<rangle> = t" unfolding rstep_iff_rstep_r_p_s rstep_r_p_s_def Let_def by force
 from `p \<in> poss s` have "p \<in> poss(plain(unsharp s))" by simp
 from `(l,r) \<in> ID R` obtain l' r' where eq: "l = plain l'" "r = plain r'" unfolding ID_def by auto
 from `(l,r) \<in> ID R` have "(plain(unsharp l),plain(unsharp r)) \<in> ID R" unfolding eq by simp
 have "plain(unsharp s) = plain(unsharp(C\<langle>l\<cdot>\<sigma>\<rangle>))" unfolding `C\<langle>l\<cdot>\<sigma>\<rangle> = s` by simp
 also have "\<dots> = (plain_ctxt(unsharp_ctxt C))\<langle>plain((unsharp l) \<cdot> Sub(unsharp \<circ> (case \<sigma> of Sub s \<Rightarrow> s)))\<rangle>" (is "_ = ?C\<langle>plain(_ \<cdot> ?\<sigma>)\<rangle>") by simp
 also have "\<dots> = ?C\<langle>(plain(unsharp l)) \<cdot> Sub(plain \<circ> unsharp \<circ> (case \<sigma> of Sub s \<Rightarrow> s))\<rangle>" (is "_ = _\<langle>_ \<cdot> ?\<sigma>'\<rangle>") unfolding o_assoc[symmetric] by simp
 finally have s: "plain(unsharp s) = ?C\<langle>plain(unsharp l) \<cdot> ?\<sigma>'\<rangle>" by simp
 have "plain(unsharp t) = plain(unsharp(C\<langle>r\<cdot>\<sigma>\<rangle>))" unfolding `C\<langle>r\<cdot>\<sigma>\<rangle> = t` by simp
 also have "\<dots> = ?C\<langle>plain((unsharp r)\<cdot>?\<sigma>)\<rangle>" by simp
 also have "\<dots> = ?C\<langle>plain(unsharp r) \<cdot> ?\<sigma>'\<rangle>" unfolding o_assoc[symmetric] by simp
 finally have t: "plain(unsharp t) = ?C\<langle>plain(unsharp r) \<cdot> ?\<sigma>'\<rangle>" by simp
 from `(plain(unsharp l),plain(unsharp r)) \<in> ID R` have "(plain(unsharp l)\<cdot>?\<sigma>',plain(unsharp r)\<cdot>?\<sigma>') \<in> CS(ID R)" by (rule csI)
 thus ?thesis unfolding s t by (rule ccI)
qed

lemma SN_elt_rstep_iff_SN_elt_rstepID:
 fixes R :: "('f,'v)trs" and s :: "('f,'v)term"
 shows "(SN_elt (rstep R) s::bool) = (SN_elt (rstep(ID R)) (plain s)::bool)"
proof
 assume "(SN_elt::('f,'v)trs \<Rightarrow> ('f,'v)term \<Rightarrow> bool) (rstep R) s"
 show "(SN_elt::('f shp,'v)trs \<Rightarrow> ('f shp,'v)term \<Rightarrow> bool) (rstep (ID R)) (plain s)"
 proof (rule ccontr)
  assume "\<not>SN_elt (rstep(ID R)) (plain s)"
  then obtain S where start: "S 0 = plain s" and seq: "\<forall>i. (S i,S(Suc i)) \<in> rstep(ID R)"
   unfolding SN_elt_def by auto
  let ?S = "\<lambda>i. plain(unsharp(S i))"
  let ?T = "\<lambda>i. unsharp(S i)"
  from seq have "\<forall>i. (?S i,?S(Suc i)) \<in> rstep(ID R)" using rstepID_imp_plain_rstep by force
  hence seq': "\<forall>i. (?T i,?T(Suc i)) \<in> rstep R" unfolding rstep_iff_plain_rstep[symmetric] by simp
  hence "?T 0 = s" unfolding `S 0 = plain s` by simp
  with seq' have "\<not>(SN_elt (rstep R) s)" unfolding SN_elt_def by auto
  with `SN_elt (rstep R) s` show False by simp
 qed
next
 assume "(SN_elt::('f shp,'v)trs \<Rightarrow> ('f shp,'v)term \<Rightarrow> bool) (rstep(ID R)) (plain s)"
 show "(SN_elt::('f,'v)trs \<Rightarrow> ('f,'v)term \<Rightarrow> bool) (rstep R) s"
 proof (rule ccontr)
  assume "\<not>(SN_elt (rstep R) s)"
  then obtain S where start: "S 0 = s" and seq: "\<forall>i. (S i,S(Suc i)) \<in> rstep R"
   unfolding SN_elt_def by auto
  let ?S = "\<lambda>i. plain(S i)"
  from seq have seq': "\<forall>i. (?S i,?S(Suc i)) \<in> rstep(ID R)" unfolding rstep_iff_plain_rstep[symmetric] by simp
  hence "?S 0 = plain s" unfolding `S 0 = s` by simp
  with seq' have "\<not>(SN_elt (rstep(ID R)) (plain s))" unfolding SN_elt_def by auto
  with `SN_elt (rstep(ID R)) (plain s)` show False by simp
 qed
qed

lemma SN_rstep_iff_SN_rstepID: "(SN(rstep R)::bool) = (SN(rstep(ID R))::bool)"
proof
 assume "SN(rstep R)" show "SN(rstep(ID R))"
 proof (rule ccontr)
  assume "\<not>SN(rstep(ID R))"
  then obtain S t where start: "S 0 = t"
   and seq: "\<forall>i. (S i,S(Suc i)) \<in> rstep(ID R)" unfolding SN_def SN_elt_def by auto
  let ?S = "\<lambda>i. plain(unsharp(S i))"
  let ?T = "\<lambda>i. unsharp(S i)"
  from seq have "\<forall>i. (?S i,?S(Suc i)) \<in> rstep(ID R)" using rstepID_imp_plain_rstep by force
  hence seq': "\<forall>i. (?T i,?T(Suc i)) \<in> rstep R" unfolding rstep_iff_plain_rstep[symmetric] by simp
  hence "?T 0 = (unsharp t)" unfolding `S 0 = t` by simp
  with seq' have "\<not>SN(rstep R)" unfolding SN_def SN_elt_def by auto
  with `SN(rstep R)` show False by simp
 qed
next
 assume "SN(rstep(ID R))" show "SN(rstep R)"
 proof (rule ccontr)
  assume "\<not>SN(rstep R)"
  then obtain S t where start: "S 0 = t"
   and seq: "\<forall>i. (S i,S(Suc i)) \<in> rstep R" unfolding SN_def SN_elt_def by auto
  let ?S = "\<lambda>i. plain(S i)"
  from seq have seq': "\<forall>i. (?S i,?S(Suc i)) \<in> rstep(ID R)" unfolding rstep_iff_plain_rstep[symmetric] by simp
  hence "?S 0 = plain t" unfolding `S 0 = t` by simp
  with seq' have "\<not>SN(rstep(ID R))" unfolding SN_def SN_elt_def by auto
  with `SN(rstep(ID R))` show False by simp
 qed
qed

lemma suptp_plain[simp]: "t \<rhd> s \<Longrightarrow> plain t \<rhd> plain s"
proof (induct rule: suptp.induct)
 case (arg s ss f)
 hence "plain s \<in> set(map plain ss)" by simp
 thus ?case unfolding plain.simps by (rule suptp.arg)
next
 case (subt_of_arg s ss u f)
 hence "plain s \<in> set(map plain ss)" by simp
 from this and `plain s \<rhd> plain u` show ?case unfolding plain.simps by (rule suptp.subt_of_arg)
qed

lemma suptp_sharp[simp]: assumes "t \<rhd> s" shows "sharp t \<rhd> plain s"
using assms proof (cases rule: suptp.cases)
 case (arg u ss f)
 hence "plain s \<in> set(map plain ss)" by simp
 thus ?thesis unfolding `t = Fun f ss` `s = u` unfolding sharp.simps by (rule suptp.arg)
next
 case (subt_of_arg u ss v f)
 hence "plain u \<in> set(map plain ss)" and "plain u \<rhd> plain v" using suptp_plain by auto
 thus ?thesis unfolding subt_of_arg sharp.simps by (rule suptp.subt_of_arg)
qed

lemma map_unsharp_plain_id[simp]: "map (unsharp \<circ> plain) ss = ss" by (induct ss) auto

lemma unsharp_sharp_id[simp]: "unsharp(sharp s) = s"
proof (cases s)
 case (Var x) thus ?thesis unfolding Var by simp
next
 case (Fun f ss) thus ?thesis unfolding Fun by simp
qed

lemma SN_elt_imp_SN_elt_sharp:
 assumes "(SN_elt::('f,'v)trs \<Rightarrow> ('f,'v)term \<Rightarrow> bool) (rstep R) s"
 shows "(SN_elt::('f shp,'v)trs \<Rightarrow> ('f shp,'v)term \<Rightarrow> bool) (rstep(ID R)) (sharp s)"
proof (rule ccontr)
 assume "\<not>(SN_elt (rstep(ID R)) (sharp s))"
 then obtain S where start: "S 0 = sharp s" and seq: "\<forall>i. (S i,S(Suc i)) \<in> rstep(ID R)" unfolding SN_elt_def by auto
 let ?S = "\<lambda>i. plain(unsharp(S i))"
 let ?T = "\<lambda>i. unsharp(S i)"
 from seq have "\<forall>i. (?S i,?S(Suc i)) \<in> rstep(ID R)" using rstepID_imp_plain_rstep by best
 hence seq': "\<forall>i. (?T i,?T(Suc i)) \<in> rstep R" unfolding rstep_iff_plain_rstep[symmetric] by simp
 have "?T 0 = s" unfolding `S 0 = sharp s` by simp
 with seq' have "\<not>(SN_elt (rstep R) s)" unfolding SN_elt_def by auto
 with `SN_elt (rstep R) s` show False by simp
qed

lemma TinfID_imp_Tinf: assumes "sharp t \<in> Tinf (ID R)" shows "t \<in> Tinf R"
proof -
 from assms have supt:"\<forall>s. (sharp t) \<rhd> s \<longrightarrow> SN_elt (rstep(ID R)) s" and "\<not>SN_elt (rstep(ID R)) (sharp t)" unfolding Tinf_def by auto
 have "\<forall>s. t \<rhd> s \<longrightarrow> SN_elt (rstep R) s"
 proof (intro allI impI)
  fix s assume "t \<rhd> s"
  hence "sharp t \<rhd> plain s" by (rule suptp_sharp)
  with supt have "SN_elt (rstep(ID R)) (plain s)" by simp
  thus "SN_elt (rstep R) s" unfolding SN_elt_rstep_iff_SN_elt_rstepID[symmetric] by simp
 qed
 from `\<not>(SN_elt (rstep(ID R)) (sharp t))` have "\<not>(SN_elt (rstep R) t)" using SN_elt_imp_SN_elt_sharp by auto
 with `\<forall>s. t \<rhd> s \<longrightarrow> SN_elt (rstep R) s` show ?thesis unfolding Tinf_def by simp
qed

lemma suptp_plain'': assumes "plain t \<rhd> plain s" shows "unsharp(plain t) \<rhd> unsharp(plain s)"
using assms proof (induct rule: suptp.induct)
 case (arg s ss f)
 hence "unsharp s \<in> set(map unsharp ss)" by simp
 thus ?case unfolding unsharp.simps by (rule suptp.arg)
next
 case (subt_of_arg s ss u f)
 hence "unsharp s \<in> set(map unsharp ss)" by simp
 from this and `unsharp s \<rhd> unsharp u` show ?case unfolding unsharp.simps by (rule suptp.subt_of_arg)
qed

lemma suptp_unsharp: "t \<rhd> s \<Longrightarrow> unsharp t \<rhd> unsharp s"
proof (induct rule: suptp.induct)
 case (arg s ss f)
 hence "unsharp s \<in> set(map unsharp ss)" by simp
 thus ?case unfolding unsharp.simps by (rule suptp.arg)
next
 case (subt_of_arg s ss u f)
 hence "unsharp s \<in> set(map unsharp ss)" (is "_ \<in> set ?ss") by simp
 from this and `unsharp s \<rhd> unsharp u` show ?case unfolding unsharp.simps by (rule suptp.subt_of_arg)
qed

lemma suptp_plain'[simp]: "plain t \<rhd> plain s \<Longrightarrow> t \<rhd> s" using suptp_plain'' by simp

lemma suptp_sharp'[simp]: assumes "sharp t \<rhd> plain s" shows "t \<rhd> s"
using assms proof (cases rule: suptp.cases)
 case (arg u ss f)
 hence "unsharp u \<in> set(map unsharp ss)" (is "_ \<in> set ?ss") by simp
 hence "s \<in> set ?ss" unfolding `plain s = u`[symmetric] by simp
 from arg have "unsharp(sharp t) = Fun !f! ?ss" by simp
 hence "t = Fun !f! ?ss" by simp
 from `s \<in> set ?ss` show ?thesis unfolding `t = Fun !f! ?ss` by (rule suptp.arg)
next
 case (subt_of_arg u ss v f)
 hence "unsharp u \<in> set(map unsharp ss)" (is "_ \<in> set ?ss") by simp
 have "unsharp v = s" unfolding `plain s = v`[symmetric] by simp
 from subt_of_arg have "unsharp(sharp t) = unsharp(Fun f ss)" by simp
 hence "t = Fun !f! ?ss" by simp
 from `u \<rhd> v` have "unsharp u \<rhd> unsharp v" by (rule suptp_unsharp)
 with `unsharp u \<in> set ?ss` show ?thesis unfolding `t = Fun !f! ?ss` `unsharp v = s`[symmetric] by (rule suptp.subt_of_arg)
qed

lemma suptp_plain_conv[iff]: "(plain s \<rhd> plain t) = (s \<rhd> t)" using iffI[OF suptp_plain' suptp_plain] by simp

lemma suptp_sharp_conv[iff]: "(sharp s \<rhd> plain t) = (s \<rhd> t)" using iffI[OF suptp_sharp' suptp_sharp] by simp
 
lemma reverseTrs[iff] : "(rstep R)^-1 = rstep (R^-1)"
proof - 
  have oneDir: "\<And>R. (rstep R)^-1 \<subseteq> rstep (R^-1)" 
  proof
    fix R st
    show "(st \<in> (rstep R)^-1) \<Longrightarrow> (st \<in> rstep (R^-1))"
    proof (cases st)
      case (Pair s t)
      assume "st \<in> (rstep R)^-1"
      with Pair have "(t,s) \<in> rstep R" by simp
      hence "(s,t) \<in> rstep (R^-1)" 
      proof
        fix l r C \<sigma> assume in_R: "(l, r) \<in> R" and t: "t = C\<langle>l\<cdot>\<sigma>\<rangle>" and s: "s = C\<langle>r\<cdot>\<sigma>\<rangle>"
        from in_R show "(s,t) \<in> rstep (R^-1)" unfolding s t by auto
      qed
      with Pair show ?thesis by simp
    qed
  qed
  hence one: "(rstep R)^-1 \<subseteq> rstep (R^-1)" by blast
  from oneDir have "(rstep (R^-1))^-1 \<subseteq> rstep ((R^-1) ^-1)" by blast
  with one show ?thesis by force
qed

lemma no_sharp_rrstep:
 fixes R :: "('f,'v)trs"
 assumes "wf_trs R" shows "(sharp s,t) \<notin> rrstep(ID R)"
proof
 assume "(sharp s,t) \<in> rrstep(ID R)"
 then obtain l r :: "('f shp,'v)term" and \<sigma> where step: "(sharp s,t) \<in> rstep_r_p_s (ID R) (l,r) \<epsilon> \<sigma>"
  unfolding rrstep_def by auto
 let ?C = "ctxt_of_pos_term \<epsilon> (sharp s)"
 have "?C = \<box>" by simp
 with step have "(l,r) \<in> ID R" and "l\<cdot>\<sigma> = sharp s" and "r\<cdot>\<sigma> = t" unfolding rstep_r_p_s_def Let_def by auto
 from `(l,r) \<in> ID R` obtain u v where "(u,v) \<in> R" and "l = plain u" unfolding ID_def by auto
 have "wf_trs (ID R)" by (rule wf_trs_imp_wf_ID[OF assms])
 from `(l,r) \<in> ID R` and `wf_trs (ID R)` obtain f ss where "l = Fun f ss" unfolding wf_trs_def by best
 from `(u,v) \<in> R` and `wf_trs R` obtain g ts where "u = Fun g ts" unfolding wf_trs_def by best
 from `l\<cdot>\<sigma> = sharp s` obtain h us where "s = Fun h us" unfolding `l = Fun f ss` sub_apply.simps by (induct s) auto
 from `l\<cdot>\<sigma> = sharp s` have "f = h\<sharp>" unfolding `l = Fun f ss` `s = Fun h us` by simp
 hence "l = Fun h\<sharp> ss" using `l = Fun f ss` by simp
 with `l = plain u` show False unfolding `u = Fun g ts` plain.simps by simp
qed

lemma poss_unsharp_conv[iff]: "(p \<in> poss(unsharp s)) = (p \<in> poss s)"
proof (induct s arbitrary: p)
 case (Var x) thus ?case by simp
next
 case (Fun f ss) thus ?case by (cases p) auto
qed

lemma ctxt_of_pos_term_unsharp_conv[iff]:
 assumes "p \<in> poss s"
 shows "unsharp_ctxt(ctxt_of_pos_term p s) = (ctxt_of_pos_term p (unsharp s))"
using assms proof (induct s arbitrary: p)
 case (Var x) thus ?case by simp
next
 case (Fun f ss) thus ?case
 proof (cases p)
  case Empty thus ?thesis by simp
 next
  case (PCons i q)
  with `p \<in> poss(Fun f ss)` have "i < length ss" and "q \<in> poss(ss!i)" unfolding PCons poss.simps by auto
  hence "ss!i \<in> set ss" by simp
  with Fun and `q \<in> poss(ss!i)`
  have IH: "unsharp_ctxt(ctxt_of_pos_term q (ss!i)) = ctxt_of_pos_term q (unsharp(ss!i))" by simp
  have "unsharp_ctxt(ctxt_of_pos_term p (Fun f ss)) = unsharp_ctxt(ctxt_of_pos_term (i<#q) (Fun f ss))" unfolding PCons by simp
  also have "\<dots> = unsharp_ctxt(More f (take i ss) (ctxt_of_pos_term q (ss!i)) (drop (Suc i) ss))" by simp
  also have "\<dots> = More !f! (map unsharp (take i ss)) (unsharp_ctxt(ctxt_of_pos_term q (ss!i))) (map unsharp (drop (Suc i) ss))" by simp
  also have "\<dots> = More !f! (map unsharp (take i ss)) (ctxt_of_pos_term q (unsharp(ss!i))) (map unsharp (drop (Suc i) ss))" unfolding IH by simp
  also have "\<dots> = More !f! (take i (map unsharp ss)) (ctxt_of_pos_term q (map unsharp ss!i)) (drop (Suc i) (map unsharp ss))" unfolding nth_map[OF `i < length ss`,symmetric] take_map drop_map nth_map by simp
  finally show ?thesis unfolding PCons by simp
 qed
qed

lemma nrrstep_imp_unsharp_nrrstep:
 assumes "(s,t) \<in> nrrstep(ID R)" shows "(unsharp s,unsharp t) \<in> nrrstep R"
proof -
 from assms obtain l r i p \<sigma> where step: "(s,t) \<in> rstep_r_p_s (ID R) (l,r) (i<#p) \<sigma>" unfolding nrrstep_def by auto
 let ?C = "ctxt_of_pos_term (i<#p) s"
 from step have "i<#p \<in> poss s" and "(l,r) \<in> ID R" and "?C\<langle>l\<cdot>\<sigma>\<rangle> = s" and "?C\<langle>r\<cdot>\<sigma>\<rangle> = t" unfolding rstep_r_p_s_def Let_def by auto
 from `(l,r) \<in> ID R` have "(unsharp l,unsharp r) \<in> R" (is "(?l,?r) \<in> R") unfolding ID_def by auto
 from `i<#p \<in> poss s` have "i<#p \<in> poss(unsharp s)" by simp
 let ?D = "unsharp_ctxt ?C"
 let ?\<sigma> = "Sub(unsharp \<circ> (case \<sigma> of Sub s \<Rightarrow> s))"
 from `i<#p \<in> poss s` have ctxt: "?D = ctxt_of_pos_term (i<#p) (unsharp s)" by simp
 have "?D\<langle>?l\<cdot>?\<sigma>\<rangle> = unsharp(?C\<langle>l\<cdot>\<sigma>\<rangle>)" unfolding unsharp_sub_distr[symmetric] by simp
 hence lhs: "?D\<langle>?l\<cdot>?\<sigma>\<rangle> = unsharp s" unfolding `?C\<langle>l\<cdot>\<sigma>\<rangle> = s` by simp
 have "?D\<langle>?r\<cdot>?\<sigma>\<rangle> = unsharp(?C\<langle>r\<cdot>\<sigma>\<rangle>)" unfolding unsharp_sub_distr[symmetric] by simp
 hence rhs: "?D\<langle>?r\<cdot>?\<sigma>\<rangle> = unsharp t" unfolding `?C\<langle>r\<cdot>\<sigma>\<rangle> = t` by simp
 from `i<#p \<in> poss(unsharp s)` and `(?l,?r) \<in> R` and lhs and rhs
  show ?thesis unfolding nrrstep_def ctxt rstep_r_p_s_def Let_def by force
qed

lemma Tinf_imp_SN_sharp:
 fixes R :: "('f,'v)trs"
 assumes "wf_trs R" and "s \<in> Tinf R"
 shows "SN_elt (rstep(ID R)) (sharp s)"
proof (rule ccontr)
 assume nontermin: "\<not>(SN_elt (rstep(ID R)) (sharp s))"
 then obtain S where "S 0 = sharp s" and seq:"\<forall>i. (S i,S(Suc i)) \<in> rstep(ID R)" unfolding SN_elt_def by auto
 hence union: "\<And>i. (S i,S(Suc i)) \<in> rrstep(ID R) \<union> nrrstep(ID R)" unfolding rstep_iff_rrstep_or_nrrstep by simp
 from seq have step: "(sharp s,S(Suc 0)) \<in> rstep(ID R)" unfolding `S 0 = sharp s`[symmetric] by simp
 then obtain l r p \<sigma> where step': "(sharp s,S(Suc 0)) \<in> rstep_r_p_s (ID R) (l,r) p \<sigma>" unfolding rstep_iff_rstep_r_p_s by auto
 let ?C = "ctxt_of_pos_term p (sharp s)"
 from step' have "p \<in> poss(sharp s)" and "(l,r) \<in> ID R" and "?C\<langle>l\<cdot>\<sigma>\<rangle> = sharp s" and "?C\<langle>r\<cdot>\<sigma>\<rangle> = S(Suc 0)" unfolding rstep_r_p_s_def Let_def by auto
 have "wf_trs (ID R)" by (rule wf_trs_imp_wf_ID[OF `wf_trs R`])
 with `(l,r) \<in> ID R` obtain g ts where "l = Fun g ts" unfolding wf_trs_def by best
 from `?C\<langle>l\<cdot>\<sigma>\<rangle> = sharp s` obtain C where "C\<langle>l\<cdot>\<sigma>\<rangle> = sharp s" by auto
 obtain f ss where "s = Fun f ss" using `C\<langle>l\<cdot>\<sigma>\<rangle> = sharp s` unfolding `l = Fun g ts` by (induct s,cases C) auto
 have "sharp s = Fun f\<sharp> (map plain ss)" (is "_ = Fun f\<sharp> ?ss") unfolding `s = Fun f ss` by simp
 hence "S 0 = Fun f\<sharp> ?ss" unfolding `S 0 = sharp s` by simp
 have "\<forall>u. sharp s \<rhd> u \<longrightarrow> SN_elt (rstep(ID R)) u"
 proof (intro allI impI)
  fix u assume "sharp s \<rhd> u"
  then obtain v where "u = plain v" using subt_of_sharp_imp_plain by auto
  from `sharp s \<rhd> u` have "s \<rhd> v" unfolding `u = plain v` suptp_sharp_conv by simp
  with `s \<in> Tinf R` have "SN_elt (rstep R) v" unfolding Tinf_def by auto
  thus "SN_elt (rstep(ID R)) u" unfolding `u = plain v` unfolding SN_elt_rstep_iff_SN_elt_rstepID[symmetric] by simp
 qed
 with nontermin have "sharp s \<in> Tinf (ID R)" unfolding Tinf_def by auto
 hence "SN_elt (nrrstep(ID R)) (sharp s)" by (rule Tinf_imp_SN_elt_nrrstep)
 hence first: "SN_elt (nrrstep(ID R)) (S 0)" unfolding `S 0 = sharp s` by simp
 obtain i where "(S i,S(Suc i)) \<in> rrstep(ID R)"
  and prefix: "\<forall>j<i. (S j,S(Suc j)) \<in> nrrstep(ID R) \<and> (S j,S(Suc j)) \<notin> rrstep(ID R)" using first_R_step[where t = S,OF union first] by auto
 have less: "\<forall>j<i.\<exists>us. S j = Fun f\<sharp> us"
 proof (intro allI impI)
  fix j assume "j < i" thus "\<exists>us. S j = Fun f\<sharp> us"
  proof (induct j)
   case 0 thus ?case using `S 0 = Fun f\<sharp> ?ss` by best
  next
   case (Suc j)
   then obtain us' where "S j = Fun f\<sharp> us'" by auto
   have "(S j,S(Suc j)) \<in> nrrstep(ID R)" using prefix and Suc by auto
   then obtain us'' where "S(Suc j) = Fun f\<sharp> us''" using nrrstep_preserves_root unfolding `S j = Fun f\<sharp> us'` by best
   thus ?case by auto
  qed
 qed
 have "\<exists>us. S i = Fun f\<sharp> us"
 proof (cases i)
  case 0 thus ?thesis using `S 0 = Fun f\<sharp> ?ss` by auto
 next
  case (Suc n)
  from less have "\<exists>us. S n = Fun f\<sharp> us" unfolding Suc by auto
  then obtain vs where "S n = Fun f\<sharp> vs" by auto
  have "(S n,S(Suc n)) \<in> nrrstep(ID R)" using prefix unfolding Suc by auto
  then obtain vs' where "S i = Fun f\<sharp> vs'" using nrrstep_preserves_root unfolding Suc `S n = Fun f\<sharp> vs` by fast
  thus ?thesis by auto
 qed
 then obtain us where "S i = Fun f\<sharp> us" by auto
 from `(S i,S(Suc i)) \<in> rrstep(ID R)` have "(Fun f\<sharp> us,S(Suc i)) \<in> rrstep(ID R)" unfolding `S i = Fun f\<sharp> us` by auto
 then obtain l r \<sigma> where "(Fun f\<sharp> us,S(Suc i)) \<in> rstep_r_p_s (ID R) (l,r) \<epsilon> \<sigma>" unfolding rrstep_def by auto
 hence "(l,r) \<in> ID R" and "l\<cdot>\<sigma> = Fun f\<sharp> us" unfolding rstep_r_p_s_def Let_def by auto
 then obtain l' r' where "(l',r') \<in> R" and "l = plain l'" unfolding ID_def by auto
 with `wf_trs R` obtain g vs where "l' = Fun g vs" unfolding wf_trs_def by best
 from `l\<cdot>\<sigma> = Fun f\<sharp> us` show False unfolding `l = plain l'` `l' = Fun g vs` by simp
qed



lemma left_var_imp_not_SN : fixes R :: "('f,'v)trs" and t :: "('f,'v)term"
  assumes "(Var y,r) \<in> R" (is "(?y,_) \<in> _") shows 
  "\<not> (SN_elt (rstep R) t)"
proof -
  have steps: "\<forall> t. \<exists> s. (t,s) \<in> rstep R"
  proof 
    fix t :: "('f,'v)term"
    let ?yt = "subst y t"
    let ?ryt = "subst y (r \<cdot> ?yt)"
    from assms have "(?y \<cdot> ?yt, r \<cdot> ?yt) \<in> CS R" by (rule  csI)
    hence "(\<box>\<langle>?y \<cdot> ?yt\<rangle>, \<box>\<langle>r \<cdot> ?yt\<rangle>) \<in> CC (CS R)" by (rule ccI)
    hence "(t, r \<cdot> ?yt) \<in> rstep R" unfolding subst_def by auto 
    thus "\<exists> s. (t,s) \<in> rstep R" ..
  qed
  show ?thesis 
  proof
    assume "SN_elt (rstep R) t"
    from this obtain s where norm: "(t,s) \<in> (rstep R)^!" using SN_imp_normalizability by force
    from steps obtain u where "(s,u) \<in> rstep R" by auto
    with norm show False by auto
  qed
qed        

lemma var_rewrite_imp_not_SN : assumes sn: "SN_elt (rstep R) t" and step: "(t,s) \<in> rstep R" 
  shows "\<not> (is_var t)" 
using assms
proof (cases t)
  case (Fun f ts) thus ?thesis by simp
next
  case (Var x)
  from step obtain l r p \<sigma> where "(Var x, s) \<in> rstep_r_p_s R (l,r) p \<sigma>" unfolding Var rstep_iff_rstep_r_p_s by best
  hence "l \<cdot> \<sigma> = Var x" and rule: "(l,r) \<in> R" unfolding rstep_r_p_s_def by (auto simp: Let_def)
  from this obtain y where "l = Var y" (is "_ = ?y") by (cases l, auto)
  with rule have "(?y, r) \<in> R" by auto
  hence "\<not> (SN_elt (rstep R) t)" by (rule left_var_imp_not_SN)
  with sn show ?thesis by blast
qed

lemma rstep_empty[simp] : "rstep {} = {}"
proof (rule set_ext, simp, rule notI, simp only: split_paired_all)
  fix l r
  assume "(l,r) \<in> rstep {}"
  thus False by (simp only: rstep_iff_rstep_r_p_s rstep_r_p_s_def, auto)
qed

lemma rstep_union : "rstep (R \<union> S) = rstep R \<union> rstep S"
proof (rule set_ext, simp only: split_paired_all)
  fix s t
  show "((s,t) \<in> rstep (R \<union> S)) = ((s,t) \<in> rstep R \<union> rstep S)"
    by (simp only: Un_iff, simp only: rstep_iff_rstep_r_p_s, unfold rstep_r_p_s_def, force simp: Let_def)
qed

lemma rstep_id : "rstep Id = Id"
  by (rule set_ext, simp only: split_paired_all, simp only: rstep_iff_rstep_r_p_s, unfold rstep_r_p_s_def, auto simp: Let_def, 
    rule exI, rule exI[where x = \<epsilon>], auto, rule exI[where x = \<iota>], auto)

lemma rstep_imp_wfrstep:
  assumes "(s,t) \<in> rstep_r_p_s R (l,r) p (Sub \<sigma>)"
  shows "(l\<cdot>Sub(clean v (funas_trs R) \<circ> \<sigma>),r\<cdot>Sub(clean v (funas_trs R) \<circ> \<sigma>)) \<in> wfrstep R"
proof -
  let ?l = "l\<cdot>(Sub \<sigma>)"
  let ?r = "r\<cdot>(Sub \<sigma>)"
  let ?u = "clean v (funas_trs R) ?l"
  let ?v = "clean v (funas_trs R) ?r"
  let ?C = "ctxt_of_pos_term p s"
  let ?\<sigma> = "Sub(clean v (funas_trs R) o \<sigma>)"
  from assms have "p \<in> poss s" and "(l,r) \<in> R"
  and "?C\<langle>l\<cdot>Sub \<sigma>\<rangle> = s" and "?C\<langle>r\<cdot>Sub \<sigma>\<rangle> = t" unfolding rstep_r_p_s_def Let_def by auto
  from `(l,r) \<in> R` have "funas_term l \<subseteq> funas_trs R" unfolding funas_trs_def funas_rule_def by auto
  moreover have "l \<in> Term(funas_term l)" by simp
  ultimately have "l \<in> Term(funas_trs R)" using wf_fun_terms_subset[of "funas_term l" "funas_trs R"] by best
  from `(l,r) \<in> R` have "funas_term r \<subseteq> funas_trs R" unfolding funas_trs_def funas_rule_def by auto
  moreover have "r \<in> Term(funas_term r)" by simp
  ultimately have "r \<in> Term(funas_trs R)" using wf_fun_terms_subset[of "funas_term r" "funas_trs R"] by best
  have "?u = l\<cdot>?\<sigma>" using `l \<in> Term(funas_trs R)` by simp
  have "?v = r\<cdot>?\<sigma>" using `r \<in> Term(funas_trs R)` by simp
  moreover have "(?u,?v) \<in> wfrstep R"
  proof -
    have "?u \<in> Term(funas_trs R)" by simp
    moreover have "?v \<in> Term(funas_trs R)" by simp
    moreover have "(?u,?v) \<in> rstep R"
    proof -
      from `(l,r) \<in> R` have "(?u,?v) \<in> CS R" unfolding `?u = l\<cdot>?\<sigma>` `?v = r\<cdot>?\<sigma>` by auto
      thus ?thesis using ccI[of ?u ?v "CS R" \<box>] by simp
    qed
    ultimately show ?thesis unfolding wfrstep_def by force
  qed
  ultimately show ?thesis unfolding `?u = l\<cdot>?\<sigma>` `?v = r\<cdot>?\<sigma>` by auto
qed

lemma rstep_imp_clean_rstep_or_Id:
  fixes R::"('f,'v)trs" and F::"('f\<times>nat)set"
  defines [simp]: "c \<equiv> clean (\<lambda>x. SOME(x::'v). True) F"
  assumes subset: "funas_trs R \<subseteq> F"
    and step: "(s,t) \<in> rstep R"
  shows "(c(s),c(t)) \<in> rstep R \<union> Id"
proof -
  from rstep_imp_C_s_r[OF step] obtain C l r \<sigma>
    where R: "(l,r) \<in> R"
    and s: "s = C\<langle>l\<cdot>\<sigma>\<rangle>"
    and t: "t = C\<langle>r\<cdot>\<sigma>\<rangle>" by auto
  from s t show ?thesis
  proof (induct C arbitrary: s t)
    case Hole
    let ?c = "\<lambda>\<sigma>. case \<sigma> of Sub s \<Rightarrow> Sub(c \<circ> s)"
    from clean_sub_apply[OF lhs_wf[OF R subset]] have s: "c(s) = l\<cdot>?c(\<sigma>)" unfolding Hole by (cases \<sigma>) simp_all
    from clean_sub_apply[OF rhs_wf[OF R subset]] have t: "c(t) = r\<cdot>?c(\<sigma>)" unfolding Hole by (cases \<sigma>) simp_all
    from R have "(c(s),c(t)) \<in> CS R" unfolding s t by auto
    hence "(c(s),c(t)) \<in> rstep R" using ccI[where C=\<box>] by simp
    thus ?case ..
  next
    case (More f ss1 D ss2)
    hence IH: "(c(D\<langle>l\<cdot>\<sigma>\<rangle>),c(D\<langle>r\<cdot>\<sigma>\<rangle>)) \<in> rstep R \<union> Id" by simp
    show ?case
    proof (cases "(f,Suc(length(ss1@ss2))) \<in> F")
      case False
      with More have "c(s) = Var(SOME(x::'v). True)" by simp
      moreover from False More have "c(t) = Var(SOME(x::'v). True)" by simp
      ultimately have "(c(s),c(t)) \<in> Id" by simp
      thus ?thesis ..
    next
      case True
      let ?s = "c(D\<langle>l\<cdot>\<sigma>\<rangle>)"
      let ?t = "c(D\<langle>r\<cdot>\<sigma>\<rangle>)"
      from IH show ?thesis
      proof
        assume "(?s,?t) \<in> Id" with True More show ?thesis by auto
      next
        let ?C = "More f (map c ss1) \<box> (map c ss2)"
        assume "(?s,?t) \<in> rstep R"
        from rstep_imp_ctxt_rstep[OF this,where C="?C"]
          have "(c(s),c(t)) \<in> rstep R" unfolding More using True by simp
        thus ?thesis ..
      qed
    qed
  qed
qed

lemma funas_term_plain[simp]: "(f,n) \<in> funas_term(plain t) \<Longrightarrow> (!f!,n) \<in> funas_term t" by (induct t) auto

lemma funasID_funas:
  fixes R::"('f,'v)trs" assumes "(f,n) \<in> funas_trs(ID R)" shows "(!f!,n) \<in> funas_trs R"
proof -
  from assms obtain l r::"('f shp,'v)term"
    where ID_R: "(l,r) \<in> ID R" and "(f,n) \<in> funas_rule(l,r)"
    unfolding funas_trs_def by auto
  from ID_R obtain l' r'::"('f,'v)term" where "l = plain l'" and "r = plain r'" and "(l',r') \<in> R"
    unfolding ID_def by auto
  from `(f,n) \<in> funas_rule(l,r)` have "(f,n) \<in> funas_term l \<or> (f,n) \<in> funas_term r" unfolding funas_rule_def by auto
  thus ?thesis
  proof
    assume "(f,n) \<in> funas_term l"
    from `(f,n) \<in> funas_term l` and `l = plain l'` have "(!f!,n) \<in> funas_term l'" by simp
    with `(l',r') \<in> R` show ?thesis unfolding funas_trs_def funas_rule_def by force
  next
    assume "(f,n) \<in> funas_term r"
    from `(f,n) \<in> funas_term r` and `r = plain r'` have "(!f!,n) \<in> funas_term r'" by simp
    with `(l',r') \<in> R` show ?thesis unfolding funas_trs_def funas_rule_def by force
  qed
qed

lemma funas_term_sharp[simp]:
  fixes t::"('f,'v)term" assumes "(f,n) \<in> funas_term(sharp t)" shows "(!f!,n) \<in> funas_term t"
proof (cases t)
  case (Var x) with assms show ?thesis by simp
next
  case (Fun g ts)
  from assms have t: "sharp t = Fun g\<sharp> (map plain ts)" unfolding Fun by simp
  hence "funas_term(sharp t) = {(g\<sharp>,length(map plain ts))} \<union> \<Union>set(map funas_term (map plain ts))" by simp
  with assms have "(f,n) = (g\<sharp>,length(map plain ts)) \<or> (f,n) \<in> \<Union>set(map funas_term (map plain ts))" (is "?A \<or> ?B") by auto
  thus ?thesis
  proof
    assume ?A with assms show ?thesis unfolding Fun by simp
  next
    assume ?B
    hence "(f,n) \<in> \<Union>set(map (funas_term\<circ>plain) ts)" by simp
    then obtain t where "t \<in> set ts" and "(f,n) \<in> funas_term(plain t)" unfolding o_def by auto
    hence "(!f!,n) \<in> funas_term t" by simp
    with `t \<in> set ts` show ?thesis unfolding Fun by auto
  qed
qed

lemma funasDP_funas:
  fixes R::"('f,'v)trs" assumes "(f,n) \<in> funas_trs(DP R)" shows "(!f!,n) \<in> funas_trs R"
proof -
  from assms obtain l u::"('f shp,'v)term"
    where DP_R: "(l,u) \<in> DP R" and "(f,n) \<in> funas_rule(l,u)"
    unfolding funas_trs_def by auto
  from DP_R obtain l' r'::"('f,'v)term" and g ts where "l = sharp l'" and "u = sharp(Fun g ts)" and "(l',r') \<in> R"
    and "r' \<unrhd> Fun g ts" and "defined R g"
    unfolding DP_def by auto
  from `(f,n) \<in> funas_rule(l,u)` have "(f,n) \<in> funas_term l \<or> (f,n) \<in> funas_term u" unfolding funas_rule_def by auto
  thus ?thesis
  proof
    assume "(f,n) \<in> funas_term l"
    with `l = sharp l'` have "(!f!,n) \<in> funas_term l'" by simp
    with `(l',r') \<in> R` show ?thesis unfolding funas_trs_def funas_rule_def by force
  next
    assume "(f,n) \<in> funas_term u"
    from `r' \<unrhd> Fun g ts` have subs: "funas_term(Fun g ts) \<subseteq> funas_term(r')" by (rule supteqp_imp_funas_subset)
    from `(f,n) \<in> funas_term u` have "(f,n) \<in> funas_term(Fun g\<sharp> (map plain ts))" unfolding `u = sharp(Fun g ts)` by simp
    hence "(f,n) = (g\<sharp>,length(map plain ts)) \<or> (f,n) \<in> \<Union>set(map funas_term (map plain ts))" (is "?A \<or> ?B") by auto
    thus ?thesis
    proof
      assume ?A
      with subs have "(!f!,n) \<in> funas_term r'" by simp
      thus ?thesis using `(l',r') \<in> R` unfolding funas_trs_def funas_rule_def by force
    next
      assume ?B
      then obtain t where "t \<in> set ts" and "(f,n) \<in> funas_term(plain t)" unfolding map_map[symmetric] o_def by auto
      hence "(!f!,n) \<in> funas_term t" by simp
      with `t \<in> set ts` have "(!f!,n) \<in> funas_term(Fun g ts)" by auto
      with `r' \<unrhd> Fun g ts` have "(!f!,n) \<in> funas_term r'" using supteqp_imp_funas_subset by best
      thus ?thesis using `(l',r') \<in> R` unfolding funas_trs_def funas_rule_def by force
    qed
  qed
qed

lemma unsharp_clean_conv[simp]:
  fixes R::"('f,'v)trs" and v::"('f,'v)term \<Rightarrow> 'v"
  defines "c \<equiv> clean v (funas_trs R)"
    and "c' \<equiv> clean (v \<circ> unsharp) ({(Plain f,n) | f n. (f,n) \<in> funas_trs R} \<union> {(Sharp f,n) | f n. (f,n) \<in> funas_trs R})" (is "_ \<equiv> clean (v \<circ> unsharp) ?F")
  shows "unsharp(c' t) = c(unsharp t)"
proof (induct t)
  case (Var x) show ?case by (simp add: c_def c'_def)
next
  case (Fun f ts)
  hence IH: "\<And>s. s \<in> set ts \<Longrightarrow> unsharp(c' s) = c(unsharp s)" by simp
  show ?case
  proof (cases "(!f!,length ts) \<in> funas_trs R")
    case False thus ?thesis unfolding c_def c'_def by auto
  next
    case True show ?thesis
    proof (cases f)
      case (Sharp g)
      with True have "(f,length ts) \<in> ?F" by simp
      with IH show ?thesis unfolding c'_def c_def Sharp by simp
    next
      case (Plain g)
      with True have "(f,length ts) \<in> ?F" by simp
      with IH show ?thesis unfolding c'_def c_def Plain by auto
    qed
  qed
qed

lemma rstepID_imp_rstep':
  assumes "(s,t) \<in> rstep(ID R)" shows "(unsharp s,unsharp t) \<in> rstep R"
proof -
  from rstep_imp_C_s_r[OF assms] obtain C \<sigma> l r
    where ID_R: "(l,r) \<in> ID R"
    and s: "s = C\<langle>l\<cdot>\<sigma>\<rangle>"
    and t: "t = C\<langle>r\<cdot>\<sigma>\<rangle>" by auto
  from ID_R obtain l' r' where R: "(l',r') \<in> R" and l: "l = plain l'" and r: "r = plain r'"
    unfolding ID_def by auto
  let ?C = "unsharp_ctxt C"
  let ?\<sigma> = "case \<sigma> of Sub s \<Rightarrow> Sub(unsharp \<circ> s)"
  have us: "unsharp s = ?C\<langle>l'\<cdot>?\<sigma>\<rangle>" unfolding s l unfolding unsharp_ctxt_distr unsharp_sub_distr unsharp_plain_id by (cases \<sigma>) auto
  have ut: "unsharp t = ?C\<langle>r'\<cdot>?\<sigma>\<rangle>" unfolding t r unfolding unsharp_ctxt_distr unsharp_sub_distr unsharp_plain_id by (cases \<sigma>) auto
  from R have "(l'\<cdot>?\<sigma>,r'\<cdot>?\<sigma>) \<in> CS R" using csI by auto
  thus ?thesis unfolding us ut using ccI[where C="?C"] by auto
qed

lemma rstepID_seq_imp_rstep_seq:
  assumes "(s,t) \<in> (rstep(ID R))^*" shows "(unsharp s,unsharp t) \<in> (rstep R)^*"
using assms proof (induct rule: rtrancl_induct)
  case base show ?case by simp
next
  case (step y z)
  from rstepID_imp_rstep'[OF `(y,z) \<in> rstep(ID R)`] have "(unsharp y,unsharp z) \<in> rstep R" by simp
  with step show ?case by simp
qed

lemma wfrstep_conv:
  fixes F::"('f\<times>nat)set" and R::"('f,'v)trs"
  assumes wf: "wf_trs R"
  shows "{(x,y) | x y. (x,y) \<in> (rstep R)^+ \<and> x \<in> Term(funas_trs R) \<and> y \<in> Term(funas_trs R)} = (wfrstep R)^+" (is "?R = ?S")
proof
  show "?R \<subseteq> ?S"
  proof (rule subsetI2)
    have subset: "funas_trs R \<subseteq> funas_trs R" by simp
    fix s t assume "(s,t) \<in> ?R"
    hence "(s,t) \<in> (rstep R)^+" and "s \<in> Term(funas_trs R)" by auto
    hence "(s,t) \<in> ?S \<and> t \<in> Term(funas_trs R)"
    proof (induct rule: trancl.induct)
      case (r_into_trancl a b)
      hence "a \<in> Term(funas_trs R)" and "(a,b) \<in> rstep R" using r_into_trancl by auto
      from rstep_preserves_wf_fun_terms[OF subset `a \<in> Term(funas_trs R)` `(a,b) \<in> rstep R` wf] have "b \<in> Term(funas_trs R)" .
      with `a \<in> Term(funas_trs R)` `(a,b) \<in> rstep R` have "(a,b) \<in> wfrstep R" unfolding wfrstep_def by blast
      thus ?case using `b \<in> Term(funas_trs R)` by simp
    next
      case (trancl_into_trancl a b c)
      hence IH: "(a,b) \<in> ?S" and "b \<in> Term(funas_trs R)" by auto
      from rstep_preserves_wf_fun_terms[OF subset `b \<in> Term(funas_trs R)` `(b,c) \<in> rstep R` wf] have "c \<in> Term(funas_trs R)" .
      with `b \<in> Term(funas_trs R)` `(b,c) \<in> rstep R` have "(b,c) \<in> wfrstep R" unfolding wfrstep_def by blast
      with `(a,b) \<in> (wfrstep R)^+` and `c \<in> Term(funas_trs R)` show ?case by auto
    qed
    thus "(s,t) \<in> (wfrstep R)^+" by simp
  qed
next
  show "?S \<subseteq> ?R"
  proof (rule subsetI2)
    fix s t assume "(s,t) \<in> ?S" thus "(s,t) \<in> ?R" by (induct rule: trancl.induct) (auto simp: wfrstep_def)
  qed
qed

lemma DP_step_in_R:
  fixes R::"('f,'v)trs" and v::"('f,'v)term \<Rightarrow> 'v"
  assumes "(s,t) \<in> DP R" shows "\<exists>C. C \<in> Ctxt(funas_trs R) \<and> (unsharp s,C\<langle>unsharp t\<rangle>) \<in> R"
proof -
  from assms obtain l r f ts
    where s: "s = sharp l" and t: "t = sharp(Fun f ts)"
    and R: "(l,r) \<in> R" and sub: "r \<unrhd> Fun f ts" and "defined R f" unfolding DP_def by auto
  from supteqp_imp_ctxt[OF sub] obtain C where r: "r = C\<langle>Fun f ts\<rangle>" by auto
  have subset: "funas_trs R \<subseteq> funas_trs R" by simp
  from rhs_wf[OF R subset] have "r \<in> Term(funas_trs R)" .
  hence "C\<langle>Fun f ts\<rangle> \<in> Term(funas_trs R)" unfolding r .
  from ctxt_apply_wf_imp_ctxt_and_term_wf[OF this] have "C \<in> Ctxt(funas_trs R)" and "Fun f ts \<in> Term(funas_trs R)" by auto
  from lhs_wf[OF R subset] have "l \<in> Term(funas_trs R)" .
  have us: "unsharp s = l" unfolding s by simp
  have ut: "unsharp t = Fun f ts" unfolding t unsharp_sharp_id[of "Fun f ts"] ..
  from R have "(unsharp s,C\<langle>unsharp t\<rangle>) \<in> R" unfolding us ut r .
  with `C \<in> Ctxt(funas_trs R)` show ?thesis by best
qed

lemma funas_DP_ID_subset:
  "funas_trs(DP R) \<union> funas_trs(ID R) \<subseteq> {(Plain f,n) | f n. (f,n) \<in> funas_trs R} \<union> {(Sharp f,n) | f n. (f,n) \<in> funas_trs R}" (is "?F \<union> ?G \<subseteq> ?H \<union> ?I")
proof (rule Un_least)
  show "?F \<subseteq> ?H \<union> ?I"
  proof (rule subsetI2)
    fix f n assume "(f,n) \<in> ?F" hence "(!f!,n) \<in> funas_trs R" by (rule funasDP_funas)
    thus "(f,n) \<in> ?H \<union> ?I" by (cases f) auto
  qed
next
  show "?G \<subseteq> ?H \<union> ?I"
  proof (rule subsetI2)
    fix f n assume "(f,n) \<in> ?G" hence "(!f!,n) \<in> funas_trs R" by (rule funasID_funas)
    thus "(f,n) \<in> ?H \<union> ?I" by (cases f) auto
  qed
qed

types ('f,'v)trsL = "('f,'v)rule list"


lemma lhs_var_imp_wfrstep_not_SN: assumes "(Var x,r) \<in> R" shows "\<not> SN(wfrstep R)"
proof -
  let ?\<sigma> = "Sub(\<lambda>x. r)"
  let ?t = "\<lambda>i. r\<cdot>(?\<sigma>^i)"
  from rhs_wf[OF assms] have wf_r: "r \<in> Term(funas_trs R)" by fast
  have clean: "?\<sigma> = Sub(\<lambda>x. clean v (funas_trs R) r)" unfolding clean_fun_terms_apply_id[OF wf_r] by simp
  have "\<forall>i. ?t i \<in> Term(funas_trs R)"
  proof
    fix i show "?t i \<in> Term(funas_trs R)"
    proof (induct i)
      case 0 from wf_r show ?case by (simp add: one_sub_def)
    next
      case (Suc i)
      from clean_sub_apply_wf[OF Suc,of v "\<lambda>x. r"] have "r\<cdot>?\<sigma>^i\<cdot>?\<sigma> \<in> Term(funas_trs R)" unfolding clean o_def .
      hence "r\<cdot>?\<sigma>^i \<circ>s ?\<sigma>^Suc 0 \<in> Term(funas_trs R)" by (simp add: one_sub_def times_sub_def)
      thus ?case unfolding sub_power_compose_distr[symmetric] by simp
    qed
  qed
  moreover have "\<forall>i. (?t i,?t(Suc i)) \<in> rstep R"
  proof
    fix i
    from csI[OF assms,of "?\<sigma>^Suc i"] have "(?t i,?t(Suc i)) \<in> CS R" by (simp add: times_sub_def)
    from ccI[OF this,of \<box>] show "(?t i,?t(Suc i)) \<in> rstep R" by simp
  qed
  ultimately have "\<forall>i. (?t i,?t(Suc i)) \<in> wfrstep R" unfolding wfrstep_def by blast
  thus ?thesis unfolding SN_defs by best
qed

lemma rhs_free_vars_imp_wfrstep_not_SN:
  assumes R: "(l,r) \<in> R" and free: "\<not> vars_term r \<subseteq> vars_term l" shows "\<not> SN(wfrstep R)"
proof -
  from free obtain x where x: "x \<in> vars_term r - vars_term l" by auto
  hence "x \<in> vars_term r" by simp
  from var_imp_supteqp[OF this] have "r \<unrhd> Var x" .
  from supteqp_imp_ctxt[OF this] obtain C where r: "C\<langle>Var x\<rangle> = r" by auto
  let ?s = "\<lambda>y. if y = x then l else Var y"
  let ?\<sigma> = "Sub ?s"
  let ?t = "\<lambda>i. ((C \<cdot>c ?\<sigma>)^i)\<langle>l\<rangle>"
  from rhs_wf[OF R] have wf_r: "r \<in> Term(funas_trs R)" by fast
  from lhs_wf[OF R] have wf_l: "l \<in> Term(funas_trs R)" by fast
  from ctxt_apply_wf_imp_ctxt_and_term_wf[OF wf_r[unfolded r[symmetric]]] have wf_C: "C \<in> Ctxt(funas_trs R)" by simp
  have clean: "?\<sigma> = Sub(\<lambda>y. clean v (funas_trs R) (?s y))"
    unfolding if_distrib clean_fun_terms_apply_id[OF wf_l] clean.simps by simp
  from x have neq: "\<forall>y\<in>vars_term l. y \<noteq> x" by auto
  have l: "l\<cdot>?\<sigma> = l"
  proof -
    have lsigma: "l\<cdot>?\<sigma> = l\<cdot>(?\<sigma>|_vars_term l)" unfolding coincidence_lemma[symmetric] by simp
    have aux: "(\<lambda>y. if y \<in> vars_term l then if y = x then l else Var y else Var y) = Var"
    proof
      fix y show "(if y \<in> vars_term l then if y = x then l else Var y else Var y) = Var y" using neq by auto
    qed
    have "l\<cdot>(?\<sigma>|_vars_term l) = l" unfolding sub_restrict.simps unfolding aux by simp
    thus ?thesis unfolding lsigma .
  qed
  have rsigma: "r\<cdot>?\<sigma> = (C \<cdot>c ?\<sigma>)\<langle>l\<rangle>" unfolding r[symmetric] by simp
  have "\<forall>i. ?t i \<in> Term(funas_trs R)"
  proof
    fix i show "?t i \<in> Term(funas_trs R)"
    proof (induct i)
      case 0 with wf_l show ?case by (simp add: one_ctxt_def)
    next
      case (Suc i)
      from ctxt_clean_sub_wf[OF wf_C,of v "\<lambda>y. ?s y"] have "C \<cdot>c ?\<sigma> \<in> Ctxt(funas_trs R)" unfolding o_def clean[symmetric] .
      from ctxt_apply_wf[OF this Suc] show ?case by (simp add: times_ctxt_def)
    qed
  qed
  moreover have "\<forall>i. (?t i,?t(Suc i)) \<in> rstep R"
  proof
    fix i
    from csI[OF R,of "?\<sigma>"] have "(l,(C \<cdot>c ?\<sigma>)\<langle>l\<rangle>) \<in> CS R" unfolding l rsigma .
    from ccI[OF this,of "(C \<cdot>c ?\<sigma>)^i"]
      have "(((C \<cdot>c ?\<sigma>)^i)\<langle>l\<rangle>, ((C \<cdot>c ?\<sigma>)^i \<circ>c (C \<cdot>c ?\<sigma>)^Suc 0)\<langle>l\<rangle>) \<in> rstep R"
      by (simp add: one_ctxt_def times_ctxt_def)
    thus "(?t i,?t(Suc i)) \<in> rstep R" unfolding ctxt_power_compose_distr[symmetric] by simp
  qed
  ultimately have "\<forall>i. (?t i,?t(Suc i)) \<in> wfrstep R" unfolding wfrstep_def by blast
  thus ?thesis unfolding SN_defs by best
qed

lemma lhs_var_imp_rstep_not_SN: assumes "(Var x,r) \<in> R" shows "\<not> SN(rstep R)"
using lhs_var_imp_wfrstep_not_SN[OF assms] unfolding wfrstep_def SN_defs by blast

lemma rhs_free_vars_imp_rstep_not_SN:
  assumes "(l,r) \<in> R" and "\<not> vars_term r \<subseteq> vars_term l"
  shows "\<not> SN(rstep R)"
using rhs_free_vars_imp_wfrstep_not_SN[OF assms] unfolding wfrstep_def SN_defs by blast


lemma SN_rstep_imp_wf_trs : assumes "SN (rstep R)" shows "wf_trs R"
proof (rule ccontr)
  assume "\<not> wf_trs R"
  then obtain l r where R: "(l,r) \<in> R"
    and not_wf: "(\<forall>f ts. l \<noteq> Fun f ts) \<or> \<not>(vars_term r \<subseteq> vars_term l)" unfolding wf_trs_def
    by auto
  from not_wf have "\<not> SN (rstep R)"
  proof
    assume free: "\<not> vars_term r \<subseteq> vars_term l"
    from rhs_free_vars_imp_rstep_not_SN[OF R free] show ?thesis .
  next
    assume "\<forall>f ts. l \<noteq> Fun f ts"
    then obtain x where l:"l = Var x" by (cases l) auto
    with R have "(Var x,r) \<in> R" unfolding l by simp
    from lhs_var_imp_rstep_not_SN[OF this] show ?thesis by simp
  qed
  with assms show False by blast
qed
  
lemma SN_wfrstep_imp_wf_trs : assumes "SN (wfrstep R)" shows "wf_trs R"
proof (rule ccontr)
  assume "\<not> wf_trs R"
  then obtain l r where R: "(l,r) \<in> R"
    and not_wf: "(\<forall>f ts. l \<noteq> Fun f ts) \<or> \<not>(vars_term r \<subseteq> vars_term l)" unfolding wf_trs_def
    by auto
  from not_wf have "\<not> SN (wfrstep R)"
  proof
    assume free: "\<not> vars_term r \<subseteq> vars_term l"
    from rhs_free_vars_imp_wfrstep_not_SN[OF R free] show ?thesis .
  next
    assume "\<forall>f ts. l \<noteq> Fun f ts"
    then obtain x where l:"l = Var x" by (cases l) auto
    with R have "(Var x,r) \<in> R" unfolding l by simp
    from lhs_var_imp_wfrstep_not_SN[OF this] show ?thesis by simp
  qed
  with assms show False by blast
qed


fun map_term :: "('f \<Rightarrow> 'g) \<Rightarrow> ('f,'v)term \<Rightarrow> ('g,'v)term"
where "map_term fg (Var x) = (Var x)"
    | "map_term fg (Fun f ts) = Fun (fg f) (map (map_term fg) ts)"

fun map_subst :: "('f \<Rightarrow> 'g) \<Rightarrow> ('f,'v)sub \<Rightarrow> ('g,'v)sub"
where "map_subst fg (Sub sigma) = Sub (\<lambda> x. map_term fg (sigma x))"

fun map_rule :: "('f \<Rightarrow> 'g) \<Rightarrow> ('f,'v)rule \<Rightarrow> ('g,'v)rule"
where "map_rule fg lr = (map_term fg (fst lr), map_term fg (snd lr))"

fun map_ctxt :: "('f \<Rightarrow> 'g) \<Rightarrow> ('f,'v)ctxt \<Rightarrow> ('g,'v)ctxt"
where "map_ctxt fg \<box> = \<box>"
   |  "map_ctxt fg (More f ss C ts) = More (fg f) (map (map_term fg) ss) (map_ctxt fg C) (map (map_term fg) ts)"

fun map_trs :: "('f \<Rightarrow> 'g) \<Rightarrow> ('f,'v)trs \<Rightarrow> ('g,'v)trs"
where "map_trs fg R = (map_rule fg) ` R" 

fun map_trsI :: "('f \<Rightarrow> 'g) \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('g,'v)trsL"
  where "map_trsI fg R = map (map_rule fg) R"

lemma map_trsI_sound[simp]: "set (map_trsI fg R) = map_trs fg (set R)" by auto
declare map_trsI.simps[simp del]

lemma map_term_comp : "map_term fg (map_term gh t) = map_term (fg o gh) t"
  by (induct t, auto)

lemma map_trs_comp : "map_trs fg (map_trs gh R) = map_trs (fg o gh) R"
proof -  
  have mr: "map_rule (fg o gh) = map_rule fg o map_rule gh"
    by (rule ext, auto simp: map_term_comp)
  thus ?thesis
    by (auto simp: map_term_comp image_compose[symmetric] mr)
qed

lemma map_subst_distrib[simp] : "map_term fg (t \<cdot> \<sigma>) = map_term fg t \<cdot> map_subst fg \<sigma>"
proof (cases \<sigma>)
  case (Sub sigma)
  have "map_term fg (t \<cdot> Sub sigma) = map_term fg t \<cdot> map_subst fg (Sub sigma)"    
    by (induct t, auto)
  with Sub show ?thesis by simp
qed

lemma map_ctxt_distrib[simp] : "map_term fg (C\<langle>t\<rangle>) = (map_ctxt fg C)\<langle>map_term fg t\<rangle>" 
  by  (induct C, auto)

lemma map_trs_mono : assumes "R \<subseteq> R'" shows "map_trs fg R \<subseteq> map_trs fg R'"
  using assms by auto

lemma map_trs_power_mono : 
  fixes R R' :: "('f,'v)trs" and fg :: "'f \<Rightarrow> 'f"
  assumes "R \<subseteq> R'" shows "((map_trs fg)^^n) R \<subseteq> ((map_trs fg)^^n) R'"
using assms by (induct n, simp, auto)

lemma rstep_map_trs_power_mono : 
  fixes R R' :: "('f,'v)trs" and fg :: "'f \<Rightarrow> 'f"
  assumes subset: "R \<subseteq> R'" shows "rstep (((map_trs fg)^^n) R) \<subseteq> rstep (((map_trs fg)^^n) R')"
  by (rule rstep_mono, rule map_trs_power_mono, rule subset)

lemma rewrite_imp_map_rewrite : assumes "(s,t) \<in> rstep R" 
 shows "(map_term fg s, map_term fg t) \<in> rstep (map_trs fg R)"
using assms
proof (induct)
  fix ss ts C
  assume "(ss,ts) \<in> CS R"
  thus "(map_term fg C\<langle>ss\<rangle>, map_term fg C\<langle>ts\<rangle>) \<in> rstep (map_trs fg R)" 
  proof
    fix s t \<sigma> 
    assume s: "ss = s \<cdot> \<sigma>" and t: "ts = t \<cdot> \<sigma>" and rule: "(s,t) \<in> R"
    from rule have lrule: "(map_term fg s, map_term fg t) \<in> map_trs fg R" by force
    from s t have lst: "map_term fg ss = map_term fg s \<cdot> map_subst fg \<sigma> \<and> map_term fg ts = map_term fg t \<cdot> map_subst fg \<sigma>" by auto
    with lrule have "(map_term fg ss, map_term fg ts) \<in> CS (map_trs fg R)" by auto
    thus "(map_term fg C\<langle>ss\<rangle>, map_term fg C\<langle>ts\<rangle>) \<in> rstep (map_trs fg R)" by auto
  qed
qed

lemma rewrites_imp_map_rewrites : assumes "(s,t) \<in> (rstep R)^*"
  shows "(map_term fg s, map_term fg t) \<in> (rstep (map_trs fg R))^*"
using assms
proof (induct, simp)
  case (step y z)
  hence "(map_term fg y, map_term fg z) \<in> rstep (map_trs fg R)" using rewrite_imp_map_rewrite by auto
  with step show ?case by auto
qed

lemma SN_map_imp_SN : 
  fixes R:: "('g,'v)trs" and t :: "('g,'v)term"
  assumes SN: "SN_elt (rstep (map_trs fg R)) (map_term fg t)" 
  shows "SN_elt (rstep R) t"
proof (rule ccontr)
  assume "\<not> SN_elt (rstep R) t"
  from this obtain f where cond: "f 0 = t \<and> (\<forall> i. (f i, f (Suc i)) \<in> rstep R)" 
  unfolding SN_elt_def by auto
  obtain g where g: "g = (\<lambda> i. map_term fg (f i))" by auto
  with cond have cond2: "g 0 = map_term fg t \<and> (\<forall> i. (g i, g (Suc i)) \<in> rstep (map_trs fg R))"
    using rewrite_imp_map_rewrite[where fg = fg] by auto
  from SN have "\<not> (\<exists> g. (g 0 = map_term fg t \<and> (\<forall> i. (g i, g (Suc i)) \<in> rstep (map_trs fg R))))" 
    unfolding SN_elt_def by auto
  with cond2 show False  by auto
qed

lemma distinct_funas_trs_list:
  fixes trs::"('f,'v)trsL" shows "distinct(funas_trs_list trs)"
  unfolding funas_trs_list_def by simp

lemma r_into_rstep_sub: assumes "(l,r) \<in> R" shows "(l\<cdot>\<sigma>,r\<cdot>\<sigma>) \<in> rstep R"
proof -
  from assms have "(l\<cdot>\<sigma>,r\<cdot>\<sigma>) \<in> CS R" by (rule csI)
  hence "(\<box>\<langle>l\<cdot>\<sigma>\<rangle>,\<box>\<langle>r\<cdot>\<sigma>\<rangle>) \<in> CC(CS R)" by (rule ccI)
  thus ?thesis by simp
qed

lemmas rstep_intros = rstep_imp_ctxt_rstep r_into_rstep_sub

lemma rstep_cases'[consumes 1,case_names root nonroot]:
assumes rstep: "(s,t) \<in> rstep R"
  and root: "\<And>l r \<sigma>. (l,r) \<in> R \<Longrightarrow> l\<cdot>\<sigma> = s \<Longrightarrow> r\<cdot>\<sigma> = t \<Longrightarrow> P"
  and nonroot: "\<And>f ss1 u ss2 v. s = Fun f (ss1@u#ss2) \<Longrightarrow> t = Fun f (ss1@v#ss2) \<Longrightarrow> (u,v) \<in> rstep R \<Longrightarrow> P"
shows "P"
proof -
  from rstep_imp_C_s_r[OF rstep] obtain C \<sigma> l r
    where R: "(l,r) \<in> R" and s: "C\<langle>l\<cdot>\<sigma>\<rangle> = s" and t: "C\<langle>r\<cdot>\<sigma>\<rangle> = t" by fast
  show ?thesis proof (cases C)
    case Hole
    from s t have "l\<cdot>\<sigma> = s" and "r\<cdot>\<sigma> = t" by (auto simp: Hole)
    with R show ?thesis by (rule root)
  next
    case (More f ss1 D ss2)
    let ?u = "D\<langle>l\<cdot>\<sigma>\<rangle>"
    let ?v  = "D\<langle>r\<cdot>\<sigma>\<rangle>"
    have "s = Fun f (ss1@?u#ss2)" by (simp add: More s[symmetric])
    moreover have "t = Fun f (ss1@?v#ss2)" by (simp add: More t[symmetric])
    moreover have "(?u,?v) \<in> rstep R" by (intro rstep_intros) (rule R)
    ultimately show ?thesis by (rule nonroot)
  qed
qed

lemma NF_var: assumes wf: "wf_trs R" shows "(Var x,t) \<notin> rstep R"
proof
  assume "(Var x,t) \<in> rstep R"
  from rstep_imp_C_s_r[OF this] obtain C l r \<sigma>
    where R: "(l,r) \<in> R" and lhs: "Var x = C\<langle>l\<cdot>\<sigma>\<rangle>" by fast
  from lhs have "Var x = l\<cdot>\<sigma>" by (induct C) auto
  then obtain y where l: "l = Var y" by (induct l) auto
  from wf R obtain f ss where "l = Fun f ss" unfolding wf_trs_def by best
  with l show False by simp
qed

lemma rstep_cases_Fun'[consumes 2,case_names root nonroot]:
assumes wf: "wf_trs R"
  and rstep: "(Fun f ss,t) \<in> rstep R"
  and root': "\<And>ls r \<sigma>. (Fun f ls,r) \<in> R \<Longrightarrow> map (\<lambda>t. t\<cdot>\<sigma>) ls = ss \<Longrightarrow> r\<cdot>\<sigma> = t \<Longrightarrow> P"
  and nonroot': "\<And>i u. i < length ss \<Longrightarrow> t = Fun f (take i ss@u#drop (Suc i) ss) \<Longrightarrow> (ss!i,u) \<in> rstep R \<Longrightarrow> P"
shows "P"
using rstep proof (cases rule: rstep_cases')
  case (root l r \<sigma>)
  with wf obtain g ls where l: "l = Fun g ls" unfolding wf_trs_def by best
  from root have [simp]: "g = f" unfolding l by simp
  from root have "(Fun f ls,r) \<in> R" and "map (\<lambda>t. t\<cdot>\<sigma>) ls = ss" and "r\<cdot>\<sigma> = t" unfolding l by auto
  thus ?thesis by (rule root')
next
  case (nonroot g ss1 u ss2 v)
  hence [simp]: "g = f" and args: "ss = ss1@u#ss2" by auto
  let ?i = "length ss1"
  from args have ss1: "take ?i ss = ss1" by simp
  from args have "drop ?i ss = u#ss2" by simp
  hence "drop (Suc 0) (drop ?i ss) = ss2" by simp
  hence ss2: "drop (Suc ?i) ss = ss2" by simp
  from args have len: "?i < length ss" by simp
  from id_take_nth_drop[OF len] have "ss = take ?i ss @ ss!?i # drop (Suc ?i) ss" by simp
  hence u: "ss!?i = u" unfolding args unfolding ss1[unfolded args] ss2[unfolded args] by simp
  from nonroot have "t = Fun f (take ?i ss@v#drop (Suc ?i) ss)" unfolding ss1 ss2 by simp
  moreover from nonroot have "(ss!?i,v) \<in> rstep R" unfolding u by simp
  ultimately show ?thesis by (rule nonroot'[OF len])
qed

lemma clean_Var_id[simp]: "clean g F \<circ> Var = Var"
proof
  fix x show "(clean g F \<circ> Var) x = Var x" unfolding o_def by simp
qed

lemma singleton_sub_restrict[simp]: "{x\<leftarrow>s}|_{x} = {x\<leftarrow>s}"
unfolding subst_def sub_restrict.simps by (rule arg_cong[of _ _ Sub],rule ext) simp

lemma singleton_sub_map[simp]:
shows "(case {x\<leftarrow>s} of Sub \<sigma> \<Rightarrow> Sub(f\<circ>\<sigma>)) = Sub(\<lambda>y. if x = y then f s else f(Var y))"
proof -
  have "(case {x\<leftarrow>s} of Sub \<sigma> \<Rightarrow> Sub(f\<circ>\<sigma>)) = Sub(f\<circ>(\<lambda>y. if x = y then s else Var y))"
    unfolding subst_def by simp
  also have "\<dots> = Sub(\<lambda>y. if x = y then f s else f(Var y))" unfolding o_def if_distrib by simp
  finally show ?thesis .
qed

lemma map_nth_conv: "map f ss = map g ts \<Longrightarrow> \<forall>i < length ss. f(ss!i) = g(ts!i)"
proof (intro allI impI)
  fix i show "map f ss = map g ts \<Longrightarrow> i < length ss \<Longrightarrow> f(ss!i) = g(ts!i)"
  proof (induct ss arbitrary: i ts)
    case Nil thus ?case by (induct ts) auto
  next
    case (Cons s ss) thus ?case
    proof (induct ts)
      case Nil thus ?case by simp
    next
      case (Cons t ts) thus ?case by (cases i) auto
    qed
  qed
qed

(* The version of List is not general enough. *)
lemma map_eq_imp_length_eq:
assumes "map f ss = map g ts"
shows "length ss = length ts"
proof -
  from assms have "length(map f ss) = length(map g ts)" by simp
  thus ?thesis unfolding length_map .
qed

lemma sub_restrict_vars[simp]:
shows "((Sub(\<lambda>z. if z \<in> V then f z else g z))|_V) = (Sub f)|_V"
unfolding sub_restrict.simps
proof (rule arg_cong[of _ _ Sub],intro ext)
  fix x show "(if x \<in> V then if x \<in> V then f x else g x else Var x)
    = (if x \<in> V then f x else Var x)" by simp
qed

lemma sub_restrict_restrict[simp]:
assumes "V \<inter> W = {}"
shows "(Sub(\<lambda>z. if z \<in> V then f z else g z)|_W) = (Sub g)|_W"
unfolding sub_restrict.simps
proof (rule arg_cong[of _ _ Sub],intro ext)
  fix x show "(if x \<in> W then if x \<in> V then f x else g x else Var x)
    = (if x \<in> W then g x else Var x)" using assms by auto
qed

lemma nth_map_conv:
assumes "length xs = length ys"
  and "\<forall>i<length xs. f(xs!i) = g(ys!i)"
shows "map f xs = map g ys"
using assms proof (induct xs arbitrary: ys)
  case Nil thus ?case by simp
next
  case (Cons x xs) thus ?case
  proof (induct ys)
    case Nil thus ?case by simp
  next
    case (Cons y ys)
    have "\<forall>i<length xs. f(xs!i) = g(ys!i)"
    proof (intro allI impI)
      fix i assume "i < length xs" thus "f(xs!i) = g(ys!i)" using Cons(4) by force
    qed
    with Cons show ?case by auto
  qed
qed

lemma map_id[simp]: "map id xs = xs" by (induct xs) auto

lemma linear_clean_subst:
fixes F::"('f\<times>nat)set"
defines [simp]: "z \<equiv> (SOME x. True)"
defines [simp]: "c \<equiv> clean (\<lambda>_. z) F"
defines [simp]: "cs \<equiv> \<lambda>s. (case s of Sub s \<Rightarrow> Sub(c \<circ> s))" 
assumes ll: "linear_term t"
  and wf: "t \<in> Term(F)"
shows "c s = t\<cdot>\<sigma> \<Longrightarrow> \<exists>\<tau>. cs(\<tau>|_(vars_term t)) = (\<sigma>|_(vars_term t)) \<and> s = t\<cdot>\<tau>"
using ll wf proof (induct t arbitrary: s)
  case (Var x)
  have "cs({x\<leftarrow>s}|_(vars_term(Var x))) = cs({x\<leftarrow>s}|_{x})" by simp
  also have "\<dots> = cs({x\<leftarrow>s})" by simp
  also have "\<dots> = Sub(\<lambda>y. if x = y then c s else c(Var y))" unfolding cs_def by simp
  also have "\<dots> = Sub(\<lambda>y. if x = y then c s else Var y)" unfolding c_def clean.simps by simp
  also have "\<dots> = Sub(\<lambda>y. if x = y then Var x \<cdot> \<sigma> else Var y)" unfolding Var by simp
  also have "\<dots> = \<sigma>|_(vars_term(Var x))"
  proof (cases \<sigma>)
    case (Sub sub) show ?thesis unfolding Sub sub_restrict.simps
      by (rule arg_cong[of _ _ Sub],rule ext) auto
  qed
  finally have "cs({x\<leftarrow>s}|_(vars_term(Var x))) = \<sigma>|_(vars_term(Var x))" .
  moreover have "s = Var x\<cdot>{x\<leftarrow>s}" by (simp add: subst_def)
  ultimately show ?case by best
next
  case (Fun f ts)
  note IH = this(1) and cs = this(2) and ll = this(3) and wf = this(4)
  from ll have distinct: "is_partition(map vars_term ts)" by simp
  have "\<exists>ss. s = Fun f ss \<and> map c ss = map (\<lambda>t. t\<cdot>\<sigma>) ts"
  proof (cases s)
    case (Var x) from cs show ?thesis by (simp add: cs Var)
  next
    case (Fun g ss) from cs show ?thesis unfolding Fun by (cases "(g,length ss) \<in> F") auto
  qed
  then obtain ss where s: "s = Fun f ss" and args: "map c ss = map (\<lambda>t. t\<cdot>\<sigma>) ts" by auto
  from map_eq_imp_length_eq[OF args] have len: "length ts = length ss" ..
  from ll have "\<forall>i<length ts. linear_term(ts!i)" by auto
  moreover from wf have "\<forall>i<length ts. (ts!i) \<in> Term(F)" by (cases) auto
  moreover from args have "\<forall>i<length ts. c(ss!i) = (ts!i)\<cdot>\<sigma>"
    unfolding len by (auto simp: map_nth_conv)
  ultimately have "\<forall>i<length ts.\<exists>\<tau>. cs(\<tau>|_(vars_term(ts!i))) = \<sigma>|_(vars_term(ts!i))
    \<and> (ss!i) = (ts!i)\<cdot>\<tau>" using IH by force
  with distinct args have "\<exists>\<tau>. cs(\<tau>|_\<Union>set(map vars_term ts)) = \<sigma>|_\<Union>set(map vars_term ts)
    \<and> ss = map (\<lambda>t. t\<cdot>\<tau>) ts" (is "\<exists>\<tau>. ?P \<tau> ts \<and> ?Q \<tau> ts ss")
  proof (induct ts arbitrary: ss)
    case Nil thus ?case by auto
  next
    case (Cons t ts)
    from Cons(3) obtain u us where ss: "ss = u#us" by (cases ss) auto
    from Cons(4) obtain \<mu> where mu_sigma: "cs(\<mu>|_vars_term t) = \<sigma>|_vars_term t" and u: "u = t\<cdot>\<mu>"
      unfolding ss by auto
    obtain mu where mu: "\<mu> = Sub mu" by (cases \<mu>) auto
    from Cons(2) have distinct: "is_partition(map vars_term ts)"
      by (simp add: is_partition_Cons)
    moreover from Cons have "map c us = map (\<lambda>t. t\<cdot>\<sigma>) ts" by (simp add: ss)
    moreover from Cons(4)
      have "\<forall>i<length ts.\<exists>\<tau>. cs(\<tau>|_vars_term(ts!i)) = \<sigma>|_vars_term(ts!i)
      \<and> us!i = ts!i \<cdot> \<tau>" unfolding ss by force
    ultimately have "\<exists>\<tau>. ?P \<tau> ts \<and> ?Q \<tau> ts us" by (rule Cons(1))
    then obtain \<tau> where IH: "?P \<tau> ts" and us: "?Q \<tau> ts us" by best
    obtain tau where tau: "\<tau> = Sub tau" by (cases \<tau>) auto
    obtain sigma where sigma: "\<sigma> = Sub sigma" by (cases \<sigma>) auto
    let ?tau = "(\<lambda>z. if z \<in> vars_term t then mu z else tau z)"
    let ?\<tau> = "Sub ?tau"
    have "?Q ?\<tau> (t#ts) ss"
    proof -
      have "u = t\<cdot>?\<tau>"
      proof -
        from u have "u = t\<cdot>\<mu>" .
        also have "\<dots> = t \<cdot> (\<mu>|_vars_term t)" unfolding coincidence_lemma[symmetric] ..
        also have "\<dots> = t \<cdot> (?\<tau>|_vars_term t)" unfolding mu sub_restrict_vars ..
        also have "\<dots> = t \<cdot> ?\<tau>" unfolding coincidence_lemma[symmetric] ..
        finally show ?thesis .
      qed
      moreover have "us = map (\<lambda>t. t\<cdot>?\<tau>) ts"
      proof -
        from us have "length us = length ts" by simp
        moreover have "\<forall>i<length us. id(us!i) = (ts!i)\<cdot>?\<tau>"
        proof (intro allI impI)
          fix i assume len: "i < length us"
          from us have usts: "map id us = map (\<lambda>t. t\<cdot>\<tau>) ts" by simp
          with len have Suci: "Suc i < length(map vars_term (t#ts))"
            and i: "i < length ts" by auto
          from map_nth_conv[OF usts] have ith: "\<forall>i<length us. id(us!i) = ts!i\<cdot>\<tau>" .
          from Cons(2)[simplified map.simps is_partition_Cons] Suci
          have "vars_term t \<inter> (map vars_term ts ! i) = {}" by force
          hence disjoint: "vars_term t \<inter> vars_term(ts!i) = {}" using i by simp
          from ith len have "id(us!i) = ts!i\<cdot>\<tau>" by simp
          also have "\<dots> = (ts!i)\<cdot>(\<tau>|_vars_term(ts!i))" unfolding coincidence_lemma[symmetric] ..
          also have "\<dots> = (ts!i)\<cdot>(?\<tau>|_vars_term(ts!i))"
            using sub_restrict_restrict[OF disjoint,of mu tau] unfolding tau by auto
          also have "\<dots> = (ts!i)\<cdot>?\<tau>" unfolding coincidence_lemma[symmetric] ..
          finally show "id(us!i) = (ts!i)\<cdot>?\<tau>" .
        qed
        ultimately have "map id us = map (\<lambda>t. t\<cdot>?\<tau>) ts" by (rule nth_map_conv)
        thus ?thesis by simp
      qed
      ultimately show ?thesis unfolding ss by simp
    qed
    moreover have "?P ?\<tau> (t#ts)" unfolding cs_def sigma sub_restrict.simps sub.simps(2) o_def
    proof (rule arg_cong[of _ _ Sub],intro ext)
      fix x show "c(if x \<in> \<Union>set (map vars_term (t#ts)) then if x \<in> vars_term t then mu x
        else tau x else Var x) = (if x \<in> \<Union>set (map vars_term (t#ts)) then sigma x else Var x)"
        (is "c ?\<mu> = ?\<sigma>")
      proof (cases "x \<in> \<Union>set(map vars_term (t#ts))")
        case False thus ?thesis by simp
      next
        case True note in_vars = this show ?thesis
        proof (cases "x \<in> vars_term t")
          case True
          have rhs: "?\<sigma> = sigma x" using True by simp
          from fun_cong[OF mu_sigma[unfolded mu sigma o_def sub.simps(2) sub.inject sub_restrict.simps cs_def]]
            have "c(if x \<in> vars_term t then mu x else Var x) = (if x \<in> vars_term t then sigma x else Var x)" by simp
          hence "c ?\<mu> = sigma x" using True by simp
          with rhs show ?thesis by simp
        next
          case False
          with in_vars have in_tail: "x \<in> \<Union>set(map vars_term ts)" by simp
          from Cons(2)[simplified map.simps is_partition_Cons]
            have disjoint: "vars_term t \<inter> \<Union>set(map vars_term ts) = {}" by auto
          with in_tail have notin_t: "x \<notin> vars_term t" by auto
          have rhs: "?\<sigma> = sigma x" using in_tail by simp
          from fun_cong[OF IH[unfolded tau  sigma sub_restrict.simps cs_def o_def sub.simps(2) sub.inject]]
            have "c(if x \<in> \<Union>set(map vars_term ts) then tau x else Var x) = (if x \<in> \<Union>set(map vars_term ts) then sigma x else Var x)" by simp
          hence "c ?\<mu> = sigma x" using in_tail notin_t by simp
          with rhs show ?thesis by simp
        qed
      qed
    qed
    ultimately show ?case by best
  qed
  thus ?case unfolding s by auto
qed

definition left_linear_trs
where "left_linear_trs R \<equiv> \<forall>(l,r)\<in>R. linear_term l"

lemma left_linear_rstep_imp_clean_rstep:
fixes R::"('f,'v)trs" and F::"('f\<times>nat)set"
defines [simp]: "z \<equiv> (SOME x. True)"
defines [simp]: "c \<equiv> clean (\<lambda>_. z) F"
defines [simp]: "cs \<equiv> \<lambda>s. (case s of Sub s \<Rightarrow> Sub(c \<circ> s))" 
assumes rstep: "(c s,t) \<in> rstep R"
  and subset: "funas_trs R \<subseteq> F"
  and wf: "wf_trs R"
  and ll: "left_linear_trs R"
shows "\<exists>u. c u = t \<and> (s,u) \<in> rstep R"
using rstep proof (induct s arbitrary: t rule: term_induct)
  case (Var x) with wf show ?case by (simp add: NF_var wf)
next
  case (Fun f ss)
  note IH = this(1) and rstep = this(2)
  show ?case
  proof (cases "(f,length ss) \<in> F")
    case False
    with rstep NF_var[OF wf,of z] show ?thesis by simp
  next
    case True
    from rstep have "(Fun f (map c ss),t) \<in> rstep R" using True by simp
    with wf show ?thesis
    proof (cases rule: rstep_cases_Fun')
      case (nonroot i u)
      let ?ss1 = "take i ss"
      let ?ss2 = "drop (Suc i) ss"
      let ?C = "More f ?ss1 \<box> ?ss2"
      from nonroot have step: "(c(ss!i),u) \<in> rstep R" and len: "i < length ss" by auto
      from IH[OF nth_mem[OF len] step]
        obtain v where u: "c(v) = u" and arg_step: "(ss!i,v) \<in> rstep R" by auto
      from id_take_nth_drop[OF len] have len_ss: "length(?ss1 @ ss!i # ?ss2) = length ss" by simp
      from nonroot(2)[unfolded take_map drop_map u[symmetric] map.simps(2)[symmetric] map_append[symmetric]]
        have "c(Fun f (?ss1 @ v # ?ss2)) = t" using True unfolding len_ss[symmetric] by auto
      moreover from arg_step have "(?C\<langle>ss!i\<rangle>,?C\<langle>v\<rangle>) \<in> rstep R" by (rule rstep_imp_ctxt_rstep)
      ultimately show ?thesis unfolding ctxt_apply.simps unfolding id_take_nth_drop[OF len,symmetric] by best
    next
      case (root ls r \<sigma>)
      hence linear: "linear_term(Fun f ls)" using ll unfolding left_linear_trs_def by auto
      moreover from root have "Fun f ls \<in> Term(F)" using lhs_wf[of "Fun f ls" r R]
        subset by simp
      moreover have "c(Fun f ss) = (Fun f ls)\<cdot>\<sigma>" unfolding sub_apply.simps root(2) using True by simp
      ultimately obtain \<tau> where cs_tau: "cs(\<tau>|_vars_term(Fun f ls)) = \<sigma>|_vars_term(Fun f ls)" and s_tau: "Fun f ss = Fun f ls\<cdot>\<tau>"
        using linear_clean_subst[of "Fun f ls" "F" "Fun f ss"] unfolding c_def z_def cs_def by best
      from root have r_wf: "r \<in> Term(F)" using rhs_wf[of "Fun f ls" r R F] subset by best
      from root and wf have subset: "vars_term r \<subseteq> vars_term(Fun f ls)" unfolding wf_trs_def by best
      have "c(r\<cdot>\<tau>) = c(r\<cdot>(\<tau>|_vars_term r))" unfolding coincidence_lemma[symmetric] ..
      also have "\<dots> = c(r\<cdot>(\<tau>|_vars_term(Fun f ls)))" using coincidence_lemma'[OF subset] by simp
      also from clean_sub_apply[OF r_wf] have "\<dots> = r \<cdot> cs(\<tau>|_vars_term(Fun f ls))" unfolding c_def cs_def by (cases \<tau>) auto
      also have "\<dots> = r \<cdot> (\<sigma>|_vars_term(Fun f ls))" unfolding cs_tau ..
      also have "\<dots> = r \<cdot> (\<sigma>|_vars_term r)" unfolding coincidence_lemma'[OF subset,symmetric] ..
      also have "\<dots> = r \<cdot> \<sigma>" unfolding coincidence_lemma[symmetric] ..
      finally have "c(r\<cdot>\<tau>) = t" using root by simp
      moreover from root have "(Fun f ls\<cdot>\<tau>,r\<cdot>\<tau>) \<in> rstep R" using r_into_rstep_sub by best
      ultimately show ?thesis unfolding s_tau by auto
    qed
  qed
qed

(* output *)
types ('f,'v)dppL = "('f shp,'v)trsL \<times>  ('f shp,'v)trsL"
fun shows_rule :: "('f::shows,'v::shows)rule \<Rightarrow> shows"
where "shows_rule(l,r) = (shows l +@+  '' -> '' +#+  shows r)"

definition shows_rules :: "('f::shows,'v::shows)trsL \<Rightarrow> shows"
where "shows_rules trs \<equiv> shows_list_gen shows_rule [] [] [newline] [] trs +@+ shows_nl"

definition shows_p :: "('f::shows,'v::shows)trsL \<Rightarrow> shows"
where "shows_p P  \<equiv> ''dependency pairs:'' +#+ shows_nl +@+ shows_rules P"

definition shows_trs :: "('f::shows,'v::shows)trsL \<Rightarrow> shows"
where "shows_trs R \<equiv> ''rewrite system:'' +#+ shows_nl +@+ shows_rules R"

definition shows_dpp :: "('f::shows,'v::shows)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> shows"
where "shows_dpp P R \<equiv> shows_p P +@+ shows_trs R"

end
